/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: mod_enhydra_director.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"
#include "edir_table.h"
#include "edir_kvtable.h"

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
#include "http_request.h"

#define EDIR_APACHE_PREBUF_SIZE		(2048)
#define EDIR_APACHE_INBUF_SIZE		(4096)
#define EDIR_APACHE_OUTBUF_SIZE		(4096)
 
#define EDIR_APACHE_CFG_MODE_SERVER	(1)
#define EDIR_APACHE_CFG_MODE_DIR		(2)

static void
edir_apache_get_request_headers(request_rec *r,
                                edir_kvtable_t *cgiHeaders,
                                edir_kvtable_t *httpHeaders);

/**
 * Enhydra Director configuration info.  
 */
typedef struct edir_apache_cfg {
    /*** Bookeeping to keep track of where in Apache config we are ***/
    int   cfg_Mode;	   /* Per-server or Per-directory config?       */
    int   inited;	   /* Whether we've initialized yet.            */
    int   cfg_Inheritable; /* Can virt. servers inherit this from main? */
    char  *cfg_Server;	   /* Server name if per-virtual-host config    */
    char  *cfg_Dir;	   /* Dir path if per-directory config          */
    /*** Per-instance config for each Enhydra Director setup. ***/
    int32 inBufSize;
    int32 outBufSize;
    char  *configFile;     /* Configuration file for this instance      */
    char  *dataFile;       /* Optional place to store scoreboard data   */
                           /*   NOTE: This _may_ be an IPC key #        */
    char  *lockFile;       /* Optional place to locate locks/mutexes    */
                           /*   NOTE: This _may_ be an IPC key #        */
    char  *pathToDaemon;   /* Optional full path to daemon executable.  */
    char  *sessionIdPrefix;/* Prefix of Session ID cookie or component  */
    edir_load_handles_t
          *handles;        /* Scoreboard handles. */
} edir_apache_cfg_t;

/**
 * Structure to pass to edir_parse_uri() to receive important URI
 * fragments.
 */
typedef struct edir_uri_data {
    char *uriBuf;	   	/* destroyed uri buffer */
    char *pathInfo;	   	/* Stripped, normalized URI path. */ 
    char *pathParams;	   	/* Full path parameters string, if any */
    char *sessionID;       	/* Session ID, if any.  */
    char *queryString;     	/* Query String, if any. */
    edir_kvtable_t *cookies;	/* Hack: let cookie getter set sessionID */
    pool *pool;		   	/* Hack: cookie getter needs this */
} edir_uri_data_t;

#define DIR_CONFIG(R) \
	((edir_apache_cfg_t *) \
	  (ap_get_module_config((R)->per_dir_config, &enhydra_director_module)));

#define SERVER_CONFIG(S) \
	((edir_apache_cfg_t *) \
	  (ap_get_module_config((S)->module_config, &enhydra_director_module)));

#define REQ_CONFIG(R) \
	((edir_apache_cfg_t *) \
	  (ap_get_module_config((R)->request_config, &enhydra_director_module)));

/*
 * Default config file.
 */
static char * DefaultConfigFile = NULL;

/*
 * Use a module-private 'main' pool, with some sub-pools to handle 
 * memory allocation and cleanup for various actions.  These pools are:
 * 	edir_apache_pool            -- The main pool.
 * 	edir_apache_pool_request    -- For per-request resources that need
 *                                   to be cleaned up each time.
 * 	edir_apache_pool_current    -- The current pool to use.  This will
 *                                   change, depending upon what's being
 *                                   built or torn down.
 * By doing it this way we can override edir_memory_standard.c with one
 * that uses the 'current' pool so that we can be reasonable sure of
 * memory cleanup when a request is finished.
 */
static pool *edir_apache_pool = NULL;
static pool *edir_apache_pool_request = NULL;
static pool *edir_apache_pool_current = NULL;

/*
 * Forward declaration to allow the module definition to be at the end
 * of this file.
 */
module enhydra_director_module;

/**
 * Convenience function to do in-place removal of leading and trailing 
 * whitespace characters.
 * <P>
 * Side effects:<BR>
 * This function modifies the string pointed to by "s".
 * 
 * @param s [in,out] String to trim in place.
 * @return A pointer to the trimmed portion of the string.
 */
static char *
trim_inplace(char *s)
{
    char *cp;
    EDIR_DEBUG_FUNCS_IN(("func: trim_inplace(\"%s\")", s ? s : "(null)"));
    while (*s && isspace(*s)) s++;
    for (cp=s; *cp; cp++) ;
    for (--cp; (cp >= s) && isspace(*cp); cp--) *cp = '\0';
    EDIR_DEBUG_FUNCS_OUT(("func: trim_inplace() returns \"%d\"",s?s:"(null)"));
    return s;
}

/**
 * Function to handle components of the request URI as it is parsed.
 * <P>
 * The edir_parse_uri() function calls this function for each part
 * of the url that it parses.  The four possible values of the <code>
 * type</code> parameter are described below. <BR>
 * <code>EDIR_PARSE_URI_PATH_INFO</code><BR>
 * The value of the <code>lhs</code> parameter will point to a character
 * string containing the stripped path portion of the URI, with leading
 * and trailing slashes removed, and with interior runs of more than
 * one slash reduced to one slash.
 * <code>EDIR_PARSE_URI_PATH_PARAM_1</code>
 * Returns a pointer to a string in <code>lhs</code> that contains a URL
 * "path parameter" that is not of the form "key=value".  The <code>rhs</code>
 * parameter is always <code>NULL</code>.
 * <code>EDIR_PARSE_URI_PATH_PARAM_2</code>
 * Returns strings comprising "path parameter" in the form of a "key=value"
 * pair, with the key in <code>lhs</code>, and the value in <code>rhs</code>.
 * Neither parameters are ever <code>NULL</code> for this type.
 * <code>EDIR_PARSE_URI_QUERY_STRING</code>
 * Returns a pointer to the beginning of the "query string" in <code>lhs</code>.
 * <P>
 * NOTE: The strings passed to this function are part of a larger
 * string buffer that is being destructively parsed by edir_parse_uri(). They
 * should not be freed, and they should only be stored in conjunction with
 * the larger URI buffer that contains them so the buffer does not
 * get freed while the strings are still in use.
 * 
 * @param ptr  [in,out] Arbitrary data structure to store things in.
 * @param type [in]     What type of component has been parsed.
 * @param lhs  [in]     "Left hand side" of "key=value" pair, or full
 * 			String of "value" item.
 * @param rhs  [in]     "Right hand side" of "key=value" pair, or NULL.
 */
static void
edir_apache_uri_parse_func(void *ptr, int type, char *lhs, char *rhs)
{
    edir_uri_data_t *uriData = (edir_uri_data_t *) ptr;
    switch (type) {
	case EDIR_PARSE_URI_PATH_INFO:
	    uriData->pathInfo = lhs;
	    break;
	case EDIR_PARSE_URI_PATH_PARAMS:
	    uriData->pathParams = ap_pstrdup(uriData->pool, lhs);
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_1:
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_2:
	    if (lhs && !strcmp(lhs, "jsessionid")) {
		uriData->sessionID = rhs;
	    }
	    break;
	case EDIR_PARSE_URI_QUERY_STRING:
	    uriData->queryString = lhs;
	    break;
    }
}

/**
 * Apache table "do" function to parse a cookie header.
 * <P>
 * This function takes a Cookie header, which is a semicolon-delimited
 * list of cookie key-value pairs.  Each key-value pair is placed
 * separately into the table passed through the "rec" pointer.
 * 
 * @param rec   [in,out] The "rec" pointer passed to ap_table_do().
 * @param key   [in]	 The key found by ap_table_do(). ("Cookie" here).
 * @param value [in]	 The value found by ap_table_do().
 * @return Zero (false) to stop the iteration, Nonzero (true) to keep going.
 */
static int
edir_apache_do_cookies(void *rec, const char *key, const char *value)
{
    static char		*funcName = "edir_apache_do_cookies";
    edir_uri_data_t	*uriData = (edir_uri_data_t *) rec;
    char		*start, *end, *name, *val, *fp = NULL;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", \"%s\", \"%s\")", funcName, \
		      rec, key ? key : "(null)", value ? value : "(null)"));
    fp = start = edir_strdup(value);
    end = strchr(start, ';');
    while (end) {
        *end++ = '\0';
        name = start;
        val = strchr(name, '=');
        if (val) *val++ = '\0';
        else val = "";
	name = trim_inplace(name);
	val = trim_inplace(val);
	edir_kvtable_add(uriData->cookies, name, val);
	if ((uriData->sessionID == NULL) &&
	    (!edir_strcasecmp(name, "jsessionid"))) {
	    uriData->sessionID = (char *) ap_pstrdup(uriData->pool, val);
	}
        start = end;
        end = strchr(start, ';');
    }
    name = start;
    val = strchr(name, '=');
    if (val)
        *val++ = '\0';
    else
        val = "";
    name = trim_inplace(name);
    val = trim_inplace(val);
    edir_kvtable_add(uriData->cookies, name, val);
    if ((uriData->sessionID == NULL) &&
	(!edir_strcasecmp(name, "jsessionid"))) {
	uriData->sessionID = (char *) ap_pstrdup(uriData->pool, val);
    }
    edir_free(fp);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1;
}

/**
 * Apache table "do" function to break up a comma-delimited header.
 * <P>
 * This function takes a comma-delimited header and splits it
 * into multiple headers with the same name.
 * 
 * @param rec   [in,out] The "rec" pointer passed to ap_table_do().
 * @param key   [in]	 The key found by ap_table_do().
 * @param value [in]	 The value found by ap_table_do().
 * @return Zero (false) to stop the iteration, Nonzero (true) to keep going.
 */
static int
edir_apache_do_commalist(void *rec, const char *key, const char *value)
{
    table	*headers = (table *) rec;
    char	*start, *end, *fp;
    EDIR_DEBUG_FUNCS_IN(("func: edir_apache_do_commalist("EDIR_PTR", \"%s\", \"%s\")",\
		      rec, key ? key : "(null)", value ? value : "(null)"));
    fp = start = edir_strdup(value);
    end = strchr(start, ',');
    while (end) {
        *end++ = '\0';
	ap_table_add(headers, key, trim_inplace(start));
        start = end;
        end = strchr(start, ',');
    }
    ap_table_add(headers, key, trim_inplace(start));
    edir_free(fp);
    EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_do_commalist() returns %d", 1));
    return 1;
}

/**
 * Create a per-directory configuration record for mod_enhydra_director.
 * <P>
 * Note that this won't be called, since Enhydra Director is not configurable
 * on a per-directory basis.
 *
 * @param p       [in,out] Apache resource pool for temporarily allocated data.
 * @param dirspec [in      Directory under which this configuration resides.
 * @return The configuration record.
 */
static void *edir_apache_create_dir_config(pool *p, char *dirspec)
{
    edir_apache_cfg_t *cfg;
    cfg = (edir_apache_cfg_t *) ap_pcalloc(p, sizeof(edir_apache_cfg_t));
    cfg->cfg_Mode = EDIR_APACHE_CFG_MODE_DIR;
    cfg->inited = 0;
    cfg->cfg_Inheritable = 0;
    cfg->cfg_Server = NULL;
    cfg->cfg_Dir = dirspec ? ap_pstrdup(p, dirspec) : NULL;
    cfg->configFile = NULL;
    cfg->dataFile = NULL;
    cfg->lockFile = NULL;
    cfg->pathToDaemon = NULL;
    cfg->inBufSize = EDIR_APACHE_INBUF_SIZE;
    cfg->outBufSize = EDIR_APACHE_OUTBUF_SIZE;
    cfg->sessionIdPrefix = ap_pstrdup(p, EDIR_LOAD_DEFAULT_SID_PREFIX);
    cfg->handles = NULL;
    return (void *)cfg;
}

/**
 * Create a per-directory configuration record for mod_enhydra_director.
 * <P>
 * Note that this won't be called, since Enhydra Director is not configurable
 * on a per-directory basis.
 *
 * @param p       [in,out] Apache resource pool for temporarily allocated data.
 * @param dirspec [in      Directory under which this configuration resides.
 * @return The configuration record.
 */
static void *
edir_apache_create_server_config(pool *p, server_rec *s)
{
    edir_apache_cfg_t	*cfg;
    cfg = (edir_apache_cfg_t *) ap_pcalloc(p, sizeof(edir_apache_cfg_t));
    cfg->cfg_Mode = EDIR_APACHE_CFG_MODE_SERVER;
    cfg->inited = 0;
    cfg->cfg_Inheritable = 0;
    if (s->server_hostname)
	cfg->cfg_Server = ap_pstrdup(p, s->server_hostname);
    else
	cfg->cfg_Server = NULL;
    cfg->cfg_Dir = NULL;
    cfg->configFile = NULL;
    cfg->dataFile = NULL;
    cfg->lockFile = NULL;
    cfg->pathToDaemon = NULL;
    cfg->inBufSize = EDIR_CLIENT_DEFAULT_INBUF_SIZE;
    cfg->outBufSize = EDIR_CLIENT_DEFAULT_OUTBUF_SIZE;
    cfg->sessionIdPrefix = ap_pstrdup(p, EDIR_LOAD_DEFAULT_SID_PREFIX);
    cfg->handles = NULL;
    return (void *)cfg;
}

/**
 * Command parse handler for the 'EnhydraDirectorVirtualInherit' command. (FLAG)
 * 
 * @param cmd     [in]     Apache command parsing context handle.
 * @param notused [unused] Unused Parameter.
 * @param bool    [in]     Boolean configuration value to set.
 * @return An error message, if a misconfiguration is detected.
 */
static const char *
cmd_Inherit(cmd_parms *cmd, void *notused, int bool)
{
    edir_apache_cfg_t *cfg = SERVER_CONFIG(cmd->server);
    cfg->cfg_Inheritable = bool;
    return NULL;
}

/**
 * Command parse handler for the 'EnhydraDirectorConfigFile' command. (TAKE1)
 * 
 * @param cmd     [in]     Apache command parsing context handle.
 * @param notused [unused] Unused Parameter.
 * @param value   [in]     String configuration value to set.
 * @return An error message, if a misconfiguration is detected.
 */
static const char *
cmd_ConfigFile(cmd_parms *cmd, void *notused, char *value)
{
    edir_apache_cfg_t *cfg = SERVER_CONFIG(cmd->server);
    cfg->configFile = value ? ap_pstrdup(cmd->pool, value) : NULL;
    return NULL;
}

/**
 * Command parse handler for the 'EnhydraDirectorDataFile' command. (TAKE1)
 * 
 * @param cmd     [in]     Apache command parsing context handle.
 * @param notused [unused] Unused Parameter.
 * @param value   [in]     String configuration value to set.
 * @return An error message, if a misconfiguration is detected.
 */
static const char *
cmd_DataFile(cmd_parms *cmd, void *notused, char *value)
{
    edir_apache_cfg_t *cfg = SERVER_CONFIG(cmd->server);
    cfg->dataFile = value ? ap_pstrdup(cmd->pool, value) : NULL;
    return NULL;
}

/**
 * Command parse handler for the 'EnhydraDirectorLockFile' command. (TAKE1)
 * 
 * @param cmd     [in]     Apache command parsing context handle.
 * @param notused [unused] Unused Parameter.
 * @param value   [in]     String configuration value to set.
 * @return An error message, if a misconfiguration is detected.
 */
static const char *
cmd_LockFile(cmd_parms *cmd, void *notused, char *value)
{
    edir_apache_cfg_t *cfg = SERVER_CONFIG(cmd->server);
    cfg->lockFile = value ? ap_pstrdup(cmd->pool, value) : NULL;
    return NULL;
}

/**
 * Command parse handler for the 'EnhydraDirectorDaemonPath' command. (TAKE1)
 * 
 * @param cmd     [in]     Apache command parsing context handle.
 * @param notused [unused] Unused Parameter.
 * @param value   [in]     String configuration value to set.
 * @return An error message, if a misconfiguration is detected.
 */
static const char *
cmd_DaemonPath(cmd_parms *cmd, void *notused, char *value)
{
    edir_apache_cfg_t *cfg = SERVER_CONFIG(cmd->server);
    cfg->pathToDaemon = value ? ap_pstrdup(cmd->pool, value) : NULL;
    return NULL;
}

/**
 * Command parse handler for the 'EnhydraDirectorDebug' command. (TAKE1)
 * 
 * @param cmd     [in]     Apache command parsing context handle.
 * @param notused [unused] Unused Parameter.
 * @param value   [in]     String configuration value to set.
 * @return An error message, if a misconfiguration is detected.
 */
static const char *
cmd_Debug(cmd_parms *cmd, void *notused, char *value)
{
    edir_atoi(value, &edir_debug);
    return NULL;
}

/**
 * Handler function for the Enhydra Director Apache module.
 * <P>
 * If the translate function determines that a particular request is 
 * a servlet request, then then request's handler field is set to refer
 * to this function.  This function then does the actual work of processing
 * the request.
 * <P>
 * If it turns out that the translate routine was lying to us, we just 
 * return DECLINED and the user sees "not found", which is correct because
 * something is not configured right.  If the request fails for some 
 * reason, we return an HTTP_* status as appropriate and log the error.
 * Otherwise, when the request is done, return OK.
 * 
 * @param r [in,out] Apache 'request_rec' structure containing information
 * 		about this request.
 * @return Integer Apache status code.  Normally 'OK' if successful.
 */
static int edir_apache_handler(request_rec *r)
{
    static char			*funcName="edir_apache_handler";
    edir_exception_t		*_exc=NULL, **exc=&_exc;
    server_rec			*s = r->server;
    char			*stickyTag = NULL, *path_info = NULL;
    char			*cp, *key, *val, **vals;
    edir_kvtable_t		*cgi_vars = NULL, *http_vars=NULL;
    table			*cookies = NULL, *merge = NULL;
    array_header		*hdrlist;
    table_entry			*hdrs;
    void			*sbuf = NULL;
    edir_apache_cfg_t		*cfg;
    edir_uri_data_t		*uriData = NULL;
    edir_connection_t		*conn = NULL;
    edir_load_handles_t		*handles = NULL;
    edir_load_sb_app_t		*app;
    edir_load_sb_server_t	*server;
    int32			*refCounts, nRefCounts, nvals;
    int				releaseConn, i, n, nhdrs, subtype, remain;
    int				stickyRetries = 0;
    time_t			startTime, elapsedTime, lastTime, now;
    int				inBufSize, outBufSize;
    octet			*inBuf, *outBuf;
    int				cond_status = HTTP_INTERNAL_SERVER_ERROR;
    int				cgi_status;
    int				do_status = 0;

    edir_apache_log_set_request_rec(r);
    EDIR_DEBUG_FUNCS_IN(("func: edir_apache_handler("EDIR_PTR")", r));

    /* 
     * Get config and set up buffers
     */
    cfg = SERVER_CONFIG(s);
    inBufSize = cfg->inBufSize;
    outBufSize = cfg->outBufSize;
    inBuf = ap_pcalloc(r->pool, inBufSize);
    outBuf = ap_pcalloc(r->pool, outBufSize);

    /*
     * Get apache ready to send us the request body, if applicable.
     * The mod_enhydra_director module supports the "chunked" encoding.
     * This function checks Content-Length and Transfer-Encoding to
     * see if we want to read a request entity and send it to Enhydra.
     * Note that we still keep the chunks when sending to Enhydra
     * because the app may assume that it must count the bytes.  This
     * just lets us know when to stop sending data without having to
     * parse the stream ourselves.
     */
    if ((i = ap_setup_client_block(r, REQUEST_CHUNKED_PASS)) != OK) {
	EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns %d", i));
	return i;
    }
    
    /*
     * Parse the URI, looking for path parameters in particular.
     */
    uriData = (edir_uri_data_t *) ap_pcalloc(r->pool, sizeof(edir_uri_data_t));
    memset((char *)uriData, 0, sizeof(edir_uri_data_t));
    uriData->pool = r->pool;
#ifdef USE_UNPARSED_URI
    uriData->uriBuf = ap_pstrdup(r->pool,
				 r->unparsed_uri ? r->unparsed_uri : "");
#else
    uriData->uriBuf = ap_pstrdup(r->pool, r->uri ? r->uri : "");
#endif

    EDIR_DEBUG_EVERYTHING(("%s: r->uri = '%s'", funcName, \
			   r->uri?r->uri:"(null)"));
    EDIR_DEBUG_EVERYTHING(("%s: uriBuf = '%s'", funcName, uriData->uriBuf));

    edir_parse_uri(uriData->uriBuf, uriData, edir_apache_uri_parse_func);

    EDIR_DEBUG_EVERYTHING(("%s: PARSED pathinfo = '%s'", funcName, \
			uriData->pathInfo ? uriData->pathInfo : "(null)"));
    EDIR_DEBUG_EVERYTHING(("%s: PARSED queryString = '%s'", funcName, \
			uriData->queryString ? uriData->queryString:"(null)"));

    if (uriData->pathInfo == NULL) {
	EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns DECLINED"));
	return DECLINED;
    }

    /*
     * Gather the cookies, separate them into separate entries, and
     * put them in a table.  This is needed so we can look for the
     * application session cookie for load balancing purposes.
     */
    uriData->cookies = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
    if (r->headers_in) {
	ap_table_do(edir_apache_do_cookies, uriData, r->headers_in,
		    "Cookie", NULL);
    }
    ap_table_unset(r->headers_in, "Cookie");

    EDIR_DEBUG_EVERYTHING(("%s: PARSED sessionID = '%s'", funcName, \
			uriData->sessionID ? uriData->sessionID : "(null)"));

    /*
     * If we have a session ID, look for sticky session tag.
     */
    if ((uriData->sessionID) && (val = strchr(uriData->sessionID, '('))) {
	if (strlen(val) >= 9) stickyTag =(char *)ap_pstrndup(r->pool,val+1,8); 
	if (stickyTag) stickyRetries = EDIR_CONNECT_STICKY_SESSION_RETRIES;
    }

    /*
     * Make sure the scoreboard hasn't been killed/reloaded.
     */
    handles = cfg->handles = edir_load_check(cfg->handles, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * See if we really only want to return scoreboard status
     */
    if (edir_load_status_check_path(handles->scoreboard, r->uri,
		&(r->connection->local_addr), 
		&(r->connection->remote_addr), 
		r->server->server_hostname)) {
	do_status = 1;
    }

    if (!do_status) {
/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	/*
	 * Lock the scoreboard briefly while we look for an app matching
	 * our prefix.
	 */
	edir_mutex_lock_exclusive(handles->mutex,
				  EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	if (*exc) EDIR_THROW(*exc);
	refCounts = edir_refcount_values(handles->refcnt, &nRefCounts, exc);
	if (*exc) {
	    edir_exception_t *exc2=NULL;
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
	    edir_exception_free(&exc2);
	    EDIR_THROW(*exc);
	}
	edir_load_choose(handles->scoreboard, uriData->pathInfo, stickyTag,
			 &app, &server, refCounts, nRefCounts, 
			 &(r->connection->local_addr), 
			 &(r->connection->remote_addr), 
			 r->server->server_hostname, exc);
	if (*exc) {
	    edir_exception_t *exc2=NULL;
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
	    edir_exception_free(&exc2);
	    EDIR_THROW(*exc);
	}
	EDIR_DEBUG_NORMAL(("Got server: %s:%d", server->hostName, \
			  server->port));
	server->useCount++; /* Up the use count on this server. */
	path_info = uriData->pathInfo + app->appPrefixLen;

	/*
	 * Up the refcount
	 */
	releaseConn = 1;
	edir_refcount_increment(handles->refcnt, server->refIndex, exc);
	if (*exc) {
	    edir_exception_t *exc2=NULL;
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
	    edir_exception_free(&exc2);
	    EDIR_THROW(*exc);
	}

	/*
	 * Give up scoreboard lock, now that we have our server.
	 */
	edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	if (*exc) EDIR_THROW(*exc);
    }
/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
    if (!(edir_debug & EDIR_DEBUG_MASK_FAKE_IT) && !do_status) {
	conn = (edir_connection_t *) edir_malloc(sizeof(edir_connection_t));
	memset((void *)conn, 0, sizeof(edir_connection_t));
	time(&startTime);
	lastTime = startTime;
	elapsedTime = 0;
	while (elapsedTime <= EDIR_CONNECT_TIMEOUT) {
	    if (*exc) edir_exception_free(exc);
	    edir_client_connect(conn, NULL, &(server->sin), server->authData,
			      outBuf, outBufSize, inBuf, inBufSize, exc);
	    if (!(*exc)) break; /* Connection established! */
	    if ((*exc)->id != EDIR_EXCEPT_IO_ERROR) EDIR_THROW(*exc);

	    /* XXX Should have 'logmask' */
	    if (edir_debug & EDIR_DEBUG_MASK_NORMAL) {
		char *trace = NULL;
		if (*exc) trace = edir_exception_stack_trace(*exc);
		edir_log_error("Warning: Couldn't connect to %s:%d.",
			       server->hostName, server->port);
		edir_log_error("%s", trace ? trace : "No details.");
		if (trace) edir_free(trace);
	    }
	    if (*exc) edir_exception_free(exc);

	    /* Mark this server as 'failing' */
/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	    edir_mutex_lock_exclusive(handles->mutex,
				      EDIR_LOAD_SB_LOCK_GLOBAL,exc);
	    if (*exc) EDIR_THROW(*exc);
	    if (!(server->failing)) {
		/*
		 * XXX FIX  - Idea: Run notification script.
		 * Especially on Unix, where it's easy, run a script if we
		 * find a server to be 'failing'.  Probably would want to
		 * only run the script occasionally to avoid spamming,
		 * and would want to run it outside any critical section
		 * to avoid blocking for prolonged periods while holding
		 * the global scoreboard lock.
		 */
		server->failing = 1;
		time(&(server->failTime));
	    }
	    server->failCount++;
	    /*
	     * If this request has a 'sticky session' token in the session
	     * cookie, then we want to try several times to go to the desired
	     * server.  After a number of attempts, though, we give up and
	     * fall back to round robin.
	     */
	    if (stickyRetries <= 0) stickyTag = NULL;
	    else stickyRetries--;
	    releaseConn = 0;
	    edir_refcount_decrement(handles->refcnt, server->refIndex, exc);
	    server->useCount--; /* Oops, undo rejected server use count. */
	    if (*exc) {
		edir_exception_t *exc2=NULL;
		edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		edir_exception_free(&exc2);
		EDIR_THROW(*exc);
	    }
	    /* Get new server */
	    refCounts = edir_refcount_values(handles->refcnt, &nRefCounts, exc);
	    if (*exc) {
		edir_exception_t *exc2=NULL;
		edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		edir_exception_free(&exc2);
		EDIR_THROW(*exc);
	    }
	    edir_load_choose(handles->scoreboard, uriData->pathInfo, stickyTag, 
			     &app, &server, refCounts, nRefCounts,
			     &(r->connection->local_addr),
			     &(r->connection->remote_addr),
			     r->server->server_hostname, exc);
	    if (*exc) {
		edir_exception_t *exc2=NULL;
		edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		edir_exception_free(&exc2);
		EDIR_THROW(*exc);
	    }
	    EDIR_DEBUG_NORMAL(("Got alternate server: %s:%d", \
			     server->hostName, server->port));
	    server->useCount++; /* Now, up the use count for new server */
	    path_info = uriData->pathInfo + app->appPrefixLen;
	    releaseConn = 1;
	    edir_refcount_increment(handles->refcnt, server->refIndex, exc);
	    if (*exc) {
		edir_exception_t *exc2=NULL;
		edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		edir_exception_free(&exc2);
		EDIR_THROW(*exc);
	    }

	    /* Release mutex so we don't hold it while waiting for conn */
            edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
            if (*exc) EDIR_THROW(*exc);
/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
            time(&now);
            /* Avoid spinning wildly on errors */
            if (EDIR_TIMEDIFF(now, lastTime) < 1) {
                sleep(1);
                time(&now);
            }
            lastTime = now;
            elapsedTime = EDIR_TIMEDIFF(now, startTime);
        } /* While not timed out */

	/*
	 * If we couldn't get a connection in time, just fail.
	 */
        if (*exc) EDIR_THROW(*exc);

	/*
	 * Check for case of timing out after freeing exception.
	 * XXX There's got to be a better way.
	 */
	if (conn->state != EDIR_STATE_OPEN) {
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR, 
		    "Can't connect to any Enhydra server for %d %s",
		    EDIR_CONNECT_TIMEOUT, "seconds -- giving up.");
	    EDIR_THROW(*exc);
	}

	/*
	 * Use Apache's built-in functions for obtaining the
	 * standard request headers.  This gathers the same
	 * headers as in a CGI program under Apache.
	 * 
	 * XXX - Probably want to add a header packet type for 
	 * numbered standard headers, and to pare down the list
	 * of headers sent.  Eventually, with connection pooling,
	 * we can cache immutable headers.
	 */
	if (*path_info || uriData->pathParams) {
	    /* this will always begin with a slash due to matching rules. */
	    if (uriData->pathParams) {
		r->path_info = ap_pstrcat(r->pool, path_info, ";",
					uriData->pathParams, NULL);
	    } else {
		r->path_info = ap_pstrdup(r->pool, path_info);
	    }
	}
	cgi_vars = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	http_vars = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	edir_apache_get_request_headers(r, cgi_vars, http_vars);
	if (uriData->pathParams) {
	    edir_kvtable_add(cgi_vars, "REQUEST_URI",
		    ap_pstrcat(r->pool, "/", uriData->pathInfo, ";",
				uriData->pathParams, NULL));
	} else {
	    edir_kvtable_add(cgi_vars, "REQUEST_URI",
		    ap_pstrcat(r->pool, "/", uriData->pathInfo, NULL));
	}

	/*
	 * Send CGI headers.
	 */
	vals = edir_kvtable_first(cgi_vars, &key, &nvals);
	while (vals != NULL) {
	    for (i=0; i<nvals; i++) {
		val = vals[i];
		edir_client_send_header(conn, key, strlen(key),
		    val ? val : "", val ? strlen(val) : 0,
		    EDIR_PKT_SUBTYPE_HEADER_HEADER,
		    outBuf, outBufSize, inBuf, inBufSize, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(cgi_vars, &key, &nvals);
	}

	/*
	 * Send HTTP request headers.
	 */
	vals = edir_kvtable_first(http_vars, &key, &nvals);
	while (vals != NULL) {
	    for (i=0; i<nvals; i++) {
		val = vals[i];
		edir_client_send_header(conn, key, strlen(key),
		    val ? val : "", val ? strlen(val) : 0,
		    EDIR_PKT_SUBTYPE_HEADER_HTTP,
		    outBuf, outBufSize, inBuf, inBufSize, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(http_vars, &key, &nvals);
	}

	/*
	 * Send cookies.
	 */
	vals = edir_kvtable_first(uriData->cookies, &key, &nvals);
	while (vals != NULL) {
	    for (i=0; i<nvals; i++) {
		val = vals[i];
		edir_client_send_header(conn, key, strlen(key),
		    val ? val : "", val ? strlen(val) : 0,
		    EDIR_PKT_SUBTYPE_HEADER_COOKIE,
		    outBuf, outBufSize, inBuf, inBufSize, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(uriData->cookies, &key, &nvals);
	}

	/* Indicate end of headers to the server */
	edir_client_send_header(conn, "EDIR_CLIENT", sizeof("EDIR_CLIENT") -1, 
			      "Apache", sizeof("Apache") - 1,
			      EDIR_PKT_SUBTYPE_HEADER_LAST,
			      outBuf, outBufSize, inBuf, inBufSize, exc);
	if (*exc) EDIR_THROW(*exc);

	edir_kvtable_destroy(cgi_vars);
	cgi_vars = NULL;
	edir_kvtable_destroy(http_vars);
	http_vars = NULL;
	edir_kvtable_destroy(uriData->cookies);
	uriData->cookies = NULL;

	/*
	 * If we have a request entity, send it to the server now.
	 * Since we support "chunked" encoding, we don't rely on
	 * Content-Length.  Apache conveniently handles finding the
	 * end of the content for us.
	 */
	if (ap_should_client_block(r)) {
	    while ((n = ap_get_client_block(r,(char*)outBuf,outBufSize))>0) {
		edir_client_send_data(conn, n, 0, outBuf, outBufSize,
				    inBuf, inBufSize, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    if (n < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
					     "read error");
		EDIR_THROW(*exc);
	    }
	}
	/* Send "no more data" indication */
	edir_client_send_data(conn, 0, 1, outBuf, outBufSize,
			      inBuf, inBufSize, exc);

	/*
	 * XXX - sending last request data automatically causes state machine
	 * to receive first response header.  FIX ME  This should be
	 * split up.  Especially if we want to split up the send
	 * and receive portions of the state machine to allow true
	 * full duplexing.
	 * 
	 * Also note that we don't check for write errors here.
	 * Since we can't report errors to the client at this point,
	 * it's not that important.  The ap_rwrite() function will
	 * complain to the logs if an error occurs.
	 */
	sbuf = edir_sbuf_create();
	cookies = ap_make_table(r->pool, 10);
	merge = ap_make_table(r->pool, 10);
	while (conn->state == EDIR_STATE_RECEIVE_HEADERS) {
            if (conn->packetSubType == EDIR_PKT_SUBTYPE_HEADER_COOKIE) {
		int len;
		edir_sbuf_clear(sbuf);
		edir_sbuf_bincatn(sbuf, conn->headerName,
                                  conn->headerNameLen);
		edir_sbuf_bincatn(sbuf, "=", 1);
		edir_sbuf_bincatn(sbuf, conn->headerValue,
			          conn->headerValueLen);
		len = edir_sbuf_length(sbuf);
		val = ap_pcalloc(r->pool, len + 1);
		len = edir_sbuf_copy_into(sbuf, val, len);
		val[len] = '\0';
		ap_table_set(cookies, "Set-Cookie", val);
            } else {
		key = ap_pstrndup(r->pool, (char *) conn->headerName,
				  conn->headerNameLen);
		val = ap_pstrndup(r->pool, (char *) conn->headerValue,
				  conn->headerValueLen);
		if (!edir_strcasecmp(key, "Status")) {
		    r->status = cgi_status = atoi(val);
		    r->status_line = val;
		} else if (!edir_strcasecmp(key, "Content-Type")) {
		    val = trim_inplace(val);
		    ap_content_type_tolower(val);
		    r->content_type = val;
		} else if (!edir_strcasecmp(key, "Location")) {
		    ap_table_set(r->headers_out, key, val);
		} else if (!edir_strcasecmp(key, "Content-Length")) {
		    ap_table_set(r->headers_out, key, val);
		} else if (!edir_strcasecmp(key, "Transfer-Encoding")) {
		    ap_table_set(r->headers_out, key, val);
		} else if (!edir_strcasecmp(key, "Last-Modified")) {
		    time_t mtime = ap_parseHTTPdate(val);
		    ap_update_mtime(r, mtime);
		    ap_set_last_modified(r);
		} else if (!edir_strcasecmp(key, "Set-Cookie")) {
		    ap_table_add(cookies, key, val);
		} else if (!edir_strcasecmp(key, "Connection")) {
		    /* Apache does this for us. */ ;
		} else {
		    ap_table_add(merge, key, val);
		}
            }
            edir_client_next_header(conn, outBuf, outBufSize, inBuf,
				    inBufSize, exc);
            if (*exc) EDIR_THROW(*exc);
	}
	if (sbuf) edir_sbuf_free(sbuf);
	sbuf = NULL;
	cond_status = OK;

	ap_overlap_tables(r->err_headers_out, merge, AP_OVERLAP_TABLES_MERGE);
	if (!ap_is_empty_table(cookies)) {
	    ap_table_unset(r->err_headers_out, "Set-Cookie");
	    r->err_headers_out = ap_overlay_tables(r->pool,
		r->err_headers_out, cookies);
	}
	if (cond_status != OK) {
	    *exc = edir_exception_create(EDIR_EXCEPT_GENERAL, 
		    "Response headers not complete.");
	    EDIR_THROW(*exc);
	}
	cond_status = HTTP_INTERNAL_SERVER_ERROR;

	ap_send_http_header(r);

	/* Finally, send the response entity */
	/*
	 * XXX FIX ME (again, sigh)  
	 * State machine goes directly from "last header" to first
	 * response data block.  Should split.
	 */
        while ((conn->state == EDIR_STATE_RECEIVE_RESPONSE) ||
               (conn->state == EDIR_STATE_RECEIVE_RESPONSE_PACKET))
        {
            n = ap_rwrite(conn->receiveData, conn->receiveDataLen, r);
	    if (n < conn->receiveDataLen) break;
            edir_client_next_data(conn, outBuf, outBufSize,
				  inBuf,inBufSize, exc);
            if (*exc) EDIR_THROW(*exc);
        }
	ap_rflush(r);

	/*
	 * Close the appserver connection.
	 */
	if (conn->state != EDIR_STATE_OPEN) {
	    EDIR_DEBUG_NORMAL(("Apache server closed write pipe.  Aborting."));
	    edir_client_abort(conn, outBuf, outBufSize, exc);
	} else {
	    EDIR_DEBUG_VERBOSE(("Request finished.  Closing connection."));
	    edir_client_close(conn, outBuf, outBufSize, exc);
	}
	conn = NULL;
	if (*exc) EDIR_THROW(*exc);
    } else { /* If not fake it - below we fake it */
	edir_kvtable_t *respHeaders;
	char clength[30];
	int32 len;
	if (sbuf) edir_sbuf_free(sbuf);
	sbuf = NULL;
	respHeaders = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	refCounts = edir_refcount_values(handles->refcnt, &nRefCounts, exc);
        sbuf = edir_load_status(handles->scoreboard, r->uri,
			        refCounts, nRefCounts,
				respHeaders, exc);
	if (*exc) EDIR_THROW(*exc);
	val = edir_kvtable_find(respHeaders, "status");
	if (!val) val = "200 OK";
	r->status_line = ap_pstrdup(r->pool, val);
	edir_kvtable_remove(respHeaders, "status");
	val = edir_kvtable_find(respHeaders, "content-type");
	if (!val) val = "text/plain";
	r->content_type = ap_pstrdup(r->pool, val);
	edir_kvtable_remove(respHeaders, "content-type");
	edir_kvtable_remove(respHeaders, "content-length");
	sprintf(clength, "%d", edir_sbuf_length(sbuf));
	ap_table_set(r->headers_out, "Content-Length", clength);
	vals = edir_kvtable_first(respHeaders, &key, &nvals);
	while (vals) {
	    ap_table_set(r->headers_out, key, vals[nvals-1]);
	    vals = edir_kvtable_next(respHeaders, &key, &nvals);
	}
	edir_kvtable_destroy(respHeaders);
	ap_send_http_header(r);
	val = edir_sbuf_tostring(sbuf, &len);
	n = ap_rwrite(val, len, r);
	ap_rflush(r);
	edir_free(val);
	edir_sbuf_free(sbuf);
	sbuf = NULL;
    } /* If not fake it */

    /*
     * Decrement the remote server's reference count.
     */
    if (!do_status && releaseConn && server) {
	edir_mutex_lock_exclusive(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL,
				  exc);
	if (*exc) EDIR_THROW(*exc);
        edir_refcount_decrement(handles->refcnt, server->refIndex, exc);
        releaseConn = 0;
        if (*exc) EDIR_THROW(*exc);
	edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	if (*exc) EDIR_THROW(*exc);
    }

    if (sbuf) edir_sbuf_free(sbuf);

    EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns OK"));
    return OK;

except:
    if (sbuf) edir_sbuf_free(sbuf);
    if (cgi_vars) edir_kvtable_destroy(cgi_vars);
    if (http_vars) edir_kvtable_destroy(http_vars);
    if (uriData) {
	if (uriData->cookies) edir_kvtable_destroy(uriData->cookies);
    }
    if (*exc) {
	char *trace = edir_exception_stack_trace(*exc);
	edir_log_error("Request failed...");
	edir_log_error("%s", trace);
	edir_free(trace);
	edir_exception_free(exc);
    }
    if (conn) {
        edir_client_abort(conn, outBuf, outBufSize, exc);
        if (*exc) edir_exception_free(exc);
        edir_free(conn); /* XXX FIX probably need an edir_conn_free() func... */
    }
    if (releaseConn && server && handles) {
        edir_refcount_decrement(handles->refcnt, server->refIndex, exc);
        if (*exc) edir_exception_free(exc);
    }
    EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns %d", \
			 cond_status));
    return cond_status;
}

/**
 * Apache module "translate" function for Enhydra Director.
 * <P>
 * This function has the task of mapping input URI's into either
 * filenames or handlers.  In this case, we map input URI's that match
 * our servlet prefixes with the "enhydra_director" handler.  This allows
 * us to configure servlet "mounts" without direct <Location> or 
 * <Directory> sections in the httpd.conf file.  Instead, we read
 * servlet/app config from enhydra_director.conf or (future) from a config 
 * server.
 * 
 * @param r [in,out] Apache Module API request structure.
 * @return Integer status code, including "OK", meaning "this call completed
 * 	the translation and do no more", or "DECLINED", meaning that we
 * 	are not a module that is interested in translating this URI, or
 * 	some other HTTP_* code to report an error.
 */
static int
edir_apache_translate(request_rec *r)
{
    static char			*funcName = "edir_apache_translate";
    edir_exception_t		*_exc=NULL, **exc=&_exc;
    edir_load_sb_app_t		*app;
    edir_apache_cfg_t		*cfg;

    edir_apache_log_set_request_rec(r);
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR")", funcName, r));

    if ((r == NULL) || (r->server == NULL)) {
	EDIR_DEBUG_VERBOSE(("%s: NULL 'r' or 'r->server'", funcName));
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
	return DECLINED;
    }

    cfg = SERVER_CONFIG(r->server);
    if (cfg == NULL) {
	EDIR_DEBUG_VERBOSE(("%s: NULL 'cfg'", funcName));
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
	return DECLINED;
    }

    /*
     * We only handle requests with nonempty URL prefixes.
     */
    if (r->uri == NULL) {
	EDIR_DEBUG_VERBOSE(("%s: No r->uri prefix", funcName));
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
	return DECLINED;
    }

    /*
     * If we haven't attached to the scoreboard yet, do so now,
     * and check for dead scoreboard.
     * However, don't check for dead on every single call to
     * this function.  Otherwise, _every_ request to this http
     * server pays for a trip to the kernel for a nonblocking
     * fcntl() call.  We only want to do that for requests that
     * this module actually ends up handling.
     */
    if (cfg->handles == NULL) {
	/* Initialize the scoreboard */
	char *cfgFile = cfg->configFile;
	if (!cfgFile) cfgFile = DefaultConfigFile;
	cfg->handles = edir_load_init(cfgFile, cfg->pathToDaemon,
				    cfg->lockFile, cfg->dataFile, exc);
	if (*exc) EDIR_THROW(*exc);
	/* Make sure we didn't open an abandoned scoreboard */
	cfg->handles = edir_load_check(cfg->handles, exc);
	if (*exc) EDIR_THROW(*exc);
    }

    /*
     * See if there is an application configured to handle the URI
     * for this request.  Real and virtual HTTP server restrictions in the
     * config are accounted for as well.
     */ 
    if ((app = edir_load_check_path(cfg->handles->scoreboard, r->uri,
		&(r->connection->local_addr), 
		&(r->connection->remote_addr), 
		r->server->server_hostname)) != NULL) {
	r->handler = ap_pstrdup(r->pool, "enhydra_director");
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns OK", funcName));
	return OK;
    }

    /*
     * See if this is a request to a 'status' URL.
     */ 
    if (edir_load_status_check_path(cfg->handles->scoreboard, r->uri,
		&(r->connection->local_addr), 
		&(r->connection->remote_addr), 
		r->server->server_hostname)) {
	r->handler = ap_pstrdup(r->pool, "enhydra_director");
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns OK", funcName));
	return OK;
    }

    /* No apps found, so assume no responsibility for this request. */
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
    return DECLINED;

except:
    /* Don't accept handling for any requests if we can't really do it */
    edir_log_error("Error initializing scoreboard daemon.");
    if (*exc) {
	char *trace;
	trace = edir_exception_stack_trace(*exc);
	edir_log_error(trace);
	edir_free(trace);
	edir_exception_free(exc);
    }
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
    return DECLINED;
}

/**
 * Module 'master server' init function.
 * <P>
 * This function is called by Apache in BOTH the (potentially) root-
 * owned 'master' process, and in EACH 'child' process, running as
 * the 'child' user.  The code in this function does anything that
 * may potentially need root privilege, such as creating the 
 * '.../logs/edir' directory in which to store the scoreboard shared
 * memory and lock files.
 * 
 * @param s [in] Server info block for the current (possibly virtual) server.
 * @param p [in] Apache resource pool to use for temporary memory.
 */
static void
edir_apache_master_init(server_rec *s, pool *p)
{
    edir_apache_cfg_t	*cfg = NULL;
    char		*dataDir, *cp;

    ap_add_version_component("Enhydra-Director/4");

    cfg = SERVER_CONFIG(s);

    /*
     * Figure out default data directory from the error log filename.
     */
    if (s->error_fname != NULL) {
	/* Preferred choice is dirname(errorlog) */
	dataDir = ap_server_root_relative(p, s->error_fname);
	dataDir = ap_pstrdup(p, dataDir);
	cp = strrchr(dataDir, '/');
	if (cp) {
	    *cp = '\0';
	} else {
	    dataDir = ap_server_root_relative(p, "logs");
	    dataDir = ap_pstrdup(p, dataDir);
	    if ((cfg->lockFile == NULL) && (cfg->dataFile == NULL)) {
		mkdir(dataDir, 0755);
	    }
	}
    } else {
	/* If we don't know error log, use reasonable default */
	dataDir = ap_server_root_relative(p, "logs");
	dataDir = ap_pstrdup(p, dataDir);
	if ((cfg->lockFile == NULL) && (cfg->dataFile == NULL)) {
	    mkdir(dataDir, 0755);
	}
    }
    dataDir = ap_pstrcat(p, dataDir, "/edir", NULL);

    /*
     * If neithr cfg->lockFile nor cfg->dataFile are set,
     * try to create the default directory such that it
     * is owned by the UID of the child and is writeable by the
     * child.  Ignore any failures and just let the child
     * complain about it.
     */
    if ((cfg->lockFile == NULL) && (cfg->dataFile == NULL)) {
	mkdir(dataDir, 0755);
	chown(dataDir, s->server_uid, s->server_gid);
    }

    /*
     * Make sure we run as the same user as the child does.
     */
    edir_load_daemon_set_uid(s->server_uid);
    edir_load_daemon_set_gid(s->server_gid);
}

/**
 * Module initializer callback.
 * <P>
 * This callback gets called to allow the Enhydra Director module to initialize
 * itself for the first time.  We set the log to use the server_rec, and
 * create a module-specific config info block and save it with the server_rec.
 * 
 * @param s [in] Server info block for the current (possibly virtual) server.
 * @param p [in] Apache resource pool to use for temporary memory.
 */
static void
edir_apache_init(server_rec *s, pool *p)
{
    static char		*funcName = "edir_apache_init";
    edir_exception_t	*_exc=NULL, **exc=&_exc;
    edir_apache_cfg_t	*cfg = NULL;
    char		*confDir = NULL, *confFile = NULL;
    char		*dataDir = NULL, *dataFile = NULL, *lockFile = NULL;
    char		*pathVar = NULL, *binDir = NULL, *libExecDir = NULL;
    char		*daemonPath = NULL, *cp;

    edir_apache_log_set_server_rec(s);

    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, s, p));

    cfg = SERVER_CONFIG(s);
    
    if (cfg && cfg->inited) {
	EDIR_DEBUG_VERBOSE(("%s: already initialized", funcName));
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
	return;
    }

    if (edir_apache_pool == NULL) {
	edir_apache_pool = ap_make_sub_pool(NULL); 	
    }
    edir_apache_pool_current = edir_apache_pool;

    /*
     * Figure out default data directory from the error log filename.
     */
    if (s->error_fname != NULL) {
	/* Preferred choice is dirname(errorlog) */
	dataDir = ap_server_root_relative(edir_apache_pool, s->error_fname);
	dataDir = ap_pstrdup(p, dataDir);
	cp = strrchr(dataDir, '/');
	if (cp) *cp = '\0';
	else dataDir = ap_server_root_relative(edir_apache_pool, "logs");
    } else {
	/* If we don't know error log, use reasonable default */
	dataDir = ap_server_root_relative(edir_apache_pool, "logs");
	dataDir = ap_pstrdup(p, dataDir);
    }
    dataDir = ap_pstrcat(edir_apache_pool, dataDir, "/edir", NULL);

    /*
     * If neithr cfg->lockFile nor cfg->dataFile are set,
     * try to create the default directory such that it
     * is owned by the UID of the child and is writeable by the
     * child.  Ignore any failures and just let the child
     * complain about it.
     */
    if ((cfg->lockFile == NULL) && (cfg->dataFile == NULL)) {
	if (mkdir(dataDir, 0755) < 0) {
	    EDIR_DEBUG_EVERYTHING(("Can't create data dir: @errstr@"));
	}
	if (chown(dataDir, s->server_uid, s->server_gid) < 0) {
	    EDIR_DEBUG_EVERYTHING(("%s %s  to Uid=%d, Gid=%d: @errstr@", \
				   "Can't chown", dataDir, \
				    s->server_uid, s->server_gid));
	}
    }

    /*
     * Make sure we run as the same user as the child does.
     */
    edir_load_daemon_set_uid(s->server_uid);
    edir_load_daemon_set_gid(s->server_gid);

    /*
     * Generate default datafile and lockfile names.
     * Note - Lock file is named ".ipc", which is sort of generic.
     * This is because the lock file may also get used by the
     * shmem routines if both mmap() shmem and fcntl() locks are used.
     * If the data file is separate from the lock file it is called '.dat'
     * and if it is a SysV shmem segment, it has a shmem key, not a file.
     */
    lockFile = ap_pstrcat(edir_apache_pool, dataDir, "/enhydra_director.ipc", NULL);
    dataFile = ap_pstrcat(edir_apache_pool, dataDir, "/enhydra_director.dat", NULL);

    /*
     * Now, try to use 'srm_confname' or 'access_confname' to get the
     * config files subdirectory.  If neither are set, then just assume
     * 'conf' under the server root.
     */
    if (s->srm_confname) {
	confDir = ap_server_root_relative(edir_apache_pool, s->srm_confname);
	if (strrchr(confDir, '/')) {
	    confDir = ap_pstrdup(p, confDir);
	    *(strrchr(confDir, '/')) = '\0'; ;
	} else {
	    confDir = ap_server_root_relative(edir_apache_pool, "conf");
	    confDir = ap_pstrdup(p, confDir);
	}
    } else if (s->access_confname) {
	confDir = ap_server_root_relative(edir_apache_pool, s->access_confname);
	if (strrchr(confDir, '/')) {
	    confDir = ap_pstrdup(p, confDir);
	    *(strrchr(confDir, '/')) = '\0'; ;
	} else {
	    confDir = ap_server_root_relative(edir_apache_pool, "conf");
	    confDir = ap_pstrdup(p, confDir);
	}
    } else {
	/* If we don't know error log, use reasonable default */
#ifdef EDIR_AC_CONFIG_DIR
	confDir = ap_pstrdup(p, EDIR_AC_CONFIG_DIR);
#else
	confDir = ap_server_root_relative(edir_apache_pool, "conf");
	confDir = ap_pstrdup(p, confDir);
#endif
    }

    /*
     * Generate default config file name.
     */
    confFile = ap_pstrcat(edir_apache_pool, confDir, "/enhydra_director.conf", NULL);

    /*
     * Finally, find the 'libexec' directory where we should normally
     * find the 'edir_daemon' program.
     */
    pathVar = getenv("PATH");
    binDir = ap_server_root_relative(edir_apache_pool, "bin");
    libExecDir = ap_server_root_relative(edir_apache_pool, "libexec");
#ifdef EDIR_AC_DAEMON_DIR
    daemonPath = ap_pstrcat(p, binDir, ":", libExecDir, 
		       pathVar ? ":" : "", pathVar ? pathVar : "", 
		       ":", EDIR_AC_DAEMON_DIR, NULL);
#else
    daemonPath = ap_pstrcat(p, binDir, ":", libExecDir, 
		       pathVar ? ":" : "", pathVar ? pathVar : "", NULL);
#endif

    /*
     * Initialize module-specific defaults for the mutex, shmem, and
     * config implementations we linked with.
     */
    edir_mutex_set_default_filename(lockFile);
    edir_shmem_set_default_filename(dataFile);
    DefaultConfigFile = ap_pstrdup(p, confFile);
    edir_load_daemon_set_default_path(daemonPath);

    /*
     * This should be the first attempt at setting up the
     * scoreboard.
     */
    if (cfg->handles == NULL) {
	/* Initialize the scoreboard */
	char *cfgFile = cfg->configFile;
	if (!cfgFile) cfgFile = DefaultConfigFile;
	cfg->handles = edir_load_init(cfgFile, cfg->pathToDaemon,
				    cfg->lockFile, cfg->dataFile, exc);
	if (*exc) EDIR_THROW(*exc);
	/* Make sure we didn't open an abandoned scoreboard */
	cfg->handles = edir_load_check(cfg->handles, exc);
	if (*exc) EDIR_THROW(*exc);
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;

except:
    edir_log_error("Error initializing scoreboard daemon.");
    if (*exc) {
	char *trace;
	trace = edir_exception_stack_trace(*exc);
	edir_log_error(trace);
	edir_free(trace);
	edir_exception_free(exc);
    }
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void (exc)", funcName));
    return;
}


/**
 * Process exit callback.
 * <P>
 * Currently we don't do anything here.  It might be nice to try to
 * keep track of the current connection and try to be "nice" if we are
 * killed in the middle, but right now we're mean and nasty and just
 * let the Enhydra server get a plain old IO exception to signal that
 * the connection dropped.
 * 
 * @param s [in] Server info block for the current (possibly virtual) server.
 * @param p [in] Apache resource pool to use for temporary memory.
 */
static void edir_apache_child_exit(server_rec *s, pool *p)
{
    edir_apache_log_set_server_rec(s);
    EDIR_DEBUG_FUNCS_IN(("func: edir_apache_child_exit("EDIR_PTR", "EDIR_PTR")", s, p));
    EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_child_exit() returns void"));
}

/**
 * Get the original URI for the request.
 * <P>
 * Borrowed from the apache util_script.c
 * 
 * @param r [in,out] Request handle.
 * @return The original URI, allocated from the request pool.
 */
static char *
original_uri(request_rec *r)
{
    char *first, *last;

    if (r->the_request == NULL) {
	return (char *) ap_pcalloc(r->pool, 1);
    }   

    first = r->the_request;     /* use the request-line */

    while (*first && !ap_isspace(*first)) {
	++first;                /* skip over the method */
    }   
    while (ap_isspace(*first)) {
	++first;                /*   and the space(s)   */
    }   

    last = first;
    while (*last && !ap_isspace(*last)) {
	++last;                 /* end at next whitespace */
    }   

    return ap_pstrndup(r->pool, first, last - first);
}   


/**
 * Gather up the request headers to send to the Enhydra server.
 * 
 * @param r          [in]     The request handle to obtain the header info from.
 * @param cgiHeaders [in,out] Key-value table for the CGI headers.
 * @param cgiHeaders [in,out] Key-value table for the HTTP headers.
 */
static void
edir_apache_get_request_headers(request_rec *r,
                                edir_kvtable_t *cgiHeaders,
                                edir_kvtable_t *httpHeaders)
{
    server_rec *s = r->server;
    conn_rec *c = r->connection;
    char *rem_logname;
    char *host, *https;
    char local_port[32], remote_port[32];
    array_header *hdrs_arr = ap_table_elts(r->headers_in);
    table_entry *hdrs = (table_entry *) hdrs_arr->elts;
    int i;
    
    sprintf(local_port, "%u", ap_get_server_port(r));
    sprintf(remote_port, "%d", ntohs(c->remote_addr.sin_port));

    for (i = 0; i < hdrs_arr->nelts; ++i) {
        if (!hdrs[i].key) {
	    continue;
	}
	edir_kvtable_add(httpHeaders, hdrs[i].key, hdrs[i].val);
    }

    edir_kvtable_add(cgiHeaders, "GATEWAY_INTERFACE", "CGI/1.1");
    edir_kvtable_add(cgiHeaders, "SERVER_PROTOCOL", r->protocol);
    edir_kvtable_add(cgiHeaders, "REQUEST_METHOD", (char *)(r->method));
    if (r->args != NULL) {
	edir_kvtable_add(cgiHeaders, "QUERY_STRING", r->args);
    }
    /* edir_kvtable_add(cgiHeaders, "REQUEST_URI", (char *) original_uri(r)); */
    edir_kvtable_add(cgiHeaders, "SERVER_SOFTWARE",
				 (char *) ap_get_server_version());
    edir_kvtable_add(cgiHeaders, "SERVER_NAME",
				 (char *) ap_get_server_name(r));
    edir_kvtable_add(cgiHeaders, "SERVER_ADDR", r->connection->local_ip);
    edir_kvtable_add(cgiHeaders, "SERVER_PORT", local_port);
    host = (char *) ap_get_remote_host(c, r->per_dir_config, REMOTE_HOST);
    if (host) edir_kvtable_add(cgiHeaders, "REMOTE_HOST", host);
    edir_kvtable_add(cgiHeaders, "REMOTE_ADDR", c->remote_ip);
    edir_kvtable_add(cgiHeaders, "DOCUMENT_ROOT", (char *) ap_document_root(r));
    edir_kvtable_add(cgiHeaders, "SERVER_ADMIN", s->server_admin);
    edir_kvtable_add(cgiHeaders, "SCRIPT_FILENAME", r->filename);
    edir_kvtable_add(cgiHeaders, "REMOTE_PORT", remote_port);
    if (c->user) edir_kvtable_add(cgiHeaders, "REMOTE_USER", c->user);
    if (c->ap_auth_type) {
	edir_kvtable_add(cgiHeaders, "AUTH_TYPE", c->ap_auth_type);
    }
    rem_logname = (char *) ap_get_remote_logname(r);
    if (rem_logname) {
	edir_kvtable_add(cgiHeaders, "REMOTE_IDENT", rem_logname);
    }
 
    if (!strcmp(r->protocol, "INCLUDED")) {
        edir_kvtable_add(cgiHeaders, "SCRIPT_NAME", r->uri);
        if (r->path_info && *r->path_info) {
            edir_kvtable_add(cgiHeaders, "PATH_INFO", r->path_info);
        }
    } else if (!r->path_info || !*r->path_info) {
        edir_kvtable_add(cgiHeaders, "SCRIPT_NAME", r->uri);
    } else {
        int path_info_start = ap_find_path_info(r->uri, r->path_info);
        edir_kvtable_add(cgiHeaders, "SCRIPT_NAME",
                         ap_pstrndup(r->pool, r->uri, path_info_start));
        edir_kvtable_add(cgiHeaders, "PATH_INFO", r->path_info);
    }

    /*
     * See if HTTPS is set
     */
    https = "off";
    if (r->subprocess_env) {
	https = (char *) ap_table_get(r->subprocess_env, "HTTPS");
	if (!https) https = "off";
    }
    if (!edir_strcasecmp(https, "on")) {
	edir_kvtable_add(cgiHeaders, "SERVER_SCHEME", "https");
    } else {
	edir_kvtable_add(cgiHeaders, "SERVER_SCHEME", "http");
    }
    

    if (r->path_info && r->path_info[0]) {
        request_rec *pa_req;
        pa_req = ap_sub_req_lookup_uri(ap_escape_uri(r->pool, r->path_info), r);
        if (pa_req->filename) {
#ifdef WIN32
            char buffer[HUGE_STRING_LEN];
#endif
            char *pt = ap_pstrcat(r->pool, pa_req->filename, pa_req->path_info,
                                  NULL);
#ifdef WIN32
            /* We need to make this a real Windows path name */
            GetFullPathName(pt, HUGE_STRING_LEN, buffer, NULL);
            edir_kvtable_add(cgiHeaders, "PATH_TRANSLATED",
			     ap_pstrdup(r->pool, buffer));
#else
            edir_kvtable_add(cgiHeaders, "PATH_TRANSLATED", pt);
#endif
        }
        ap_destroy_sub_req(pa_req);
    }

}

/* 
 * List of configuration file directives known to Enhydra Director.
 * <P>
 * These directives are only alowed at the top level resource level,
 * and are not allowed within directory configuration.  Enhydra Director
 * dynamically figures out prefix to handler mappings from its own
 * configuration file.
 * 
 * @see edir_apache_translate()
 */
static const command_rec edir_apache_cmds[] =
{
    {
        "EnhydraDirectorInherit", cmd_Inherit, NULL, RSRC_CONF, FLAG,
        "Whether default config is inherited by virtual servers."
    },
    {
        "EnhydraDirectorConfigFile", cmd_ConfigFile, NULL, RSRC_CONF, TAKE1,
        "Full path to the Enhydra Director configuration file"
    },
    {
        "EnhydraDirectorDataFile", cmd_DataFile, NULL, RSRC_CONF, TAKE1,
        "Full path to the Enhydra Director scoreboard file"
    },
    {
        "EnhydraDirectorLockFile", cmd_LockFile, NULL, RSRC_CONF, TAKE1,
        "Full path to the Enhydra Director scoreboard lock file"
    },
    {
        "EnhydraDirectorDaemonPath", cmd_DaemonPath, NULL, RSRC_CONF, TAKE1,
        "Full path to the Enhydra Director daemon executable file"
    },
    {
        "EnhydraDirectorDebug", cmd_Debug, NULL, RSRC_CONF, TAKE1,
        "Turn on debug logging."
    },
    {NULL}
};

/**
 * Lists the content handlers for Enhydra Director.
 * <P>
 * Right now, only the "enhydra_director" handler is known, and it maps
 * to the 'edir_apache_handler()' function above.
 */
static const handler_rec edir_apache_handlers[] =
{
    {"enhydra_director", edir_apache_handler},
    {NULL}
};

/**
 * Enhydra Director Apache module definition block.
 * <P>
 * The number in brackets indicates the order in which the routine is called
 * during request processing.  Note that not all routines are necessarily
 * called (such as if a resource doesn't have access restrictions).
 */
module enhydra_director_module =
{
    STANDARD_MODULE_STUFF,
    edir_apache_master_init,          /* module initializer as server UID. */
    edir_apache_create_dir_config,    /* per-directory config creator */
    NULL,                             /* dir config merger */
    edir_apache_create_server_config, /* server config creator */
    NULL,                             /* server config merger */
    edir_apache_cmds,                 /* command table */
    edir_apache_handlers,             /* [7] list of handlers */
    edir_apache_translate,            /* [2] filename-to-URI translation */
    NULL,                             /* [5] check/validate user_id */
    NULL,                             /* [6] check user_id is valid *here* */
    NULL,                             /* [4] check access by host address */
    NULL,                             /* [7] MIME type checker/setter */
    NULL,                             /* [8] fixups */
    NULL,                             /* [10] logger */
#if MODULE_MAGIC_NUMBER >= 19970103
    NULL,                             /* [3] header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
    edir_apache_init,                 /* process initializer as child UID */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
    edir_apache_child_exit,           /* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
    NULL                              /* [1] post read_request handling */
#endif
};

