/*
 * Enhydra Java Application Server
 * The Initial Developer of the Original Code is Lutris Technologies Inc.
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc.
 * All Rights Reserved.
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License.
 *
 * $Id: mod_enhydra_director.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

/* Ported to Apache/Win32 by rkoma
 */

#include "edir.h"

#include "edir_table.h"
#include "edir_kvtable.h"
#include "edir_apache_log_line.h"
#include "edir_string.h"

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
#include "http_request.h"
#include "util_date.h"

#include "mod_enhydra_director.h"
#include "mod_enhydra_director_cmds.h"


static void
edir_apache_get_request_headers(request_rec *r,
								edir_kvtable_t *cgiHeaders,
								edir_kvtable_t *httpHeaders);


CRITICAL_SECTION vLock;

/**
 * Convenience function to do in-place removal of leading and trailing
 * whitespace characters.
 * <P>
 * Side effects:<BR>
 * This function modifies the string pointed to by "s".
 *
 * @param s [in,out] String to trim in place.
 * @return A pointer to the trimmed portion of the string.
 */
static char *
trim_inplace(char *s)
{
	char *cp;
	EDIR_DEBUG_FUNCS_IN(("func: trim_inplace(\"%s\")", s ? s : "(null)"));
	while (*s && isspace(*s)) s++;
	for (cp=s; *cp; cp++) ;
	for (--cp; (cp >= s) && isspace(*cp); cp--) *cp = '\0';
	EDIR_DEBUG_FUNCS_OUT(("func: trim_inplace() returns \"%d\"",s?s:"(null)"));
	return s;
}

/**
 * Function to handle components of the request URI as it is parsed.
 * <P>
 * The edir_parse_uri() function calls this function for each part
 * of the url that it parses.  The four possible values of the <code>
 * type</code> parameter are described below. <BR>
 * <code>EDIR_PARSE_URI_PATH_INFO</code><BR>
 * The value of the <code>lhs</code> parameter will point to a character
 * string containing the stripped path portion of the URI, with leading
 * and trailing slashes removed, and with interior runs of more than
 * one slash reduced to one slash.
 * <code>EDIR_PARSE_URI_PATH_PARAM_1</code>
 * Returns a pointer to a string in <code>lhs</code> that contains a URL
 * "path parameter" that is not of the form "key=value".  The <code>rhs</code>
 * parameter is always <code>NULL</code>.
 * <code>EDIR_PARSE_URI_PATH_PARAM_2</code>
 * Returns strings comprising "path parameter" in the form of a "key=value"
 * pair, with the key in <code>lhs</code>, and the value in <code>rhs</code>.
 * Neither parameters are ever <code>NULL</code> for this type.
 * <code>EDIR_PARSE_URI_QUERY_STRING</code>
 * Returns a pointer to the beginning of the "query string" in <code>lhs</code>.
 * <P>
 * NOTE: The strings passed to this function are part of a larger
 * string buffer that is being destructively parsed by edir_parse_uri(). They
 * should not be freed, and they should only be stored in conjunction with
 * the larger URI buffer that contains them so the buffer does not
 * get freed while the strings are still in use.
 *
 * @param ptr  [in,out] Arbitrary data structure to store things in.
 * @param type [in]     What type of component has been parsed.
 * @param lhs  [in]     "Left hand side" of "key=value" pair, or full
 * 			String of "value" item.
 * @param rhs  [in]     "Right hand side" of "key=value" pair, or NULL.
 */
static void
edir_apache_uri_parse_func(void *ptr, int type, char *lhs, char *rhs)
{
	edir_uri_data_t *uriData = (edir_uri_data_t *) ptr;
	switch (type) {
	case EDIR_PARSE_URI_PATH_INFO:
		uriData->pathInfo = lhs;
		break;
	case EDIR_PARSE_URI_PATH_PARAMS:
		uriData->pathParams = ap_pstrdup(uriData->pool, lhs);
		break;
	case EDIR_PARSE_URI_PATH_PARAM_1:
		break;
	case EDIR_PARSE_URI_PATH_PARAM_2:
		if (lhs && !strcmp(lhs, "jsessionid")) {
			uriData->sessionID = rhs;
		}
		break;
	case EDIR_PARSE_URI_QUERY_STRING:
		uriData->queryString = lhs;
		break;
	}
}

/**
 * Apache table "do" function to parse a cookie header.
 * <P>
 * This function takes a Cookie header, which is a semicolon-delimited
 * list of cookie key-value pairs.  Each key-value pair is placed
 * separately into the table passed through the "rec" pointer.
 * 
 * @param rec   [in,out] The "rec" pointer passed to ap_table_do().
 * @param key   [in]	 The key found by ap_table_do(). ("Cookie" here).
 * @param value [in]	 The value found by ap_table_do().
 * @return Zero (false) to stop the iteration, Nonzero (true) to keep going.
 */
static int
edir_apache_do_cookies(void *rec, const char *key, const char *value)
{
	static char		*funcName = "edir_apache_do_cookies";
	edir_uri_data_t	*uriData = (edir_uri_data_t *) rec;
	char			*start, *end, *name, *val, *fp = NULL;

	EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", \"%s\", \"%s\")", funcName, \
						rec, key ? key : "(null)", value ? value : "(null)"));
    fp = start = edir_strdup(value);
	end = strchr(start, ';');
	while (end) {
		*end++ = '\0';
		name = start;
		val = strchr(name, '=');
		if (val) *val++ = '\0';
		else val = "";
	name = trim_inplace(name);
	val = trim_inplace(val);
	edir_kvtable_add(uriData->cookies, name, val);
	if ((uriData->sessionID == NULL) &&
		(!edir_strcasecmp(name, "jsessionid"))) {
		uriData->sessionID = (char *) ap_pstrdup(uriData->pool, val);
	}
		start = end;
		end = strchr(start, ';');
	}
	name = start;
	val = strchr(name, '=');
	if (val)
		*val++ = '\0';
	else
		val = "";
	name = trim_inplace(name);
	val = trim_inplace(val);
	edir_kvtable_add(uriData->cookies, name, val);
	if ((uriData->sessionID == NULL) &&
	(!edir_strcasecmp(name, "jsessionid"))) {
		uriData->sessionID = (char *) ap_pstrdup(uriData->pool, val);
	}
	edir_free(fp);
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	return 1;
}

/**
 * Apache table "do" function to break up a comma-delimited header.
 * <P>
 * This function takes a comma-delimited header and splits it
 * into multiple headers with the same name.
 * 
 * @param rec   [in,out] The "rec" pointer passed to ap_table_do().
 * @param key   [in]	 The key found by ap_table_do().
 * @param value [in]	 The value found by ap_table_do().
 * @return Zero (false) to stop the iteration, Nonzero (true) to keep going.
 */
static int
edir_apache_do_commalist(void *rec, const char *key, const char *value)
{
	table	*headers = (table *) rec;
	char	*start, *end, *fp;

	EDIR_DEBUG_FUNCS_IN(("func: edir_apache_do_commalist("EDIR_PTR", \"%s\", \"%s\")",\
						rec, key ? key : "(null)", value ? value : "(null)"));
	fp = start = edir_strdup(value);
	end = strchr(start, ',');
	while (end) {
		*end++ = '\0';
		ap_table_add(headers, key, trim_inplace(start));
		start = end;
		end = strchr(start, ',');
	}
	ap_table_add(headers, key, trim_inplace(start));
	edir_free(fp);
	EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_do_commalist() returns %d", 1));
	return 1;
}

/**
 * Handler function for the Enhydra Director Apache module.
 * <P>
 * If the translate function determines that a particular request is
 * a servlet request, then then request's handler field is set to refer
 * to this function.  This function then does the actual work of processing
 * the request.
 * <P>
 * If it turns out that the translate routine was lying to us, we just
 * return DECLINED and the user sees "not found", which is correct because
 * something is not configured right.  If the request fails for some
 * reason, we return an HTTP_* status as appropriate and log the error.
 * Otherwise, when the request is done, return OK.
 *
 * @param r [in,out] Apache 'request_rec' structure containing information
 * 		about this request.
 * @return Integer Apache status code.  Normally 'OK' if successful.
 */
static int edir_apache_handler(request_rec *r)
{
	static char				*funcName="edir_apache_handler";
	edir_exception_t		*_exc=NULL, **exc=&_exc;
	server_rec				*s = r->server;
	char					*stickyTag = NULL, *path_info = NULL;
	char					*key, *val, **vals;
	edir_kvtable_t			*cgi_vars = NULL, *http_vars=NULL;
	table					*cookies = NULL, *merge = NULL;
	void					*sbuf = NULL;
	edir_apache_cfg_t		*cfg;
	edir_uri_data_t			*uriData = NULL;
	edir_connection_t		*conn = NULL;
	edir_load_sb_app_t		*app;
	edir_load_sb_server_t	*server;
	int32					nvals;
	int						releaseConn, i, n;
	int						stickyRetries = 0;
	time_t					startTime, elapsedTime, lastTime, now;
	int						inBufSize, outBufSize;
	octet					*inBuf, *outBuf;
	int						cond_status = HTTP_INTERNAL_SERVER_ERROR;
	int						cgi_status;
	int						do_status = 0;

	edir_apache_log_set_request_rec(r);
	EDIR_DEBUG_FUNCS_IN(("func: edir_apache_handler("EDIR_PTR")", r));

	/*
	 * Get config and set up buffers
	 */
	cfg = SERVER_CONFIG(s);
	inBufSize = cfg->inBufSize;
	outBufSize = cfg->outBufSize;
	inBuf = ap_pcalloc(r->pool, inBufSize);
	outBuf = ap_pcalloc(r->pool, outBufSize);

	/*
	 * Get apache ready to send us the request body, if applicable.
	 * The mod_enhydra_director module supports the "chunked" encoding.
	 * This function checks Content-Length and Transfer-Encoding to
	 * see if we want to read a request entity and send it to Enhydra.
	 * Note that we still keep the chunks when sending to Enhydra
	 * because the app may assume that it must count the bytes.  This
	 * just lets us know when to stop sending data without having to
	 * parse the stream ourselves.
	 */
	if ((i = ap_setup_client_block(r, REQUEST_CHUNKED_PASS)) != OK) {
		EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns %d", i));
		return i;
	}

	/*
	 * Parse the URI, looking for path parameters in particular.
	 */
//	DebugBreak();
	uriData = (edir_uri_data_t *) ap_pcalloc(r->pool, sizeof(edir_uri_data_t));
	memset((char *)uriData, 0, sizeof(edir_uri_data_t));
	uriData->pool = r->pool;
#ifdef USE_UNPARSED_URI
	uriData->uriBuf = ap_pstrdup(r->pool,
				r->unparsed_uri ? r->unparsed_uri : "");
#else
	uriData->uriBuf = ap_pstrdup(r->pool, r->uri ? r->uri : "");
#endif

	EDIR_DEBUG_EVERYTHING(("%s: r->uri = '%s'", funcName, \
				r->uri?r->uri:"(null)"));
	EDIR_DEBUG_EVERYTHING(("%s: uriBuf = '%s'", funcName, uriData->uriBuf));

	edir_parse_uri(uriData->uriBuf, uriData, edir_apache_uri_parse_func);

	EDIR_DEBUG_EVERYTHING(("%s: PARSED pathinfo = '%s'", funcName, \
				uriData->pathInfo ? uriData->pathInfo : "(null)"));
	EDIR_DEBUG_EVERYTHING(("%s: PARSED queryString = '%s'", funcName, \
				uriData->queryString ? uriData->queryString:"(null)"));

	if (uriData->pathInfo == NULL) {
		EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns DECLINED"));
		return DECLINED;
	}

	/*
	 * Gather the cookies, separate them into separate entries, and
	 * put them in a table.  This is needed so we can look for the
	 * application session cookie for load balancing purposes.
	 */
	uriData->cookies = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	if (r->headers_in) {
		ap_table_do(edir_apache_do_cookies, uriData, r->headers_in,
				"Cookie", NULL);
	}
	ap_table_unset(r->headers_in, "Cookie");

	EDIR_DEBUG_EVERYTHING(("%s: PARSED sessionID = '%s'", funcName, \
			uriData->sessionID ? uriData->sessionID : "(null)"));

	/*
	 * If we have a session ID, look for sticky session tag.
	 */
	if ((uriData->sessionID) && (val = strchr(uriData->sessionID, '('))) {
		if (strlen(val) >= 9) stickyTag =(char *)ap_pstrndup(r->pool,val+1,8);
		if (stickyTag) stickyRetries = EDIR_CONNECT_STICKY_SESSION_RETRIES;
	}

	/* TODO : Proveri da li ovde treba da zakljucavam scoreboard.
U NSAPI modulu se to ne radi. */
	/*
	 * See if we really only want to return scoreboard status
	 */
	if (edir_load_status_check_path(scoreboard, r->uri,
		&(r->connection->local_addr),
		&(r->connection->remote_addr),
		r->server->server_hostname)) {
		do_status = 1;
	}

	if (!do_status) {
/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	/*
	 * Lock the scoreboard briefly while we look for an app matching
	 * our prefix.
	 */
		EnterCriticalSection(&vLock);
//		DebugBreak();
		edir_load_choose(scoreboard, uriData->pathInfo, stickyTag,
			 &app, &server, refCounts, numServers,
			 &(r->connection->local_addr),
			 &(r->connection->remote_addr),
			 r->server->server_hostname, exc);
		if (*exc) {
			LeaveCriticalSection(&vLock);
			EDIR_THROW(*exc);
		}
		EDIR_DEBUG_NORMAL(("Got server: %s:%d", server->hostName, \
			server->port));
		server->useCount++; /* Up the use count on this server. */
		path_info = uriData->pathInfo + app->appPrefixLen;

		/*
		* Up the refcount
		*/
		releaseConn = 1;
		refCounts[server->refIndex]++;

		/*
		* Give up scoreboard lock, now that we have our server.
		*/
		LeaveCriticalSection(&vLock);
	}
/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
	if (!(edir_debug & EDIR_DEBUG_MASK_FAKE_IT) && !do_status) {
		conn = (edir_connection_t *) edir_malloc(sizeof(edir_connection_t));
		memset((void *)conn, 0, sizeof(edir_connection_t));
		time(&startTime);
		lastTime = startTime;
		elapsedTime = 0;
		while (elapsedTime <= EDIR_CONNECT_TIMEOUT) {
			if (*exc) edir_exception_free(exc);
			edir_client_connect(conn, NULL, &(server->sin), server->authData,
				outBuf, outBufSize, inBuf, inBufSize, exc);
			if (!(*exc)) break; /* Connection established! */
			if ((*exc)->id != EDIR_EXCEPT_IO_ERROR) EDIR_THROW(*exc);

			/* XXX Should have 'logmask' */
			if (edir_debug & EDIR_DEBUG_MASK_NORMAL) {
				char *trace = NULL;
				if (*exc) trace = edir_exception_stack_trace(*exc);
				edir_log_error("Warning: Couldn't connect to %s:%d.",
					server->hostName, server->port);
				edir_log_error("%s", trace ? trace : "No details.");
				if (trace) edir_free(trace);
			}
			if (*exc) edir_exception_free(exc);

			/* Mark this server as 'failing' */
/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
            EnterCriticalSection(&vLock);
			if (!(server->failing)) {
				/*
				* XXX FIX  - Idea: Run notification script.
				* Especially on Unix, where it's easy, run a script if we
				* find a server to be 'failing'.  Probably would want to
				* only run the script occasionally to avoid spamming,
				* and would want to run it outside any critical section
				* to avoid blocking for prolonged periods while holding
				* the global scoreboard lock.
				*/
				server->failing = 1;
				time(&(server->failTime));
			}
			server->failCount++;
			/*
			* If this request has a 'sticky session' token in the session
			* cookie, then we want to try several times to go to the desired
			* server.  After a number of attempts, though, we give up and
			* fall back to round robin.
			*/
			if (stickyRetries <= 0) stickyTag = NULL;
			else stickyRetries--;
			releaseConn = 0;
			if (refCounts[server->refIndex] > 0)
			refCounts[server->refIndex]--;
			server->useCount--; /* Oops, undo rejected server use count. */
			EDIR_DEBUG_NORMAL(("Failed to get %sserver: %s:%d\n", (stickyTag) ? "sticky " : "", server->hostName, server->port));
			/* Get new server */
			edir_load_choose(scoreboard, uriData->pathInfo, stickyTag,
					&app, &server, refCounts, numServers,
					&(r->connection->local_addr),
					&(r->connection->remote_addr),
					r->server->server_hostname, exc);
			if (*exc) {
				LeaveCriticalSection(&vLock);
				EDIR_THROW(*exc);
			}
			EDIR_DEBUG_NORMAL(("Got alternate server: %s:%d", \
					server->hostName, server->port));
			server->useCount++; /* Now, up the use count for new server */
			path_info = uriData->pathInfo + app->appPrefixLen;
			releaseConn = 1;

			refCounts[server->refIndex]++;

			/* Release lock so we don't hold it while waiting for conn */
			LeaveCriticalSection(&vLock);
/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
			time(&now);
			/* Avoid spinning wildly on errors */
			if (EDIR_TIMEDIFF(now, lastTime) < 1) {
#ifdef WIN32
				Sleep(1000);
#else
				sleep(1);
#endif
				time(&now);
			}
			lastTime = now;
			elapsedTime = EDIR_TIMEDIFF(now, startTime);
		} /* While not timed out */

		/*
		* If we couldn't get a connection in time, just fail.
		*/
		if (*exc) EDIR_THROW(*exc);

		/*
		* Check for case of timing out after freeing exception.
		* XXX There's got to be a better way.
		*/
		if (conn->state != EDIR_STATE_OPEN) {
			*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR, 
					"Can't connect to any Enhydra server for %d %s",
					EDIR_CONNECT_TIMEOUT, "seconds -- giving up.");
			EDIR_THROW(*exc);
		}

		/*
		* Use Apache's built-in functions for obtaining the
		* standard request headers.  This gathers the same
		* headers as in a CGI program under Apache.
		* 
		* XXX - Probably want to add a header packet type for 
		* numbered standard headers, and to pare down the list
		* of headers sent.  Eventually, with connection pooling,
		* we can cache immutable headers.
		*/
		if (*path_info || uriData->pathParams) {
			/* this will always begin with a slash due to matching rules. */
			if (uriData->pathParams) {
				r->path_info = ap_pstrcat(r->pool, path_info, ";",
						uriData->pathParams, NULL);
			} else {
				r->path_info = ap_pstrdup(r->pool, path_info);
			}
		}
		cgi_vars = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
		http_vars = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
		edir_apache_get_request_headers(r, cgi_vars, http_vars);
		if (uriData->pathParams) {
			edir_kvtable_add(cgi_vars, "REQUEST_URI",
					ap_pstrcat(r->pool, "/", uriData->pathInfo, ";",
					uriData->pathParams, NULL));
		} else {
			edir_kvtable_add(cgi_vars, "REQUEST_URI",
			ap_pstrcat(r->pool, "/", uriData->pathInfo, NULL));
		}

		/*
		* Send CGI headers.
		*/
		vals = edir_kvtable_first(cgi_vars, &key, &nvals);
		while (vals != NULL) {
			for (i=0; i<nvals; i++) {
				val = vals[i];
				edir_client_send_header(conn, key, strlen(key),
				val ? val : "", val ? strlen(val) : 0,
				EDIR_PKT_SUBTYPE_HEADER_HEADER,
				outBuf, outBufSize, inBuf, inBufSize, exc);
				if (*exc) EDIR_THROW(*exc);
			}
			vals = edir_kvtable_next(cgi_vars, &key, &nvals);
		}

		/*
		* Send HTTP request headers.
		*/
		vals = edir_kvtable_first(http_vars, &key, &nvals);
		while (vals != NULL) {
			for (i=0; i<nvals; i++) {
				val = vals[i];
				edir_client_send_header(conn, key, strlen(key),
				val ? val : "", val ? strlen(val) : 0,
				EDIR_PKT_SUBTYPE_HEADER_HTTP,
				outBuf, outBufSize, inBuf, inBufSize, exc);
			if (*exc) EDIR_THROW(*exc);
			}
			vals = edir_kvtable_next(http_vars, &key, &nvals);
		}

		/*
		* Send cookies.
		*/
		vals = edir_kvtable_first(uriData->cookies, &key, &nvals);
		while (vals != NULL) {
			for (i=0; i<nvals; i++) {
				val = vals[i];
				edir_client_send_header(conn, key, strlen(key),
				val ? val : "", val ? strlen(val) : 0,
				EDIR_PKT_SUBTYPE_HEADER_COOKIE,
				outBuf, outBufSize, inBuf, inBufSize, exc);
				if (*exc) EDIR_THROW(*exc);
			}
			vals = edir_kvtable_next(uriData->cookies, &key, &nvals);
		}

		/* Indicate end of headers to the server */
		edir_client_send_header(conn, "EDIR_CLIENT", sizeof("EDIR_CLIENT") -1,
						"Apache", sizeof("Apache") - 1,
						EDIR_PKT_SUBTYPE_HEADER_LAST,
						outBuf, outBufSize, inBuf, inBufSize, exc);
		if (*exc) EDIR_THROW(*exc);

		edir_kvtable_destroy(cgi_vars);
		cgi_vars = NULL;
		edir_kvtable_destroy(http_vars);
		http_vars = NULL;
		edir_kvtable_destroy(uriData->cookies);
		uriData->cookies = NULL;

		/*
		* If we have a request entity, send it to the server now.
		* Since we support "chunked" encoding, we don't rely on
		* Content-Length.  Apache conveniently handles finding the
		* end of the content for us.
		*/
		if (ap_should_client_block(r)) {
			while ((n = ap_get_client_block(r,(char*)outBuf,outBufSize))>0) {
				edir_client_send_data(conn, n, 0, outBuf, outBufSize,
						inBuf, inBufSize, exc);
				if (*exc) EDIR_THROW(*exc);
			}
			if (n < 0) {
				*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
							"read error");
				EDIR_THROW(*exc);
			}
		}
		/* Send "no more data" indication */
		edir_client_send_data(conn, 0, 1, outBuf, outBufSize,
				inBuf, inBufSize, exc);

		/*
		* XXX - sending last request data automatically causes state machine
		* to receive first response header.  FIX ME  This should be
		* split up.  Especially if we want to split up the send
		* and receive portions of the state machine to allow true
		* full duplexing.
		*
		* Also note that we don't check for write errors here.
		* Since we can't report errors to the client at this point,
		* it's not that important.  The ap_rwrite() function will
		* complain to the logs if an error occurs.
		*/
		sbuf = edir_sbuf_create();
		cookies = ap_make_table(r->pool, 10);
		merge = ap_make_table(r->pool, 10);
		while (conn->state == EDIR_STATE_RECEIVE_HEADERS) {
			if (conn->packetSubType == EDIR_PKT_SUBTYPE_HEADER_COOKIE) {
				int len;
				edir_sbuf_clear(sbuf);
				edir_sbuf_bincatn(sbuf, conn->headerName,
										conn->headerNameLen);
				edir_sbuf_bincatn(sbuf, "=", 1);
				edir_sbuf_bincatn(sbuf, conn->headerValue,
										conn->headerValueLen);
				len = edir_sbuf_length(sbuf);
				val = ap_pcalloc(r->pool, len + 1);
				len = edir_sbuf_copy_into(sbuf, val, len);
				val[len] = '\0';
				ap_table_set(cookies, "Set-Cookie", val);
			} else {
				key = ap_pstrndup(r->pool, (char *) conn->headerName,
						conn->headerNameLen);
				val = ap_pstrndup(r->pool, (char *) conn->headerValue,
						conn->headerValueLen);
				if (!edir_strcasecmp(key, "Status")) {
					r->status = cgi_status = atoi(val);
					r->status_line = val;
				} else if (!edir_strcasecmp(key, "Content-Type")) {
					val = trim_inplace(val);
					ap_content_type_tolower(val);
					r->content_type = val;
				} else if (!edir_strcasecmp(key, "Location")) {
					ap_table_set(r->headers_out, key, val);
				} else if (!edir_strcasecmp(key, "Content-Length")) {
					ap_table_set(r->headers_out, key, val);
				} else if (!edir_strcasecmp(key, "Transfer-Encoding")) {
					ap_table_set(r->headers_out, key, val);
				} else if (!edir_strcasecmp(key, "Last-Modified")) {
					time_t mtime = ap_parseHTTPdate(val);
					ap_update_mtime(r, mtime);
					ap_set_last_modified(r);
				} else if (!edir_strcasecmp(key, "Set-Cookie")) {
					ap_table_add(cookies, key, val);
				} else if (!edir_strcasecmp(key, "Connection")) {
					/* Apache does this for us. */ ;
				} else {
					ap_table_add(merge, key, val);
				}
			}
			edir_client_next_header(conn, outBuf, outBufSize, inBuf,
					inBufSize, exc);
			if (*exc) EDIR_THROW(*exc);
		}
		if (sbuf) edir_sbuf_free(sbuf);
		sbuf = NULL;
		cond_status = OK;

		ap_overlap_tables(r->err_headers_out, merge, AP_OVERLAP_TABLES_MERGE);
		if (!ap_is_empty_table(cookies)) {
			ap_table_unset(r->err_headers_out, "Set-Cookie");
			r->err_headers_out = ap_overlay_tables(r->pool,
			r->err_headers_out, cookies);
		}
		if (cond_status != OK) {
			*exc = edir_exception_create(EDIR_EXCEPT_GENERAL, 
				"Response headers not complete.");
			EDIR_THROW(*exc);
		}
		cond_status = HTTP_INTERNAL_SERVER_ERROR;

		ap_send_http_header(r);

		/* Finally, send the response entity */
		/*
		* XXX FIX ME (again, sigh)  
		* State machine goes directly from "last header" to first
		* response data block.  Should split.
		*/
		while ((conn->state == EDIR_STATE_RECEIVE_RESPONSE) ||
			   (conn->state == EDIR_STATE_RECEIVE_RESPONSE_PACKET))
		{
			n = ap_rwrite(conn->receiveData, conn->receiveDataLen, r);
			if (n < conn->receiveDataLen) break;
			edir_client_next_data(conn, outBuf, outBufSize,
					inBuf,inBufSize, exc);
			if (*exc) EDIR_THROW(*exc);
		}
		ap_rflush(r);

		/*
		 * Close the appserver connection.
		*/
		if (conn->state != EDIR_STATE_OPEN) {
			EDIR_DEBUG_NORMAL(("Apache server closed write pipe.  Aborting."));
			edir_client_abort(conn, outBuf, outBufSize, exc);
		} else {
			EDIR_DEBUG_VERBOSE(("Request finished.  Closing connection."));
			edir_client_close(conn, outBuf, outBufSize, exc);
		}
		conn = NULL;
		if (*exc) EDIR_THROW(*exc);
	} else { /* If not fake it - below we fake it */
		edir_kvtable_t *respHeaders;
		char clength[30];
		int32 len;
		if (sbuf) edir_sbuf_free(sbuf);
		sbuf = NULL;
		respHeaders = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
		sbuf = edir_load_status(scoreboard, r->uri,
					refCounts, numServers,
					respHeaders, exc);
		if (*exc) EDIR_THROW(*exc);

		val = edir_kvtable_find(respHeaders, "status");
		if (!val) val = "200 OK";
		r->status_line = ap_pstrdup(r->pool, val);
		edir_kvtable_remove(respHeaders, "status");

		val = edir_kvtable_find(respHeaders, "content-type");
		if (!val) val = "text/plain";
		r->content_type = ap_pstrdup(r->pool, val);
		edir_kvtable_remove(respHeaders, "content-type");
		edir_kvtable_remove(respHeaders, "content-length");
		sprintf(clength, "%d", edir_sbuf_length(sbuf));
		ap_table_set(r->headers_out, "Content-Length", clength);

		vals = edir_kvtable_first(respHeaders, &key, &nvals);
		while (vals) {
			ap_table_set(r->headers_out, key, vals[nvals-1]);
			vals = edir_kvtable_next(respHeaders, &key, &nvals);
		}
		edir_kvtable_destroy(respHeaders);
		ap_send_http_header(r);
		val = edir_sbuf_tostring(sbuf, &len);
		n = ap_rwrite(val, len, r);
		ap_rflush(r);
		edir_free(val);
		edir_sbuf_free(sbuf);
		sbuf = NULL;
	} /* If not fake it */

	/*
	 * Decrement the remote server's reference count.
	 */
	if (!do_status && releaseConn && server) {
		EnterCriticalSection(&vLock);
		if (refCounts[server->refIndex] > 0)
			refCounts[server->refIndex]--;
			releaseConn = 0;
			LeaveCriticalSection(&vLock);
	}

	if (sbuf) edir_sbuf_free(sbuf);

	EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns OK"));
	return OK;

except:
	if (sbuf) edir_sbuf_free(sbuf);
	if (cgi_vars) edir_kvtable_destroy(cgi_vars);
	if (http_vars) edir_kvtable_destroy(http_vars);
	if (uriData) {
		if (uriData->cookies) edir_kvtable_destroy(uriData->cookies);
	}
	if (*exc) {
		char *trace = edir_exception_stack_trace(*exc);
		edir_log_error("Request failed...");
		edir_log_error("%s", trace);
		edir_free(trace);
		edir_exception_free(exc);
	}
	if (conn) {
		edir_client_abort(conn, outBuf, outBufSize, exc);
		if (*exc) edir_exception_free(exc);
		edir_free(conn); /* XXX FIX probably need an edir_conn_free() func... */
	}
	if (releaseConn && server) {
		EnterCriticalSection(&vLock);
		if (refCounts[server->refIndex] > 0)
		refCounts[server->refIndex]--;
		releaseConn = 0;
		LeaveCriticalSection(&vLock);
	}
	EDIR_DEBUG_FUNCS_OUT(("func: edir_apache_handler() returns %d", \
				 cond_status));
	return cond_status;
}

/**
 * Apache module "translate" function for Enhydra Director.
 * <P>
 * This function has the task of mapping input URI's into either
 * filenames or handlers.  In this case, we map input URI's that match
 * our servlet prefixes with the "enhydra_director" handler.  This allows
 * us to configure servlet "mounts" without direct <Location> or
 * <Directory> sections in the httpd.conf file.  Instead, we read
 * servlet/app config from enhydra_director.conf or (future) from a config
 * server.
 *
 * @param r [in,out] Apache Module API request structure.
 * @return Integer status code, including "OK", meaning "this call completed
 * 	the translation and do no more", or "DECLINED", meaning that we
 * 	are not a module that is interested in translating this URI, or
 * 	some other HTTP_* code to report an error.
 */
static int
edir_apache_translate(request_rec *r)
{
	static char				*funcName = "edir_apache_translate";
	edir_exception_t		*_exc=NULL, **exc=&_exc;
	edir_load_sb_app_t		*app;
	edir_apache_cfg_t		*cfg;

	edir_apache_log_set_request_rec(r);
	EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR")", funcName, r));

	if ((r == NULL) || (r->server == NULL)) {
		EDIR_DEBUG_VERBOSE(("%s: NULL 'r' or 'r->server'", funcName));
		EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
		return DECLINED;
	}

	cfg = SERVER_CONFIG(r->server);
	if (cfg == NULL) {
		EDIR_DEBUG_VERBOSE(("%s: NULL 'cfg'", funcName));
		EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
		return DECLINED;
	}

	/*
	 * We only handle requests with nonempty URL prefixes.
	 */
	if (r->uri == NULL) {
		EDIR_DEBUG_VERBOSE(("%s: No r->uri prefix", funcName));
		EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
		return DECLINED;
	}


	/*
	 * See if there is an application configured to handle the URI
	 * for this request.  Real and virtual HTTP server restrictions in the
	 * config are accounted for as well.
	 */
	if ((app = edir_load_check_path(scoreboard, r->uri,
		&(r->connection->local_addr),
		&(r->connection->remote_addr),
		r->server->server_hostname)) != NULL) {
		r->handler = ap_pstrdup(r->pool, "enhydra_director");
		EDIR_DEBUG_FUNCS_OUT(("func: %s() returns OK", funcName));
		return OK;
	}

	/*
	 * See if this is a request to a 'status' URL.
	 */
	if (edir_load_status_check_path(scoreboard, r->uri,
		&(r->connection->local_addr),
		&(r->connection->remote_addr),
		r->server->server_hostname)) {
		r->handler = ap_pstrdup(r->pool, "enhydra_director");
		EDIR_DEBUG_FUNCS_OUT(("func: %s() returns OK", funcName));
		return OK;
	}

	/* No apps found, so assume no responsibility for this request. */
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
	return DECLINED;

except:
	/* Don't accept handling for any requests if we can't really do it */
	edir_log_error("Error initializing scoreboard daemon.");
	if (*exc) {
		char *trace;
		trace = edir_exception_stack_trace(*exc);
		edir_log_error(trace);
		edir_free(trace);
		edir_exception_free(exc);
	}
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns DECLINED", funcName));
	return DECLINED;
}



/**
 * Get the original URI for the request.
 * <P>
 * Borrowed from the apache util_script.c
 *
 * @param r [in,out] Request handle.
 * @return The original URI, allocated from the request pool.
 */
static char *
original_uri(request_rec *r)
{
	char *first, *last;

	if (r->the_request == NULL) {
		return (char *) ap_pcalloc(r->pool, 1);
	}   

	first = r->the_request;		/* use the request-line */

	while (*first && !ap_isspace(*first)) {
		++first;				/* skip over the method */
	}   
	while (ap_isspace(*first)) {
		++first;				/*   and the space(s)   */
	}   

	last = first;
	while (*last && !ap_isspace(*last)) {
		++last;				/* end at next whitespace */
	}   

	return ap_pstrndup(r->pool, first, last - first);
}   


/**
 * Gather up the request headers to send to the Enhydra server.
 * 
 * @param r          [in]     The request handle to obtain the header info from.
 * @param cgiHeaders [in,out] Key-value table for the CGI headers.
 * @param cgiHeaders [in,out] Key-value table for the HTTP headers.
 */
static void
edir_apache_get_request_headers(request_rec *r,
								edir_kvtable_t *cgiHeaders,
								edir_kvtable_t *httpHeaders)
{
	server_rec *s = r->server;
	conn_rec *c = r->connection;
	char *rem_logname;
	char *host, *https;
	char local_port[32], remote_port[32];
	array_header *hdrs_arr = ap_table_elts(r->headers_in);
	table_entry *hdrs = (table_entry *) hdrs_arr->elts;
	int i;

	sprintf(local_port, "%u", ap_get_server_port(r));
	sprintf(remote_port, "%d", ntohs(c->remote_addr.sin_port));

	for (i = 0; i < hdrs_arr->nelts; ++i) {
		if (!hdrs[i].key) {
			continue;
		}
		edir_kvtable_add(httpHeaders, hdrs[i].key, hdrs[i].val);
	}

	edir_kvtable_add(cgiHeaders, "GATEWAY_INTERFACE", "CGI/1.1");
	edir_kvtable_add(cgiHeaders, "SERVER_PROTOCOL", r->protocol);
	edir_kvtable_add(cgiHeaders, "REQUEST_METHOD", (char *)(r->method));
	if (r->args != NULL) {
		edir_kvtable_add(cgiHeaders, "QUERY_STRING", r->args);
	}
	/* edir_kvtable_add(cgiHeaders, "REQUEST_URI", (char *) original_uri(r)); */
	edir_kvtable_add(cgiHeaders, "SERVER_SOFTWARE",
				 (char *) ap_get_server_version());
	edir_kvtable_add(cgiHeaders, "SERVER_NAME",
				 (char *) ap_get_server_name(r));
	edir_kvtable_add(cgiHeaders, "SERVER_ADDR", r->connection->local_ip);
	edir_kvtable_add(cgiHeaders, "SERVER_PORT", local_port);
	host = (char *) ap_get_remote_host(c, r->per_dir_config, REMOTE_HOST);
	if (host) edir_kvtable_add(cgiHeaders, "REMOTE_HOST", host);
	edir_kvtable_add(cgiHeaders, "REMOTE_ADDR", c->remote_ip);
	edir_kvtable_add(cgiHeaders, "DOCUMENT_ROOT", (char *) ap_document_root(r));
	edir_kvtable_add(cgiHeaders, "SERVER_ADMIN", s->server_admin);
	edir_kvtable_add(cgiHeaders, "SCRIPT_FILENAME", r->filename);
	edir_kvtable_add(cgiHeaders, "REMOTE_PORT", remote_port);
	if (c->user) edir_kvtable_add(cgiHeaders, "REMOTE_USER", c->user);
	if (c->ap_auth_type) {
		edir_kvtable_add(cgiHeaders, "AUTH_TYPE", c->ap_auth_type);
	}
	rem_logname = (char *) ap_get_remote_logname(r);
	if (rem_logname) {
		edir_kvtable_add(cgiHeaders, "REMOTE_IDENT", rem_logname);
	}
 
	if (!strcmp(r->protocol, "INCLUDED")) {
		edir_kvtable_add(cgiHeaders, "SCRIPT_NAME", r->uri);
		if (r->path_info && *r->path_info) {
			edir_kvtable_add(cgiHeaders, "PATH_INFO", r->path_info);
		}
	} else if (!r->path_info || !*r->path_info) {
		edir_kvtable_add(cgiHeaders, "SCRIPT_NAME", r->uri);
	} else {
		int path_info_start = ap_find_path_info(r->uri, r->path_info);
		edir_kvtable_add(cgiHeaders, "SCRIPT_NAME",
						ap_pstrndup(r->pool, r->uri, path_info_start));
		edir_kvtable_add(cgiHeaders, "PATH_INFO", r->path_info);
	}

	/*
	 * See if HTTPS is set
	 */
	https = "off";
	if (r->subprocess_env) {
		https = (char *) ap_table_get(r->subprocess_env, "HTTPS");
		if (!https) https = "off";
	}
	if (!edir_strcasecmp(https, "on")) {
		edir_kvtable_add(cgiHeaders, "SERVER_SCHEME", "https");
	} else {
		edir_kvtable_add(cgiHeaders, "SERVER_SCHEME", "http");
	}

	if (r->path_info && r->path_info[0]) {
		request_rec *pa_req;
		pa_req = ap_sub_req_lookup_uri(ap_escape_uri(r->pool, r->path_info), r);
		if (pa_req->filename) {
#ifdef WIN32
			char buffer[HUGE_STRING_LEN];
#endif
			char *pt = ap_pstrcat(r->pool, pa_req->filename, pa_req->path_info,
								NULL);
#ifdef WIN32
			/* We need to make this a real Windows path name */
			GetFullPathName(pt, HUGE_STRING_LEN, buffer, NULL);
			edir_kvtable_add(cgiHeaders, "PATH_TRANSLATED",
			ap_pstrdup(r->pool, buffer));
#else
			edir_kvtable_add(cgiHeaders, "PATH_TRANSLATED", pt);
#endif
		}
		ap_destroy_sub_req(pa_req);
	}

}


/**
 * Lists the content handlers for Enhydra Director.
 * <P>
 * Right now, only the "enhydra_director" handler is known, and it maps
 * to the 'edir_apache_handler()' function above.
 */
static const handler_rec edir_apache_handlers[] =
{
	{"enhydra_director", edir_apache_handler},
	{NULL}
};

/**
 * Enhydra Director Apache module definition block.
 * <P>
 * The number in brackets indicates the order in which the routine is called
 * during request processing.  Note that not all routines are necessarily
 * called (such as if a resource doesn't have access restrictions).
 */
module MODULE_VAR_EXPORT enhydra_director_module =
{
	STANDARD_MODULE_STUFF,
	edir_apache_master_init,			/* module initializer as server UID. */
	edir_apache_create_dir_config,		/* per-directory config creator */
	NULL,								/* dir config merger */
	edir_apache_create_server_config,	/* server config creator */
	NULL,								/* server config merger */
	edir_apache_cmds,					/* command table */
	edir_apache_handlers,				/* [7] list of handlers */
	edir_apache_translate,				/* [2] filename-to-URI translation */
	NULL,								/* [5] check/validate user_id */
	NULL,								/* [6] check user_id is valid *here* */
	NULL,								/* [4] check access by host address */
	NULL,								/* [7] MIME type checker/setter */
	NULL,								/* [8] fixups */
	NULL,								/* [10] logger */
#if MODULE_MAGIC_NUMBER >= 19970103
	NULL,								/* [3] header parser */
#endif
//#if MODULE_MAGIC_NUMBER >= 19970719
	edir_apache_init,					/* process initializer as child UID */
//#endif
#if MODULE_MAGIC_NUMBER >= 19970728
	edir_apache_child_exit,				/* process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
	NULL								/* [1] post read_request handling */
#endif
};

