/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: cgi_main.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

#define EDIR_CONFIG_FILE "./enhydra_director.conf"

#define EDIR_CGI_INBUF_SIZE		(4096)
#define EDIR_CGI_OUTBUF_SIZE		(4096)

/**
 * Global list of CGI headers.
 */
edir_kvtable_t *cgi_headers = NULL;

/**
 * Global list of HTTP_* headers.
 */
edir_kvtable_t *http_headers = NULL;

/**
 * Global list of cookies.
 */
edir_kvtable_t *cookies = NULL;

/**
 * Program name.
 */
static char *progName = NULL;

/**
 * Trim whitespace from the beginning and end of a string.
 * <P>
 * The input string is modified by this function.  The returned character
 * pointer must not be freed, as it may point to somewhere in the middle
 * of the original string.
 *
 * @param s [in,out] The string to trim.  Contents are modified!
 * @return Pointer to the beginning of the trimmed string within the original.
 */
static char *
cgi_trim(char *s)
{
    char *cp;
    while (*s && isspace(*s)) s++;
    for (cp=s; *cp; cp++) ;
    for (--cp; (cp >= s) && isspace(*cp); cp--) *cp = '\0';
    return s;
}

/**
 * Add the contents of a cookie header to the cookies list.
 * <P>
 * This function parses the cookie header, breaking it up
 * into its individual cookies.
 * 
 * @param value [in,out] The cookie header value to parse. (gets modified).
 */
void
cgi_add_cookie(char *value)
{
    char *start, *end, *name, *val;
    EDIR_DEBUG_FUNCS_IN(("func: cgi_add_cookie(\"%s\")", \
		        value ? value : "(null)"));
    /* parse key-value string as semicolon separated list. */
    start = value;
    end = strchr(start, ';');
    while (end) {
	*end++ = '\0';
	name = start;
	val = strchr(name, '=');
	if (val) *val++ = '\0';
	else val = "";
	edir_kvtable_add(cookies, cgi_trim(name), cgi_trim(val));
	start = end;
	end = strchr(start, ';');
    }
    name = start;
    val = strchr(name, '=');
    if (val)
	*val++ = '\0';
    else
	val = "";
    edir_kvtable_add(cookies, cgi_trim(name), cgi_trim(val));
    EDIR_DEBUG_FUNCS_OUT(("func: cgi_add_cookie() returns void"));
}

/**
 * Process a CGI environment variable and place it into the
 * correct key-value table to pass to the server.
 * <P>
 * The parameters are modified by this function, so it should be
 * considered 'destructive' to the strings.
 * 
 * @param name [in,out]  name  The CGI environment variable name.
 * @param value [in,out] value The value.
 */
void
cgi_process_header(char *name, char *value)
{
    char *cp;
    EDIR_DEBUG_FUNCS_IN(("func: cgi_process_header("EDIR_PTR", \"%s\", \"%s\")", \
		      name ? name : "(null)", value ? value : "(null)"));
    if (!strcmp(name, "QUERY_STRING")) {
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "REQUEST_URI", sizeof("REQUEST_") - 1)) {
	/* Special case - Servlet engine specifies this as without QS */
	cp = strchr(value, '?');
	if (cp) *cp = '\0';
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "REQUEST_", sizeof("REQUEST_") - 1)) {
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "REMOTE_", sizeof("REMOTE_") - 1)) {
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "SERVER_", sizeof("SERVER_") - 1)) {
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "PATH_", sizeof("PATH_") - 1)) {
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "AUTH_", sizeof("AUTH_") - 1)) {
	edir_kvtable_add(cgi_headers, name, value);
    } else if (!strncmp(name, "CONTENT_", sizeof("CONTENT_") - 1)) {
	cp = name;
	for (*cp++ = toupper(*cp); *cp; cp++) {
	    if (*cp == '_') {
		*cp++ = '-';
		if (*cp == '\0') break;
		*cp = toupper(*cp);
	    } else {
		*cp = tolower(*cp);
	    }
	}
	edir_kvtable_add(http_headers, name, value);
    } else if (!strncmp(name, "HTTP_", sizeof("HTTP_") - 1)) {
	if (!strcmp(name, "HTTP_COOKIE")) {
	    cgi_add_cookie(value);
	} else {
	    cp = name + sizeof("HTTP_") - 1;
	    for (*cp++ = toupper(*cp); *cp; cp++) {
		if (*cp == '_') {
		    *cp++ = '-';
		    if (*cp == '\0') break;
		    *cp = toupper(*cp);
		} else {
		    *cp = tolower(*cp);
		}
	    }
	    edir_kvtable_add(http_headers,
			     name + sizeof("HTTP_") - 1, value);
	}
    }
    EDIR_DEBUG_FUNCS_OUT(("func: cgi_process_headers() returns void"));
}

/**
 * Send the CGI environment to the Enhydra server.
 * <P>
 * Passes the HTTP headers, CGI variables, and cookies to the
 * Enhydra server as the first act of a request to the server.
 * 
 * @param conn   [in,out] Connection handle to the enhydra server.
 * @param outBuf [in,out] The output buffer for the connection.
 * @param olen   [in]     Size of the output buffer for the connection.
 * @param inBuf  [in,out] The input buffer for the connection.
 * @param ilen   [in]     Size of the input buffer for the connection.
 * @param exc    [in,out] Exception handle filled in if an error occurs.
 */
extern void
cgi_send_headers(edir_connection_t *conn, octet *outBuf, int olen,
		 octet *inBuf, int ilen, edir_exception_t **exc)
{
    static char *funcName = "cgi_send_headers";
    char	*key, *val, **vals;
    int32	i, nvals;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", %d, "EDIR_PTR", %d, "EDIR_PTR")", \
		      funcName, conn, outBuf, olen, inBuf, ilen, *exc));
    /* first send cgi headers */
    vals = edir_kvtable_first(cgi_headers, &key, &nvals);
    while (vals != NULL) {
	for (i=0; i<nvals; i++) {
	    val = vals[i];
	    EDIR_DEBUG_VERBOSE(("SendCgiHeader('%s','%s');", key, val));
	    edir_client_send_header(conn, key, strlen(key),
		    val, strlen(val), EDIR_PKT_SUBTYPE_HEADER_HEADER,
		    outBuf, olen, inBuf, ilen, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	vals = edir_kvtable_next(cgi_headers, &key, &nvals);
    }

    /* next send http headers */
    vals = edir_kvtable_first(http_headers, &key, &nvals);
    while (vals != NULL) {
	for (i=0; i<nvals; i++) {
	    val = vals[i];
	    EDIR_DEBUG_VERBOSE(("SendHttpHeader('%s','%s');", key, val));
	    edir_client_send_header(conn, key, strlen(key),
		    val, strlen(val), EDIR_PKT_SUBTYPE_HEADER_HTTP,
		    outBuf, olen, inBuf, ilen, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	vals = edir_kvtable_next(http_headers, &key, &nvals);
    }

    /* next send cookies */
    vals = edir_kvtable_first(cookies, &key, &nvals);
    while (vals != NULL) {
	for (i=0; i<nvals; i++) {
	    val = vals[i];
	    EDIR_DEBUG_VERBOSE(("SendCookie('%s','%s');", key, val));
	    edir_client_send_header(conn, key, strlen(key),
		    val, strlen(val), EDIR_PKT_SUBTYPE_HEADER_COOKIE,
		    outBuf, olen, inBuf, ilen, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	vals = edir_kvtable_next(cookies, &key, &nvals);
    }

    edir_client_send_header(conn,
			    "EDIR_CLIENT", sizeof("EDIR_CLIENT") - 1,
			    "CGI", sizeof("CGI") - 1,
			    EDIR_PKT_SUBTYPE_HEADER_LAST,
			    outBuf, olen, inBuf, ilen, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;

except:
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Structure to pass to edir_parse_uri() to receive important URI
 * fragments.
 */
typedef struct edir_uri_data {
    char *uriBuf;	   /* destroyed uri buffer */
    char *pathInfo;	   /* Stripped, normalized URI path. */ 
    char *pathParams;	   /* Full path parameters string, if any. */
    char *sessionID;       /* Session ID, if any.  */
    char *queryString;     /* Query String, if any. */
} edir_uri_data_t;

/**
 * Function to handle components of the request URI as it is parsed.
 * <P>
 * The edir_parse_uri() function calls this function for each part
 * of the url that it parses.  The four possible values of the <code>
 * type</code> parameter are described below. <BR>
 * <code>EDIR_PARSE_URI_PATH_INFO</code><BR>
 * The value of the <code>lhs</code> parameter will point to a character
 * string containing the stripped path portion of the URI, with leading
 * and trailing slashes removed, and with interior runs of more than
 * one slash reduced to one slash.
 * <code>EDIR_PARSE_URI_PATH_PARAM_1</code>
 * Returns a pointer to a string in <code>lhs</code> that contains a URL
 * "path parameter" that is not of the form "key=value".  The <code>rhs</code>
 * parameter is always <code>NULL</code>.
 * <code>EDIR_PARSE_URI_PATH_PARAM_2</code>
 * Returns strings comprising "path parameter" in the form of a "key=value"
 * pair, with the key in <code>lhs</code>, and the value in <code>rhs</code>.
 * Neither parameters are ever <code>NULL</code> for this type.
 * <code>EDIR_PARSE_URI_QUERY_STRING</code>
 * Returns a pointer to the beginning of the "query string" in <code>lhs</code>.
 * <P>
 * NOTE: The strings passed to this function are part of a larger
 * string buffer that is being destructively parsed by edir_parse_uri(). They
 * should not be freed, and they should only be stored in conjunction with
 * the larger URI buffer that contains them so the buffer does not
 * get freed while the strings are still in use.
 * 
 * @param ptr  [in,out] Arbitrary data structure to store things in.
 * @param type [in]     What type of component has been parsed.
 * @param lhs  [in]     "Left hand side" of "key=value" pair, or full
 * 			String of "value" item.
 * @param rhs  [in]     "Right hand side" of "key=value" pair, or NULL.
 */
static void
edir_cgi_uri_parse_func(void *ptr, int type, char *lhs, char *rhs)
{
    edir_uri_data_t *uriData = (edir_uri_data_t *) ptr;
    switch (type) {
	case EDIR_PARSE_URI_PATH_INFO:
	    uriData->pathInfo = lhs;
	    break;
	case EDIR_PARSE_URI_PATH_PARAMS:
	    uriData->pathParams = edir_strdup(lhs);
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_1:
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_2:
	    if (lhs && !strcmp(lhs, "jsessionid")) {
		uriData->sessionID = rhs;
	    }
	    break;
	case EDIR_PARSE_URI_QUERY_STRING:
	    uriData->queryString = lhs;
	    break;
    }
}

/**
 * Main loop of the Enhydra Director CGI module for most HTTP servers.
 * <P>
 * This function opens up an external shared-memory scoreboard 
 * file, chooses a server based on Enhydra Director's load balancing
 * algorithm, and then attempts to connect to the server and
 * process an incoming request.
 * 
 * @param configFile [in]  Optional configuration file name.NULL for default.
 * @param lockFile   [in]  Optional scoreboard lock name.  NULL for default.
 * @param dataFile   [in]  Optional scoreboard data name.  NULL for default.
 * @param urlPrefix  [in]  URL prefix to use.
 * @param exc        [out] Exception handle filled in upon error.
 * @exception EDIR_EXCEPT_*  Exceptions thrown at any level.
 */
void
enhydra_director(char *configFile, char *daemonFile, char *lockFile,
	         char *dataFile, char *urlPrefix, edir_exception_t **exc)
{
    static char			*funcName = "enhydra_director";
    char			*header, *cp;
    char			*sticky = NULL;
    int				n, requestBytesLeft, numMutex;
    int32			cLength;
    int				preBufferCount, isPost, stickyRetries = 0;
    octet			inBuf[EDIR_CGI_INBUF_SIZE];
    octet			outBuf[EDIR_CGI_INBUF_SIZE];
    int32			iv;
    char			*headerName, *headerValue, *val;
    char			*stickyTag = NULL, *path_info = NULL;
    char			*method;
    edir_uri_data_t		*uriData;
    edir_connection_t		*conn = NULL;
    edir_load_handles_t		*handles = NULL;
    edir_load_sb_app_t		*app;
    edir_load_sb_server_t	*server;
    struct sockaddr_in		local_addr, remote_addr;
    char			*server_name=NULL, *server_addr = NULL;
    char			*vhost_name=NULL;
    int32			nRefCounts = 0;
    int32			*refCounts = NULL;
    int				releaseConn = 0;
    int				do_status = 0;
    time_t			startTime, elapsedTime, lastTime, now;

    EDIR_DEBUG_FUNCS_IN(("%s(\"%s\", \"%s\", \"%s\", \"%s\", \"%s\", "EDIR_PTR")",\
		    "func: enhydra_director", \
		    configFile ? configFile : "(null)", \
		    daemonFile ? daemonFile : "(null)", \
		    lockFile ? lockFile : "(null)", \
		    dataFile ? dataFile : "(null)", \
		    urlPrefix ? urlPrefix : "(null)", *exc));

    /*
     * Determine the servlet prefix from the incoming request
     * URI so we can find the application and its server.
     */
    if (urlPrefix == NULL) {
	header = edir_kvtable_find(cgi_headers, "REQUEST_URI");
	if (header == NULL) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CGI,
					 "Can't get servlet %s",
				         "prefix from REQUEST_URI");
	    EDIR_THROW(*exc);
	}
	urlPrefix = edir_strdup(header);
    }
    uriData = (edir_uri_data_t *) edir_malloc(sizeof(edir_uri_data_t));
    memset(uriData, 0, sizeof(edir_uri_data_t));
    uriData->uriBuf = edir_strdup(urlPrefix);

    EDIR_DEBUG_EVERYTHING(("%s: urlPrefix = '%s'", urlPrefix));

    edir_parse_uri(uriData->uriBuf, uriData, edir_cgi_uri_parse_func);

    EDIR_DEBUG_EVERYTHING(("%s: PARSED pathinfo = '%s'", funcName, \
                        uriData->pathInfo ? uriData->pathInfo : "(null)"));
    EDIR_DEBUG_EVERYTHING(("%s: PARSED sessionID = '%s'", funcName, \
                        uriData->sessionID ? uriData->sessionID : "(null)"));
    EDIR_DEBUG_EVERYTHING(("%s: PARSED queryString = '%s'", funcName, \
                        uriData->queryString ? uriData->queryString:"(null)"));

    if ((uriData->pathInfo == NULL) || strlen(uriData->pathInfo) < 1) {
	*exc = edir_exception_create(EDIR_EXCEPT_CGI, "Can't get servlet %s",
				   "path info from REQUEST_URI");
	EDIR_THROW(*exc);
    }

    if (uriData->sessionID == NULL) {
	uriData->sessionID = edir_kvtable_find(cookies, "jsessionid");
    }

    EDIR_DEBUG_VERBOSE(("%s: Session ID = '%s'", funcName, \
		     uriData->sessionID ? uriData->sessionID : "(null)"));

    /*
     * If we have a session ID, look for sticky session tag.
     */
    if ((uriData->sessionID) && (val = strchr(uriData->sessionID, '('))) {
        if (strlen(val) >= 9) {
	    stickyTag = (char *) edir_malloc(9);
	    strncpy(stickyTag, val + 1, 8);
	    stickyTag[8] = '\0';
	    stickyRetries = EDIR_CONNECT_STICKY_SESSION_RETRIES;
	}
    }

    /*
     * Get the local host name and vhost name, as well as the port.
     */
    vhost_name = edir_kvtable_find(http_headers, "Host");
    server_addr = edir_kvtable_find(cgi_headers, "SERVER_ADDR");
    if (server_addr == NULL) {
	server_addr = edir_kvtable_find(cgi_headers, "LOCAL_ADDR");
    }
    if (server_addr == NULL) {
	server_addr = "127.0.0.1"; /* XXX */
    }
    server_name = edir_kvtable_find(cgi_headers, "SERVER_NAME");
    if (server_name == NULL) {
	server_name = edir_kvtable_find(cgi_headers, "LOCAL_NAME");
    }
    if (server_name == NULL) {
	server_name = edir_kvtable_find(cgi_headers, "SERVER_HOST");
    }
    if (server_name == NULL) {
	server_name = server_addr; /* XXX */
    }
    if (vhost_name == NULL) vhost_name = server_name;
    local_addr.sin_family = AF_INET;
    local_addr.sin_port = 0;
    if ((header = edir_kvtable_find(cgi_headers, "SERVER_PORT")) != NULL) {
	if (edir_atoi(header, &iv) && (iv >= 0) && (iv <= 65535)) {
	    local_addr.sin_port = htons(iv);
	}
    }
    if (!((server_addr) && edir_ip_addr(server_addr, &(local_addr.sin_addr)))) {
	local_addr.sin_addr.s_addr = INADDR_ANY;
    }

    /*
     * Get remote address, but not port
     */
    remote_addr.sin_family = AF_INET;
    remote_addr.sin_port = 0;
    if ((header = edir_kvtable_find(cgi_headers, "REMOTE_ADDR")) != NULL) {
	if (!edir_ip_addr(header, &(remote_addr.sin_addr))) {
	    remote_addr.sin_addr.s_addr = INADDR_ANY;
	}
    }

    EDIR_DEBUG_EVERYTHING(("%s: Sticky Tag = '%s'", funcName, \
		        stickyTag ? stickyTag : "(null)"));
    
    /*
     * Initialize our attachment to the shared scoreboard and
     * its associated mutex objects.  If this call succeeds we then
     * have a connection to the scoreboard with a shared lock on
     * the "watchdog" lock.  We keep the watchdog lock until we
     * finish the request.  This allows the 'edir_daemon' program to
     * monitor usage of the scoreboard and to time out and exit
     * after a period of inactivity.
     */
    handles = edir_load_init(configFile, daemonFile, lockFile, dataFile, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * Check the scoreboard out to make sure the daemon's still alive.
     */
    handles = edir_load_check(handles, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * Check to see if we only want to do the 'status page' thing.
     */
    if (edir_load_status_check_path(handles->scoreboard, uriData->pathInfo,
				    &local_addr, &remote_addr, vhost_name)) {
        do_status = 1;
    }

    if (!do_status) {
	/*
	 * Get connection host and port using the prevailing
	 * load balancing function.
	 */
	/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	edir_mutex_lock_exclusive(handles->mutex,EDIR_LOAD_SB_LOCK_GLOBAL,exc);
	if (*exc) EDIR_THROW(*exc);
	    
	refCounts = edir_refcount_values(handles->refcnt, &nRefCounts, exc);
	if (*exc) {
	    edir_exception_t *exc2=NULL;
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
	    edir_exception_free(&exc2);
	    EDIR_THROW(*exc);
	}
	edir_load_choose(handles->scoreboard, uriData->pathInfo, stickyTag,
		       &app, &server, refCounts, nRefCounts,
		       &local_addr, &remote_addr, vhost_name,  exc);
	if (*exc) {
	    edir_exception_t *exc2=NULL;
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
	    edir_exception_free(&exc2);
	    EDIR_THROW(*exc);
	}
	EDIR_DEBUG_VERBOSE(("Got server: %s:%d", \
			 server->hostName ? server->hostName : "(null)", \
			 server->port));
	server->useCount++;
	path_info = uriData->pathInfo + app->appPrefixLen;

	/*
	 * Up the refcount
	 */
	releaseConn = 1;
	edir_refcount_increment(handles->refcnt, server->refIndex, exc);
	if (*exc) {
	    edir_exception_t *exc2=NULL;
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
	    edir_exception_free(&exc2);
	    EDIR_THROW(*exc);
	}
	edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	if (*exc) EDIR_THROW(*exc);
	/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
    }

    if (!do_status && !(edir_debug & EDIR_DEBUG_MASK_FAKE_IT)) {
	conn = (edir_connection_t *) edir_malloc(sizeof(edir_connection_t));
	time(&startTime);
	lastTime = startTime;
	elapsedTime = 0;
	while (elapsedTime <= EDIR_CONNECT_TIMEOUT) {
	    if (*exc) edir_exception_free(exc);
	    edir_client_connect(conn, NULL, &(server->sin), server->authData,
		  outBuf, EDIR_CGI_OUTBUF_SIZE,inBuf,EDIR_CGI_INBUF_SIZE, exc);
	    if (!(*exc)) break; /* Success! */
	    if ((*exc)->id != EDIR_EXCEPT_IO_ERROR) EDIR_THROW(*exc);

	    /* XXX Should have 'logmask' */
	    if (edir_debug & EDIR_DEBUG_MASK_NORMAL) {
		char *trace = NULL;
		if (*exc) trace = edir_exception_stack_trace(*exc);
		edir_log_error("Warning: Error connecting to Enhydra.");
		edir_log_error("%s", trace ? trace : "No details.");
		if (trace) edir_free(trace);
	    }
	    if (*exc) edir_exception_free(exc);

	    /* Mark this server as 'failing' */
	    /* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	    edir_mutex_lock_exclusive(handles->mutex,
				    EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	    if (*exc) EDIR_THROW(*exc);
	    if (!(server->failing)) {
                /*
		 * XXX FIX  - Idea: Run notification script.
		 * Especially on Unix, where it's easy, run a script if we
		 * find a server to be 'failing'.  Probably would want to
		 * only run the script occasionally to avoid spamming,
		 * and would want to run it outside any critical section
		 * to avoid blocking for prolonged periods while holding
		 * the global scoreboard lock.
		 */
		server->failing = 1;
		time(&(server->failTime));
	    }
	    server->failCount++;
            /*
	     * If this request has a 'sticky session' token in the session
	     * cookie, then we want to try several times to go to the desired
	     * server.  After a number of attempts, though, we give up and
	     * fall back to round robin.  If getterData.retries is greater
	     * than zero, that means 'keep trying this server'.
	     */
	    if (stickyRetries <= 0) {
		releaseConn = 0;
		edir_refcount_decrement(handles->refcnt, server->refIndex, exc);
		server->useCount--; /* Undo rejected server use count. */
		if (*exc) {
		    edir_exception_t *exc2=NULL;
		    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		    edir_exception_free(&exc2);
		    EDIR_THROW(*exc);
		}
		/* Get new server */
		refCounts = edir_refcount_values(handles->refcnt,
						 &nRefCounts, exc);
		if (*exc) {
		    edir_exception_t *exc2=NULL;
		    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		    edir_exception_free(&exc2);
		    EDIR_THROW(*exc);
		}
		edir_load_choose(handles->scoreboard, uriData->pathInfo,
				 stickyTag, &app, &server, refCounts,
				 nRefCounts, &local_addr, &remote_addr,
				 vhost_name, exc);
		if (*exc) {
		    edir_exception_t *exc2=NULL;
		    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		    edir_exception_free(&exc2);
		    EDIR_THROW(*exc);
		}
		EDIR_DEBUG_NORMAL(("Got alternate server: %s:%d", \
				 server->hostName, server->port));
		server->useCount++; /* Now, up the use count for new server */
		path_info = uriData->pathInfo + app->appPrefixLen;
		releaseConn = 1;
		edir_refcount_increment(handles->refcnt, server->refIndex, exc);
		if (*exc) {
		    edir_exception_t *exc2=NULL;
		    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, &exc2);
		    edir_exception_free(&exc2);
		    EDIR_THROW(*exc);
		}
	    } else stickyRetries--;
	    /* Release mutex so we don't hold it while waiting for conn */
	    edir_mutex_unlock(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	    if (*exc) EDIR_THROW(*exc);
	    /* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
	    time(&now);
	    /* Avoid spinning wildly on errors */
	    if (EDIR_TIMEDIFF(now, lastTime) < 1) {
		sleep(1);
		time(&now);
	    }
	    lastTime = now;
	    elapsedTime = EDIR_TIMEDIFF(now, startTime);
	} /* While not timed out */

	/*
	 * If we couldn't get a connection in time, just fail.
	 */
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Check for case of timing out after freeing exception.
	 * XXX There's got to be a better way.
	 */
	if (conn->state != EDIR_STATE_OPEN) {
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR, 
		    "Can't connect to any Enhydra server for %d %s",
		    EDIR_CONNECT_TIMEOUT, "seconds -- giving up.");
	    EDIR_THROW(*exc);
	}

	/*
	 * Send request headers
	 */
	if (*path_info || uriData->pathParams) {
	    char *pinfo;
	    if (uriData->pathParams) {
		pinfo = edir_asprintf("%s;%s", path_info, uriData->pathParams);
		edir_kvtable_set(cgi_headers, "PATH_INFO", pinfo);
		edir_free(pinfo);
		edir_free(uriData->pathParams);
		uriData->pathParams = NULL;
	    } else {
		edir_kvtable_set(cgi_headers, "PATH_INFO", path_info);
	    }
	}
	edir_free(uriData->uriBuf);
	uriData->uriBuf = NULL;
	edir_free(uriData);
	uriData = NULL;
	cgi_send_headers(conn, outBuf, EDIR_CGI_OUTBUF_SIZE,
			 inBuf, EDIR_CGI_INBUF_SIZE, exc);
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Send request entity, if any.
	 */
	method = edir_kvtable_find(cgi_headers, "REQUEST_METHOD");
	if (method != NULL) {
	    if (!strcasecmp(method, "POST")) {
		header = edir_kvtable_find(http_headers, "Content-Length");
		if ((header != NULL) && edir_atoi(header, &cLength)) {
		    int32 bytesLeft;
		    bytesLeft = cLength;
		    while (bytesLeft > 0) {
			n = read(0, outBuf, (bytesLeft < EDIR_CGI_OUTBUF_SIZE) ?
				  bytesLeft : EDIR_CGI_OUTBUF_SIZE);
			if (n < 0) {
			    if (errno == EINTR) continue;
			    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
							"read error");
			    EDIR_THROW(*exc);
			}
			if (n < 1) break;
			edir_client_send_data(conn, n, 0,
					    outBuf, EDIR_CGI_OUTBUF_SIZE,
					    inBuf, EDIR_CGI_INBUF_SIZE, exc);
			if (*exc) EDIR_THROW(*exc);
			bytesLeft -= n;
		    }
		}
	    }
	}
	edir_client_send_data(conn, 0, 1, outBuf, EDIR_CGI_OUTBUF_SIZE,
			    inBuf, EDIR_CGI_INBUF_SIZE, exc);
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Receive and output response headers.
	 */
	while (conn->state == EDIR_STATE_RECEIVE_HEADERS) {
	    if (conn->packetSubType == EDIR_PKT_SUBTYPE_HEADER_COOKIE) {
		fwrite("Set-Cookie: ", 1, 12, stdout);
		fwrite(conn->headerName, 1, conn->headerNameLen, stdout);
		fwrite("=", 1, 1, stdout);
		fwrite(conn->headerValue, 1, conn->headerValueLen, stdout);
		fwrite("\r\n", 1, 2, stdout);
	    } else {
		fwrite(conn->headerName, 1, conn->headerNameLen, stdout);
		fwrite(": ", 1, 2, stdout);
		fwrite(conn->headerValue, 1, conn->headerValueLen, stdout);
		fwrite("\r\n", 1, 2, stdout);
	    }
	    edir_client_next_header(conn, outBuf, EDIR_CGI_OUTBUF_SIZE,
				  inBuf, EDIR_CGI_INBUF_SIZE, exc);
	    if (*exc) EDIR_THROW(*exc);
	}

	/*
	 * CRLF for end of headers.
	 */
	fwrite("\r\n", 1, 2, stdout);

	fflush(stdout);

	/* 
	 * Receive and output response entity.
	 */
	while ((conn->state == EDIR_STATE_RECEIVE_RESPONSE) ||
	       (conn->state == EDIR_STATE_RECEIVE_RESPONSE_PACKET))
	{
	    fwrite(conn->receiveData, 1, conn->receiveDataLen, stdout);
	    edir_client_next_data(conn, outBuf, EDIR_CGI_OUTBUF_SIZE, inBuf,
				EDIR_CGI_INBUF_SIZE, exc);
	    if (*exc) EDIR_THROW(*exc);
	}

	fflush(stdout);

	/*
	 * Close the appserver connection.
	 */
	edir_client_close(conn, outBuf, EDIR_CGI_OUTBUF_SIZE, exc);
	conn = NULL;
	if (*exc) EDIR_THROW(*exc);
    } else { /* fake it */
	void *sbuf;
        edir_kvtable_t *respHeaders;
        char clength[30], *key, **vals;
        int32 len, nvals;
        respHeaders = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	refCounts = edir_refcount_values(handles->refcnt, &nRefCounts, exc);
        sbuf = edir_load_status(handles->scoreboard, uriData->pathInfo,
                                refCounts, nRefCounts, respHeaders, exc);
        if (*exc) EDIR_THROW(*exc);
        val = edir_kvtable_find(respHeaders, "status");
        if (!val) val = "200 OK";
	printf("Status: %s\r\n", val);
	edir_kvtable_remove(respHeaders, "status");
        val = edir_kvtable_find(respHeaders, "content-type");
        if (!val) val = "text/plain";
	printf("Content-Type: %s\r\n", val);
	edir_kvtable_remove(respHeaders, "content-type");
	edir_kvtable_remove(respHeaders, "content-length");
        sprintf(clength, "%d", edir_sbuf_length(sbuf));
	printf("Content-Length: %s\r\n", clength);
        vals = edir_kvtable_first(respHeaders, &key, &nvals);
        while (vals) {
	    printf("%s: %s\r\n", key, vals[nvals-1]);
            vals = edir_kvtable_next(respHeaders, &key, &nvals);
        }
        edir_kvtable_destroy(respHeaders);
        val = edir_sbuf_tostring(sbuf, &len);
	printf("\r\n");
	fwrite(val, 1, len, stdout);
	fflush(stdout);
        edir_free(val);
        edir_sbuf_free(sbuf);
    }

    /*
     * Decrement this server's connection count
     */
    if (!do_status && releaseConn && server) {
	edir_refcount_decrement(handles->refcnt, server->refIndex, exc);
	releaseConn = 0;
	if (*exc) EDIR_THROW(*exc);
    }

    /*
     * Release our shared memory and mutex handles -- we're done.
     */
    edir_load_shutdown(handles, exc);
    handles = NULL;
    if (*exc) EDIR_THROW(*exc);

    EDIR_DEBUG_VERBOSE(("Success."));
    EDIR_DEBUG_FUNCS_OUT(("func: enhydra_director() returns void"));
    return;

except:
    if (conn) {
	edir_exception_t *_exc2=NULL, **exc2=&_exc2;
	edir_client_abort(conn, outBuf, EDIR_CGI_OUTBUF_SIZE, exc2);
	if (*exc2) edir_exception_free(exc2);
	edir_free(conn);
    }
    if (releaseConn && server && handles) {
	edir_exception_t *_exc2=NULL, **exc2=&_exc2;
	edir_refcount_decrement(handles->refcnt, server->refIndex, exc2);
	if (*exc2) edir_exception_free(exc2);
    }
    if (handles) {
	edir_exception_t *_exc2=NULL, **exc2=&_exc2;
	edir_load_shutdown(handles, exc2);
	if (*exc2) edir_exception_free(exc2);
    }
    EDIR_DEBUG_VERBOSE(("Failed."));
    EDIR_DEBUG_FUNCS_OUT(("func: enhydra_director() returns void"));
    return;
}

/**
 * Usage message generator.
 * <P>
 * Side effect:  Calls exit() to terminate the program.
 * 
 * @param progName [in] Name of the program to put in the message.
 */
static void
usage()
{
    edir_log_error("Usage: %s [-D <DebugMask>] [-F] [-d <DataFile>] %s%s",
		    progName ? progName : "enhydra_director",
		    "[-f <ConfigFile>] [-l <LockFile>] [-p <Prefix>] ",
		    "[-x <DaemonPath>]");
    edir_log_error("       -D   Set debug logging mask to <DebugMask>");
    edir_log_error("       -F   (debug) Fake the connection to the server");
    edir_log_error("       -d   Use <DataFile> for the scoreboard");
    edir_log_error("       -l   Use <LockFile> for the scoreboard lock");
    edir_log_error("       -f   Use <ConfigFile> for the configuration");
    edir_log_error("       -p   Use <Prefix> as the URL prefix");
    edir_log_error("       -x   Full path to the daemon program");
    edir_log_error("       -?   Print this usage message");
    exit(1);
}

main(int argc, char *argv[], char *envp[])
{
    static char		*funcName = "main";
    edir_exception_t	*_exc = NULL, **exc = &_exc;
    edir_load_handles_t	*handles = NULL;
    char		**cp, ch, *str;
    char		*buf, *name, *value;
    int			len, buflen, i;
    char		*configFile = NULL;
    char		*daemonFile = NULL;
    char		*dataFile = NULL;
    char		*lockFile = NULL;
    char		*urlPrefix = NULL;
    uint32		fake_it = 0;
    
    /* Get basename of arg0 for progName */
    progName = strrchr(argv[0], '/');
    if (progName) progName++;
    else progName = argv[0];
    
    /*
     * Allow hard-coded config parameters using #define macros
     * at the beginning of this file.  This is useful for a last-ditch
     * case where there is no clean way to pass arguments to 
     * a CGI program.
     */
#ifdef EDIR_DEBUG_MASK
    edir_debug = EDIR_DEBUG_MASK;
#endif
#ifdef DEBUG_FAKE_IT
    fake_it = EDIR_DEBUG_MASK_FAKE_IT;
#endif
#ifdef EDIR_CONFIG_FILE
    configFile = EDIR_CONFIG_FILE;
#endif
#ifdef EDIR_DAEMON_FILE
    daemonFile = EDIR_DAEMON_FILE;
#endif
#ifdef EDIR_DATA_FILE
    dataFile = EDIR_DATA_FILE;
#endif
#ifdef EDIR_LOCK_FILE
    lockFile = EDIR_LOCK_FILE;
#endif
#ifdef EDIR_URL_PREFIX
    urlPrefix = EDIR_URL_PREFIX;
#endif

    /*
     * Get file options from environment first, then override
     * from the command line.  This will allow some CGI
     * implementations to configure through environment variables
     * passed to the HTTP server.
     */ 
    if (str = getenv("EDIR_DEBUG_MASK")) edir_atoi(str, &edir_debug);
    if (str = getenv("EDIR_CONFIG_FILE")) configFile = str;
    if (str = getenv("EDIR_DAEMON_FILE")) daemonFile = str;
    if (str = getenv("EDIR_DATA_FILE")) dataFile = str;
    if (str = getenv("EDIR_LOCK_FILE")) lockFile = str;
    if (str = getenv("EDIR_URL_PREFIX")) urlPrefix = str;
    if (str = getenv("EDIR_DEBUG_FAKE_IT")) fake_it = EDIR_DEBUG_MASK_FAKE_IT;

    /*
     * Command line arguments supercede defaults or the environment.
     */
    while ((ch = getopt(argc, argv, "?D:Fd:f:l:p:x:")) != -1) {
	switch (ch) {
	    case 'D':
		edir_atoi(optarg, &edir_debug);
		break;
	    case 'F':
		fake_it = EDIR_DEBUG_MASK_FAKE_IT;
		break;
	    case 'd':
		dataFile = optarg;
		break;
	    case 'f':
		configFile = optarg;
		break;
	    case 'l':
		lockFile = optarg;
		break;
	    case 'p':
		urlPrefix = optarg;
		break;
	    case 'x':
		daemonFile = optarg;
		break;
	    case '?':
	    default:
		usage();
	}
    }
    argc -= optind;
    argv += optind;
    
    if (argc > 0) usage();

    /* Add 'fake it' mask to debug mask */
    edir_debug |= fake_it;
    
    /* Set log to "daemon" mode to log to syslog. */
    if (!isatty(0)) edir_log_setdaemon(progName, 1);

    /*
     * We don't want to get dumped because of a bad write to a
     * pipe, or because of belligerent treatment of CGI progs by
     * some HTTP servers.
     */
#ifdef SIGPIPE
    signal(SIGPIPE, SIG_IGN);
#endif
#ifdef SIGTERM
    signal(SIGTERM, SIG_IGN);
#endif

    /* Figure out buffer size. */
    for (buflen=0, cp = envp; *cp; cp++) {
	len = *cp ? strlen(*cp) : 0;
	if (len > buflen) buflen = len;
    }
    buflen += 100; /* leave some spare. */
    buf = (char *) edir_malloc(buflen);

    cgi_headers = edir_kvtable_create(EDIR_KVTABLE_TO_UPPER);
    http_headers = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
    cookies = edir_kvtable_create(EDIR_KVTABLE_TO_UPPER_PRESERVE);
    
    for (cp = envp; *cp; cp++) {
	if (*cp) {
	    strcpy(buf, *cp);
	    name = buf;
	    value = strchr(buf, '=');
	    if (!name || !value) continue; /* Goofy env. var. */
	    *value++ = '\0';
	    name = cgi_trim(name);
	    value = cgi_trim(value);
	    cgi_process_header(name, value);
	}
    }
    
    enhydra_director(configFile, daemonFile, lockFile, dataFile, urlPrefix, exc);
    if (*exc) EDIR_THROW(*exc);

    exit(0);

except:
    if (*exc) {
	char *trace;
	trace = edir_exception_stack_trace(*exc);
	edir_log_error("%s", trace);
	edir_free(trace);
	edir_exception_free(exc);
	exit(1);
    }
    return;
}



