/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_conf_xml.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "parser.h"
#include "edir.h"
#include "edir_xml.h"
#include "edir_conf_xml.h"

/*
 * Kludge to allow building with both libxml 1.8.6 and 2.2.1.
 */
#ifdef EDIR_LIBXML_186
#define children childs
#endif

static char * DefaultFileName = EDIR_CONF_XML_DEFAULT_FILE;

extern edir_conf_t *
edir_conf_xml_walk_dom(xmlDocPtr doc, edir_exception_t **exc);

static edir_conf_t *
edir_conf_xml_walk_enhydradirectorconfig(xmlNode *pnode,
			    edir_exception_t **exc);

static void
edir_conf_xml_walk_options(xmlNode *pnode, edir_conf_t *config,
			   edir_exception_t **exc);

static edir_conf_app_t *
edir_conf_xml_walk_application(xmlNode *pnode, edir_exception_t **exc);

static edir_conf_stat_t *
edir_conf_xml_walk_status(xmlNode *pnode, edir_exception_t **exc);

static edir_conf_appserver_t *
edir_conf_xml_walk_appserver(xmlNode *pnode, edir_exception_t **exc);

static edir_conf_restrict_t *
edir_conf_xml_walk_restrict(xmlNode *pnode, edir_exception_t **exc);

/**
 * Process configuration info from the named source.
 * <P>
 * This implementation accepts an XML configuration file and verifies
 * it for syntactic correctness with the EnhydraDirectorConfig DTD.
 * 
 * @param fileName [in] String name identifying the configuration file.
 * @param exc [out] Exception handle filled in if an error occurs.
 * @return A pointer to an <code>edir_conf_t</code> structure containing
 * 	the configuration data to be placed in the scoreboard.
 * @exception EDIR_EXCEPT_CONFIG If illegal values are found in the
 * 	configuration, or the XML input file is not valid according to
 * 	XML rules and the EnhydraDirectorConfig DTD.
 * @exception EDIR_EXCEPT_IO_ERROR If there is an I/O error while opening
 * 	and/or reading the configuration file.
 */
extern edir_conf_t *
edir_conf_process(char *fileName, edir_exception_t **exc)
{
    static char			*funcName = "edir_conf_process";
    xmlDocPtr			doc = NULL;
    xmlParserCtxtPtr		ctxt = NULL;
    edir_conf_xml_errdata_t	*errorData = NULL;
    char			buf[64], **line = NULL, **lines = NULL;
    edir_conf_t			*config = NULL;
    int				lineNum = 1, status;

    EDIR_DEBUG_FUNCS_IN(("func: %s(\"%s\", "EDIR_PTR")", funcName, \
			fileName ? fileName : "(null)", *exc));

    line = lines = edir_conf_read_file_lines(fileName, exc);
    if (*exc) EDIR_THROW(*exc);

    if (*lines == NULL) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, config));
	config = (edir_conf_t *) edir_malloc(sizeof(edir_conf_t));
	memset(config, 0, sizeof(edir_conf_t));
	return config;
    }

    edir_xml_init();

    ctxt = xmlCreatePushParserCtxt(NULL, NULL, buf, 0, fileName);
    if (!ctxt->wellFormed) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG, 
			"Malformed configuration file: %s", fileName);
	EDIR_THROW(*exc);
    }

    errorData = edir_xml_set_error_filter(ctxt);

    while (*line) {
	ctxt->input->line = lineNum++;
	status = xmlParseChunk(ctxt, *line, strlen(*line), 0);
	line++;
    }

    status = xmlParseChunk(ctxt, buf, 0, 1);

    if (errorData->sbuf) {
	char *msg;
	msg = edir_sbuf_tostring(errorData->sbuf, NULL);
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG, "%s", msg);
	edir_free(msg);
	edir_xml_clear_error_filter(ctxt);
	EDIR_THROW(*exc);
    }

    edir_xml_clear_error_filter(ctxt);

    /* errorData gets freed by edir_xml_clear_error_filter */
    errorData = NULL;

    doc = ctxt->myDoc;
    if (doc == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG, 
		    "Parse of \"%s\" failed - doc ptr is NULL!", fileName);
	EDIR_THROW(*exc);
    }

    config = edir_conf_xml_walk_dom(doc, exc);
    if (*exc) EDIR_THROW(*exc);

    xmlFreeParserCtxt(ctxt);
    xmlFreeDoc(doc);

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, config));
    return config;

except:
    if (ctxt) xmlFreeParserCtxt(ctxt);
    if (lines) {
	for (line=lines; *line; line++) {
	    edir_free(*line);
	}
	edir_free(lines);
    }
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns NULL", funcName));
    return NULL;
}

/**
 * Walk the DOM tree and gather the information needed to build the scoreboard.
 * <P>
 * This function process the DOM tree pointed to by <code>doc</code> and
 * places the relevant configuration parameters into an
 * <code>edir_conf_t</code> structure.  This is not as efficient as 
 * trying to directly build a scoreboard with the SAX API, but it is
 * much more intuitive and lets the libxml validating parser completely
 * handle all syntax checking.
 * <P>
 * At this point the DOM tree should only contain legal tokens, so the
 * main problems that can occur hear have to do with incorrect or
 * invalid data values, such as passing unparseable numbers in numeric
 * fields.  More work needs to be done to see what standard XML based
 * technologies can be used to further validate data types beyond CDATA.
 * 
 * @param doc [in]     The DOM tree to walk.
 * @param exc [in,out] Exception handle filled in if an error is detected.
 * @return Pointer to an <code>edir_conf_t</code> structure.
 * @exception EDIR_EXCEPT_CONFIG If illegal values are found in the
 * 	configuration.
 * 
 */
extern edir_conf_t *
edir_conf_xml_walk_dom(xmlDocPtr doc, edir_exception_t **exc)
{
    static char		*funcName = "edir_conf_xml_walk_dom";
    xmlNode 		*node;
    edir_conf_t		*config = NULL;

#ifdef EDIR_LIBXML_186
    if ((doc == NULL) || (doc->root == NULL)) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
				    "NULL DOM Tree! @lastnum@ @laststr@");
	EDIR_THROW(*exc);
    }
#else
    if ((doc == NULL) || (doc->children == NULL)) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
				    "NULL DOM Tree! @lastnum@ @laststr@");
	EDIR_THROW(*exc);
    }
#endif

    /*
     * Skip comment nodes and get to the <EnhydraDirectorConfig> element.
     */
#ifdef EDIR_LIBXML_186
    for (node = doc->root; node; node = node->next) {
	if (node->type == XML_ELEMENT_NODE) {
	    if (!edir_strcasecmp((char *) node->name, "EnhydraDirectorConfig")) {
		break;
	    }
	}
    }
#else
    for (node = doc->children; node; node = node->next) {
	if (node->type == XML_ELEMENT_NODE) {
	    if (!edir_strcasecmp((char *) node->name, "EnhydraDirectorConfig")) {
		break;
	    }
	}
    }
#endif
    if (node == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"Missing EnhydraDirectorConfig section!");
	EDIR_THROW(*exc);
    }

    config = edir_conf_xml_walk_enhydradirectorconfig(node, exc);
    if (*exc) EDIR_THROW(*exc);

    return config;

except:
    if (config) edir_conf_free(config);
    return NULL;
}

/**
 * Walk the DOM tree for the top level of the configuration tree.
 * <P>
 * Initializes the application and status port lists and fills
 * them in with data from lower levels in the tree.
 * 
 * @param pnode [in]     The parent node in the tree.
 * @param exc   [in,out] Exception handle filled in if an error occurs.
 * @return A pointer to an <code>edir_conf_t</code> block containing the
 * 	parsed configuration tree.
 * @exception EDIR_EXCEPT_CONFIG If illegal configuration values are found.
 */
static edir_conf_t *
edir_conf_xml_walk_enhydradirectorconfig(xmlNode *pnode, edir_exception_t **exc)
{
    static char		  *funcName="edir_conf_xml_walk_enhydradirectorconfig";
    xmlNode		  *node;
    edir_conf_t		  *config = NULL;
    int32		  numApps = 0, numStats = 0;
    edir_conf_app_t	  *ap, **app;
    edir_conf_appserver_t *sp, **spp;
    EDIR_DEBUG_FUNCS_IN(("func %s("EDIR_PTR", "EDIR_PTR")", funcName, pnode, *exc));

    config = edir_conf_create();

    /*
     * Not an error to have an empty configuration.
     */
    if (pnode->children == NULL) {
	config->apps = edir_conf_apps_create(0);
	config->stats = edir_conf_stats_create(0);
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, config));
	return config;
    }

    /*
     * First, count up the number of apps, stats in order to
     * allocate the arrays for them.
     */
    for (node=pnode->children; node; node = node->next) {
	if ((node->type == XML_ELEMENT_NODE) && (node->name != NULL)) {
	    if (!edir_strcasecmp((char *) node->name, "Application")) {
		numApps++;
	    } else if (!edir_strcasecmp((char *) node->name, "Status")) {
		numStats++;
	    }
	}
    }

    /*
     * Allocate arrays large enough to hold all Status and Application
     * element pointers.
     */
    config->apps = edir_conf_apps_create(numApps);
    config->stats = edir_conf_stats_create(numStats);

    /*
     * Now, fill in options, apps, and stat ports.
     */
    numApps = numStats = 0;
    for (node=pnode->children; node; node = node->next) {
	if ((node->type == XML_ELEMENT_NODE) && (node->name != NULL)) {
	    if (!edir_strcasecmp((char *) node->name, "Options")) {
		edir_conf_xml_walk_options(node, config, exc);
		if (*exc) EDIR_THROW(*exc);
	    } else if (!edir_strcasecmp((char *) node->name, "Application")) {
		config->apps[numApps++] = 
		    edir_conf_xml_walk_application(node, exc);
		if (*exc) EDIR_THROW(*exc);
	    } else if (!edir_strcasecmp((char *) node->name, "Status")) {
		config->stats[numStats++] = 
		    edir_conf_xml_walk_status(node, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	}
    }

    /*
     * Compute default weights.
     */
    for (app = config->apps; *app; app++) {
	int32 weights, nserv, avg;
	weights = 0;
	nserv = 0;
	ap = *app;
	if (ap->servers) {
	    for (spp = ap->servers; *spp; spp++) {
		sp = *spp;
		if (!(sp->weightIsDefault)) {
		    nserv++;
		    weights += sp->weight;
		}
	    }
	    if ((nserv < 1) || (weights < 1)) avg = 1;
	    else avg = weights / nserv;
	    if (avg < 1) avg = 1;
	    for (spp = ap->servers; *spp; spp++) {
		sp = *spp;
		if (sp->weightIsDefault) sp->weight = avg;
	    }
	}
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, config));
    return config;

except:
    if (config) edir_conf_free(config);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns NULL <exc>", funcName));
    return NULL;
}

/**
 * Walk the DOM tree for an "Options" element.
 * <P>
 * Fills in a configuration block's options with any options specified
 * in the element.
 * 
 * @param pnode  [in]     The parent node in the tree.
 * @param config [in]     Config block in which to place the options.
 * @param exc    [in,out] Exception handle filled in if an error occurs.
 * @exception EDIR_EXCEPT_CONFIG If illegal configuration values are found.
 */
static void
edir_conf_xml_walk_options(xmlNode *pnode, edir_conf_t *config,
			   edir_exception_t **exc)
{
    static char		*funcName="edir_conf_xml_walk_options";
    char		*cp;
    int32		iv;
    EDIR_DEBUG_FUNCS_IN(("func %s("EDIR_PTR", "EDIR_PTR", "EDIR_PTR")",
			funcName, pnode, config, *exc));

    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "daemon")) {
	config->daemonProg = cp;
    } 
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "daemoninterval")) {
	if ((!edir_atoi(cp, &iv)) || (iv < 0)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<Options> property \"daemoninterval\" must be integer %s",
		"greater than or equal to zero");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	config->daemonInterval = iv;
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "daemontimeout")) {
	if ((!edir_atoi(cp, &iv)) || (iv < 0)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<Options> property \"daemontimeout\" must be integer %s",
		"greater than or equal to zero");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	config->daemonTimeout = iv;
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "debug")) {
	if (!edir_atoi(cp, &iv)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<Options> property \"debug\" must be integer.");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	config->debug = 1;
	config->debugMask |= iv;
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "daemondebug")) {
	if (!edir_atoi(cp, &iv)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<Options> property \"daemondebug\" must be integer.");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	config->daemonDebug = 1;
	config->daemonDebugMask |= iv;
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "retrytime")) {
	if ((!edir_atoi(cp, &iv)) || (iv < 0)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<Options> property \"retrytime\" must be integer %s",
		"greater than or equal to zero");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	config->failRetrySet = 1;
	config->failRetry = iv;
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;

except:
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void <exc>", funcName));
    return;
}

/**
 * Walk the DOM tree for an "Application" element.
 * <P>
 * Creates a new <code>edir_conf_app_t</code> structure and fills it
 * in with data from the current and child tree nodes.
 * 
 * @param pnode  [in]     The parent node in the tree.
 * @param exc    [in,out] Exception handle filled in if an error occurs.
 * @exception EDIR_EXCEPT_CONFIG If illegal configuration values are found.
 */
static edir_conf_app_t *
edir_conf_xml_walk_application(xmlNode *pnode, edir_exception_t **exc)
{
    static char		*funcName="edir_conf_xml_walk_application";
    xmlNode		*node;
    edir_conf_app_t	*app = NULL;
    char		*cp, *dp, *ep, *fcp;
    int32		numRestricts = 0, numAppServers = 0;
    EDIR_DEBUG_FUNCS_IN(("func %s("EDIR_PTR", "EDIR_PTR")", funcName, pnode, *exc));

    app = edir_conf_app_create();

    if (fcp = cp = (char *) xmlGetProp(pnode, (unsigned char *) "prefix")) {
	app->appPrefix = edir_strdup(cp);
	dp = app->appPrefix;
	for (ep=cp; *ep; ) ep++;
	while ((cp < ep) && (isspace(*cp))) cp++;
	while ((cp < ep) && (*cp == '/')) cp++;
	while (cp < ep) {
	    if (*cp == '/') {
		*dp++ = '/';
		while ((cp < ep) && (*cp == '/')) cp++;
	    } else {
		*dp++ = *cp++;
	    }
	}
	*dp = '\0';
	for (--dp; (dp >= app->appPrefix) && isspace(*dp); dp--) *dp = '\0';
	for (; (dp >= app->appPrefix) && (*dp == '/'); dp--) *dp = '\0';
	edir_free(fcp);
    }  else {
	/* Should never get here if DTD validation worked properly. */
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"Required <Application> property \"prefix\" is missing.");
	EDIR_THROW(*exc);
    }

    /* Count up the AppServers and Restricts first, to allocate space */
    for ( node=pnode->children; node; node = node->next) {
	if ((node->type == XML_ELEMENT_NODE) && (node->name != NULL)) {
	    if (!edir_strcasecmp((char *) node->name, "AppServer")) {
		numAppServers++;
	    } else if (!edir_strcasecmp((char *) node->name, "Restrict")) {
		numRestricts++;
	    }
	}
    }

    if (numAppServers < 1) {
	/* Should never get here if DTD validation worked properly. */
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
	    "At least one <AppServer> is required per <Application> element.");
	EDIR_THROW(*exc);
    }

    app->servers = edir_conf_appservers_create(numAppServers);
    app->restricts = edir_conf_restricts_create(numRestricts);

    /* Now fill in the App servers and restriction entries */
    numAppServers = numRestricts = 0;
    for (node=pnode->children; node; node = node->next) {
	if ((node->type == XML_ELEMENT_NODE) && (node->name != NULL)) {
	    if (!edir_strcasecmp((char *) node->name, "AppServer")) {
		app->servers[numAppServers++] =
		    edir_conf_xml_walk_appserver(node, exc);
		if (*exc) EDIR_THROW(*exc);
	    } else if (!edir_strcasecmp((char *) node->name, "Restrict")) {
		app->restricts[numRestricts++] =
		    edir_conf_xml_walk_restrict(node, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	}
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, app));
    return app;

except:
    if (app) edir_conf_app_free(app);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns NULL <exc>", funcName));
    return NULL;
}

/**
 * Walk the DOM tree for an "Status" element.
 * <P>
 * Creates a new <code>edir_conf_app_t</code> structure and fills it
 * in with data from the current and child tree nodes.
 * 
 * @param pnode  [in]     The parent node in the tree.
 * @param exc    [in,out] Exception handle filled in if an error occurs.
 * @exception EDIR_EXCEPT_CONFIG If illegal configuration values are found.
 */
static edir_conf_stat_t *
edir_conf_xml_walk_status(xmlNode *pnode, edir_exception_t **exc)
{
    static char		*funcName="edir_conf_xml_walk_status";
    edir_conf_stat_t	*stat = NULL;
    xmlNode		*node;
    int32		numRestricts = 0;
    char		*cp, *dp, *ep, *fcp;
    EDIR_DEBUG_FUNCS_IN(("func %s("EDIR_PTR", "EDIR_PTR")", funcName, pnode, *exc));

    stat = edir_conf_stat_create();

    if (fcp = cp = (char *) xmlGetProp(pnode, (unsigned char *) "prefix")) {
	stat->statPrefix = edir_strdup(cp);
	dp = stat->statPrefix;
	for (ep=cp; *ep; ) ep++;
	while ((cp < ep) && (isspace(*cp))) cp++;
	while ((cp < ep) && (*cp == '/')) cp++;
	while (cp < ep) {
	    if (*cp == '/') {
		*dp++ = '/';
		while ((cp < ep) && (*cp == '/')) cp++;
	    } else {
		*dp++ = *cp++;
	    }
	}
	*dp = '\0';
	for (--dp; (dp >= stat->statPrefix) && isspace(*dp); dp--) *dp = '\0';
	for (; (dp >= stat->statPrefix) && (*dp == '/'); dp--) *dp = '\0';
	edir_free(fcp);
    }  else {
	/* Should never get here if DTD validation worked properly. */
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"Required <Status> property \"prefix\" is missing.");
	EDIR_THROW(*exc);
    }

    /* Count up Restricts first, to allocate space */
    for (node=pnode->children; node; node = node->next) {
	if ((node->type == XML_ELEMENT_NODE) && (node->name != NULL)) {
	    if (!edir_strcasecmp((char *) node->name, "Restrict")) {
		numRestricts++;
	    }
	}
    }

    stat->restricts = edir_conf_restricts_create(numRestricts);

    /* Now fill in the restriction entries */
    numRestricts = 0;
    for (node=pnode->children; node; node = node->next) {
	if ((node->type == XML_ELEMENT_NODE) && (node->name != NULL)) {
	    if (!edir_strcasecmp((char *) node->name, "Restrict")) {
		stat->restricts[numRestricts++] =
		    edir_conf_xml_walk_restrict(node, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	}
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, stat));
    return stat;

except:
    if (stat) edir_conf_stat_free(stat);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns NULL <exc>", funcName));
    return NULL;
}

/**
 * Walk the DOM tree for an "AppServer" element.
 * <P>
 * Creates and fills in an <conf>edir_conf_app_t</code> structure with
 * data from the current node.
 * 
 * @param pnode  [in]     The parent node in the tree.
 * @param exc    [in,out] Exception handle filled in if an error occurs.
 * @return A newly created and filled-in application config block.
 * @exception EDIR_EXCEPT_CONFIG If illegal configuration values are found.
 */
static edir_conf_appserver_t *
edir_conf_xml_walk_appserver(xmlNode *pnode, edir_exception_t **exc)
{
    static char			*funcName="edir_conf_xml_walk_appserver";
    edir_conf_appserver_t	*server = NULL;
    char			*cp;
    int32			iv;
    EDIR_DEBUG_FUNCS_IN(("func %s("EDIR_PTR", "EDIR_PTR")", funcName, pnode, *exc));

    server = edir_conf_appserver_create();

    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "host")) {
	server->hostName = cp;
    } else {
	/* Should never get here if DTD validation worked properly. */
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"Required <AppServer> property \"host\" is missing.");
	EDIR_THROW(*exc);
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "port")) {
	if ((!edir_atoi(cp, &iv)) || (iv < 0) || (iv > 65535)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<AppServer> property \"port\" must be integer %s",
		"between 0 and 65535.");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	server->port = (uint16) iv;
    } else {
	/* Should never get here if DTD validation worked properly. */
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"Required <AppServer> property \"port\" is missing.");
	edir_free(cp);
	EDIR_THROW(*exc);
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "weight")) {
	if ((!edir_atoi(cp, &iv)) || (iv < 0)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<AppServer> property \"weight\" must be integer %s",
		"greater than or equal to 0.");
	    edir_free(cp);
	    EDIR_THROW(*exc);
	}
	edir_free(cp);
	server->weight = iv;
	server->weightIsDefault = 0;
    }
    if (cp = (char *) xmlGetProp(pnode, (unsigned char *) "auth")) {
	server->authData = cp;
    }

    /* 
     * Convert the host and port to a socket address structure.
     */
    edir_sockaddr_in(server->hostName, server->port, &(server->sin), exc);
    if (*exc) {
	edir_exception_t *exc2;
	exc2 = edir_exception_create(EDIR_EXCEPT_CONFIG, 
		"Invalid host/port %s:%d: %s", server->hostName, server->port,
		(*exc)->message ? (*exc)->message : "");
	edir_exception_free(exc);
	EDIR_THROW(exc2);
    }
    server->ipName = (char *) edir_malloc(16);
    edir_sockaddr_to_str(&(server->sin), server->ipName);

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, server));
    return server;

except:
    if (server) edir_conf_appserver_free(server);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns NULL <exc>", funcName));
    return NULL;
}

/**
 * Walk the DOM tree for a "Restrict" element.
 * <P>
 * Creates and fills in an <conf>edir_conf_restrict_t</code> structure with
 * data from the current node.
 * 
 * @param pnode  [in]     The parent node in the tree.
 * @param exc    [in,out] Exception handle filled in if an error occurs.
 * @return A newly created and filled-in application restriction config block.
 * @exception EDIR_EXCEPT_CONFIG If illegal configuration values are found.
 */
static edir_conf_restrict_t *
edir_conf_xml_walk_restrict(xmlNode *pnode, edir_exception_t **exc)
{
    static char			*funcName="edir_conf_xml_walk_restrict";
    edir_conf_restrict_t	*restrict_ent = NULL;
    char			*port = NULL, *bitStr;
    int32			iv;
    uint32			uiv;
    EDIR_DEBUG_FUNCS_IN(("func %s("EDIR_PTR", "EDIR_PTR")", funcName, pnode, *exc));

    restrict_ent = edir_conf_restrict_create();

    if (restrict_ent->hostName = (char *)
		xmlGetProp(pnode, (unsigned char *) "server")) {
	restrict_ent->type = EDIR_CONF_RESTRICT_SERVER;
    } else if (restrict_ent->hostName = (char *)
		    xmlGetProp(pnode, (unsigned char *) "virtualserver")) {
	restrict_ent->type = EDIR_CONF_RESTRICT_VIRTUAL;
    } else if (restrict_ent->hostName = (char *)
		    xmlGetProp(pnode, (unsigned char *) "client")) {
	restrict_ent->type = EDIR_CONF_RESTRICT_CLIENT;
    } else {
	restrict_ent->type = EDIR_CONF_RESTRICT_SERVER;
    }
    port = (char *) xmlGetProp(pnode, (unsigned char *) "port");
    if ((restrict_ent->hostName == NULL) && (port == NULL)) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
	    "<Restrict> requires {server,virtualserver,client} or port or both.");
	EDIR_THROW(*exc);
    }

    if (port) {
	if ((!edir_atoi(port, &iv)) || (iv < 0)||(iv > 65535)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		"<Restrict> property \"port\" must be integer %s",
		"between 0 and 65535.");
	    EDIR_THROW(*exc);
	}
	restrict_ent->port = (uint16) iv;
    } else restrict_ent->port = 0;
    edir_free(port);
    port = NULL;

    /* 
     * Convert the host and port to a socket address structure if this
     * is not a virtual host.
     */
    if (restrict_ent->type != EDIR_CONF_RESTRICT_VIRTUAL) {
	if (bitStr = strchr(restrict_ent->hostName, '/')) {
	    *bitStr++ = '\0';
	    if (!edir_ip_addr(restrict_ent->hostName, &(restrict_ent->sin.sin_addr))) {
		*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		    "<Restrict> property \"%s\" must be dotted-decimal %s",
		    (restrict_ent->type == EDIR_CONF_RESTRICT_CLIENT) ?
			    "client" : "server",
		    "when using '/' to specify network bits.");
		EDIR_THROW(*exc);
	    }
	    restrict_ent->sin.sin_port = htons(restrict_ent->port);
	    if ((!edir_atoi(bitStr, &iv)) || (iv < 0) || (iv > 32)) {
		*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
		    "IP mask bits in <Restrict> must be integer %s",
		    "between 0 and 32.");
		EDIR_THROW(*exc);
	    }
	    restrict_ent->netBits = iv;
	    for (uiv=0; iv > 0; iv--) {
		uiv >>= 1;
		uiv |= 0x80000000;
	    }
	    restrict_ent->netMask = (uint32) htonl((int32) uiv);
	} else {
	    edir_sockaddr_in(restrict_ent->hostName, restrict_ent->port,
			     &(restrict_ent->sin), exc);
	    if (*exc) {
		edir_exception_t *exc2;
		exc2 = edir_exception_create(EDIR_EXCEPT_CONFIG, 
			"Invalid host/port %s:%d: %s", restrict_ent->hostName,
			restrict_ent->port, (*exc)->message ? (*exc)->message : "");
		edir_exception_free(exc);
		EDIR_THROW(exc2);
	    }
	    restrict_ent->netBits = 32;
	    restrict_ent->netMask = 0xffffffff;
	}
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, restrict_ent));
    return restrict_ent;

except:
    if (restrict_ent) edir_conf_restrict_free(restrict_ent);
    if (port) edir_free(port);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns NULL <exc>", funcName));
    return NULL;
}

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void
edir_conf_set_default_filename(char  *filename)
{
    DefaultFileName = edir_strdup(filename);
}

