/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_daemon.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_WAIT_H
#include <wait.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#define EDIR_DUP2_CLOSE 1
#define EDIR_DUP2_NOCLOSE 0
static int edir_dup2_carefully(int fd1, int fd2, int closeflag, char *progname);

/**
 * Conjure up a daemon.
 * <P>
 * Causes the process to fork.  The child then detaches itself from
 * this process's process group and controlling tty and becomes a true
 * background daemon process.  Note that once a daemon has been
 * spawned successfully, the current process can't wait() on it
 * because it is not a child process.
 * <P>
 * The spawned daemon process will automatically start logging messages
 * to syslog instead of the standard output and error.
 * 
 * @param progname [in] The name to use to identify syslog messages.
 * @param readFd [in,out] If not NULL, filled in with read descriptor
 * 		for a pipe to the daemon process.
 * @param writeFd [in,out] If not NULL, filled in with a write descriptor
 * 		for a pipe to the daemon process.
 * @param exc [out] Exception context filled in if an error occurs.
 * @return If this is the new daemon process, returns zero, if this is
 * 	the parent and upon success, returns the PID of the child.  Returns
 * 	-1 on failure.
 * @exception EDIR_EXCEPT_DAEMON If an error occured while spawning the daemon
 * 		process.
 */
extern int
edir_daemon_fork(char *progname, int *readFd, int *writeFd, edir_exception_t **exc)
{
    static char		*funcName = "edir_daemon_fork";
    int			fd, nfd;
    int			pipe1[2], pipe2[2];
    int			childPid, grandChildPid, nbytes;
#define PARENT_WRFD (pipe1[1])
#define CHILD_RDFD (pipe1[0])
#define PARENT_RDFD (pipe2[0])
#define CHILD_WRFD (pipe2[1])
    pipe1[0] = pipe1[1] = pipe2[0] = pipe2[1] = -1;
    if (pipe(pipe1) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Can't open pipe to daemon: @errstr@ (errnum=@errnum@)");
	EDIR_THROW(*exc);
    }
    if (pipe(pipe2) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Can't open pipe to daemon: @errstr@ (errnum=@errnum@)");
	EDIR_THROW(*exc);
    }
    if ((childPid = fork()) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON, 
		"Can't fork to create daemon: @errstr@ (errnum=@errnum@)");
	EDIR_THROW(*exc);
    } else if (childPid > 0) {
	int st=0, p;
	/* We're parent - close child's pipe descriptors */
	close(CHILD_RDFD); close(CHILD_WRFD);
	CHILD_RDFD = CHILD_WRFD = -1;
	 
	/* Wait for first child to die to prevent zombies. */
        while ((p = edir_daemon_waitpid(childPid, &st, 0)) == -1) {
	    if (errno == ECHILD) break;
	    if (errno != EINTR) {
		*exc = edir_exception_create(EDIR_EXCEPT_DAEMON, 
		    "%s: @errstr@ (errnum=@errnum@)", 
		    "Error waiting for daemon to go into background: ");
		EDIR_THROW(*exc);
	    }
	}
	if (p == -1) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		    "PID of spawned daemon was invalid");
	    EDIR_THROW(*exc);
	}
	if (WIFEXITED(st) && (WEXITSTATUS(st) != 0)) {
	    /* First child failed during creation. */
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Daemon creation failed with status %d", WEXITSTATUS(st));
	    EDIR_THROW(*exc);
	} else if (WIFSIGNALED(st)) {
	    /* First child dumped core during creation. */
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Daemon creation terminated by signal %d", WTERMSIG(st));
	    EDIR_THROW(*exc);
	} 

	/* Read integer pid from the grandchild */
	errno = 0;
	while ((nbytes = read(PARENT_RDFD, &grandChildPid,
			     sizeof(grandChildPid))) < 0) {
	    if (errno != EINTR) {
		*exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		    "%s: @errstr@ (errnum=@errnum@)",
		    "Error reading grandchild's PID from pipe");
		kill(childPid, 15);
		sleep(2);
		if (kill(childPid, 0) != -1)
		    kill(childPid, 9); /* Make sure. */
		EDIR_THROW(*exc);
	    }
	}
	if (nbytes != sizeof(grandChildPid)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Unexpected short read from pipe. (got %d expected %d)",
		nbytes, sizeof(grandChildPid));
	    kill(childPid, 15); /* Try to kill nicely with SIGTERM */
	    sleep(2);
	    if (kill(childPid, 0) != -1)
		kill(childPid, 9); /* Now get nasty */
	    EDIR_THROW(*exc);
	}

	/*
	 * Must supply both.  Makes no sense for both processes
	 * to have readers or both to have writers.
	 */
	if (readFd && writeFd) {
	    *readFd = PARENT_RDFD;
	    *writeFd = PARENT_WRFD;
	} else {
	    if (readFd) *readFd = -1;
	    if (writeFd) *writeFd = -1;
	    close(PARENT_RDFD);
	    close(PARENT_WRFD);
	}

	return grandChildPid;
    }

    /*
     * In the first child process now.  We must create a new process
     * group and disassociate from the controlling tty.
     */

    /* Close parent's pipe descriptors */
    close(PARENT_RDFD); close(PARENT_WRFD);
    PARENT_RDFD = PARENT_WRFD = -1;
    edir_log_setdaemon(progname, 1);
#ifdef SIGTTOU
    signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
    signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif
#ifdef SETPGRP_VOID
    /* Posix and SysV setpgrp() */
    if (setpgrp() == -1) {
	edir_log_error("Can't disassociate from process group! Exiting.");
	exit(1);
    }
#else /* SETPGRP_VOID */
    /* BSD setpgrp() */
    if (setpgrp(0, getpid()) == -1) {
	edir_log_error("Can't disassociate from process group! Exiting.");
	exit(1);
    }
#ifdef TIOCNOTTY
    if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
	ioctl(fd, TIOCNOTTY, (char *)NULL); /* BSD: Detach controlling tty */
	close(fd);
    }
#endif /* TIOCNOTTY */
#endif /* SETPGRP_VOID */

    /*
     * Spawn a second child.  This is more important for SysV, since
     * it makes the daemon a non-process-group-leader so it won't
     * get a controlling TTY.  We do it for BSD as well, though, to
     * return an exit status of zero to the parent so it knows we
     * successfully started.
     */
    signal(SIGHUP, SIG_IGN);
    if ((childPid = fork()) < 0) {
	edir_log_error("Can't fork second child! Exiting.");
	exit(1);
    } else if (childPid > 0) {
	exit(0); /* First child dies */
    }

    /*
     * Now, we're completely in the background.  Close all fd's to
     * prevent interference with the parent.
     */
    
    /* First, send our PID to grandpa. */
    grandChildPid = getpid(); /* We're the grandchild now */
    errno = 0;
    while ((nbytes = write(CHILD_WRFD, &grandChildPid,
			   sizeof(grandChildPid))) < 0) {
	if (errno != EINTR) {
	    /*
	     * Unfortunately, can't return error status in parent anymore
	     * Parent will throw error anyway when it gets short read or
	     * the PID read fails.
	     */
	    edir_log_error("FATAL: %s: @errstr@ (errnum=@errnum@)", 
			 "Can't send daemon PID to spawning grandparent.");
	    exit(1);
	}
    }
    if (nbytes != sizeof(grandChildPid)) {
	/*
	 * Unfortunately, can't return error status in parent anymore
	 * Parent will throw error anyway when it gets short read or
	 * the PID read fails.
	 */
	edir_log_error("Unexpected short write to pipe. (sent %d expected %d)",
		     nbytes, sizeof(grandChildPid));
	exit(1);
    }

    /*
     * If not interested in pipe fd's close them.
     */
    if (!(readFd && writeFd)) {
	if (readFd) *readFd = -1;
	if (writeFd) *writeFd = -1;
	close(CHILD_RDFD);
	close(CHILD_WRFD);
	CHILD_RDFD = CHILD_WRFD = -1;
    }

    /*
     * Close all open fd's except our pipe fd's
     */
#if defined(HAVE_GETDTABLESIZE)
    nfd = getdtablesize();
#elif defined(NOFILE)
    nfd = NOFILE;
#else
    nfd = 60;
#endif
    for (fd=0; fd < nfd; fd++) {
	if ((fd != CHILD_RDFD) && (fd != CHILD_WRFD)) {
	    close(fd);
	}
    }
    /*
     * Now, if we're interested in pipe fd's, make the read
     * pipe stdin and the write pipe stderr using dup().
     * We know that there are only two fd's open -- the pipe fd's.
     * The ugly code below is because of the fact that we don't
     * know where the current two fd's are.  Thay may or may not
     * be in slots 0 and 1, or may be in the wrong order, etc.
     * We carefully rearrange the fd's using dup(), then copy
     * fd 1 to fd 2 (stderr).
     */
    if (readFd && writeFd) {
	if (CHILD_RDFD != 0) {
	    if (CHILD_WRFD == 0) {
		CHILD_WRFD = edir_dup2_carefully(CHILD_WRFD, -1,
					    EDIR_DUP2_NOCLOSE, progname);
		edir_dup2_carefully(CHILD_RDFD, 0, EDIR_DUP2_CLOSE, progname);
	    } else {
		edir_dup2_carefully(CHILD_RDFD, 0, EDIR_DUP2_NOCLOSE, progname);
	    }
	}
	if (CHILD_WRFD != 1) {
	    if (CHILD_RDFD == 1) {
		edir_dup2_carefully(CHILD_WRFD, 1, EDIR_DUP2_CLOSE, progname);
	    } else {
		edir_dup2_carefully(CHILD_WRFD, 1, EDIR_DUP2_NOCLOSE, progname);
	    }
	}
	if (CHILD_WRFD != 2) {
	    if (CHILD_RDFD == 2) {
		edir_dup2_carefully(CHILD_WRFD, 2, EDIR_DUP2_CLOSE, progname);
	    } else {
		edir_dup2_carefully(CHILD_WRFD, 2, EDIR_DUP2_NOCLOSE, progname);
	    }
	}
	if (CHILD_RDFD > 2) while ((close(CHILD_RDFD)<0)&&(errno==EINTR));
	if (CHILD_WRFD > 2) while ((close(CHILD_WRFD)<0)&&(errno==EINTR));
    } else {
	/* No fd's are open.  Make stdin, stdout, stderr be /dev/null */
	while (((fd = open("/dev/null", O_RDONLY)) < 0) && (errno == EINTR)) ;
	if (fd != 0) {
	    edir_log_error("Can't open /dev/null as stdin: @errstr@");
	    exit(1);
	}
	while (((fd = open("/dev/null", O_WRONLY)) < 0) && (errno == EINTR)) ;
	if (fd != 1) {
	    edir_log_error("Can't open /dev/null as stdout: @errstr@");
	    exit(1);
	}
	edir_dup2_carefully(1, 2, EDIR_DUP2_NOCLOSE, progname);
    }
    /* This is always the case */
    *readFd = 0;
    *writeFd = 1;

    /* Make sure to reopen log, probably using fd 3 */
    edir_log_setdaemon(progname, 1);
    errno = 0;

    /* Don't let umask interfere with explicit file perms. */
    umask(0);

    return 0; /* Return with "I am the daemon" status */

except:
    /* Nothing to clean up. */
    return -1;
}

/**
 * Portable waitpid() replacement.  
 * <P>
 * Tries to use the best available version of wait...() to implement
 * waitpid().  Note that if only wait3() or wait() are available,
 * this function for ALL returned children until the desired child exits.
 * 
 * @param pid [in] The PID to wait for.
 * @param status [in,out] If not NULL, points to an integer that is 
 * 	filled in with the exit status of the process that terminated.
 * @param opt [in] Standard waitpid() options flags.
 * @return The PID of the waited-for process, or -1 on failure.
 */
extern int
edir_daemon_waitpid(int pid, int *status, int opt)
{
#if defined(HAVE_WAITPID)
    return waitpid(pid, status, opt);
#elif defined(HAVE_WAIT4)
    return wait4(pid, status, opt, NULL);
#elif defined(HAVE_WAIT3)
    int p;
    if (kill(pid, 0) < 0) {
	errno = ECHILD;
	return -1;
    }
    while ((p = wait3(status, opt, NULL)) != pid) {
	if (p <= 0) break;
	if (pid == -1) break;
    }
    return p;
#else
    int p;
    if (kill(pid, 0) < 0) {
	errno = ECHILD;
	return -1;
    }
    while ((p = wait(status)) != pid) {
	if (p <= 0) break;
	if (pid == -1) break;
    }
    return p;
#endif
}

/**
 * Convenience function to dup fd1.  
 * <P>
 * If fd2 is greater than equal to zero then it is closed with error
 * checking.  In either case fd1 is duplicated to the next available
 * fd, which may or may not be fd2.
 * <P>
 * This function exits on failure!  It is only intended for use while
 * setting up a newly spawned daemon's file descriptors, in which case
 * it cannot throw an exception to the parent.
 * 
 * @param fd1 [in] The fd to duplicate.
 * @param fd2 [in] The fd to close if not -1.
 * @return The duplicate file descriptor.
 */
static int 
edir_dup2_carefully(int fd1, int fd2, int closeflag, char *progname)
{
    int retFd = -1;
    if (closeflag && (fd2 >= 0)) {
	while (close(fd2) < 0) {
	    if (errno != EINTR) {
		edir_log_setdaemon(progname, 1);
		edir_log_error("%s %d %s: @errstr@ (errnum=@errnum@)",
		     "Can't close fd ", fd2, "before dup() to it");
		exit(1);
	    }
	}
    }
    while (((retFd=dup(fd1)) < 0) && (errno == EINTR)) ;
    if (retFd < 0) {
	edir_log_setdaemon(progname, 1);
	edir_log_error("Can't dup fd %d: @errstr@ (errnum=@errnum@)", fd1);
	exit(1);
    }
    if ((fd2 >= 0) && (retFd != fd2)) {
	edir_log_setdaemon(progname, 1);
	edir_log_error("Incorrectly duped fd %d to %d -- expected %d",
		     fd1, retFd, fd2);
	exit(1);
    }
    return retFd;
}

