/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_exception.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

/**
 * Create a new exception object with the given ID and message.
 * Message is specified as with "printf()".
 * 
 * @param id  [in] The exception ID.
 * @param fmt [in] Printf-style format string for the message.
 * @param ... [in] Printf-style variadic argument list.
 * @return A newly allocated exception object that must be
 * 	freed using "edir_exception_free()".
 */
extern edir_exception_t *
edir_exception_create(int id, char *fmt, ...)
{
    va_list ap;
    edir_exception_t *exc;
    char *newfmt;
#ifdef WIN32
    DWORD myErrno = errno;
    DWORD lastErr, wsaErr;
    lastErr = GetLastError();
    wsaErr = WSAGetLastError();
#else 
    int myErrno = errno;
#endif
    exc = (edir_exception_t *) edir_malloc(sizeof(edir_exception_t));
    memset(exc, 0, sizeof(edir_exception_t));
    exc->id = id;
    exc->stackpos = 0;
    exc->chain = NULL;
#ifdef WIN32
    newfmt = edir_log_fmt_add_error_string(fmt, myErrno, lastErr, wsaErr);
#else
    newfmt = edir_log_fmt_add_error_string(fmt, myErrno);
#endif
    va_start(ap, fmt);
    exc->message = edir_vasprintf(newfmt, ap);
    va_end(ap);
    edir_free(newfmt);
    return exc;
}

/**
 * Used by the EDIR_THROW macros to fill in useful information
 * such as file name, line number, etc.
 * 
 * WARNING! NEVER pass dynamically allocated strings to this method.
 * It NEVER frees them because it assumes it is getting static
 * string literals, as generated by __FILE__, etc.
 * 
 * @param exc Exception instance to fill in.
 * @param file	[in]	Filename from __FILE__ macro.
 * @param line	[in]	Line number from __LINE__ macro.
 * @param func	[in]	Function name from required "funcName" var.
 */
extern void
edir_exception_fill(edir_exception_t *exc, char *file, int line, char *func)
{
    char *cp;
    if (exc) {
        if (exc->stackpos < EDIR_EXC_MAX_STACK) {
#ifdef WIN32
	    if (cp = strrchr(file, '\\')) file = cp + 1;
	    if (cp = strrchr(file, '/')) file = cp + 1;
#else
	    if (cp = strrchr(file, '/')) file = cp + 1;
#endif
	    exc->stackLineNum[exc->stackpos] = line;
	    exc->stackFuncName[exc->stackpos] = func;
	    exc->stackFileName[exc->stackpos] = file;
	    exc->stackpos++;
	}
    }
}

/**
 * Free an exception object and its allocated memory resources.
 * 
 * Code that handles an exception without reraising it must call
 * this function to free the exception's memory.
 * 
 * This function takes pointer-to-pointer so that it can NULL
 * the pointer for safety's sake.
 * 
 * @param exc The exception data to free.
 */ 
extern void
edir_exception_free(edir_exception_t **excp)
{
    if (!excp) return;
    if (*excp) {
	if ((*excp)->message) edir_free((*excp)->message);
	(*excp)->message = NULL;
	edir_free(*excp);
    }
    *excp = NULL;
}

/**
 * Returns a text name for a given exception number.
 * The string returned should not be freed or modified.
 * 
 * @param exc [in] The exception object to get a name for.
 * @return The exception's text name.
 */
extern char *
edir_exception_name(edir_exception_t *exc)
{
    switch (exc->id) {
	case EDIR_EXCEPT_NONE:
	    return "EDIR_EXCEPT_NONE";
	case EDIR_EXCEPT_NULL_POINTER:
	    return "EDIR_EXCEPT_NULL_POINTER";
	case EDIR_EXCEPT_IO_ERROR:
	    return "EDIR_EXCEPT_IO_ERROR";
	case EDIR_EXCEPT_PROTO:
	    return "EDIR_EXCEPT_PROTO";
	case EDIR_EXCEPT_PROTO_BAD_STATE:
	    return "EDIR_EXCEPT_PROTO_BAD_STATE";
	case EDIR_EXCEPT_PROTO_AUTH:
	    return "EDIR_EXCEPT_PROTO_AUTH";
	case EDIR_EXCEPT_CONFIG:
	    return "EDIR_EXCEPT_CONFIG";
	case EDIR_EXCEPT_SHMEM:
	    return "EDIR_EXCEPT_SHMEM";
	case EDIR_EXCEPT_MUTEX:
	    return "EDIR_EXCEPT_MUTEX";
	case EDIR_EXCEPT_MUTEX_EXISTS:
	    return "EDIR_EXCEPT_MUTEX_EXISTS";
	case EDIR_EXCEPT_MUTEX_DELETED:
	    return "EDIR_EXCEPT_MUTEX_DELETED";
	case EDIR_EXCEPT_CGI:
	    return "EDIR_EXCEPT_CGI";
	case EDIR_EXCEPT_SERVLET:
	    return "EDIR_EXCEPT_SERVLET";
	case EDIR_EXCEPT_REFCOUNT:
	    return "EDIR_EXCEPT_REFCOUNT";
	case EDIR_EXCEPT_DAEMON:
	    return "EDIR_EXCEPT_DAEMON";
	case EDIR_EXCEPT_GENERAL:
	    return "EDIR_EXCEPT_GENERAL";
	default:
	    return "Unknown";
    }
}

/**
 * Assembles a string stack trace for the specified exception.
 * Returns a string that must be freed using edir_free().
 * 
 * @param exc The exception to obtain a stack trace for.
 * @return Newly allocated string containing the trace.
 */
extern char *
edir_exception_stack_trace(edir_exception_t *exc)
{
    char *str, *chain = NULL;
    void *sbuf;
    int i;
    if (exc->chain) chain = edir_exception_stack_trace(exc->chain);
    sbuf = edir_sbuf_create();
    
    edir_sbuf_strcatn(sbuf, exc->message);
    edir_sbuf_strcatn(sbuf, "\n");
    edir_sbuf_strcatn(sbuf, "Exception Details: ");
    edir_sbuf_strcatn(sbuf, edir_exception_name(exc));
    edir_sbuf_strcatn(sbuf, "\n");
    for (i=0; i<exc->stackpos; i++) {
	edir_sbuf_strcatn(sbuf, "    ");
	edir_sbuf_strcatn(sbuf, exc->stackFileName[i]);
	edir_sbuf_appendf(sbuf, "(Line %d): ", exc->stackLineNum[i]);
	edir_sbuf_strcatn(sbuf, exc->stackFuncName[i]);
	edir_sbuf_strcatn(sbuf, "(...)\n");
    }
    if (chain) {
	edir_sbuf_strcatn(sbuf, "Chained: \n");
	edir_sbuf_strcatn(sbuf, chain);
    }
    str = edir_sbuf_tostring(sbuf, NULL);
    if (chain) edir_free(chain);
    edir_sbuf_free(sbuf);
    return str;
}

