/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_exception.h,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#ifndef __EDIR_EXCEPTION_H__
#define __EDIR_EXCEPTION_H__ 1
#ifdef __cplusplus
extern "C" {
#endif

/**
 * Maximum stack trace length.  Stack trace consists of static arrays
 * in the exception structure.
 */
#define EDIR_EXC_MAX_STACK			32

/**
 * This means that there is no error.  May be used to indicate that
 * no exception instance is present in various situations.
 */
#define	EDIR_EXCEPT_NONE			0

/**
 * Indicates that a NULL pointer was detected (by means other than
 * core dumping due to SIGSEGV :)
 */
#define	EDIR_EXCEPT_NULL_POINTER		1

/**
 * Indicates an I/O error occured during an operation, such as receiving
 * from a socket, etc.
 */
#define	EDIR_EXCEPT_IO_ERROR			2

/**
 * Indicates that a general protocol failure occured with Enhydra Director
 */
#define	EDIR_EXCEPT_PROTO			3

/**
 * Indicates that a module-generated event (such as sending headers)
 * occured at a point in which a particular connection was not ready.
 */
#define	EDIR_EXCEPT_PROTO_BAD_STATE		4

/**
 * Authentication failure while connecting to the server.
 */
#define	EDIR_EXCEPT_PROTO_AUTH			5

/**
 * Configuration error.
 */
#define	EDIR_EXCEPT_CONFIG			6

/**
 * Shared memory or scoreboard error.
 */
#define	EDIR_EXCEPT_SHMEM			7

/**
 * Mutex error.
 */
#define	EDIR_EXCEPT_MUTEX			8

/**
 * Tried to create a mutex that would overwrite an existing one that
 * is currently in use.
 */
#define	EDIR_EXCEPT_MUTEX_EXISTS		9

/**
 * Mutex was deleted, possibly because owning daemon died/restarted.
 */
#define	EDIR_EXCEPT_MUTEX_DELETED		10

/**
 * CGI agent error.
 */
#define	EDIR_EXCEPT_CGI				11

/**
 * Servlet config exception.  For example, no application configured
 * for a prefix.
 */
#define	EDIR_EXCEPT_SERVLET			12

/**
 * Reference count exception.
 */
#define	EDIR_EXCEPT_REFCOUNT			13

/**
 * Daemon creation exception.
 */
#define	EDIR_EXCEPT_DAEMON			14

/**
 * General application error.
 */
#define	EDIR_EXCEPT_GENERAL			15

/**
 * Exception structure for the Enhydra Director state machine and
 * its friends.  It allows Java-like exception disposition in certain
 * parts of the code.  This allows the state machines for the protocol
 * to be somewhat similar between C and Java.
 * A few rules about this structure:
 * 	1.  For function name use a static char * variable in each function.
 * 	2.  For file name use __FILE__ macro.
 * 	3.  For line number use __LINE__ macro.
 */
typedef struct edir_exception {
    int			id;
    int			stackpos;
    char		*message;
    int			stackLineNum[EDIR_EXC_MAX_STACK];
    char 		*stackFuncName[EDIR_EXC_MAX_STACK];
    char 		*stackFileName[EDIR_EXC_MAX_STACK];
    struct edir_exception	*chain;
} edir_exception_t;

#define EDIR_THROW(EXC) \
	{ edir_exception_fill((EXC), __FILE__, __LINE__, funcName); goto except; }
#define EDIR_THROWTO(EXC,TO) \
	{ edir_exception_fill((EXC), __FILE__, __LINE__, funcName); goto TO ; }

extern edir_exception_t *
edir_exception_create(int num, char *fmt, ...);

extern void
edir_exception_fill(edir_exception_t *exc, char *file, int line, char *func);

extern void
edir_exception_free(edir_exception_t **exc);

extern char *
edir_exception_name(edir_exception_t *exc);

extern char *
edir_exception_stack_trace(edir_exception_t *exc);


#ifdef __cplusplus
} /* Extern "C" */
#endif
#endif /* __EDIR_EXCEPTION_H__ */

