/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_ipc.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

/**
 * Internal data pointed to by a "CONFIG" IPC handle.
 * <P>
 * Contains a string value.
 */
typedef struct edir_ipc_config_ctx {
    char *value;
} edir_ipc_config_ctx_t;

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * For CONFIG handles, returns a reference to the value string.
 * <P>
 * The string pointed to MUST NOT be modified!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_ipc_config_handle_get_object(edir_ipc_handle_t *handle)
{
    edir_ipc_config_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_ipc_config_ctx_t *) handle->pData;
    return (void *)(ctx->value);
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * For CONFIG handles, returns a copy of the value string allocated
 * using edir_malloc().
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char *
edir_ipc_config_handle_get_object_name(edir_ipc_handle_t *handle)
{
    edir_ipc_config_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_ipc_config_ctx_t *) handle->pData;
    if (ctx->value == NULL) return NULL;
    return edir_strdup(ctx->value);
}

/**
 * Convert a string into an IPC handle of type EDIR_IPC_HANDLE_TYPE_CONFIG.
 * <P>
 * This allows explicit configuration values to be used as well as
 * values parsed from a config source/file.
 * 
 * @param str [in] The string to wrap in a handle object.
 * @return A new handle object with a copy of the string.
 */
extern edir_ipc_handle_t *
edir_ipc_make_config_handle(char *str)
{
    edir_ipc_handle_t	*handle = NULL;
    edir_ipc_config_ctx_t	*ctx = NULL;
    if (str == NULL) return NULL;
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_CONFIG;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_STRING;
    handle->nrefs = 1; /* Don't care. not used */
    ctx = (edir_ipc_config_ctx_t *) edir_malloc(sizeof(edir_ipc_config_ctx_t));
    handle->pData = (void *)ctx;
    handle->share = NULL;
    ctx->value = edir_strdup(str);
    return handle;
}

/**
 * Convert a string into an IPC handle of type EDIR_IPC_HANDLE_TYPE_CONFIG.
 * <P>
 * This allows explicit configuration values to be used as well as
 * values parsed from a config source/file.
 * 
 * @param str [in] The string to wrap in a handle object.
 * @return A new handle object with a copy of the string.
 */
extern void
edir_ipc_destroy_config_handle(edir_ipc_handle_t *handle)
{
    edir_ipc_config_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) {
	edir_log_error("Warning: %s: tried to free invalid handle",
		     "edir_conf_free_ipc_handle()");
	return;
    }
    ctx = (edir_ipc_config_ctx_t *) (handle->pData);
    if (ctx->value) edir_free(ctx->value);
    edir_free(ctx);
    edir_free(handle);
}

