/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_kvtable.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"
#include "edir_kvtable.h"

#define EDIR_KVTABLE_ARRAY_SIZE_INCREMENT 50

/*
 * Forward declarations.
 */
static void edir_kvtable_normalize_key(edir_kvtable_t *table, char *key);

/**
 * Internal structure to store key information about a key-value mapping.
 */
typedef struct edir_kvent {
    char	keyBuf[EDIR_KVTABLE_MAX_KEY_LEN+1];
    char	origKeyBuf[EDIR_KVTABLE_MAX_KEY_LEN+1];
    char	*key;
    char	*origKey;
    void	*sbuf;
} edir_kvent_t;

/**
 * Create a key-value mapping table.
 * <P>
 * This function returns a pointer to a table that can be used to
 * associate string keys with one or more string values.  When the 
 * table is no longer needed it can be freed using
 * <code>edir_kvtable_destroy()</code>.
 * <P>
 * The <code>options</code> parameter specifies how the table treats
 * the lexicographic case of keys.  It may have one of the following
 * values:<BR>
 * <PRE>
 * <B>EDIR_KVTABLE_DEFAULT</B>
 *	Keys are used as-is with no conversion.  All searches are
 *	case-sensitive.
 * <B>EDIR_KVTABLE_TO_UPPER</B>
 *	Converts all keys to uppercase before storing or looking up
 *	values.  The enumeration routines will return the converted keys.
 *	Searches are not case sensitive.
 * <B>EDIR_KVTABLE_TO_UPPER_PRESERVE</B>
 * 	Same as <CODE>EDIR_KVTABLE_TO_UPPER</CODE>, except that the
 * 	original key is preserved and returned by the enumeration
 * 	functions.
 *	Searches are not case sensitive.
 * <B>EDIR_KVTABLE_TO_LOWER</B>
 *	Converts all keys to lowercase before storing or looking up
 *	values.  The enumeration routines will return the converted keys.
 *	Searches are not case sensitive.
 * <B>EDIR_KVTABLE_TO_LOWER_PRESERVE</B>
 * 	Same as <CODE>EDIR_KVTABLE_TO_UPPER</CODE>, except that the
 * 	original key is preserved and returned by the enumeration
 * 	functions.
 *	Searches are not case sensitive.
 * </PRE>
 * 	
 * @param options [in] Case handling preference for keys. (See above)
 * @return Pointer to a new key-value table.
 */
extern edir_kvtable_t *
edir_kvtable_create(int32 options)
{
    edir_kvtable_t	*table;
    table = (edir_kvtable_t *) edir_malloc(sizeof(edir_kvtable_t));
    memset((void *)table, 0, sizeof(edir_kvtable_t));
    edir_InitHashTable(&(table->table), EDIR_STRING_KEYS);
    table->options = options;
    table->key = table->keybuf;
    *(table->key) = '\0';
    table->arrayLen = 0;
    table->array = NULL;
    return table;
}

/**
 * Destroy a key-value table.
 * <P>
 * Destroys a key-value table created using <code>edir_kvtable_create()</code>
 * and frees all memory associated with it.  The table must not be used
 * after this function is called on it.
 * 	
 * @param table [in,out] Table to destroy.
 */
extern void
edir_kvtable_destroy(edir_kvtable_t *table)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    memset((void *)&(table->search), 0, sizeof(edir_HashSearch));
    ent = edir_FirstHashEntry(&(table->table), &(table->search));
    while (ent != NULL) {
	if (kvent = (edir_kvent_t *) edir_GetHashValue(ent)) {
	    if (kvent->sbuf) edir_sbuf_free(kvent->sbuf);
	    edir_free(kvent);
	}
	ent = edir_NextHashEntry(&(table->search));
    }
    edir_DeleteHashTable(&(table->table));
    if (table->array) edir_free(table->array);
    memset((void *)table, 0, sizeof(edir_kvtable_t));
    edir_free(table);
}

/**
 * Add a new value to a key-value table.
 * <P>
 * Adds <code>value</code> to the list of values associated with
 * <code>key</code> in the table.
 * <P>
 * If <code>key</code> does not exist, it is added to the table.
 * 
 * @param table [in,out] The table to add the value to.
 * @param key   [in]     The key under which to store <code>value</code>.
 * @param value [in]     The value to store.
 */
extern void
edir_kvtable_add(edir_kvtable_t *table, char *key, char *value)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    int			new ;

    edir_kvtable_normalize_key(table, key);
    ent = edir_CreateHashEntry(&(table->table), table->key, &new);
    if (new) {
	kvent = (edir_kvent_t *) edir_malloc(sizeof(edir_kvent_t));
	kvent->sbuf = edir_sbuf_create();
	strncpy(kvent->keyBuf, table->key, EDIR_KVTABLE_MAX_KEY_LEN);
	kvent->keyBuf[EDIR_KVTABLE_MAX_KEY_LEN] = '\0';
	if (table->options & EDIR_KVTABLE_PRESERVE) {
	    strncpy(kvent->origKeyBuf, key, EDIR_KVTABLE_MAX_KEY_LEN);
	    kvent->origKeyBuf[EDIR_KVTABLE_MAX_KEY_LEN] = '\0';
	    kvent->origKey = kvent->origKeyBuf;
	} else {
	    kvent->origKey = kvent->keyBuf;
	}
	edir_SetHashValue(ent, kvent);
    } else {
	kvent = (edir_kvent_t *) edir_GetHashValue(ent);
    }
    edir_sbuf_strcat(kvent->sbuf, value);
}

/**
 * Set a value in a key-value table.
 * <P>
 * Replaces the values (if any) associated with <code>key</code> with
 * the new value specified by <code>value</code>.  Any previous
 * values are discarded.
 * <P>
 * If <code>key</code> does not exist, it is added to the table.
 * 
 * @param table [in,out] The table to set the value in.
 * @param key   [in]     The key under which to store <code>value</code>.
 * @param value [in]     The value to store.
 */
extern void
edir_kvtable_set(edir_kvtable_t *table, char *key, char *value)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    int			new ;

    edir_kvtable_normalize_key(table, key);
    ent = edir_CreateHashEntry(&(table->table), table->key, &new);
    if (new) {
	kvent = (edir_kvent_t *) edir_malloc(sizeof(edir_kvent_t));
	kvent->sbuf = edir_sbuf_create();
	strncpy(kvent->keyBuf, table->key, EDIR_KVTABLE_MAX_KEY_LEN);
	kvent->keyBuf[EDIR_KVTABLE_MAX_KEY_LEN] = '\0';
	if (table->options & EDIR_KVTABLE_PRESERVE) {
	    strncpy(kvent->origKeyBuf, key, EDIR_KVTABLE_MAX_KEY_LEN);
	    kvent->origKeyBuf[EDIR_KVTABLE_MAX_KEY_LEN] = '\0';
	    kvent->origKey = kvent->origKeyBuf;
	} else {
	    kvent->origKey = kvent->keyBuf;
	}
	edir_SetHashValue(ent, kvent);
    } else {
	kvent = (edir_kvent_t *) edir_GetHashValue(ent);
	edir_sbuf_clear(kvent->sbuf);
    }
    edir_sbuf_strcat(kvent->sbuf, value);
}

/**
 * Find a single value for a given key in a key-value table.
 * <P>
 * Returns the first value associated with <code>key</code>, or <code>
 * NULL</code> if no value is found.
 * 
 * @param table [in,out] The table in which to search for the value.
 * @param key   [in]     The key to look up.
 * @return The value associated with <code>key</code> or <code>NULL</code>
 * 	if not found.
 */
extern char *
edir_kvtable_find(edir_kvtable_t *table, char *key)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    struct iovec	*iovp;
    int32		niovs;
    edir_kvtable_normalize_key(table, key);
    ent = edir_FindHashEntry(&(table->table), table->key);
    if (!ent) return NULL;
    kvent = (edir_kvent_t *) edir_GetHashValue(ent);
    if (!kvent) {
	edir_DeleteHashEntry(ent);
	return NULL;
    }
    iovp = edir_sbuf_get_iov(kvent->sbuf, &niovs);
    return (char *) iovp[0].iov_base;
}

/**
 * Find a all values for a given key in a key-value table.
 * <P>
 * Returns a NULL-terminated argv-style array of pointers to character
 * containing all of the values associated with <code>key</code>, or 
 * <code>NULL</code> if no values are found.
 * <P>
 * If <code>nvals</code> is not NULL, it points to an integer that is
 * filled in with the number of values found.
 * <P>
 * Note that the pointer returned points to an internal static array
 * held by the table structure.  It should not be freed, and its contents
 * are undefined after subsequent calls to edir_kvtable_*() routines.
 * 
 * @param table [in,out] The table in which to search for the value.
 * @param key   [in]     The key to look up.
 * @param nvals [in,out] Optional pointer to an integer that gets filled
 * 			 in with the number of values found.
 * @return Argv-style pointer to array of pointers to null-terminated
 * 	character strings that contain values associated with <code>
 * 	key</code> or <code>NULL</code> if no values are found.
 */
extern char **
edir_kvtable_find_all(edir_kvtable_t *table, char *key, int32 *nvals)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    struct iovec	*iovp;
    int32		niovs, i;
    if (nvals) *nvals = 0;
    edir_kvtable_normalize_key(table, key);
    ent = edir_FindHashEntry(&(table->table), table->key);
    if (!ent) return NULL;
    kvent = (edir_kvent_t *) edir_GetHashValue(ent);
    if (!kvent) {
	edir_DeleteHashEntry(ent);
	return NULL;
    }
    iovp = edir_sbuf_get_iov(kvent->sbuf, &niovs);
    if (table->arrayLen < niovs) {
	if (table->array) edir_free(table->array);
	table->arrayLen = niovs + EDIR_KVTABLE_ARRAY_SIZE_INCREMENT;
	/* Guarantee extra space for argv-style null terminator. */
	table->array =
	    (char **) edir_malloc((table->arrayLen + 1) * sizeof(char *));
    }
    for (i=0; i<niovs; i++) table->array[i] = (char *) iovp[i].iov_base;
    if (nvals) *nvals = niovs;
    return table->array;
}

/**
 * Find the first key-value mapping in a key-value table.
 * <P>
 * This function sets up an internal search context so that
 * subsequent calls to <code>edir_kvtable_next()</code> will enumerate the
 * rest of the key-value mappings.
 * <P>
 * Returns a NULL-terminated argv-style array of pointers to character
 * containing all of the values associated with the key that was found.
 * If the <code>keyp</code> parameter is not NULL, it is filled in with
 * a pointer to the key associated with the newly found values.
 * <P>
 * If there are no key-value mappings yet stored in the table, this
 * function returns <code>NULL</code>.
 * <P>
 * If <code>keyp</code> is not NULL, a pointer to a character string
 * If <code>nvals</code> is not NULL, it points to an integer that is
 * filled in with the number of values found.
 * matching the key for this value is placed in <code>*keyp</code>.
 * <P>
 * Note that the pointer returned points to an internal static array
 * held by the table structure.  It should not be freed, and its contents
 * are undefined after subsequent calls to edir_kvtable_*() routines.
 * 
 * @param table [in,out] The table in which to search for the value.
 * @param keyp  [in]     Option pointer to a character pointer to receive
 * 			 the key for the returned value.
 * @param nvals [in,out] Optional pointer to an integer that gets filled
 * 			 in with the number of values found.
 * @return Argv-style pointer to array of pointers to null-terminated
 * 	character strings that contain values associated with <code>
 * 	key</code> or <code>NULL</code> if no values are found.
 */
extern char **
edir_kvtable_first(edir_kvtable_t *table, char **keyp, int32 *nvals)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    int32		niovs, i;
    struct iovec	*iovp;
    if (nvals) *nvals = 0;
    if (keyp) *keyp = NULL;
    memset((void *)&(table->search), 0, sizeof(edir_HashSearch));
    ent = edir_FirstHashEntry(&(table->table), &(table->search));
    if (!ent) return NULL;
    kvent = (edir_kvent_t *) edir_GetHashValue(ent);
    while (kvent == NULL) {
	ent = edir_NextHashEntry(&(table->search));
	if (!ent) return NULL;
	kvent = (edir_kvent_t *) edir_GetHashValue(ent);
    }
    iovp = edir_sbuf_get_iov(kvent->sbuf, &niovs);
    if (table->arrayLen < niovs) {
	if (table->array) edir_free(table->array);
	table->arrayLen = niovs + EDIR_KVTABLE_ARRAY_SIZE_INCREMENT;
	/* Guarantee extra space for argv-style null terminator. */
	table->array =
	    (char **) edir_malloc((table->arrayLen + 1) * sizeof(char *));
    }
    for (i=0; i<niovs; i++) table->array[i] = (char *) iovp[i].iov_base;
    table->array[i] = NULL;
    if (nvals) *nvals = niovs;
    if (keyp) *keyp = kvent->origKey;
    return table->array;
}

/**
 * Find the next key-value mapping in a key-value table.
 * <P>
 * This function may be called after a call to <code>edir_kvtable_first()
 * </code> to return the next key-value mapping in an enumeration.
 * <P>
 * Returns a NULL-terminated argv-style array of pointers to character
 * containing all of the values associated with the key that was found.
 * If the <code>keyp</code> parameter is not NULL, it is filled in with
 * a pointer to the key associated with the newly found values.
 * <P>
 * If there are no key-value mappings yet stored in the table, this
 * function returns <code>NULL</code>.
 * <P>
 * If <code>keyp</code> is not NULL, a pointer to a character string
 * matching the key for this value is placed in <code>*keyp</code>.
 * If <code>nvals</code> is not NULL, it points to an integer that is
 * filled in with the number of values found.
 * <P>
 * Note that the pointer returned points to an internal static array
 * held by the table structure.  It should not be freed, and its contents
 * are undefined after subsequent calls to edir_kvtable_*() routines.
 * 
 * @param table [in,out] The table in which to search for the value.
 * @param keyp  [in]     Option pointer to a character pointer to receive
 * 			 the key for the returned value.
 * @param nvals [in,out] Optional pointer to an integer that gets filled
 * 			 in with the number of values found.
 * @return Argv-style pointer to array of pointers to null-terminated
 * 	character strings that contain values associated with <code>
 * 	key</code> or <code>NULL</code> if no values are found.
 */
extern char **
edir_kvtable_next(edir_kvtable_t *table, char **keyp, int32 *nvals)
{
    edir_kvent_t	*kvent = NULL;
    edir_HashEntry	*ent;
    int32		niovs, i;
    struct iovec	*iovp;
    while (kvent == NULL) {
	ent = edir_NextHashEntry(&(table->search));
	if (!ent) return NULL;
	kvent = (edir_kvent_t *) edir_GetHashValue(ent);
    }
    iovp = edir_sbuf_get_iov(kvent->sbuf, &niovs);
    if (table->arrayLen < niovs) {
	if (table->array) edir_free(table->array);
	table->arrayLen = niovs + EDIR_KVTABLE_ARRAY_SIZE_INCREMENT;
	/* Guarantee extra space for argv-style null terminator. */
	table->array =
	    (char **) edir_malloc((table->arrayLen + 1) * sizeof(char *));
    }
    for (i=0; i<niovs; i++) table->array[i] = (char *) iovp[i].iov_base;
    table->array[i] = NULL;
    if (nvals) *nvals = niovs;
    if (keyp) *keyp = kvent->origKey;
    return table->array;
}

/**
 * Remove all values associated with <code>key</code> from a key-value table.
 * <P>
 * Removes <code>key</code> from the table along with all values 
 * associated with it.  Subsequent searches for this key will yield
 * NULL.  If the <code>key</code> is not found, this function does nothing.
 * 
 * @param table [in,out] The table from which to remove the key.
 * @param key   [in]     The key to remove.
 */
extern void
edir_kvtable_remove(edir_kvtable_t *table, char *key)
{
    edir_kvent_t	*kvent;
    edir_HashEntry	*ent;
    edir_kvtable_normalize_key(table, key);
    ent = edir_FindHashEntry(&(table->table), table->key);
    if (ent) {
	if (kvent = (edir_kvent_t *) edir_GetHashValue(ent)) {
	    if (kvent->sbuf) edir_sbuf_free(kvent->sbuf);
	    edir_free(kvent);
	}
	edir_DeleteHashEntry(ent);
    }
}

/**
 * Convert keys according to table options flags.
 * <P>
 * Performs any necessary conversions of keys (to upper or
 * lower case) and copies the converted characters into a fixed
 * buffer in the table structure for use by the other functions.
 * 
 * @param table [in,out] The table for which we're converting.
 * @param key   [in]     The key to convert.
 */
static void
edir_kvtable_normalize_key(edir_kvtable_t *table, char *key)
{
    char		*cp, *dp, *ep;
    if (table->options & EDIR_KVTABLE_TO_UPPER) {
	cp = key;
	dp = table->keybuf;
	ep = dp + EDIR_KVTABLE_MAX_KEY_LEN;
	while ((dp < ep) && *cp) *dp++ = toupper(*cp++);
	*dp = '\0';
	table->key = table->keybuf;
    } else if (table->options & EDIR_KVTABLE_TO_LOWER) {
	cp = key;
	dp = table->keybuf;
	ep = dp + EDIR_KVTABLE_MAX_KEY_LEN;
	while ((dp < ep) && *cp) *dp++ = tolower(*cp++);
	*dp = '\0';
	table->key = table->keybuf;
    } else {
	strncpy(table->keybuf, key, EDIR_KVTABLE_MAX_KEY_LEN);
	table->keybuf[EDIR_KVTABLE_MAX_KEY_LEN] = '\0';
	table->key = table->keybuf;
    }
}

