/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_choose.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

#define		SMALL_PREFIX_LEN	200

#if defined(WIN32)
# define __ALLOCA(N) _alloca(N);
# define __UNALLOCA(P)
#elif defined(HAVE_ALLOCA)
# define __ALLOCA(N) alloca(N);
# define __UNALLOCA(P)
#else
# define __ALLOCA(N) edir_malloc(N)
# define __UNALLOCA(P) { if (P) edir_free(P); }
#endif

/**
 * Check a path prefix to see if the EnhydraDirector module should
 * handle the current request.
 * <P>
 * Compares the URI prefix specified by <code>path</code> to the
 * prefix for each application.  This function returns a pointer to
 * a matching application according to the following criteria:
 * <BR><PRE>
 * 	1.  If there are restrictions for an application slot, then
 * 	    the <code>local_addr</code> and <code>virt_host</code>
 * 	    parameters must satisfy at least one of the restrictions.
 * 	2.  Among the applications meeting criteria #1, the <code>
 * 	    path</code> parameter must match at least one application
 * 	    prefix up to the length of the prefix.
 * 	3.  If there is more than one prefix match among applications
 * 	    which meet both #1 and #2, then the longest match is chosen.
 * 	4.  If no match is found, this function returns NULL.
 * </PRE><BR>
 * 
 * @param scoreboard  [in] The scoreboard to check.
 * @param path        [in] The path prefix to check.
 * @param local_addr  [in] Socket address structure for server side of the
 * 			   request connection.
 * @param remote_addr [in] Socket address structure for client side of the
 * 			   request connection.
 * @param virt_host   [in] The value of the "Host" HTTP header or the current
 * 			   virtual server name.
 * @return Pointer to the application slot, if found.  If no matching
 * 	application is found, returns <code>NULL</code>.
 */ 
extern edir_load_sb_app_t *
edir_load_check_path(edir_load_sb_offsets_t *scoreboard, char *path, 
		     struct sockaddr_in *local_addr,
		     struct sockaddr_in *remote_addr,
		     char *virt_host)
{
    EDIR_IF_DEBUG( static char   *funcName = "edir_load_check_path"; )
    char			smallPrefix[SMALL_PREFIX_LEN + 1];
    char			*prefix = smallPrefix;
    char			*allocBuf = NULL;
    char			*cp, *dp, *ep, ch;
    int32			i, j, num, len, maxLen=-1;
    edir_load_sb_app_t		*app, *selectedApp = NULL;
    edir_load_sb_restrict_t	*restricts, *rp;
    int32			clientMatchRequired;
    int32			clientMatch;
    int32			serverMatchRequired;
    int32			serverMatch;
    int32			vhostMatchRequired;
    int32			vhostMatch;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", \"%s\", "EDIR_PTR", \"%s\")", \
			funcName, scoreboard, path ? path : "(null)", \
			local_addr, virt_host ? virt_host : "(null)"));

    /*
     * Make sure prefix buffer is big enough.
     */
    if (scoreboard->maxAppPrefixLen > SMALL_PREFIX_LEN) {
	allocBuf = (char *) __ALLOCA(scoreboard->maxAppPrefixLen + 1);
	prefix = allocBuf;
    }
    
    /**
     * Trim and clean up the path, and copy only the number of bytes
     * needed for comparison.
     */
    cp = path;
    dp = prefix;
    ep = dp + scoreboard->maxAppPrefixLen;
    while (*cp && (*cp == '/')) cp++;
    while (*cp && (dp < ep)) {
	if (*cp == '/') {
	    *dp++ = '/';
	    while (*cp == '/') cp++;
	} else {
	    *dp++ = *cp++;
	}
    }
    *dp = '\0';

    /**
     * Search for an application which matches the prefix and host
     * information.
     */
    num = scoreboard->numApps;
    app = EDIR_SB_APPS(scoreboard);
    for (i = 0; i < num; i++, app++) {
	len = app->appPrefixLen;
	ch = prefix[len];
	EDIR_DEBUG_EVERYTHING(("Checking prefix \"%s\" against app \"%s\" (maxLen=%d)", \
			      prefix, app->appPrefix, maxLen));
	if ( ((len == 0) && (len > maxLen)) ||
	     ((len > maxLen) && (!strncmp(prefix, app->appPrefix, len)) &&
	      ((ch == '\0') || (ch == '/') || (ch == '?') || (ch == ';') ||
	      (ch == '#') || (ch == ' ') || (ch == '\t'))) ) {
	    /* Potential match -- check restrictions */
	    if (((local_addr == NULL) && (virt_host == NULL)) ||
		 (app->numRestricts < 1)) {
		EDIR_DEBUG_EVERYTHING(("App has no server restrictions"));
		selectedApp = app;
		maxLen = len;
	    } else {
		/* Check restrictions - first match enables this request */
		EDIR_DEBUG_EVERYTHING(("App has server restrictions"));
		restricts = EDIR_SB_RESTRICTS(scoreboard, app);
		clientMatchRequired = 0;
		clientMatch = 0;
		serverMatchRequired = 0;
		serverMatch = 0;
		vhostMatchRequired = 0;
		vhostMatch = 0;
		for (j=0; j<app->numRestricts; j++) {
		    rp = &(restricts[j]);
		    EDIR_DEBUG_EVERYTHING(("Restrict type: %s", \
			    (rp->type == EDIR_CONF_RESTRICT_CLIENT) ? "CLIENT" : \
				(rp->type == EDIR_CONF_RESTRICT_SERVER) ? \
					"SERVER" : "VIRTUAL"));
		    if (rp->type == EDIR_CONF_RESTRICT_SERVER) {
			serverMatchRequired = 1;
			if ((rp->sin.sin_port != 0) &&
			    (local_addr->sin_port != rp->sin.sin_port)) {
			    EDIR_DEBUG_EVERYTHING(("Port %d doesn't match %d", \
						  ntohs(local_addr->sin_port), \
						  ntohs(rp->sin.sin_port)));
			    ;
			} else if ((rp->sin.sin_addr.s_addr != INADDR_ANY) &&
			       ((local_addr->sin_addr.s_addr & rp->netMask)
				!= (rp->sin.sin_addr.s_addr & rp->netMask))) {
#ifdef DEBUG
			    char ipbuf1[20];
			    char ipbuf2[20];
			    edir_sockaddr_to_str(local_addr, ipbuf1);
			    edir_sockaddr_to_str(&(rp->sin), ipbuf2);
			    EDIR_DEBUG_EVERYTHING(("IP %s %s %s", \
					ipbuf1, "doesn't match", ipbuf2));
#endif
			} else {
			    /* We fit within server restrictions */
			    serverMatch = 1;
			}
		    } else if (rp->type == EDIR_CONF_RESTRICT_CLIENT) {
			clientMatchRequired = 1;
			if ((rp->sin.sin_addr.s_addr != INADDR_ANY) &&
			       ((remote_addr->sin_addr.s_addr & rp->netMask)
				!= (rp->sin.sin_addr.s_addr & rp->netMask))) {
#ifdef DEBUG
			    char ipbuf1[20];
			    char ipbuf2[20];
			    edir_sockaddr_to_str(local_addr, ipbuf1);
			    edir_sockaddr_to_str(&(rp->sin), ipbuf2);
			    EDIR_DEBUG_EVERYTHING(("IP %s %s %s", \
					ipbuf1, "doesn't match", ipbuf2));
#endif
			} else {
			    /* We fit within client restrictions */
			    clientMatch = 1;
			}
		    } else { 
			vhostMatchRequired = 1;
			if ((strlen(rp->hostName) > 0) && virt_host &&
			    edir_strcasecmp(rp->hostName, virt_host)) {
			    EDIR_DEBUG_EVERYTHING(("VHost %s %s %s", \
						  rp->hostName, \
						  "doesn't match", \
						  virt_host));
			} else {
			    /* We fit within virtual server restrictions */
			    vhostMatch = 1;
			}
		    }
		    if (!serverMatchRequired) serverMatch = 1;
		    if (!clientMatchRequired) clientMatch = 1;
		    if (!vhostMatchRequired) vhostMatch = 1;
		    if (serverMatch && clientMatch && vhostMatch) {
			/* We are allowed to connect with this app! */
			selectedApp = app;
			maxLen = len;
		    }
		} /* end for - Restriction iterator */
	    } /* end if we have restrictions */
	} /* end if initial match */
    } /* end for - Application iterator */
    __UNALLOCA(allocBuf);
#ifdef DEBUG
    if (selectedApp) {
	EDIR_DEBUG_VERBOSE(("Application prefix \"%s\" matched \"%s\"", \
			    selectedApp->appPrefix, path));
    } else {
	EDIR_DEBUG_VERBOSE(("No application prefixes match \"%s\"", path));
    }
#endif
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, selectedApp));
    return selectedApp;
}

/**
 * Look up the server list for the given prefix and select an Enhydra server
 * to handle the request.
 * <P>
 * This function implements a simple weighted round-robin scheme.  Under
 * light load, the servers will be selected in round-robin order.  However,
 * under higher load, when more than one request must be dispatched to
 * a server, a weight factor is used to determine the proportional number
 * of simultaneous active connections each application server gets.  This
 * allows newer, faster servers to be added, and to allow administrators
 * to assign higher loads to such servers such that each server is loaded
 * in rough proportion to its capacity.
 * <P>
 * The exact algorithm is as follows:<BF>
 * <PRE>
 * 	let maxWeight = MAX({The weights of all servers for an application})
 * 	let MAXINT = {The maximum value of a 32 bit unsigned integer (2^32)}
 * 	let factor = MAXINT / maxWeight
 * 	let maxScore = -1;
 * 	let maxServer = nil;
 * 	for server in {All Servers For This App}
 * 		let numConn = {The number of current connections to this server}
 * 		let weight = {The weight factor configured for this server}
 * 		if numConn > 0 then
 * 			score = (weight * factor) / numConn;
 * 		else
 * 			score = MAXINT
 * 		end if
 * 		if score > maxScore then
 * 			maxScore = score
 * 			maxServer = server
 * 		end if
 * 	end for
 * 	if (maxServer not nil) return maxServer
 * </PRE><BR>
 * This scheme allows any unsigned 32 bit integer number to be 
 * chosen as a weight without danger of causing unpredictable behavior due
 * to integer overflow on multiplication.  Also, the proportional 
 * relationships between the server weights is preserved.
 * <P>
 * The prefix passed in must contain one or more path component.
 * Leading and trailing slash characters must be trimmed, and internal
 * sequences of two or more slash characters must be reduced to a single
 * slash.  The comparison of the passed-in prefix is done in a case
 * sensitive manner.
 * <P>
 * Finally, if sticky sessions are enabled for an application, the
 * client module will find a sticky session tag as part of the session
 * ID.  This tag is compared, up to a length of 8 bytes, to a stored
 * server "signature" for each server configured for the selected
 * application.  If a match is found, this function ALWAYS returns
 * that server, since sticky sessions will usually fail of they go
 * to a different server.
 * 
 * @param scoreboard   [in,out] Pointer to the scoreboard memory area.
 * @param prefix       [in]     Servlet prefix string, with no leading or
 * 			        trailing slash characters.
 * @param stickyTag    [in]     Sticky session tag.  NULL if none found.
 * @param app          [out]    Pointer to app list, filled in upon success.
 * @param server       [out]    Pointer to server list, filled in upon success.
 * @param refCounts    [in]     Array of int32 reference counts corresponding
 * 			        to the 'refIndex' field of each server.
 * @param numRefCounts [in]     Number of reference counts in 'refCounts'.
 * @param local_addr   [in]     The local socket address of this connection.
 * @param virt_host    [in]     Virtual host name to use in checking
 * 				restrictions.  Should be "Host" header value.
 * @param exc          [out]    Exception handle filled in if an error occurs.
 * @exception EDIR_EXCEPT_CONFIG
 * 		If incorrect or nonsensical configuration values are found
 * 		in the scoreboard.
 * @exception EDIR_EXCEPT_SERVLET
 * 		If an application and/or server cannot be found for the
 * 		given prefix
 */
extern void
edir_load_choose(void *scoreboard, char *prefix, char *stickyTag,
	         edir_load_sb_app_t **app, edir_load_sb_server_t **server, 
	         int32 *refCounts, int32 refCountsLen,  
	         struct sockaddr_in *local_addr,
	         struct sockaddr_in *remote_addr,
		 char *virt_host, edir_exception_t **exc)
{
    static char *		funcName = "edir_load_choose";
    int32			i, numServers, numConn, slotMod;
    uint32			score, maxScore; /* MUST be unsigned! */
    uint32			factor;
    uint32			weight, maxWeight;
    edir_load_sb_offsets_t	*sb = NULL;
    edir_load_sb_server_t	*servers, *serverSlot, *maxServer, *lrfServer;
    edir_load_sb_app_t		*matchedApp;
    int32			*serverIdxList;
    time_t			diff, now, lrfTime;
    EDIR_DEBUG_FUNCS_IN(("%s("EDIR_PTR", %s, %s, "EDIR_PTR", "EDIR_PTR", "EDIR_PTR", "\
			 "%d, "EDIR_PTR", \"%s\", "EDIR_PTR")",\
		         "func: edir_load_choose", scoreboard, \
		         prefix ? prefix : "(null)", \
		         stickyTag ? stickyTag : "(null)", \
		         app, server, refCounts, refCountsLen, \
			 local_addr, virt_host, *exc));

    if (!EDIR_SCOREBOARD_VALID(scoreboard)) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
	    "Invalid scoreboard pointer "EDIR_PTR"", scoreboard);
	EDIR_THROW(*exc);
    }

    /*
     * Set up scoreboard pointers.
     */
    sb = (edir_load_sb_offsets_t *) scoreboard;
    servers = EDIR_SB_SERVERS(sb);
    if (refCountsLen < sb->numServers) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
	    "Number of reference counts does not match scoreboard's");
	EDIR_THROW(*exc);
    }

    /*
     * Look up the application for the prefix.
     */
    matchedApp = edir_load_check_path(sb, prefix, local_addr,
				      remote_addr, virt_host);
    if (matchedApp == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET,
		    "No application configured for prefix '%s'", prefix);
	EDIR_THROW(*exc);
    } 

    /*
     * Grab the server list for this application, and compute the
     * weighted score for each one.  The server with the highest
     * score that is not "failing" is selected.
     */
    serverIdxList = EDIR_SB_SERVERINDEX(sb, matchedApp);
    /* If numServers is zero we'll get divide by zero without this check */
    if ((numServers = matchedApp->numServers) < 1) {
	*exc = edir_exception_create(EDIR_EXCEPT_CONFIG,
	    "No servers configured for prefix '%s'", matchedApp->appPrefix);
	EDIR_THROW(*exc);
    }
    /* Start search on different server each time to "round robin". */
    slotMod = matchedApp->serverCounter++ % numServers;
    maxWeight = 0;
    for (i=0; i < numServers; i++) {
	weight = (servers[serverIdxList[i]]).weight;
	if (weight > maxWeight) maxWeight = weight;
    }
    /* If maxWeight is zero we'll get divide by zero without this check */
    if (maxWeight < 1) {
	/*
	 * In this case all servers have weight zero, so their score will
	 * always be zero, regardless of maxWeight or scale.  Just fake
	 * maxWeight to 1 to prevent divide by zero.
	 */
	maxWeight = 1;
    }
    factor = (~((uint32)0)) / maxWeight;
    /*
     * Loop through all servers in circular fashion, starting at
     * "slotMod" position.  Determine which non-failing server has the
     * highest score.
     */
    i = slotMod;
    maxScore = 0;
    maxServer = NULL;
    lrfServer = NULL;
    lrfTime = -1;
    time(&now);
    do {
	serverSlot = &(servers[serverIdxList[i]]);
	/* Retry sticky server, even if failing */
	if (stickyTag && (!strncmp(serverSlot->stickyTag, stickyTag, 8))) {
	    /* Matched session affinity! Return this server. */
	    *app = matchedApp;
	    *server = serverSlot;
	    sb->counter++; /* Up global usage count. */
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
	    return;
	}
	/* Check if failing but fail timeout expired */
	if (serverSlot->failing) {
	    diff = EDIR_TIMEDIFF(now, serverSlot->failTime);
	    if ((sb->failRetry > 0) && (diff >= sb->failRetry)) {
		serverSlot->failing = 0;
	    } else {
		/* Don't do next good server twice. */
		if (maxServer == NULL) matchedApp->serverCounter++;
		/* Check 'least recently failing' */
		if (diff > lrfTime) {
		    lrfTime = diff;
		    lrfServer = serverSlot; /* Least recently failed. */
		}
		i = (i + 1) % numServers;
		continue;
	    }
	}
	if ((numConn = refCounts[serverSlot->refIndex]) > 0) {
	    score = (serverSlot->weight * factor) / numConn;
	} else {
	    /* If no current usage, give the higest possible score. */
	    /* Must do this to avoid divide-by-zero error. */
	    /*
	     * Modification: if weight is zero, score is always zero.
	     * This allows a weight of zero to mean "never go to
	     * this server except to fail over."
	     */
	    if (serverSlot->weight == 0)
		score = 0;
	    else
		score = (~((uint32)0));
	}
	/*
	 * Modification: Use '>=' instead of '>'.
	 * Variable 'maxScore' is unsigned, so it can't start at '-1'.
	 * The effect of this is that the last server of equal scores
	 * is used instead of the first.  Six one way, half dozen the other.
	 */
	if (score >= maxScore) {
	    maxScore = score;
	    maxServer = serverSlot;
	}
	i = (i + 1) % numServers;
    } while (i != slotMod);
    /* If no unfailed server, retry least recently failed server */
    if (maxServer == NULL) {
	if (lrfServer == NULL) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SERVLET,
					"Can't find any servers");
	    EDIR_THROW(*exc);
	}
	maxServer = lrfServer;
	maxServer->failing = 0; /* Retry a bit early. */
    }

    *app = matchedApp;
    *server = maxServer;

    sb->counter++; /* Up global usage count. */

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;	

except:
    /* Nothing to clean */
    if (sb) sb->counter++; /* Up global usage count. */
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

