/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_daemon.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#define EDIR_LOAD_DEFAULT_CGI_TIMEOUT 30

static char  saveDir[MAXPATHLEN + 1];
static octet failCode = (octet) 1;
static octet successCode = (octet) 0;
static int   alive = 1, restart = 0;
static int   use_cgi_timeout = 0;

/**
 * Signal handler for HUP signal.
 * <P>
 * This signal is used to tell the daemon to restart itself and 
 * re-read the configuration.
 * 
 * @param arg [in]  The number of the signal that invoked this handler.
 */
static void
handle_sighup(int arg)
{
    restart = 1;
}

/**
 * Signal handler for TERM signal.
 * <P>
 * This signal is used to tell the daemon to clean up and exit immediately.
 * 
 * @param arg [in]  The number of the signal that invoked this handler.
 */
static void
handle_sigterm(int arg)
{
    alive = 0;
}

/**
 * The main loop of the Enhydra Director daemon.
 * <P>
 * Loads the configuration, sets up the scoreboard, and continually
 * monitors the scoreboard, watching for inactivity, as evidenced by
 * the ability to immediately gain an exclusive lock on the TESTCLIENT
 * lock for the scoreboard.  
 * 
 * @param configFile [in]  The name of the configuration file, or NULL for
 *                         the default file.
 * @param configFile [in]  The name of the configuration file, or NULL for
 *                         the default file.
 */
static void
edir_load_daemon_main(char *configFile, char *dataFile, char *lockFile)
{
    static char			*funcName = "edir_load_daemon_main";
    edir_exception_t		*_exc=NULL, **exc=&_exc;
    edir_ipc_handle_t		*shmemKey = NULL, *mutexKey = NULL;
    edir_ipc_handle_t		*mutex = NULL, *shmem = NULL, *refcnt = NULL;
    edir_load_sb_offsets_t	*sb = NULL;
    edir_load_sb_app_t		*apps = NULL;
    edir_load_sb_server_t	*servers = NULL;
    edir_conf_t			*appcfg = NULL;
    int				n, locked, shmSize, status, mutexexists=0;
    uint32			lastCounter = 0;
    time_t			now, lastUsed;
    time_t			timeout = 0;
    time_t			interval = 1;
    time_t			timediff = 0;
    time_t			failRetry = 30; /* default 30 sec. */

    if (use_cgi_timeout) timeout = EDIR_LOAD_DEFAULT_CGI_TIMEOUT;

    /*
     * Read the configuration file.
     */
    chdir(saveDir);
    if (configFile == NULL) configFile = EDIR_CONF_DEFAULT_CONFIG_FILE;
    appcfg = edir_conf_process(configFile, exc);
/*XXX    chdir("/"); */
    if (*exc) EDIR_THROW(*exc);

    if (appcfg->daemonTimeout >= 0) timeout = appcfg->daemonTimeout;
    if (appcfg->daemonInterval >= 0) interval = appcfg->daemonInterval;
    if (appcfg->daemonDebug) edir_debug |= appcfg->daemonDebugMask;
    if (appcfg->failRetrySet) failRetry = appcfg->failRetry;

    /* 
     * Find out how big to make scoreboard.
     */
    shmSize = edir_load_sb_compute_size(appcfg);

    /*
     * Create the global lock object.  This will have three locks,
     * the TESTDAEMON lock (#0), the TESTCLIENT lock (#1), and the
     * GLOBAL lock (#2).  All are exclusively locked by this program.
     */
    if (lockFile) mutexKey = edir_ipc_make_config_handle(lockFile);
    mutex = edir_mutex_create(mutexKey, EDIR_LOAD_SB_NUM_LOCKS, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * Create the scoreboard shared memory region.
     */
    if (dataFile) shmemKey = edir_ipc_make_config_handle(dataFile);
    if (shmemKey)
	shmem = edir_shmem_create(shmemKey, shmSize, edir_load_sb_cleanup, exc);
    else
	shmem = edir_shmem_create(mutex, shmSize, edir_load_sb_cleanup, exc);
    if (*exc) EDIR_THROW(*exc);
    sb = (edir_load_sb_offsets_t *) edir_shmem_handle_get_object(shmem);
    if (sb == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON, "Got a NULL scoreboard!");
	EDIR_THROW(*exc);
    }

    /*
     * Fill in the scoreboard with initial configuration data
     * XXX FIX ME Should free config data.
     */
    edir_load_sb_from_config(appcfg, (void *)sb);
    apps = EDIR_SB_APPS(sb);
    servers = EDIR_SB_SERVERS(sb);

    /*
     * Obtain reference count handle for this scoreboard.
     * XXX - FIX - Always takes a scoreboard shmem, so this function
     * really ought to figure out the number of servers by itself.
     */
    refcnt = edir_refcount_create(shmem, sb->numServers, exc);

    /*
     * Initialize timestamps while we have the exclusive locks.
     */
    sb->daemonTTL = timeout;
    sb->daemonInterval = interval;
    time(&lastUsed);
    sb->counter = 0;

    /*
     * Make sure the "scoreboard invalid" flag is not set.
     */
    sb->dead = 0;

    /*
     * Set the "retry timer" parameter.
     */
    sb->failRetry = failRetry;

    /*
     * Unlock the locks and open the doors for business.
     * KEEP the TESTDAEMON lock, because that is what the clients
     * use to see if we're alive or not.
     */
    edir_mutex_unlock(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, exc);
    if (*exc) EDIR_THROW(*exc);
    edir_mutex_unlock(mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
    if (*exc) EDIR_THROW(*exc);
    
    /*
     * Now, send the OK status back to whoever spawned this daemon.
     */
    if (!isatty(0)) {
	while (((n = write(1, &successCode, 1)) < 0) && (errno == EINTR)) ;
	if (n < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
			"Can't send success code: @errstr@");
	    EDIR_THROW(*exc);
	}
	if (n == 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
			"Can't send success code: write() sent 0 bytes");
	    EDIR_THROW(*exc);
	}
    }

    /*
     * Wait for spawner to signal that it has acquired the first
     * TESTCLIENT lock.  This is important for implementations like
     * Apache modules where our timeout is zero and we want to 
     * make sure the daemon doesn't die before it is used.
     */
    if (!isatty(0)) {
	while (((n = read(0, &status, 1)) < 0) && (errno != EINTR)) ;
	if (n < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
			"Can't read final ACK from parent: @errstr@");
	    EDIR_THROW(*exc);
	}
	if (n == 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
			"Can't read final ACK from parent: got 0 bytes");
	    EDIR_THROW(*exc);
	}
    }

    /*
     * Close stdin, stdout, and stderr, which are the pipe connections
     * to the parent.  At this point we are detached and we go into
     * the main loop.
     */
    if (!isatty(0)) {
	close(0);
	close(1);
	close(2);
	open("/dev/null", O_RDONLY); /* Stdin is /dev/null */
	open("/dev/null", O_WRONLY); /* Stdout is /dev/null */
	dup(1); /* Stderr is /dev/null */
    }

    /*
     * Set nonlocking mode on the TESTCLIENT lock.
     * If we try to wait for an exclusive lock on the TESTCLIENT
     * mutex then new clients may block until we get and
     * release the mutex, and this will take effectively
     * forever because existing clients never release the
     * watchdog lock until done.  The correct method is to
     * periodically try for an exclusive lock in nonblocking mode.
     */
    edir_mutex_setblocking(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, 0);

    /*
     * The daemon's main loop.
     */
    while (alive) {
	locked = edir_mutex_lock_exclusive(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, exc);
	if (*exc) EDIR_THROW(*exc);
	if (locked) {
	    EDIR_DEBUG_EVERYTHING(("Got watch lock.  Scoreboard not in use."));
	    if (!alive) {
		EDIR_DEBUG_SPARSE(("Daemon terminated with SIGTERM."));
		break; /* Don't release the lock. */
	    }
	    /* First make sure there was no usage during last interval */
	    if (sb->counter != lastCounter) {
		EDIR_DEBUG_EVERYTHING(("Use counter changed.  %s", \
				     "Scoreboard WAS used."));
		time(&lastUsed);
		lastCounter = sb->counter;
	    }
	    time(&now);
	    timediff = EDIR_TIMEDIFF(now, lastUsed);
	    if (timediff >= timeout) {
		EDIR_DEBUG_SPARSE(("Timeout expired.  Daemon exiting."));
		sb->daemonTTL = 0;
		break; /* Don't release the lock. */
	    } else {
		sb->daemonTTL = timeout - timediff;
	    }
	    edir_mutex_unlock(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, exc);
	    if (*exc) EDIR_THROW(*exc);
	} else {
	    EDIR_DEBUG_EVERYTHING(("Watch lock busy.  Scoreboard is in use."));
	    if (lastCounter != sb->counter) time(&lastUsed);
	    lastCounter = sb->counter;
	    time(&now);
	    timediff = EDIR_TIMEDIFF(now, lastUsed);
	    if (timediff >= timeout) {
		sb->daemonTTL = 0;
	    } else {
		sb->daemonTTL = timeout - timediff;
	    }
	}
	if (sb->daemonInterval > 0) sleep(sb->daemonInterval);
    }

    edir_log("Daemon shutting down.");
    sb->dead = 1;
    if (refcnt) edir_refcount_destroy(refcnt, exc);
    if (*exc) edir_exception_free(exc);
    if (shmem) edir_shmem_destroy(shmem, exc);
    if (*exc) edir_exception_free(exc);
    if (mutex) {
	edir_mutex_unlock(mutex, EDIR_MUTEX_UNLOCK_ALL, exc);
	if (*exc) edir_exception_free(exc);
	edir_mutex_destroy(mutex, exc);
	if (*exc) edir_exception_free(exc);
    }
    if (mutexKey) edir_ipc_destroy_config_handle(mutexKey);
    if (shmemKey) edir_ipc_destroy_config_handle(mutexKey);
    if (*exc) edir_exception_free(exc);
    exit(0);

except:
    if (sb) sb->dead = 1; /* Always flag sb dead on exit */
    if (*exc) {
	char *trace = NULL;
	mutexexists = ((*exc)->id == EDIR_EXCEPT_MUTEX_EXISTS);
	/* Mask "mutex exists" errors.  They're quite normal */
	if (mutexexists) {
	    EDIR_DEBUG_VERBOSE(("Mutex already exists while starting daemon"));
	    EDIR_DEBUG_EVERYTHING(("%s", ((*exc) && \
				(trace = edir_exception_stack_trace(*exc)) ? \
				trace : "(No exception trace)")));
	} else {
	    edir_log_error("Exception in Enhydra Director daemon");
	    edir_log_error("%s", ((*exc) &&
				(trace = edir_exception_stack_trace(*exc)) ? 
				trace : "(No exception trace)"));
	}
	if (trace) edir_free(trace);
	edir_exception_free(exc);
    }
    if (!isatty(0)) {
	if (mutexexists) {
	    /* Looks like someone else already beat us to the draw */
	    /* Just "lie" and tell the spawner to try for the mutex */
	    while (((n = write(1, &successCode, 1)) < 0) && (errno == EINTR)) ;
	    if (n == 1)
		while (((n = read(0, &status, 1)) < 0) && (errno == EINTR)) ;
	} else {
	    while (((n = write(1, &failCode, 1)) < 0) && (errno == EINTR)) ;
	}
    }
    if (refcnt) edir_refcount_destroy(refcnt, exc);
    if (*exc) edir_exception_free(exc);
    if (shmem) edir_shmem_destroy(shmem, exc);
    if (*exc) edir_exception_free(exc);
    if (mutex) {
	edir_mutex_unlock(mutex, EDIR_MUTEX_UNLOCK_ALL, exc);
	if (*exc) edir_exception_free(exc);
	edir_mutex_destroy(mutex, exc);
	if (*exc) edir_exception_free(exc);
    }
    if (mutexKey) edir_ipc_destroy_config_handle(mutexKey);
    if (shmemKey) edir_ipc_destroy_config_handle(mutexKey);
    exit(0);
}

/**
 * Usage message generator.
 * <P>
 * Side effect:  Calls exit() to terminate the program.
 * 
 * @param progName [in] Name of the program to put in the message.
 */
static void
usage(char *progName)
{
    edir_log_error("Usage: %s [-D <DebugMask>] [-d <DataFile>] %s",
		    progName ? progName : "edir_daemon",
		    "[-f <ConfigFile>] [-l <LockFile>]");
    while ((write(1, &failCode, 1) < 0) && (errno == EINTR)) ;
    exit(1);
}


/**
 * Main entry point into the Enhydra Director daemon program.
 * <P>
 * Parses command line arguments and environment variables
 * before calling the main loop.
 * 
 * @param argc [in]  Number of command line arguments.
 * @param argv [in]  Array of command line arguments of length "argc".
 * @param envp [in]  NULL terminated array of character pointers which
 *                   point to string of the form "VAR=VALUE".
 * @return	     Integer exit code.  Zero indicates success and nonzero
 *                   indicates failure. 
 */
extern int
main(int argc, char *argv[], char *envp[])
{
    static char		*funcName = "main";
    char		*progName;
    edir_exception_t	*_exc = NULL, **exc = &_exc;
    edir_load_handles_t	*handles = NULL;
    char		*cp, ch;
    char		*buf, *name, *value;
    int			len, buflen, i;
    char		*debugStr = NULL;
    char		*configFile = NULL;
    char		*dataFile = NULL;
    char		*lockFile = NULL;

    /*
     * Get basename of arg0 for progName 
     */
    progName = strrchr(argv[0], '/');
    if (progName) progName++;
    else progName = argv[0];

    /*
     * We're running as a daemon, so
     */
    if (!isatty(0)) edir_log_setdaemon(progName, 1);

#ifdef HAVE_GETCWD
    if (getcwd(saveDir, MAXPATHLEN) != NULL) {
	saveDir[MAXPATHLEN] = '\0';
	chdir("/");
    } else {
	strcpy(saveDir, ".");
    }
#else
    strcpy(saveDir, ".");
#endif

    /*
     * We DON'T want to get terminated because of write errors on pipes.
     */
    signal(SIGPIPE, SIG_IGN);
    signal(SIGHUP, handle_sighup);
    signal(SIGTERM, handle_sigterm);

    /*
     * Allow hard-coded config parameters using #define macros
     * at the beginning of this file.  This is useful for a last-ditch
     * case where there is no clean way to pass arguments to 
     * a CGI program.
     */
#ifdef EDIR_DEBUG_MASK
    edir_debug = EDIR_DEBUG_MASK;
#endif
#ifdef EDIR_CONFIG_FILE
    configFile = EDIR_CONFIG_FILE;
#endif
#ifdef EDIR_DATA_FILE
    dataFile = EDIR_DATA_FILE;
#endif
#ifdef EDIR_LOCK_FILE
    lockFile = EDIR_LOCK_FILE;
#endif

    /*
     * Get file options from environment first, then override
     * from the command line.  This will allow some CGI
     * implementations to configure through environment variables
     * passed to the HTTP server.
     */ 
    if (cp = getenv("EDIR_DEBUG")) {
	edir_atoi(cp, &edir_debug);
	edir_debug &= ~EDIR_DEBUG_MASK_FAKE_IT;
    }
    if (cp = getenv("EDIR_CONFIG_FILE")) edir_conf_set_default_filename(cp);
    if (cp = getenv("EDIR_DATA_FILE")) edir_shmem_set_default_filename(cp);
    if (cp = getenv("EDIR_LOCK_FILE")) edir_mutex_set_default_filename(cp);

    /*
     * Command line arguments supercede all defaults.
     */
    while ((ch = getopt(argc, argv, "?CD:d:f:l:")) != -1) {
	switch (ch) {
	    case 'C':
		use_cgi_timeout = 1;
		break;
	    case 'D':
		edir_atoi(optarg, &edir_debug);
		edir_debug &= ~EDIR_DEBUG_MASK_FAKE_IT;
		break;
	    case 'd':
		dataFile = optarg;
		break;
	    case 'f':
		configFile = optarg;
		break;
	    case 'l':
		lockFile = optarg;
		break;
	    case '?':
	    default:
		usage(progName);
	}
    }
    argc -= optind;
    argv += optind;
    
    if (argc > 0) usage(progName);
    
    if (debugStr) edir_debug |= atoi(debugStr);

    edir_load_daemon_main(configFile, dataFile, lockFile);

    exit(0);
}
