/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_multiproc.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

static char *DefaultDaemonPath = EDIR_LOAD_DEFAULT_DAEMON_PATH;
static char *DefaultDaemonName = EDIR_LOAD_DEFAULT_DAEMON_NAME;

/**
 * The user ID of the daemon process, if we started as root.
 */
static uid_t	daemonUID = -1;
static int	daemonUIDSet = 0;

/**
 * The group ID of the daemon process, if we started as root.
 */
static gid_t	daemonGID = -1;
static int	daemonGIDSet = 0;

static edir_ipc_handle_t * 
edir_load_start_watchdog_parent(char *configFile, edir_ipc_handle_t *mutexKey,
			      char *dataFile, int rfd, int wfd,
			      int childPid, edir_exception_t **exc)
{
    static char		*funcName = "edir_load_start_watchdog_parent";
    static octet	failCode = (octet) 1;
    static octet	successCode = (octet) 0;
    edir_ipc_handle_t	*mutex = NULL;
    int			n, numMutexes;
    octet		status;
    void		(*savesig)(int);

    /*
     * wait for a "go ahead" byte from the read pipe.
     */
    while ((n = read(rfd, &status,  sizeof(status))) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		    "Watchdog failed to return status: @errstr@");
	    EDIR_THROW(*exc);
	}
    }
    if (n != sizeof(status)) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Unexpected short read of status byte (%d bytes)", n);
	EDIR_THROW(*exc);
    }
    
    if (status != successCode) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Watchdog init failed with status %d", status);
	EDIR_THROW(*exc);
    }
    
    /*
     * Startup was OK.  Obtain the mutex and wait for a shareable lock.
     */
    mutex = edir_mutex_obtain(mutexKey, &numMutexes, exc);
    if (*exc) EDIR_THROW(*exc);
    if (numMutexes != EDIR_LOAD_SB_NUM_LOCKS) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Expected %d mutexes, got %d",
		numMutexes, EDIR_LOAD_SB_NUM_LOCKS);
	EDIR_THROW(*exc);
    }
    
    /*
     * We got the mutex, wait for a shared lock on the TESTCLIENT mutex.
     */
    edir_mutex_lock_shared(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, exc);
    if (*exc) EDIR_THROW(*exc);
    
    /*
     * Write the "I'm OK too" byte to the pipe
     */
    savesig = signal(SIGPIPE, SIG_IGN);
    while ((write(wfd, &successCode, sizeof(successCode))<0)&&(errno==EINTR));
    signal(SIGPIPE, savesig);

    return mutex;

except:
    if (mutex) {
	edir_exception_t *exc2 = NULL;
	edir_mutex_release(mutex, &exc2);
	if (exc2) edir_exception_free(&exc2);
    }
    savesig = signal(SIGPIPE, SIG_IGN);
    while ((write(wfd, &failCode, sizeof(failCode)) < 0) && (errno == EINTR));
    signal(SIGPIPE, savesig);
    return NULL;
}

static void
edir_load_start_watchdog_child(char *configFile, char *daemonFile,
			     char *lockFile, char *dataFile)
{
    static char		*funcName = "edir_load_start_watchdog_child";
    static octet	failCode = (octet) 1;
    edir_exception_t	*_exc = NULL, **exc=&_exc;
    char		*prog = NULL, *base=NULL, *path=NULL;
    char		*argv[30];
    char		dbuf[16];
    char		*defLockFile=NULL, *defDataFile=NULL;
    
    int			argc = 0, n;
    int			save_errno;
    
    if ((prog = daemonFile) == NULL) prog = DefaultDaemonName;
    
    base = strrchr(prog, '/');
    if (base) {
	base++;
	path = NULL;
    } else {
	base = prog;
	path = getenv("PATH");
	if (!path) path = "/bin:/usr/bin";
	path = edir_asprintf("%s%s%s",
			   DefaultDaemonPath ? DefaultDaemonPath : "",
			   DefaultDaemonPath ? ":" : "", path);
    }

    /*
     * Yech.  For some modules, like apache, we want to be able to
     * alter the defaults to be reasonable locations like <serverroot>/...
     * This puts such defaults into the passed environment
     */
    defLockFile = edir_mutex_default_name();
    defDataFile = edir_shmem_default_name();
#if defined(HAVE_SETENV)
    if (defLockFile) setenv("EDIR_LOCK_FILE", defLockFile, 1);
    if (defDataFile) setenv("EDIR_DATA_FILE", defDataFile, 1);
#elif defined(HAVE_PUTENV)
    if (defLockFile) {
	char *tmp;
	tmp = edir_asprintf("EDIR_LOCK_FILE=%s", defLockFile);
	putenv(tmp);
    }
    if (defDataFile) {
	char *tmp;
	tmp = edir_asprintf("EDIR_DATA_FILE=%s", defDataFile);
	putenv(tmp);
    }
#endif
	
    argv[argc++] = prog;

    /* Pass debugging mask if set */
    if (edir_debug &~ EDIR_DEBUG_MASK_FAKE_IT) {
	sprintf(dbuf, "0x%08x", (edir_debug &~ EDIR_DEBUG_MASK_FAKE_IT));
	argv[argc++] = "-D";
	argv[argc++] = dbuf;
    }

#ifdef EDIR_CGI_MODULE
    argv[argc++] = "-C";
#endif
    argv[argc++] = "-f";
    argv[argc++] = configFile;
    if (lockFile) {
	argv[argc++] = "-l";
	argv[argc++] = lockFile;
    }
    if (dataFile) {
	argv[argc++] = "-d";
	argv[argc++] = dataFile;
    }
    argv[argc] = NULL; 
    
    /* First, try the program name as-is */
    execv(prog, argv);
    
    /* Now maybe try exec from PATH */
    if (path) {
#if defined(HAVE_SETENV)
	setenv("PATH", path, 1);
#elif defined(HAVE_PUTENV)
	char *tmp = path;
	path = edir_asprintf("PATH=%s", tmp);
	edir_free(tmp);
	putenv(path);
#endif
	execvp(base, argv);
    }
    *exc = edir_exception_create(EDIR_EXCEPT_DAEMON,
		"Exec of '%s' failed: @errstr@", prog);
    EDIR_THROW(*exc);
    
except:
    if (path) edir_free(path);
    if (*exc) {
	char *trace;
	trace = edir_exception_stack_trace(*exc);
	edir_log_error("%s", trace);
	edir_free(trace);
	edir_exception_free(exc);
    }
    while (((n = write(1, &failCode, sizeof(failCode))) < 0)&&(errno==EINTR));
    if (n == 0) edir_log_error("Wrote wrong # bytes (%d)", n);
    exit(1);
}

/**
 * Attempt to start up the watchdog daemon and obtain a shared
 * lock on the TESTCLIENT mutex.
 * <P>
 * If the scoreboard is not available, attempt to start up
 * the watchdog daemon (if applicable) and create the scoreboard
 * and its mutexes.
 * 
 * @param configFile [in]  The name of the configuration file to be passed
 *                         to the watchdog if necessary.
 * @param daemonFile [in]  Full or relative path to the daemon program.
 * @param lockFile   [in]  The name of the lock file.  This can be a file
 *                         name of a string IPC key number, depending on
 *                         the locking method in use.
 * @param dataFile   [in]  The name of the scoreboard file.  This can be a file
 *                         name of a string IPC key number, depending on
 *                         the locking method in use.  By default, the
 *                         scoreboard shares the same file as the lock file,
 *                         if such sharing is possible.
 * @param exc        [out] Exception pointer filled in if an error occurs.
 * @return                 A handle to the mutex for the new scoreboard,
 * 			   with a shared lock on the TESTCLIENT mutex.
 */
static edir_ipc_handle_t *
edir_load_start_watchdog(char *configFile, char *daemonFile, char *lockFile,
		       edir_ipc_handle_t *mutexKey,
		       char *dataFile, edir_exception_t **exc)
{
    static char		*funcName = "edir_load_start_watchdog";
    edir_ipc_handle_t	*mutex = NULL;
    int			childPid;
    int			rfd=-1, wfd=-1;

    /*
     * Conjure up a daemon.
     */
    childPid = edir_daemon_fork("edir_daemon", &rfd, &wfd, exc);
    if (*exc) EDIR_THROW(*exc);
    if (childPid < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_DAEMON, 
			"Daemon creation returned %d", childPid);
	EDIR_THROW(*exc);
    } else if (childPid == 0) {
	/* In daemon process.  rfd is always zero and wfd is always 1,2 */
	/*
	 * First, see if we need to become a 'nobody'
	 */
	if (daemonGIDSet) {
	    if ((geteuid() == 0) || (getuid() != geteuid())) {
		if (setgid(daemonGID) < 0) {
		    /* Be very anal retentive in root or setuid env! */
		    exit(1);
		}
		/* Even more anal retention in root case */
		if (getgid() != daemonGID) exit(2);
	    } else {
		/* If just normal non-root non-setuid user, go ahead and try
		 * The user can only run as himself anyway.
		 */
		setgid(daemonGID);
	    }
	}
	if (daemonUIDSet) {
	    if ((geteuid() == 0) || (getuid() != geteuid())) {
		if (setuid(daemonUID) < 0) {
		    /* Be very anal retentive in root or setuid env! */
		    exit(1);
		}
		/* Even more anal retention in root case */
		if (getuid() != daemonUID) exit(2);
	    } else {
		/* If just normal non-root non-setuid user, go ahead and try
		 * The user can only run as himself anyway.
		 */
		setuid(daemonUID);
	    }
	}
	edir_load_start_watchdog_child(configFile, daemonFile,
				     lockFile, dataFile);
	/* Should never get here -- we call exec() then exit(). */
	edir_log_error("Exec failed for daemon!");
	exit(1);
    } 
    mutex = edir_load_start_watchdog_parent(configFile, mutexKey, dataFile,
					  rfd, wfd, childPid, exc);
    if (*exc) EDIR_THROW(*exc);
    if (rfd >= 0) close(rfd);
    if (wfd >= 0) close(wfd);
    return mutex;

except:
    if (rfd >= 0) close(rfd);
    if (wfd >= 0) close(wfd);
    return NULL;
}

/**
 * Obtain the mutex object and scoreboard in order to access
 * load balancing and configuration data.
 * <P>
 * If the scoreboard is not available, attempt to start up
 * the watchdog daemon (if applicable) and create the scoreboard
 * and its mutexes.
 * 
 * @param configFile [in]  The name of the configuration file to be passed
 *                         to the watchdog if necessary.
 * @param daemonFile [in]  Full or relative path to the daemon program.
 * @param lockFile   [in]  The name of the lock file.  This can be a file
 *                         name of a string IPC key number, depending on
 *                         the locking method in use.
 * @param dataFile   [in]  The name of the scoreboard file.  This can be a file
 *                         name of a string IPC key number, depending on
 *                         the locking method in use.  By default, the
 *                         scoreboard shares the same file as the lock file,
 *                         if such sharing is possible.
 * @param exc        [out] Exception pointer filled in if an error occurs.
 * @return                 A set of handles to the mutex object, the shared
 *                         memory object, the refcount object, the scoreboard,
 *                         etc.
 */
extern edir_load_handles_t *
edir_load_init(char *configFile, char *daemonFile, char *lockFile,
	     char *dataFile, edir_exception_t **exc)
{
    static char * funcName = "edir_load_init";
    edir_load_sb_offsets_t *sb = NULL;
    edir_load_handles_t *handles = NULL;
    edir_ipc_handle_t *mutex = NULL,  *mutexKey = NULL;
    edir_ipc_handle_t *shmem = NULL,  *shmemKey = NULL;
    edir_ipc_handle_t *refcnt = NULL;
    int tries, numMutexes;
    
    if (lockFile) mutexKey = edir_ipc_make_config_handle(lockFile);
    if (dataFile) shmemKey = edir_ipc_make_config_handle(dataFile);
    
    /*
     * Get the mutex object for the scoreboard.  If we can't
     * get the mutex, maybe the watchdog daemon is dead, so try
     * to start one.  If that still fails, then give up.
     */
    mutex = edir_mutex_obtain(mutexKey, &numMutexes, exc);
    if (!(*exc) && (numMutexes != EDIR_LOAD_SB_NUM_LOCKS)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Expected %d mutexes, got %d",
		numMutexes, EDIR_LOAD_SB_NUM_LOCKS);
	EDIR_THROW(*exc);
    }
    for (tries = EDIR_LOAD_DAEMON_MAX_TRIES; *exc && (tries > 0); tries--) {
	EDIR_IF_DEBUG( char *trace = NULL; )
	/* If an exception occurs, try to start a new watchdog */
	edir_exception_free(exc);
	mutex = edir_load_start_watchdog(configFile, daemonFile, lockFile,
				       mutexKey, dataFile, exc);
    
	EDIR_DEBUG_NORMAL(("Exception while retrying daemon startup"));
	EDIR_DEBUG_VERBOSE(("%s", ((*exc) && \
			 (trace = edir_exception_stack_trace(*exc))) ? \
			 trace : "(No exception trace)"));
	EDIR_IF_DEBUG( if (trace) edir_free(trace); )

	if (!(*exc)) break; /* Success! */
	edir_exception_free(exc);
	sleep(2); /* Give possible other daemon starter some time */

	mutex = edir_mutex_obtain(mutexKey, &numMutexes, exc);
	if (!(*exc) && (numMutexes != EDIR_LOAD_SB_NUM_LOCKS)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Expected %d mutexes, got %d",
		    numMutexes, EDIR_LOAD_SB_NUM_LOCKS);
	    EDIR_THROW(*exc);
	}
	if (*exc && edir_debug) {
	    char *trace;
	    trace = edir_exception_stack_trace(*exc);
	    edir_log_debug("%s", trace);
	    edir_free(trace);
	}
    }
    if (*exc) EDIR_THROW(*exc);

    /*
     * Since we're not the daemon, we only ever try to get shareable
     * locks in nonblocking mode on the TESTDAEMON lock, as a way of
     * checking whether the daemon is alive.
     */
    edir_mutex_setblocking(mutex, EDIR_LOAD_SB_LOCK_TESTDAEMON, 0);
    
    /*
     * Obtain a shareable lock on the "watchdog" mutex. This
     * mutex acts as a blind reference count.  The watchdog
     * daemon periodically attempts to lock this mutex
     * exclusively as evidence that nobody is interested in
     * the scoreboard.
     */
    edir_mutex_lock_shared(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, exc);
    if (*exc) EDIR_THROW(*exc);
    
    /*
     * Try to get the memory region containing the scoreboard.
     * On some servers, this will be an interprocess shared memory
     * region, and on other (thread based) servers, this will
     * just be a malloc-ed region shared among threads.
     */
    if (shmemKey == NULL)
	shmem = edir_shmem_obtain(mutex, exc);
    else
	shmem = edir_shmem_obtain(shmemKey, exc);
    if (*exc) EDIR_THROW(*exc);
    
    sb = (edir_load_sb_offsets_t *) edir_shmem_handle_get_object(shmem);
    if (!EDIR_SCOREBOARD_VALID(sb)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid scoreboard");
	EDIR_THROW(*exc);
    }

    /*
     * Add any additional debugging bits from the scoreboard debug mask,
     * by OR-ing them with the current debug mask.
     */
    if (sb->debug) edir_debug |= sb->debugMask;
    
    refcnt = edir_refcount_obtain(shmem, exc);
    if (*exc) EDIR_THROW(*exc);
    
    if (mutexKey) edir_ipc_destroy_config_handle(mutexKey);
    if (shmemKey) edir_ipc_destroy_config_handle(shmemKey);
    
    handles = (edir_load_handles_t *) edir_malloc(sizeof(edir_load_handles_t));
    handles->configFile = configFile ? edir_strdup(configFile) : NULL;
    handles->daemonFile = daemonFile ? edir_strdup(daemonFile) : NULL;
    handles->lockFile = lockFile ? edir_strdup(lockFile) : NULL;
    handles->dataFile = dataFile ? edir_strdup(dataFile) : NULL;
    handles->mutex = mutex;
    handles->shmem = shmem;
    handles->refcnt = refcnt;
    handles->scoreboard = sb;
    handles->apps = EDIR_SB_APPS(sb);
    handles->servers = EDIR_SB_SERVERS(sb);
    handles->destroyOnShutdown = 0;
    
    return handles;
    
except:
    if (refcnt) {
	edir_exception_t *exc2 = NULL;
	edir_refcount_release(refcnt, &exc2);
	if (exc2) edir_exception_free(&exc2);
    }
    if (shmem) {
	edir_exception_t *exc2 = NULL;
	edir_shmem_release(shmem, &exc2);
	if (exc2) edir_exception_free(&exc2);
    }
    if (mutex) {
	edir_exception_t *exc2 = NULL;
	edir_mutex_release(mutex, &exc2);
	if (exc2) edir_exception_free(&exc2);
    }
    if (mutexKey) edir_ipc_destroy_config_handle(mutexKey);
    if (shmemKey) edir_ipc_destroy_config_handle(shmemKey);
    return NULL;
}

/**
 * Release all mutexes and shared memory resources for the scoreboard.
 * 
 * @param handles [in]  The "handles" return value fromm edir_load_init.
 * @param exc     [out] Exception pointer filled in if an error occurs.
 */
extern void
edir_load_shutdown(edir_load_handles_t *handles, edir_exception_t **exc)
{
    static char		*funcName = "edir_load_shutdown";
    edir_exception_t	*_ex2=NULL, **ex2=&_ex2;
    int			unlocked = 0;

    if (!handles) {
	*exc = edir_exception_create(EDIR_EXCEPT_NULL_POINTER, "Null handles");
	EDIR_THROW(*exc);
    }

    /* 
     * Bump the scoreboard use counter.  No locking here.  There is
     * an assumption here that even if there is a race condition
     * the counter value will be _different_ than it was before,
     * and that's all the 'edir_daemon' cares about.  Normally the
     * result of a race condition with N threads is that the
     * counter gets bumped one time instead of N times.  Since
     * this counter is informational only, both to people and to
     * edir_daemon, it's not worth the overhead of a mutex protecting it.
     */
    if (handles->scoreboard) handles->scoreboard->counter++;

    if (handles->destroyOnShutdown) {
	if (handles->refcnt) edir_refcount_destroy(handles->refcnt, exc);
    } else {
	if (handles->refcnt) edir_refcount_release(handles->refcnt, exc);
    }
    handles->refcnt = NULL;
    if (*exc) EDIR_THROW(*exc);

    if (handles->destroyOnShutdown) {
	if (handles->shmem) {
	    edir_shmem_set_destroy_func(handles->shmem, edir_load_sb_cleanup);
	    edir_shmem_destroy(handles->shmem, exc);
	}
    } else {
	if (handles->shmem) edir_shmem_release(handles->shmem, exc);
    }
    handles->shmem = NULL;
    handles->scoreboard = NULL;
    handles->apps = NULL;
    handles->servers = NULL;
    if (*exc) EDIR_THROW(*exc);

    if (handles->mutex)
	edir_mutex_unlock(handles->mutex, EDIR_MUTEX_UNLOCK_ALL, exc);
    unlocked = 1;
    if (*exc) EDIR_THROW(*exc);

    if (handles->destroyOnShutdown) {
	if (handles->mutex) edir_mutex_destroy(handles->mutex, exc);
    } else {
	if (handles->mutex) edir_mutex_release(handles->mutex, exc);
    }
    handles->mutex = NULL;
    if (*exc) EDIR_THROW(*exc);

    if (handles->configFile) edir_free(handles->configFile);
    if (handles->daemonFile) edir_free(handles->daemonFile);
    if (handles->lockFile) edir_free(handles->lockFile);
    if (handles->dataFile) edir_free(handles->dataFile);
    memset((char *)handles, 0, sizeof(edir_load_handles_t));

    return;

except:
    if (handles) {
	if (handles->destroyOnShutdown) {
	    if (handles->refcnt) edir_refcount_destroy(handles->refcnt, ex2);
	} else {
	    if (handles->refcnt) edir_refcount_release(handles->refcnt, ex2);
	}
	if (*ex2) edir_exception_free(ex2);
	if (handles->destroyOnShutdown) {
	    edir_shmem_set_destroy_func(handles->shmem, edir_load_sb_cleanup);
	    if (handles->shmem) edir_shmem_destroy(handles->shmem, ex2);
	} else {
	    if (handles->shmem) edir_shmem_release(handles->shmem, ex2);
	}
	if (*ex2) edir_exception_free(ex2);
	if ((handles->mutex) && (!unlocked))
	    edir_mutex_unlock(handles->mutex, EDIR_MUTEX_UNLOCK_ALL, ex2);
	if (*ex2) edir_exception_free(ex2);
	if (handles->destroyOnShutdown) {
	    if (handles->mutex) edir_mutex_destroy(handles->mutex, ex2);
	} else {
	    if (handles->mutex) edir_mutex_release(handles->mutex, ex2);
	}
	if (*ex2) edir_exception_free(ex2);
	if (handles->configFile) edir_free(handles->configFile);
	if (handles->daemonFile) edir_free(handles->daemonFile);
	if (handles->lockFile) edir_free(handles->lockFile);
	if (handles->dataFile) edir_free(handles->dataFile);
	memset((char *)handles, 0, sizeof(edir_load_handles_t));
    }
    return;
}

/**
 * Check the scoreboard for death or abandonment by a crashed or killed daemon.
 * <P>
 * If scoreboard died normally (sb->dead != 0),
 * then just release the scoreboard and re-obtain.
 * <P>
 * To test for abandonment, we try to get a shared lock on the TESTDAEMON
 * lock, which the daemon always keeps locked exclusively.  This
 * attempt is non-blocking.  If the lock is unavaliable, the daemon
 * is still alive and is holding the exclusive lock.
 * <P>
 * If we get the shared lock, we may be in company of N other processes
 * that also got the same shared lock.  All of these processes "know" the
 * daemon is dead.  We only want ONE to stop and restart the daemon.  At this
 * point, we lock the GLOBAL lock, _exclusively_ and in blocking mode.
 * In all other places where the GLOBAL lock is locked, it is
 * given back almost immediately (after time(&somevar) usually).  
 * <P>
 * Once we get the GLOBAL lock, check the (sb->dead) flag.  If it is
 * nonzero, then someone else has already killed the old scoreboard and
 * we just detach and reattach.  Otherwise, set the "destroy function" 
 * for the shared memory (which marks sb->dead AFTER unlink), and then
 * try to reattach using edir_load_init().
 * <P>
 * Once the scoreboard has been destroyed and marked dead, mutex object
 * is destroyed as well, which releases the GLOBAL lock.  Other
 * waiting processes discover the scoreboard to already be dead, detach,
 * and attempt reattach using the usual startup protocol.
 * 
 * @param handles [in]  The "handles" return value fromm edir_load_init.
 * @param exc     [out] Exception pointer filled in if an error occurs.
 * @return A pointer to either the old handles block, or a new one if
 * 	the scoreboard was reattached.
 */
extern edir_load_handles_t *
edir_load_check(edir_load_handles_t *handles, edir_exception_t **exc)
{
    static char		*funcName = "edir_load_check";
    char		*confFile = NULL, *daemonFile = NULL;
    char		*lockFile = NULL, *dataFile = NULL;
    int			owner, dead;

    if (!handles) {
	*exc = edir_exception_create(EDIR_EXCEPT_NULL_POINTER, "Null handles");
	EDIR_THROW(*exc);
    }

    /*
     * Try for the TESTDAEMON lock.  If we get it, the daemon has kicked 
     * the bucket and we try to restart it.
     */
    dead = handles->scoreboard->dead;

    if (!dead) {
	owner = edir_mutex_lock_shared(handles->mutex,
				     EDIR_LOAD_SB_LOCK_TESTDAEMON,exc);
	if (*exc) EDIR_THROW(*exc);
        if (!owner) return handles;  /* if Daemon OK, return */
    }

    /* D'oh!  The daemon has crashed or has been killed! */
    /* Or we picked up a stale, leftover scoreboard */
    if (owner) {
	EDIR_DEBUG_SPARSE(("Got TESTDAEMON lock -- daemon has terminated!"));
	/*
	 * This lock is always in _blocking_ mode.
	 */
	edir_mutex_lock_exclusive(handles->mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
	if (*exc) EDIR_THROW(*exc);
    } else {
	EDIR_DEBUG_SPARSE(("Scoreboard is dead -- Will obtain new one."));
    }

    /*
     * Since we are about to destroy the old handles block, we move
     * the old config strings out of the block so they won't be freed.
     * This code is exactly why the strings were saved there in
     * the first place, so no one else will be disturbed by this action.
     */
    confFile = handles->configFile;
    handles->configFile = NULL;
    daemonFile = handles->daemonFile;
    handles->daemonFile = NULL;
    lockFile = handles->lockFile;
    handles->lockFile = NULL;
    dataFile = handles->dataFile;
    handles->dataFile = NULL;;

    /*
     * Check sb->dead to see if we need to restart. 
     */
    if (handles->scoreboard->dead) {
	EDIR_DEBUG_SPARSE(("Scoreboard is dead -- will use new one (if any)"));
	sleep(1); /* Give restarter a little head start. */
    } else {
	edir_log_error("Warning: %s",
		     "Enhydra Director daemon terminated -- restarting...");
	/* Destroy IPC objects on shutdown instead of just releasing them */
	handles->destroyOnShutdown = 1;
    }

    edir_load_shutdown(handles, exc);
    /* Just ignore exceptions.  This scoreboard is toast anyway */
    if (*exc) edir_exception_free(exc);
    handles = edir_load_init(confFile, daemonFile, lockFile, dataFile, exc);
    if (*exc) EDIR_THROW(*exc);
    /* These strings were strdup-ed */
    if (confFile) edir_free(confFile);
    if (daemonFile) edir_free(daemonFile);
    if (lockFile) edir_free(lockFile);
    if (dataFile) edir_free(dataFile);

    return handles;

except:
    /* Just try to kill scoreboard handles and throw the exception. */
    if (handles) {
	edir_exception_t *exc2=NULL;
       	edir_load_shutdown(handles, &exc2);
	if (exc2) edir_exception_free(&exc2);
    }
    /* These strings were strdup-ed */
    if (confFile) edir_free(confFile);
    if (daemonFile) edir_free(daemonFile);
    if (lockFile) edir_free(lockFile);
    if (dataFile) edir_free(dataFile);
    return NULL;
}

/**
 * Allow a module to set up the "default" daemon program location.
 * 
 * @param prog [in] The name of the program to run as the daemon.
 */
extern void
edir_load_daemon_set_default_path(char *path)
{
    DefaultDaemonPath = edir_strdup(path);
}

/**
 * Set the UID of the watchdog daemon.  This is useful for
 * Apache, which initializes as root on most systems.
 * 
 * @param uid [in] The user ID to change to on init.
 */
extern void
edir_load_daemon_set_uid(uid_t uid)
{
    daemonUIDSet = 1;
    daemonUID = uid;
}

/**
 * Set the GID of the watchdog daemon.  This is useful for
 * Apache, which initializes as root on most systems.
 * 
 * @param gid [in] The group ID to change to on init.
 */
extern void
edir_load_daemon_set_gid(gid_t gid)
{
    daemonGIDSet = 1;
    daemonGID = gid;
}

