/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_scoreboard.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"
#include "edir_conf.h"
#include "edir_load_scoreboard.h"

/**
 * Calculate scoreboard size from configuration data.
 * <P>
 * Counts the entries in the specified configuration object and 
 * computes to total size in bytes needed to hold a position-independent
 * scoreboard block.  This function is needed to allow a scoreboard's
 * shared memory area to be allocated prior to filling it in.
 * 
 * @param conf [in] Configuration data that eventually will be copied to
 * 			the scoreboard.
 * @return Number of bytes required by the scoreboard.
 */
extern int32
edir_load_sb_compute_size(edir_conf_t *conf)
{
    int32				numApps = 0;
    int32				numServers = 0;
    int32				numStats = 0;
    int32				serversThisApp = 0;
    int32				restrictsThisApp = 0;
    int32				restrictsThisStats = 0;
    int32				nbytes;
    edir_conf_app_t			**app, *ap;
    edir_conf_stat_t			**stp, *st;
    edir_conf_appserver_t		**spp;
    edir_conf_restrict_t		**rpp;

    /*
     * Compute the number of app and server slots.
     */
    for (app = conf->apps; *app; app++) {
	numApps++;
	for (spp = (*app)->servers; *spp; spp++) numServers++;
    }

    /*
     * Compute the number of stats debugging prefixes.
     */
    for (stp = conf->stats; *stp; stp++) {
	numStats++;
    }

    /*
     * Compute offset and size of server list.
     * 128 bit aligned.
     */
    nbytes = ((sizeof(edir_load_sb_offsets_t) + 15) / 16) * 16;

    /*
     * Compute application list size and position.
     * 128 bit aligned.
     */
    nbytes += numServers * sizeof(edir_load_sb_server_t);
    nbytes = ((nbytes + 15) / 16) * 16; /* 128 bit align */

    /*
     * Compute the server list size and position.
     * 128 bit aligned.
     */
    nbytes += numApps * sizeof(edir_load_sb_app_t);
    nbytes = ((nbytes + 15) / 16) * 16; /* 128 bit align */

    /*
     * Compute the debug stats prefix list size and position.
     * 128 bit aligned.
     */
    nbytes += numStats * sizeof(edir_load_sb_stats_t);
    nbytes = ((nbytes + 15) / 16) * 16; /* 128 bit align */

    /*
     * Compute the size of the application server indexes and restrictions.
     */
    for (app = conf->apps; *app; app++) {
	ap = *app;

	/*
	 * Compute the size of the server index array for this app.
	 */
	serversThisApp = 0;
	for (spp = ap->servers; *spp; spp++) {
	    serversThisApp++;
	}
	nbytes += serversThisApp * sizeof(int32);
	nbytes = ((nbytes + 15) / 16) * 16; /* 128 bit align */

	/*
	 * Compute the size of the restrictions for this app.
	 */
	restrictsThisApp = 0;
	for (rpp = ap->restricts; *rpp; rpp++) {
	    restrictsThisApp++;
	}
	nbytes += restrictsThisApp * sizeof(edir_load_sb_restrict_t);
	nbytes = ((nbytes + 15) / 16) * 16; /* 128 bit align */
    }

    /*
     * Compute the size of the stats debugging prefixes and restricts.
     */
    for (stp = conf->stats; *stp; stp++) {
	st = *stp;
	/*
	 * Compute the size of the restrictions for this stats.
	 */
	restrictsThisStats = 0;
	for (rpp = st->restricts; *rpp; rpp++) {
	    restrictsThisStats++;
	}
	nbytes += restrictsThisStats * sizeof(edir_load_sb_restrict_t);
	nbytes = ((nbytes + 15) / 16) * 16; /* 128 bit align */
    }

    return nbytes;
}

/**
 * Copy configuration data into a presized and allocated scoreboard area.
 * <P>
 * The newly created scoreboard will be position-independent so that it
 * can be dynamically mapped as an interprocess shared memory segment
 * by client modules such as Apache and CGI.  For the sake of consistency,
 * we also use the same scoreboard format for single-processs environments
 * like NSAPI and ISAPI, since the slight overhead of position independence
 * is worth not having to maintain code for two or more scoreboard layouts.
 * 
 * @param conf       [in]  The configuration data to place into the scoreboard.
 * @param scoreboard [out] Presized and allocated scoreboard memory area.
 */
extern void
edir_load_sb_from_config(edir_conf_t *conf, void *scoreboard)
{
    int32				numApps = 0;
    int32				numServers = 0;
    int32				numStats = 0;
    int32				serversThisApp = 0;
    int32				restrictsThisApp = 0;
    int32				serverIdxOffset = 0;
    int32				restrictsThisStats = 0;
    int32				offset;
    edir_conf_app_t			**app, *ap;
    edir_conf_stat_t			**stp, *st;
    edir_conf_appserver_t		**spp, *sp;
    edir_conf_restrict_t		**rpp, *rp;
    edir_load_sb_offsets_t		*sb;
    edir_load_sb_server_t		*servers, *serverSlot;
    edir_load_sb_app_t			*apps, *appSlot;
    edir_load_sb_stats_t		*stats, *statSlot;
    edir_load_sb_restrict_t		*restricts, *restrictEntry;
    int32				*serverIdxList;
    char 				*sbPtr, *serverIdxPtr;

    /*
     * Initialize the scoreboard header area.
     */
    sbPtr = (char *) scoreboard;
    sb = (edir_load_sb_offsets_t *) scoreboard;
#ifndef EDIR_MULTITHREADED
    /* Clear scoreboard _after_ the mutex header (which is already filled) */
    memset(((char *)sb) + sizeof(edir_mutex_file_header_t), 0,
	   sizeof(edir_load_sb_offsets_t) - sizeof(edir_mutex_file_header_t));
    sb->mutexHeader.magic = EDIR_LOAD_SB_MAGIC;
    sb->daemonPid = getpid();
#else /* EDIR_MULTITHREADED */
    memset(sb, 0, sizeof(edir_load_sb_offsets_t));
#endif

    /*
     * Set the global configuration values.
     */
    sb->debug = conf->debug;
    sb->debugMask = conf->debugMask;
    if (conf->failRetrySet) {
	sb->failRetry = conf->failRetry;
    } else {
	sb->failRetry = EDIR_CONF_DEFAULT_FAIL_RETRY;
    }
    sb->daemonInterval = conf->daemonInterval;

    /*
     * Compute the number of app and server slots.
     */
    for (app = conf->apps; *app; app++) {
	numApps++;
	for (spp = (*app)->servers; *spp; spp++) numServers++;
    }

    /*
     * Compute the number of stats debugging prefixes.
     */
    for (stp = conf->stats; *stp; stp++) {
	numStats++;
    }

    /*
     * Compute server list size and position.
     * 128 bit aligned.
     */
    offset = ((sizeof(edir_load_sb_offsets_t) + 15) / 16) * 16;
    sb->serverListOffset = offset;
    sb->numServers = numServers;
    servers = (edir_load_sb_server_t *) (sbPtr + offset);

    /*
     * Compute application list size and position.
     * This is immediately after the server list.
     * 128 bit aligned.
     */
    offset = sb->serverListOffset;
    offset += numServers * sizeof(edir_load_sb_server_t);
    offset = ((offset + 15) / 16) * 16; /* 128 bit align */
    sb->appListOffset = offset;
    sb->numApps = numApps;
    apps = (edir_load_sb_app_t *) (sbPtr + offset);

    /*
     * Compute debug stats prefix list size and position.
     * This is immediately after the application list.
     * 128 bit aligned.
     */
    offset = sb->appListOffset;
    offset += numApps * sizeof(edir_load_sb_app_t);
    offset = ((offset + 15) / 16) * 16; /* 128 bit align */
    sb->statsListOffset = offset;
    sb->numStats = numStats;
    stats = (edir_load_sb_stats_t *) (sbPtr + offset);

    /*
     * Compute the beginning of the server index and restriction list area.
     * This is immediately after the debug stats prefix list.
     * 128 bit aligned.
     */
    offset = sb->statsListOffset;
    offset += numStats * sizeof(edir_load_sb_stats_t);
    offset = ((offset + 15) / 16) * 16; /* 128 bit align */
    serverIdxOffset = offset;
    serverIdxPtr = sbPtr + offset;

    /*
     * Fill in the scoreboard.
     */
    for (numApps=0, numServers=0, app = conf->apps; *app; app++, numApps++) {
	ap = *app;
	appSlot = &(apps[numApps]);
	memset((char *)appSlot, 0, sizeof(edir_load_sb_app_t));
	strncpy(appSlot->appPrefix, ap->appPrefix, EDIR_NAME_MAX);
	appSlot->appPrefixLen = strlen(appSlot->appPrefix);
	if (appSlot->appPrefixLen > sb->maxAppPrefixLen) {
	    sb->maxAppPrefixLen = appSlot->appPrefixLen;
	}
	appSlot->serverCounter = 0;
	appSlot->serverIndexOffset = offset;
	serverIdxList = (int32 *) serverIdxPtr;
	serversThisApp = 0;
	for (spp = ap->servers; *spp; spp++) {
	    sp = *spp;
	    serverSlot = &(servers[numServers]);
	    memset((char *)serverSlot, 0, sizeof(edir_load_sb_server_t));
	    serverIdxList[serversThisApp] = numServers;
	    serverSlot->refCount = 0;
	    serverSlot->useCount = 0;
	    serverSlot->failCount = 0;
	    serverSlot->refIndex = numServers;
	    if (sp->authData)
		strncpy(serverSlot->authData, sp->authData, EDIR_AUTH_MAX);
	    else
		strcpy(serverSlot->authData, "");
	    strncpy(serverSlot->hostName, sp->hostName, EDIR_NAME_MAX);
	    strncpy(serverSlot->ipName, sp->ipName, 15);
	    serverSlot->port = sp->port;
	    memcpy(&(serverSlot->sin), &(sp->sin), sizeof(sp->sin));
	    edir_cookie_sticky_session_tag(sp->sin.sin_addr.s_addr,
		    sp->sin.sin_port, serverSlot->stickyTag);
	    serverSlot->weight = sp->weight;
	    serverSlot->weightIsDefault = sp->weightIsDefault;
	    serverSlot->failing = 0;
	    serverSlot->failTime = 0;
	    numServers++;
	    serversThisApp++;
	}
	appSlot->numServers = serversThisApp;
	offset += serversThisApp * sizeof(int32);
	offset = ((offset + 15) / 16) * 16; /* 128 bit align */
	serverIdxPtr = sbPtr + offset; /* Now 128-bit aligned */

	/*
	 * Position the restriction list right after the server indices
	 * for each app.  Be careful to include the case of 'no
	 * restrictions, in which case we still have one server entry
	 * due to the structure definition.
	 */
	appSlot->restrictOffset = offset;
	restricts = (edir_load_sb_restrict_t *) serverIdxPtr;
	restrictsThisApp = 0;
	for (rpp = ap->restricts; *rpp; rpp++) {
	    rp = *rpp;
	    restrictEntry = &(restricts[restrictsThisApp]);
	    memset((char *)restrictEntry, 0, sizeof(edir_load_sb_restrict_t));
	    restrictEntry->type = rp->type;
	    if (rp->hostName == NULL) {
		strcpy(restrictEntry->hostName, "");
	    } else {
		strncpy(restrictEntry->hostName, rp->hostName, EDIR_NAME_MAX);
	    }
	    restrictEntry->port = rp->port;
	    memcpy(&(restrictEntry->sin), &(rp->sin), sizeof(struct sockaddr_in));
	    restrictEntry->netBits = rp->netBits;
	    restrictEntry->netMask = rp->netMask;
	    restrictsThisApp++;
	}
	appSlot->numRestricts = restrictsThisApp;
	offset += restrictsThisApp * sizeof(edir_load_sb_restrict_t);
	offset = ((offset + 15) / 16) * 16; /* 128 bit align */
	serverIdxPtr = sbPtr + offset; /* Now 128-bit aligned */
    }

    /*
     * Add debug stats prefix entries.
     */
    for (numStats=0, stp = conf->stats; *stp; stp++, numStats++) {
	st = *stp;
	statSlot = &(stats[numStats]);
	memset((char *)statSlot, 0, sizeof(edir_load_sb_stats_t));
	strncpy(statSlot->statPrefix, st->statPrefix, EDIR_NAME_MAX);
	statSlot->statPrefixLen = strlen(statSlot->statPrefix);
	if (statSlot->statPrefixLen > sb->maxStatsPrefixLen) {
	    sb->maxStatsPrefixLen = statSlot->statPrefixLen;
	}
	/*
	 * Align the offset of the Restrictions list for this stats prefix.
	 */
	statSlot->restrictOffset = offset;
	restricts = (edir_load_sb_restrict_t *) serverIdxPtr;
	restrictsThisStats = 0;
	for (rpp = st->restricts; *rpp; rpp++) {
	    rp = *rpp;
	    restrictEntry = &(restricts[restrictsThisStats]);
	    memset((char *)restrictEntry, 0, sizeof(edir_load_sb_restrict_t));
	    restrictEntry->type = rp->type;
	    if (rp->hostName == NULL) {
		strcpy(restrictEntry->hostName, "");
	    } else {
		strncpy(restrictEntry->hostName, rp->hostName, EDIR_NAME_MAX);
	    }
	    restrictEntry->port = rp->port;
	    memcpy(&(restrictEntry->sin), &(rp->sin), sizeof(struct sockaddr_in));
	    restrictEntry->netBits = rp->netBits;
	    restrictEntry->netMask = rp->netMask;
	    restrictsThisStats++;
	}
	statSlot->numRestricts = restrictsThisStats;
	offset += restrictsThisStats * sizeof(edir_load_sb_restrict_t);
	offset = ((offset + 15) / 16) * 16; /* 128 bit align */
	serverIdxPtr = sbPtr + offset; /* Now 128-bit aligned */
    }
}

#ifndef EDIR_MULTITHREADED
/**
 * Clean up callback to set the "dead" flag on a scoreboard
 * after it is deleted but before it is detached.
 * 
 * @param ptr [in]  Pointer to scoreboard to clean up.
 * @param exc [out] Exception pointer if exception is thrown.
 */
static void
edir_load_sb_cleanup_impl(void *ptr, edir_exception_t **exc)
{
    if (ptr) ((edir_load_sb_offsets_t *)ptr)->dead = 1;
}
edir_shmem_callback_t edir_load_sb_cleanup = edir_load_sb_cleanup_impl;
#endif

/**
 * Copy this scoreboard's reference counts into an array of int32.
 * <P>
 * This is used when 'simple' reference counting is used, as opposed
 * to external reference counting using System V Semaphores or the like.
 * 
 * @param scoreboard [in]  Scoreboard whose reference counts are needed.
 * @param refCounts  [out] Pointer to an array of int32 to receive the 
 * 			   reference counts.
 * @param num	     [in]  Maximum number of reference counts that will fit
 * 			   in 'refCounts'.
 * @return		   The number of refCounts copied.
 */
extern int32
edir_load_sb_get_refcounts(void *scoreboard, int32 *refCounts, int32 num)
{
    edir_load_sb_server_t		*servers;
    int32				count, i;
    count = ((edir_load_sb_offsets_t *)scoreboard)->numServers;
    count = EDIR_MIN(count, num);
    servers = EDIR_SB_SERVERS((edir_load_sb_offsets_t *)scoreboard);
    for (i=0; i<count; i++) refCounts[i] = servers[i].refCount;
    return i;
}

/**
 * Returns the number of servers listed in this scoreboard
 * <P>
 * Returns the 'numServers' field of the server list.  This is also
 * the number of reference counts.
 * <P>
 * A 'server' is defined to be a unique combination of 'IP address' and
 * 'port number'.
 * 
 * @param scoreboard [in]  Scoreboard to check.
 * @return		   The number of servers in this scoreboard.
 */
extern int32
edir_load_sb_num_servers(void *scoreboard)
{
    return ((edir_load_sb_offsets_t *)scoreboard)->numServers;
}

/**
 * Convenience function for debugging: Log scoreboard info to standard log.
 * <P>
 * Logs the data currently contained in the scoreboard in a human-readable
 * format.  Useful for debugging.
 * 
 * @param scoreboard [in] Scoreboard whose data is to be dumped.
 * @param refCounts  [in] Optional array of reference counts.  If not specified,
 * 			  The scoreboards reference counts are used.
 * @param num        [in] The number of reference counts in 'refCounts'.
 */
extern void
edir_load_sb_debug_print(void *scoreboard, int32 *refCounts, int32 num)
{
    char *str;
    str = edir_load_sb_debug_string(scoreboard, refCounts, num);
    edir_log("%s", str);
    edir_free(str);
}

/**
 * Format the contents of a scoreboard into a human-readable string.
 * <P>
 * The formatted string is suitable for output to standard output
 * or to log files.
 * <P>
 * XXX - FIX - Might be nice to format to a standard XML DTD, so that
 * this information could be passed to other programs, like Enhydra
 * Admin, for instance. :)
 * 
 * @param scoreboard [in] Scoreboard whose data is to be dumped.
 * @param refCounts  [in] Optional array of reference counts.  If not specified,
 * 			  The scoreboards reference counts are used.
 * @param num        [in] The number of reference counts in 'refCounts'.
 * @return The formatted string.
 */
extern char *
edir_load_sb_debug_string(void *scoreboard, int32 *refCounts, int32 num)
{
    void *sbuf;
    char *str;
    sbuf = edir_load_sb_debug_sbuf(scoreboard, refCounts, num);
    str = edir_sbuf_tostring(sbuf, NULL);
    edir_sbuf_free(sbuf);
    return str;
}

/**
 * Format the contents of a scoreboard into a human-readable string.
 * <P>
 * This function formats the string into a stringbuffer object that
 * can later be converted into a string witout a large number of
 * copies.
 * <P>
 * XXX - FIX - Might be nice to format to a standard XML DTD, so that
 * this information could be passed to other programs, like Enhydra
 * Admin, for instance. :)
 * 
 * @param scoreboard [in] Scoreboard whose data is to be dumped.
 * @param refCounts  [in] Optional array of reference counts.  If not specified,
 * 			  The scoreboards reference counts are used.
 * @param num        [in] The number of reference counts in 'refCounts'.
 * @return The formatted string, in an 'sbuf' handle useable by the
 * 		'edir_sbuf...()' functions.
 */
extern void *
edir_load_sb_debug_sbuf(void *scoreboard, int32 *refCounts, int32 num)
{
    edir_exception_t			*_exc, **exc;
    int32				i, j;
    int32				*alloc_rc = NULL;
    edir_load_sb_offsets_t		*sb;
    edir_load_sb_server_t		*serverSlot, *servers;
    edir_load_sb_app_t			*appSlot, *apps;
    edir_load_sb_stats_t		*statSlot, *stats;
    edir_load_sb_restrict_t		*restricts;
    int32				*serverIdxList;
    char 				*comma;
    void				*sbuf;

    exc = &_exc;
    *exc = NULL;

    sbuf = edir_sbuf_create();

    edir_sbuf_appendf(sbuf, "Enhydra Director Runtime Configuration:\n");

    sb = (edir_load_sb_offsets_t *) scoreboard;
    apps = EDIR_SB_APPS(sb);
    servers = EDIR_SB_SERVERS(sb);
    stats = EDIR_SB_STATS(sb);

    if ((refCounts == NULL) || (num < 0)) {
	num = edir_load_sb_num_servers(scoreboard);
	alloc_rc = (int32 *) edir_malloc(num * sizeof(int32));
	edir_load_sb_get_refcounts(scoreboard, alloc_rc, num);
	refCounts = alloc_rc;
    }
    
    edir_sbuf_appendf(sbuf, "Global Settings:\n");
    if (sb->daemonPid > 0) 
#ifndef EDIR_MULTITHREADED
	edir_sbuf_appendf(sbuf, "    Daemon PID              = %d\n", sb->daemonPid);
    if (sb->daemonTTL >= 0) 
	edir_sbuf_appendf(sbuf, "    Daemon TTL              = %d %s          \n",
		sb->daemonTTL, 
		(sb->daemonTTL == 1) ? "second" : "seconds");
    if (sb->daemonInterval >  0)
	edir_sbuf_appendf(sbuf, "    Check frequency         = %d %s            \n",
		sb->daemonInterval, 
		(sb->daemonInterval == 1) ? "second" : "seconds");
#endif /* EDIR_MULTITHREADED */
    if (sb->failRetry >= 0) 
	edir_sbuf_appendf(sbuf, "    Fail retry              = %d %s          \n",
		sb->failRetry, 
		(sb->failRetry == 1) ? "second" : "seconds");
    edir_sbuf_appendf(sbuf, "    Global access count = %d\n", sb->counter);
    edir_sbuf_appendf(sbuf, "Application Servers:\n");
    for (i=0; i<sb->numServers; i++) {
	serverSlot = &(servers[i]);
	edir_sbuf_appendf(sbuf, "    Server slot #%d:           \n", i);
	edir_sbuf_appendf(sbuf, "        Host Name           = %s      \n",
		serverSlot->hostName);
	edir_sbuf_appendf(sbuf, "        Remote Address      = %s:%d      \n",
		serverSlot->ipName, (int) (serverSlot->port));
#ifdef PRINT_SIN
	edir_sbuf_appendf(sbuf, "        sin.sin_addr        = 0x%08x       \n",
		ntohl((int32) (serverSlot->sin.sin_addr.s_addr)));
	edir_sbuf_appendf(sbuf, "        sin.sin_port        = %d         \n",
		(int) ntohs(serverSlot->sin.sin_port));
#endif
	edir_sbuf_appendf(sbuf, "        Weight,Tag,Status   = %-7d%s %-8s ",
		serverSlot->weight, (serverSlot->weightIsDefault) ? "*" : " ",
		serverSlot->stickyTag);
	if (serverSlot->failing) {
	    if (sb->failRetry < 0) {
		edir_sbuf_strcatn(sbuf, "FAILED: Permanently\n");
	    } else {
		time_t tm;
		time(&tm);
		tm = EDIR_TIMEDIFF(tm, serverSlot->failTime);
		if (tm > sb->failRetry) tm = 0;
		else tm = sb->failRetry - tm;
		edir_sbuf_appendf(sbuf, "FAILED: Retry in %d %s\n",
			tm, (tm == 1) ? "second" : "seconds");
	    }
	} else {
	    edir_sbuf_strcatn(sbuf, "OK\n");
	}
	edir_sbuf_appendf(sbuf, "        Used,Failed,Refs    = %-8d %-8d ",
		serverSlot->useCount, serverSlot->failCount);
	if (refCounts && (serverSlot->refIndex < num)) {
	    edir_sbuf_appendf(sbuf, "%d       \n",
			    refCounts[serverSlot->refIndex]);
	} else {
	    edir_sbuf_appendf(sbuf, "???      \n");
	}
    }
    edir_sbuf_appendf(sbuf, "Applications:\n");
    for (i=0; i<sb->numApps; i++) {
	appSlot = &(apps[i]);
	edir_sbuf_appendf(sbuf, "    Application slot #%d:      \n", i);
	edir_sbuf_appendf(sbuf, "        URI Prefix          = /%s     \n",
		appSlot->appPrefix);
	edir_sbuf_appendf(sbuf, "        Server Counter      = %d      \n",
		appSlot->serverCounter);
	serverIdxList = EDIR_SB_SERVERINDEX(sb, appSlot);
	edir_sbuf_strcatn(sbuf, "        App Servers         = ");
	for (comma="", j=0; j<appSlot->numServers; j++, comma=", ") {
	    edir_sbuf_appendf(sbuf, "%s%d", comma, serverIdxList[j]);
	}
	edir_sbuf_strcatn(sbuf, "\n");
	restricts = EDIR_SB_RESTRICTS(sb, appSlot);
	for (j=0; j<appSlot->numRestricts; j++) {
	    edir_sbuf_strcatn(sbuf, "        Restrict To ");
	    if (restricts[j].type == EDIR_CONF_RESTRICT_VIRTUAL) {
		edir_sbuf_strcatn(sbuf, "VServer = ");
	    } else if (restricts[j].type == EDIR_CONF_RESTRICT_SERVER) {
		edir_sbuf_strcatn(sbuf, "Server  = ");
	    } else if (restricts[j].type == EDIR_CONF_RESTRICT_CLIENT) {
		edir_sbuf_strcatn(sbuf, "Client  = ");
	    } else {
		edir_sbuf_strcatn(sbuf, "??????  = ");
	    }
	    if (strlen(restricts[j].hostName) < 1) {
		edir_sbuf_strcatn(sbuf, "<Any>");
	    } else {
		edir_sbuf_appendf(sbuf, "%s", restricts[j].hostName);
	    }
	    if (restricts[j].type != EDIR_CONF_RESTRICT_VIRTUAL) {
		if (restricts[j].netBits < 32) {
		    edir_sbuf_appendf(sbuf, "/%d", restricts[j].netBits);
		}
	    }
	    if (restricts[j].port < 1) {
		edir_sbuf_strcatn(sbuf, ":<Any>\n");
	    } else {
		edir_sbuf_appendf(sbuf, ":%d\n", restricts[j].port);
	    }
	}
    }
    edir_sbuf_appendf(sbuf, "Status Prefixes:\n");
    for (i=0; i<sb->numStats; i++) {
	statSlot = &(stats[i]);
	edir_sbuf_appendf(sbuf, "    Status Prefix #%d:      \n", i);
	edir_sbuf_appendf(sbuf, "        URI Prefix          = /%s     \n",
		statSlot->statPrefix);
	restricts = EDIR_SB_STATRESTRICTS(sb, statSlot);
	for (j=0; j<statSlot->numRestricts; j++) {
	    edir_sbuf_strcatn(sbuf, "        Restrict To ");
	    if (restricts[j].type == EDIR_CONF_RESTRICT_VIRTUAL) {
		edir_sbuf_strcatn(sbuf, "VServer = ");
	    } else if (restricts[j].type == EDIR_CONF_RESTRICT_SERVER) {
		edir_sbuf_strcatn(sbuf, "Server  = ");
	    } else if (restricts[j].type == EDIR_CONF_RESTRICT_CLIENT) {
		edir_sbuf_strcatn(sbuf, "Client  = ");
	    } else {
		edir_sbuf_strcatn(sbuf, "??????  = ");
	    }
	    if (strlen(restricts[j].hostName) < 1) {
		edir_sbuf_strcatn(sbuf, "<Any>");
	    } else {
		edir_sbuf_appendf(sbuf, "%s", restricts[j].hostName);
	    }
	    if (restricts[j].type != EDIR_CONF_RESTRICT_VIRTUAL) {
		if (restricts[j].netBits < 32) {
		    edir_sbuf_appendf(sbuf, "/%d", restricts[j].netBits);
		}
	    }
	    if (restricts[j].port < 1) {
		edir_sbuf_strcatn(sbuf, ":<Any>\n");
	    } else {
		edir_sbuf_appendf(sbuf, ":%d\n", restricts[j].port);
	    }
	}
    }
    if (alloc_rc) edir_free(alloc_rc);
    return sbuf;
}

