/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_scoreboard.h,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#ifndef __EDIR_LOAD_SCOREBOARD_H__
#define __EDIR_LOAD_SCOREBOARD_H__ 1
#ifdef __cplusplus
extern "C" {
#endif

#define EDIR_LOAD_SB_MAGIC		(0x00feef00)
 
/*
 * Define the individual locks within the mutex object for this
 * scoreboard.
 */
#define EDIR_LOAD_SB_NUM_LOCKS		(3)
#define EDIR_LOAD_SB_LOCK_TESTDAEMON	(0)
#define EDIR_LOAD_SB_LOCK_TESTCLIENT	(1)
#define EDIR_LOAD_SB_LOCK_GLOBAL	(2)
 
/**
 * Position independent EnhydraDirector scoreboard header block.
 * <P>
 * This header holds global options and the offsets of all server,
 * application, and stats configurations.  It and its associated
 * structures are created at module startup time from the contents
 * of the EnhydraDirector configuration file or data.
 * 
 * @member mutexHeader	Optional header containing a magic number.  This
 * 			is used in external file-based implementations to
 * 			allow identification and locking of the scoreboard
 * 			shared memory.  (CGI and Apache implementations).
 * @member daemonInterval For the Apache and CGI implementations, the 
 * 			interval in seconds between checks for scoreboard
 * 			inactivity.
 * @member daemonTTL	The remaining number of seconds of inactivity until
 * 			the daemon (CGI,Apache) exits due to inactivity.
 * @member failRetry	Number of seconds to keep retrying a failed server
 * 			when the sticky session preference is found in
 * 			the session ID.
 * @member counter	Total number of uses since startup.
 * @member daemonPID	Process ID of the daemon, if applicable.
 * @member dead		Flag to mark the scoreboard as 'dead' when the
 * 			daemon exits.
 * @member numServers	Total number of servers configured for this instance.
 * @member serverListOffset Byte offset to the list of configured servers.
 * 			This list is an aligned array of edir_load_sb_server_t.
 * @member numApps	Total number of apps configured for this instance.
 * @member appListOffset Byte offset to the list of configured apps.
 * 			This list is an aligned array of edir_load_sb_app_t.
 * @member maxAppPrefixLen The maximum length of an application prefix.
 * 			Allows quicker determination of prefix matches by
 * 			URL translation functions.
 * @member numStats	Total number of stats debug prefixes configured for
 * 			this instance.
 * @member statsListOffset Byte offset to the list of configured stats prefixes.
 * 			This list is an aligned array of edir_load_sb_stats_t.
 * @member maxStatsPrefixLen The maximum length of a stats prefix.
 * 			Allows quicker determination of prefix matches by
 * 			URL translation functions.
 * @member debug	Whether to debug.
 * @member DebugMask	Debugging mask to select which types of debugging.
 */
typedef struct edir_load_sb_offsets {
#ifndef EDIR_MULTITHREADED
    edir_mutex_file_header_t	mutexHeader;
#endif
    time_t			daemonInterval;
    time_t			daemonTTL;
    time_t			failRetry;
    uint32			counter;
    int32			daemonPid;
    int32			dead;
    int32			numServers;
    int32			serverListOffset;
    int32			numApps;
    int32			appListOffset;
    int32			maxAppPrefixLen;
    int32			numStats;
    int32			statsListOffset;
    int32			maxStatsPrefixLen;
    int32			debug;
    int32			debugMask;
} edir_load_sb_offsets_t;

/**
 * Position independent appserver entry.
 * <P>
 * These entries contain all runtime configuration data for a single
 * instance of an Enhydra application server.  This includes the 
 * IP and port of the server, as well as load balancing weight,
 * authentication string, reference counts, etc.
 * 
 * @param authData	Optional cleartext password shared between the
 * 			client and server.  XXX - Need more secure scheme.
 * @param hostName	The display name of the server host.  Usually the
 * 			DNS name, but can be the raw IP if the DNS name is not
 * 			available.
 * @param ipName	The IP address of the server as a string.
 * @param sin		Socket address structure containing the IP address and
 * 			port for program usage.
 * @param stickyTag	Precomputed 'sticky session' tag encoded using the
 * 			IP and port number.  If found in a session ID for
 * 			a request, indicates the 'preferred' server.
 * @param weight	Proportional load balancing weight for this server.
 * @param port		Port the server is listening on, in host byte order.
 * @param weightIsDefault Zero if the weight for this server was explicitly	 *                      specified, otherwise, non-zero (use default weight).
 * @param failing	True if there has been a recent connection failure	
 * 			to this server.
 * @param failTime	Time at which this server last failed if <code>failing
 * 			</code> is true.
 * @param failCount	Number of connection failures for this server since
 * 			start up.
 * @param useCount	Number of non-failing uses of this server entry since
 * 			start up.
 * @param refIndex	Index into possible external reference count array.
 * 			This is currently always the same as the index of this
 * 			server within the servers list.
 * @param refCount	Reference count, if the scoreboard itself holds
 * 			the refcounts (the CGI and Apache implementations).
 */
typedef struct edir_load_sb_server {
    char			authData[EDIR_AUTH_MAX + 1];
    char			hostName[EDIR_NAME_MAX + 1];
    char			ipName[EDIR_NAME_MAX + 1];
    struct sockaddr_in		sin;
    char			stickyTag[9];
    int32			weight;
    uint16			port;
    int16			weightIsDefault;
    int16			failing;
    time_t			failTime;
    uint32			failCount; /* Total # failures */
    uint32			useCount;  /* Odometer */
    int32			refIndex;
    uint32			refCount;  /* If using non-locked refcount */
} edir_load_sb_server_t;

/**
 * Position independent application entry.
 * <P>
 * Defines the URL prefix for a connected application, along with a list
 * of servers and host restrictions.
 * 
 * @param appPrefix	The URL prefix for the application.
 * @param appPrefixLen	Length of the URL prefix.  (To avoid many strlen's)
 * @param serverModulus Modulus integer for round robin schedule.
 * @param numServers	The number of servers configured to handle this
 * 			application's requests.
 * @param serverIndexOffset Offset to the list of indexes to servers that
 * 			handle this request.
 * @param numRestricts  Number of host restrictions for this application.
 * @param restrictOffset Offset to the list of host restrictions.
 */
typedef struct edir_load_sb_app {
    char			appPrefix[EDIR_NAME_MAX + 1];
    int32			appPrefixLen;
    int32			serverModulus;
    int32			serverCounter;
    int32			numServers;
    int32			serverIndexOffset;
    int32			numRestricts;
    int32			restrictOffset;
} edir_load_sb_app_t;

/**
 * Position independent "Statistics Prefix" pseudo-application entry.
 * <P>
 * These entries denote prefixes used for debugging and observation
 * of the current scoreboard status.  Within these sections, 
 * host restrictions can be added to prevent unauthorized access.
 * 
 * @member statPrefix	  The URL prefix to use for debug stats.
 * @member statPrefixLen  The URL prefix to use for debug stats.
 * @member numRestricts	  Number of host restrictions.
 * @member restrictOffset Byte offset of host restrictions.
 */
typedef struct edir_load_sb_stats {
    char			statPrefix[EDIR_NAME_MAX + 1];
    int32			statPrefixLen;
    int32			numRestricts;
    int32			restrictOffset;
} edir_load_sb_stats_t;

/**
 * Position independent "Host Restriction" entry.
 * <P>
 * These entries restrict an application to a specified real or virtual
 * host and port number.
 * 
 * @member type		Whether the host is a real or virtual host.
 * @member hostName	The name of the host to restrict to.
 * @member port		The host-byte-order port to restrict to.
 * @member sin		Socket address for real host restriction.
 */
typedef struct edir_load_sb_restrict {
    int				type;
    char			hostName[EDIR_NAME_MAX + 1];
    uint16			port;
    struct sockaddr_in		sin;
    int32			netBits;
    uint32			netMask;
} edir_load_sb_restrict_t;

#define EDIR_SB_APPS(SB) \
    ((edir_load_sb_app_t *)(((char *)(SB))+((SB)->appListOffset)))

#define EDIR_SB_SERVERS(SB) \
    ((edir_load_sb_server_t *)(((char *)(SB))+((SB)->serverListOffset)))

#define EDIR_SB_STATS(SB) \
    ((edir_load_sb_stats_t *)(((char *)(SB))+((SB)->statsListOffset)))

#define EDIR_SB_SERVERINDEX(SB, AP) \
    ((int32 *)(((char *)(SB))+((AP)->serverIndexOffset)))

#define EDIR_SB_RESTRICTS(SB, AP) \
    ((edir_load_sb_restrict_t *)(((char *)(SB))+((AP)->restrictOffset)))

#define EDIR_SB_STATRESTRICTS(SB, ST) \
    ((edir_load_sb_restrict_t *)(((char *)(SB))+((ST)->restrictOffset)))

#ifdef EDIR_MULTITHREADED
#define EDIR_SCOREBOARD_VALID(P) ((P) != NULL)
#else
#define EDIR_SCOREBOARD_VALID(P) \
    (((P)!=NULL) && (((edir_load_sb_offsets_t *)(P))->mutexHeader.magic == EDIR_LOAD_SB_MAGIC))
#endif

extern int32 edir_load_sb_compute_size(edir_conf_t *conf);

extern void edir_load_sb_from_config(edir_conf_t *conf, void *scoreboard);

extern int32 edir_load_sb_get_refcounts(void *sb, int32 *rfcnts, int32 n);

extern int32 edir_load_sb_num_servers(void *scoreboard);

extern void edir_load_sb_debug_print(void *scoreboard, int32 *rc, int32 n);

extern char * edir_load_sb_debug_string(void *scoreboard, int32 *rc, int32 n);

extern void * edir_load_sb_debug_sbuf(void *scoreboard, int32 *rc, int32 n);

#ifndef EDIR_MULTITHREADED
extern edir_shmem_callback_t edir_load_sb_cleanup;
#endif

#ifdef __cplusplus
} /* Extern "C" */
#endif
#endif /*__EDIR_LOAD_SCOREBOARD_H__ */

