/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_statcmd.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

/**
 * Usage message generator.
 * <P>
 * Side effect:  Calls exit() to terminate the program.
 * 
 * @param progName [in] Name of the program to put in the message.
 */
static void
usage(char *progName)
{
    edir_log_error("Usage: %s [-D] [-v] [-d <DataFile>] %s",
		    progName ? progName : "edir_status",
		    "[-l <LockFile>]");
    edir_log_error("       -D   Enable debugging output");
    edir_log_error("       -v   Enable verbose output");
    edir_log_error("       -d   Use scoreboard file (or IPC key) <DataFile>");
    edir_log_error("       -l   Use lock file (or IPC key) <LockFile>");
    exit(1);
}

/**
 * Main program entry point for the "edir_status" command.
 * <P>
 * This command tries to attach to an Enhydra Director scoreboard and dump
 * its current contents.  It does not touch any of the usage counters,
 * nor does it cause the scoreboard daemon (edir_daemon) to reset its
 * "last used" timer.
 * 
 * @param argc [in]  Command line argument count.
 * @param argv [in]  Command line argument list.
 * @return 	     Program exit code, if exit() is not called.
 */
main(int argc, char *argv[]) 
{
    static char			*funcName = "main";
    char			*progName;
    char			*message = NULL;
    edir_exception_t		*_exc=NULL, **exc=&_exc;
    char			ch, *endptr, *cp;
    char			*shmemName = NULL, *mutexName = NULL;
    edir_ipc_handle_t		*mutexKey = NULL, *shmemKey = NULL;
    edir_ipc_handle_t		*mutex = NULL, *shmem = NULL, *refcnt = NULL;
    edir_load_sb_offsets_t	*sb = NULL;
    char			*debugStr=NULL, *dataFile=NULL, *lockFile=NULL;
    int				numMutex;
    int				locked=0, debug=0, verbose=0, exitCode=0;
    int32			*refcounts, nrefcounts;

    /*
     * Get basename of progName 
     */
    progName = strrchr(argv[0], '/');
    if (progName) progName++; else progName = argv[0];

    /*
     * Get file options from environment first, then override
     * from the command line.  This will allow some CGI
     * implementations to configure through environment variables
     * passed to the HTTP server.
     */ 
    if (cp = getenv("EDIR_DEBUG")) {
	debugStr = cp;
    }
#ifdef EDIR_DEBUG_ENABLED
    else {
	debugStr = EDIR_DEBUG_ENABLED;
    }
#endif
    if (cp = getenv("EDIR_DATA_FILE")) {
	edir_shmem_set_default_filename(cp);
    }
#ifdef EDIR_DATA_FILE
    else {
	edir_shmem_set_default_filename(EDIR_DATA_FILE);
    }
#endif
    if (cp = getenv("EDIR_LOCK_FILE")) {
	edir_mutex_set_default_filename(cp);
    }
#ifdef EDIR_LOCK_FILE
    else {
	edir_mutex_set_default_filename(EDIR_LOCK_FILE);
    }
#endif

    /*
     * Command line arguments supercede all defaults.
     */
    while ((ch = getopt(argc, argv, "?Dvd:l:")) != -1) {
	switch (ch) {
	    case 'D':
		debug = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'd':
		dataFile = optarg;
		break;
	    case 'l':
		lockFile = optarg;
		break;
	    case '?':
	    default:
		usage(progName);
	}
    }
    argc -= optind;
    argv += optind;

    if (argc > 0) usage(progName);

    if (debug) {
	edir_log_debug("dataFile = %s", dataFile ? dataFile : "(default)");
	edir_log_debug("lockFile = %s", lockFile ? lockFile : "(default)");
    }

    if (lockFile) mutexKey = edir_ipc_make_config_handle(lockFile);
    if (dataFile) shmemKey = edir_ipc_make_config_handle(dataFile);

    /*
     * Get the scoreboard mutex.
     */
    mutex = edir_mutex_obtain(mutexKey, &numMutex, exc);
    if (*exc) {
	message = "Scoreboard not found.";
	EDIR_THROW(*exc);
    }

    if (verbose) edir_log("Obtained scoreboard mutex.");

    mutexName = edir_mutex_handle_get_object_name(mutex);
    if (!mutexName) mutexName = edir_strdup("(null)");

    if (debug) {
	edir_log_debug("mutex name  = %s", mutexName);
	edir_log_debug("mutex count = %d", numMutex);
    }

    /*
     * Obtain a lock on the scoreboard mutex.
     */
    edir_mutex_lock_shared(mutex, EDIR_LOAD_SB_LOCK_TESTCLIENT, exc);
    if (*exc) {
	message = "Can't lock scoreboard.";
	EDIR_THROW(*exc);
    }

    if (verbose) edir_log("Have shared lock on watchdog mutex.");

    /*
     * Obtain and attach the scoreboard shared memory region.
     */
    if (shmemKey)
	shmem = edir_shmem_obtain(shmemKey, exc);
    else
	shmem = edir_shmem_obtain(mutex, exc);
    if (*exc) {
	message = "Can't read scoreboard.";
	EDIR_THROW(*exc);
    }
    sb = (edir_load_sb_offsets_t *) edir_shmem_handle_get_object(shmem);

    if (verbose) edir_log("Scoreboard shared memory attached.");

    shmemName = edir_shmem_handle_get_object_name(shmem);
    if (!shmemName) shmemName = edir_strdup("(null)");

    if (debug) {
	edir_log_debug("shmem name = %s", shmemName);
	edir_log_debug("shmem addr = "EDIR_PTR"", sb);
    }

    if (!EDIR_SCOREBOARD_VALID(sb)) {
	*exc = edir_exception_create(EDIR_EXCEPT_GENERAL, "Invalid scoreboard.");
	message = "Scoreboard data not valid.";
	EDIR_THROW(*exc);
    }

    if (debug) edir_log_debug("Scoreboard looks valid.");

    /*
     * Check for dead scoreboard.
     */
    edir_mutex_setblocking(mutex, EDIR_LOAD_SB_LOCK_TESTDAEMON, 0);
    locked = edir_mutex_lock_shared(mutex, EDIR_LOAD_SB_LOCK_TESTDAEMON, exc);
    if (*exc) {
	message = "Error checking TESTDAEMON lock";
	EDIR_THROW(*exc);
    }
    if (locked) edir_mutex_unlock(mutex, EDIR_LOAD_SB_LOCK_TESTDAEMON, exc);
    if (*exc) {
	message = "Error releasing TESTDAEMON lock";
	EDIR_THROW(*exc);
    }
    if (sb->dead) {
	edir_log_error("WARNING!!! Scoreboard has been marked as DEAD.");
    } else if (locked) {
	edir_log_error("WARNING!!! Scoreboard has been ABANDONED.");
	edir_log_error("This happens if 'edir_daemon' is %s",
			"terminated unexpectedly.");
	edir_log_error("To clean up:");
	if (!strcmp(mutexName, shmemName)) {
	    edir_log_error("    Remove file/object '%s'", mutexName);
	} else {
	    edir_log_error("    Remove files/objects '%s' and '%s'",
			 mutexName, shmemName);
	}
    }

    /*
     * Obtain the reference count object for the scoreboard.
     */
    refcnt = edir_refcount_obtain(shmem, exc);
    if (*exc) {
	message = "Cannot read reference counts from scoreboard.";
	EDIR_THROW(*exc);
    }

    if (verbose) edir_log("Got reference count handle.");

    /*
     * Obtain the scoreboard global lock while reading data.
     */
    edir_mutex_lock_exclusive(mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
    if (*exc) {
	message = "Can't lock global lock.";
	EDIR_THROW(*exc);
    }

    /*
     * Get the reference count array.
     */
    refcounts = edir_refcount_values(refcnt, &nrefcounts, exc);
    if (*exc) {
	message = "Error while reading reference counts.";
	EDIR_THROW(*exc);
    }

    /*
     * Print out the scoreboard.
     */
    edir_load_sb_debug_print(sb, refcounts, nrefcounts);

    /*
     * Obtain the scoreboard global lock while reading data.
     */
    edir_mutex_unlock(mutex, EDIR_LOAD_SB_LOCK_GLOBAL, exc);
    if (*exc) {
	message = "Can't unlock global lock.";
	EDIR_THROW(*exc);
    }

except:
    if (*exc) {
	if (message) edir_log_error("%s", message);
	if (verbose || debug) {
	    char *trace;
	    trace = edir_exception_stack_trace(*exc);
	    edir_log_error("%s", trace);
	    edir_free(trace);
	}
	edir_exception_free(exc);
	exitCode = 1;
    }

    /*
     * Detach refcount object.
     */
    if (refcnt) edir_refcount_release(refcnt, exc);
    if (*exc) {
	if (verbose || debug) {
	    char *trace;
	    edir_log_error("Error releasing refcount handle.");
	    trace = edir_exception_stack_trace(*exc);
	    edir_log_error("%s", trace);
	    edir_free(trace);
	} else {
	    if (!exitCode) edir_log_error("Error releasing refcount handle.");
	}
	exitCode = 1;
	edir_exception_free(exc);
    }

    /*
     * Detach shared memory.
     */
    if (shmem) edir_shmem_release(shmem, exc);
    if (*exc) {
	if (verbose || debug) {
	    char *trace;
	    edir_log_error("Error releasing shared memory.");
	    trace = edir_exception_stack_trace(*exc);
	    edir_log_error("%s", trace);
	    edir_free(trace);
	} else {
	    if (!exitCode) edir_log_error("Error releasing shared memory.");
	}
	exitCode = 1;
	edir_exception_free(exc);
    }

    /*
     * Unlock mutex.
     */
    if (mutex) edir_mutex_unlock(mutex, EDIR_MUTEX_UNLOCK_ALL, exc);
    if (*exc) {
	if (verbose || debug) {
	    char *trace;
	    edir_log_error("Error unlocking watchdog mutex.");
	    trace = edir_exception_stack_trace(*exc);
	    edir_log_error("%s", trace);
	    edir_free(trace);
	} else {
	    if (!exitCode) edir_log_error("Error unlocking watchdog mutex.");
	}
	exitCode = 1;
	edir_exception_free(exc);
    }

    /*
     * Release mutex handle.
     */
    if (mutex) edir_mutex_release(mutex, exc);
    if (*exc) {
	if (verbose || debug) {
	    char *trace;
	    edir_log_error("Error releasing mutex handle.");
	    trace = edir_exception_stack_trace(*exc);
	    edir_log_error("%s", trace);
	    edir_free(trace);
	} else {
	    if (!exitCode) edir_log_error("Error releasing mutex handle.");
	}
	exitCode = 1;
	edir_exception_free(exc);
    }

    if (mutexKey) edir_ipc_destroy_config_handle(mutexKey);

    if (shmemKey) edir_ipc_destroy_config_handle(shmemKey);
    
    if (mutexName) edir_free(mutexName);

    if (shmemName) edir_free(shmemName);
    
    exit(exitCode);
}

