/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_load_status.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"
#include "edir_conf.h"
#include "edir_load_scoreboard.h"

#define		SMALL_PREFIX_LEN	200

#if defined(WIN32)
# define __ALLOCA(N) _alloca(N);
# define __UNALLOCA(P)
#elif defined(HAVE_ALLOCA)
# define __ALLOCA(N) alloca(N);
# define __UNALLOCA(P)
#else
# define __ALLOCA(N) edir_malloc(N)
# define __UNALLOCA(P) { if (P) edir_free(P); }
#endif

/**
 * Format the contents of a scoreboard into a human-readable string.
 * <P>
 * This function formats the string into a stringbuffer object that
 * can later be converted into a string witout a large number of
 * copies.
 * <P>
 * XXX - FIX - Might be nice to format to a standard XML DTD, so that
 * this information could be passed to other programs, like Enhydra
 * Admin, for instance. :)
 * 
 * @param scoreboard [in] Scoreboard whose data is to be dumped.
 * @param refCounts  [in] Optional array of reference counts.  If not specified,
 * 			  The scoreboards reference counts are used.
 * @param num        [in] The number of reference counts in 'refCounts'.
 * @return An opaque handle to a string/data (sbuf) buffer.
 */
extern void *
edir_load_status(edir_load_sb_offsets_t *sb, char *uri,
		 int32 *refCounts, int32 numRefCounts,
		 edir_kvtable_t *responseHeaders,
		 edir_exception_t **exc)
{
    static char		*funcName = "edir_load_status";
    void		*sbuf;
    char		length[40];
    EDIR_DEBUG_FUNCS_IN(("%s%s("EDIR_PTR", \"%s\", "EDIR_PTR", %d, "EDIR_PTR", "EDIR_PTR")",\
			"func: ", funcName, sb, uri ? uri : "(null)", \
			refCounts, numRefCounts, responseHeaders, *exc));

    sbuf = edir_load_sb_debug_sbuf(sb, refCounts, numRefCounts);
    sprintf(length, "%d", edir_sbuf_length(sbuf));
    edir_kvtable_set(responseHeaders, "Content-Type", "text/plain");
    edir_kvtable_set(responseHeaders, "Content-Length", length);

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName, sbuf));
    return sbuf;
}

/**
 * Check a URI to see if it references a 'status' URL prefix.
 * If so, a module should dump scoreboard status to the client browser.
 * 
 * @param scoreboard  [in] The scoreboard to check.
 * @param path        [in] The path prefix to check.
 * @param local_addr  [in] Socket address structure for server side of the
 * 			   request connection.
 * @param remote_addr [in] Socket address structure for client side of the
 * 			   request connection.
 * @param virt_host   [in] The value of the "Host" HTTP header or the current
 * 			   virtual server name.
 * @return True if this is a status prefix, otherwise false.
 */ 
extern int32
edir_load_status_check_path(edir_load_sb_offsets_t *scoreboard, char *path, 
			    struct sockaddr_in *local_addr,
			    struct sockaddr_in *remote_addr,
			    char *virt_host)
{
    EDIR_IF_DEBUG( static char  *funcName = "edir_load_status_check_path"; )
    char			smallPrefix[SMALL_PREFIX_LEN + 1];
    char			*prefix = smallPrefix;
    char			*allocBuf = NULL;
    char			*cp, *dp, *ep, ch;
    int32			i, j, num, len;
    edir_load_sb_stats_t	*stp, *selectedStats = NULL;
    edir_load_sb_restrict_t	*restricts, *rp;
    int32			clientMatchRequired;
    int32			clientMatch;
    int32			serverMatchRequired;
    int32			serverMatch;
    int32			vhostMatchRequired;
    int32			vhostMatch;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", \"%s\", "EDIR_PTR", \"%s\")", \
			funcName, scoreboard, path ? path : "(null)", \
			local_addr, virt_host ? virt_host : "(null)"));

    /*
     * Make sure prefix buffer is big enough.
     */
    if (scoreboard->maxStatsPrefixLen > SMALL_PREFIX_LEN) {
	allocBuf = (char *) __ALLOCA(scoreboard->maxStatsPrefixLen + 1);
	prefix = allocBuf;
    }
    
    /**
     * Trim and clean up the path, and copy only the number of bytes
     * needed for comparison.
     */
    cp = path;
    dp = prefix;
    ep = dp + scoreboard->maxStatsPrefixLen;
    while (*cp && (*cp == '/')) cp++;
    while (*cp && (dp < ep)) {
	if (*cp == '/') {
	    *dp++ = '/';
	    while (*cp == '/') cp++;
	} else {
	    *dp++ = *cp++;
	}
    }
    *dp = '\0';

    /**
     * Search for an stats entry which matches the prefix and host
     * information.
     */
    num = scoreboard->numStats;
    stp = EDIR_SB_STATS(scoreboard);
    for (i = 0; (selectedStats == NULL) && (i < num); i++, stp++) {
	len = stp->statPrefixLen;
	ch = prefix[len];
	EDIR_DEBUG_EVERYTHING(("Checking prefix \"%s\" against %s \"%s\"", \
			      prefix, "stats prefix", stp->statPrefix));
	if ((!strncmp(prefix, stp->statPrefix, len)) &&
	    ((ch == '\0') || (ch == '/') || (ch == '?') || (ch == ';') ||
	     (ch == '#') || (ch == ' ') || (ch == '\t'))) {
	    /* Potential match -- check restrictions */
	    if (((local_addr == NULL) && (virt_host == NULL)) ||
		 (stp->numRestricts < 1)) {
		EDIR_DEBUG_EVERYTHING(("Stat prefix: no server restrictions"));
		selectedStats = stp;
		break; /* Done - we have a match. */
	    } else {
		/* Check restrictions - first match enables this request */
		EDIR_DEBUG_EVERYTHING(("Stat prefix has server restrictions"));
		restricts = EDIR_SB_STATRESTRICTS(scoreboard, stp);
		clientMatchRequired = 0;
		clientMatch = 0;
		serverMatchRequired = 0;
		serverMatch = 0;
		vhostMatchRequired = 0;
		vhostMatch = 0;
		for (j=0; j<stp->numRestricts; j++) {
		    rp = &(restricts[j]);
		    EDIR_DEBUG_EVERYTHING(("Restrict type: %s", \
			    (rp->type == EDIR_CONF_RESTRICT_CLIENT) ? "CLIENT" : \
				(rp->type == EDIR_CONF_RESTRICT_SERVER) ? \
					"SERVER" : "VIRTUAL"));
		    if (rp->type == EDIR_CONF_RESTRICT_SERVER) {
			serverMatchRequired = 1;
			if ((rp->sin.sin_port != 0) &&
			    (local_addr->sin_port != rp->sin.sin_port)) {
			    EDIR_DEBUG_EVERYTHING(("Port %d doesn't match %d", \
						  ntohs(local_addr->sin_port), \
						  ntohs(rp->sin.sin_port)));
			    ;
			} else if ((rp->sin.sin_addr.s_addr != INADDR_ANY) &&
			       ((local_addr->sin_addr.s_addr & rp->netMask)
				!= (rp->sin.sin_addr.s_addr & rp->netMask))) {
#ifdef DEBUG
			    char ipbuf1[20];
			    char ipbuf2[20];
			    edir_sockaddr_to_str(local_addr, ipbuf1);
			    edir_sockaddr_to_str(&(rp->sin), ipbuf2);
			    EDIR_DEBUG_EVERYTHING(("IP %s %s %s", \
					ipbuf1, "doesn't match", ipbuf2));
#endif
			} else {
			    /* We fit within server restrictions */
			    serverMatch = 1;
			}
		    } else if (rp->type == EDIR_CONF_RESTRICT_CLIENT) {
			clientMatchRequired = 1;
			if ((rp->sin.sin_addr.s_addr != INADDR_ANY) &&
			       ((remote_addr->sin_addr.s_addr & rp->netMask)
				!= (rp->sin.sin_addr.s_addr & rp->netMask))) {
#ifdef DEBUG
			    char ipbuf1[20];
			    char ipbuf2[20];
			    edir_sockaddr_to_str(local_addr, ipbuf1);
			    edir_sockaddr_to_str(&(rp->sin), ipbuf2);
			    EDIR_DEBUG_EVERYTHING(("IP %s %s %s", \
					ipbuf1, "doesn't match", ipbuf2));
#endif
			} else {
			    /* We fit within client restrictions */
			    clientMatch = 1;
			}
		    } else { 
			vhostMatchRequired;
			if ((strlen(rp->hostName) > 0) && virt_host &&
			    edir_strcasecmp(rp->hostName, virt_host)) {
			    EDIR_DEBUG_EVERYTHING(("VHost %s %s %s", \
						  rp->hostName, \
						  "doesn't match", \
						  virt_host));
			} else {
			    /* We fit within virtual server restrictions */
			    vhostMatch = 1;
			}
		    }
		    if (!serverMatchRequired) serverMatch = 1;
		    if (!clientMatchRequired) clientMatch = 1;
		    if (!vhostMatchRequired) vhostMatch = 1;
		    if (serverMatch && clientMatch && vhostMatch) {
			/* We have a match - display stats */
			selectedStats = stp;
		    }
		} /* end for - Restriction iterator */
	    } /* end if we have restrictions */
	} /* end if initial match */
    } /* end for - Stats prefix iterator */
    __UNALLOCA(allocBuf);
#ifdef DEBUG
    if (selectedStats) {
	EDIR_DEBUG_VERBOSE(("Status prefix \"%s\" matched \"%s\"", \
			    selectedStats->statPrefix, path));
    } else {
	EDIR_DEBUG_VERBOSE(("No status prefixes match \"%s\"", path));
    }
#endif
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns "EDIR_PTR"", funcName,
			 selectedStats ? 1 : 0));
    return selectedStats ? 1 : 0;
}
