/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_log.c,v 1.1 2003/04/14 12:25:56 sinisa Exp $
 */

#include "edir.h"

#ifdef WIN32
static char *edir_get_winsock_errstr(DWORD errorcode, char *buf, int bufsize);
static char *edir_get_win32_errstr(DWORD errorcode, char *buf, int bufsize);
#endif

/**
 * Format log or error output and write the resulting text to the log.
 * <P>
 * Maximum length of a log line, including all token replacements, is
 * 4096 characters.  
 * 
 * @param level  [in] One of LOG_NOTICE, LOG_ERR, or LOG_DEBUG.
 * @param errnum [in] Unix-style 'errno' value.
 * @param last   [in] [WIN32 ONLY] Win32 GetLastError() result.
 * @param wsa    [in] [WIN32 ONLY] Win32 WSAGetLastError() result.
 * @param fmt    [in] Format string with '@errstr@', etc. tokens.
 * @param ap     [in] Printf-style variadic argument list.
 */
#ifdef WIN32
static void
_edir_log(int level, DWORD errnum, DWORD last, DWORD wsa, char *fmt, va_list ap)
#else
static void
_edir_log(int level, int errnum, char *fmt, va_list ap)
#endif
{
    char buf[4096], *newfmt;
#ifndef WIN32
    char *cp1, *cp2;
#endif
#ifdef WIN32
    newfmt = edir_log_fmt_add_error_string(fmt, errnum, last, wsa);
#else
    newfmt = edir_log_fmt_add_error_string(fmt, errnum);
#endif
    edir_vsnprintf(buf, 4096, newfmt, ap); 
    buf[4095] = '\0';
    edir_free(newfmt);
#ifdef WIN32
    edir_log_line(level, buf);
#else
    for (cp1 = buf; cp1 != NULL; cp1 = cp2) {
	cp2 = strchr(cp1, '\n');
	if (cp2) *cp2++ = '\0';
	edir_log_line(level, cp1);
	if (cp2 && (*cp2 == '\0')) break;
    }
#endif
}

/**
 * Write an informative message to the standard log.
 * <P>
 * Writes a printf-style message to the standard log stream or file.
 * In addition to printf-style, the following tokens are replaced with
 * error codes or text strings as follows:<BR>
 * <PRE>
 * 	@errnum@	The current value of 'errno'.
 * 	@errstr@	Error string from 'strerror(errno)'.
 * 	@lasterr@	(Win32 only) The result of GetLastError().
 * 	@laststr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * 	@wsaerr@	(Win32 only) The result of GetLastError().
 * 	@wsastr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * </PRE><BR>
 * 
 * @param fmt [in] Printf-style format string with extended tokens (above).
 * @param ... [in] Printf-style variadic argument list.
 */
extern void
edir_log(char *fmt, ...)
{
#ifdef WIN32
    DWORD myErrnum, lastErr, wsaErr;
#else
    int myErrnum;
#endif
    va_list ap;
    myErrnum = errno;
#ifdef WIN32
    lastErr = GetLastError();
    wsaErr = WSAGetLastError();
#endif
    va_start(ap, fmt);
#ifdef WIN32
    _edir_log(LOG_NOTICE, myErrnum, lastErr, wsaErr, fmt, ap);
#else
    _edir_log(LOG_NOTICE, myErrnum, fmt, ap);
#endif
    va_end(ap);
}

/**
 * Write an error message to the standard log
 * <P>
 * Writes a printf-style error message to the standard log stream or file.
 * In addition to printf-style, the following tokens are replaced with
 * error codes or text strings as follows:<BR>
 * <PRE>
 * 	@errnum@	The current value of 'errno'.
 * 	@errstr@	Error string from 'strerror(errno)'.
 * 	@lasterr@	(Win32 only) The result of GetLastError().
 * 	@laststr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * 	@wsaerr@	(Win32 only) The result of GetLastError().
 * 	@wsastr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * </PRE><BR>
 * 
 * @param fmt [in] Printf-style format string with extended tokens (above).
 * @param ... [in] Printf-style variadic argument list.
 */
extern void
edir_log_error(char *fmt, ...)
{
#ifdef WIN32
    DWORD myErrnum, lastErr, wsaErr;
#else
    int myErrnum;
#endif
    va_list ap;
    myErrnum = errno;
#ifdef WIN32
    lastErr = GetLastError();
    wsaErr = WSAGetLastError();
#endif
    va_start(ap, fmt);
#ifdef WIN32
    _edir_log(LOG_ERR, myErrnum, lastErr, wsaErr, fmt, ap);
#else
    _edir_log(LOG_ERR, myErrnum, fmt, ap);
#endif
    va_end(ap);
}

#ifdef DEBUG
static void
LogToFile(char *buf)
{
	FILE *fp;
	fp = fopen("C:/fook.txt", "a");
	if (fp) {
		fwrite(buf, 1, strlen(buf), fp);
		fwrite("\r\n", 1, 2, fp);
		fclose(fp);
	}

}
#endif

/**
 * Write a debug message to the standard log
 * <P>
 * Writes a printf-style debug message to the standard log stream or file.
 * In addition to printf-style, the following tokens are replaced with
 * error codes or text strings as follows:<BR>
 * <PRE>
 * 	@errnum@	The current value of 'errno'.
 * 	@errstr@	Error string from 'strerror(errno)'.
 * 	@lasterr@	(Win32 only) The result of GetLastError().
 * 	@laststr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * 	@wsaerr@	(Win32 only) The result of GetLastError().
 * 	@wsastr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * </PRE><BR>
 * 
 * @param fmt [in] Printf-style format string with extended tokens (above).
 * @param ... [in] Printf-style variadic argument list.
 */
extern void
edir_log_debug(char *fmt, ...)
{
#ifdef WIN32
    DWORD myErrnum, lastErr, wsaErr;
#else
    int myErrnum;
#endif
    va_list ap;
/* LogToFile(fmt); */
    myErrnum = errno;
#ifdef WIN32
    lastErr = GetLastError();
    wsaErr = WSAGetLastError();
#endif
    va_start(ap, fmt);
#ifdef WIN32
    _edir_log(LOG_DEBUG, myErrnum, lastErr, wsaErr, fmt, ap);
#else
    _edir_log(LOG_DEBUG, myErrnum, fmt, ap);
#endif
    va_end(ap);
}

/**
 * Replaces extended tokens in a format string.
 * <P>
 * Creates a newly allocated string in which the following extended tokens
 * are substituted with their values.
 * <PRE>
 * 	@errnum@	The current value of 'errno'.
 * 	@errstr@	Error string from 'strerror(errno)'.
 * 	@lasterr@	(Win32 only) The result of GetLastError().
 * 	@laststr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * 	@wsaerr@	(Win32 only) The result of GetLastError().
 * 	@wsastr@	(Win32 only) Formatted English text for the error code
 * 			returned by GetLastError().
 * </PRE><BR>
 * 
 * @param fmt        [in] Printf-style format string with extended
 *                        tokens (above).
 * @param errnum     [in] Unix-style 'errno' value.
 * @param winlasterr [in] [WIN32 ONLY] Win32 GetLastError() result.
 * @param winsockerr [in] [WIN32 ONLY] Win32 WSAGetLastError() result.
 * @return A newly allocated string containing the processed format
 * 	   string.  NOTE: This string has not yet been run through
 * 	   a 'printf' routine yet, so it is still a valid format string.
 */
#ifdef WIN32
extern char *
edir_log_fmt_add_error_string(char *fmt, DWORD errnum,
			      DWORD winlasterr, DWORD winsockerr)
#else
extern char *
edir_log_fmt_add_error_string(char *fmt, int errnum)
#endif
{
#define _MAXERRLEN 255
    char *cp, *dp, *ep;
    void *sbuf;
    char *errNum = NULL, *errStr = NULL;
    char errNumBuf[16];
    char errStrBuf[_MAXERRLEN + 1];
    int  errNumLen, errStrLen;
#ifdef WIN32
    char *lastErr = NULL, *lastStr = NULL;
    char lastErrBuf[16];
    char lastStrBuf[_MAXERRLEN + 1];
    int  lastErrLen, lastStrLen;
    char *wsaErr = NULL, *wsaStr = NULL;
    char wsaErrBuf[16];
    char wsaStrBuf[_MAXERRLEN + 1];
    int  wsaErrLen, wsaStrLen;
#endif

    sbuf = edir_sbuf_create();

    for (cp=fmt, dp = strchr(cp, '@'); dp; cp = dp, dp = strchr(cp, '@')) {
	edir_sbuf_bincatn(sbuf, cp, dp - cp);
	dp++;
	switch (*dp) {
	    /* errno, both Unix and Win32 */
	    case 'e':
		if (!strncmp(dp, "errnum@", 7)) {
		    dp += 7;
		    if (!errNum) {
			sprintf(errNumBuf, "%d", errnum);
			errNum = errNumBuf;
			errNumLen = strlen(errNum);
		    }
		    edir_sbuf_bincatn(sbuf, errNum, errNumLen);
		} else if (!strncmp(dp, "errstr@", 7)) {
		    dp += 7;
		    if (!errStr) {
#if defined(HAVE_STRERROR) || defined(WIN32)
			errStr = strerror(errnum);
			if (!errStr) errStr = "";
			strncpy(errStrBuf, errStr, _MAXERRLEN);
#else /* HAVE_STRERROR or WIN32 */
			snprintf(errStrBuf, _MAXERRLEN, "(Error code %d)",
				 errnum);
#endif /* HAVE_STRERROR or WIN32 */
			errStrBuf[_MAXERRLEN] = '\0';
			errStr = errStrBuf;
			errStrLen = strlen(errStr);
			/* Prevent dangerous printf() format strings */
			for (ep = errStr; *ep; ep++) if (*ep == '%') *ep = '#';
		    }
		    edir_sbuf_bincatn(sbuf, errStr, errStrLen);
		} else {
		    edir_sbuf_bincatn(sbuf, "@", 1);
		}
		break;
#ifdef WIN32
	    /* Win32 GetLastError status */
	    case 'l':
		if (!strncmp(dp, "lasterr@", 8)) {
		    dp += 8;
		    if (!lastErr) {
			sprintf(lastErrBuf, "%d", winlasterr);
			lastErr = lastErrBuf;
			lastErrLen = strlen(lastErr);
		    }
		    edir_sbuf_bincatn(sbuf, lastErr, lastErrLen);
		} else if (!strncmp(dp, "laststr@", 8)) {
		    dp += 8;
		    if (!lastStr) {
			lastStr = edir_get_win32_errstr(winlasterr,
					lastStrBuf, _MAXERRLEN);
			lastStrLen = strlen(lastStr);
		    }
		    edir_sbuf_bincatn(sbuf, lastStr, lastStrLen);
		} else {
		    edir_sbuf_bincatn(sbuf, "@", 1);
		}
		break;
	    /* Win32 WSAGetLastError status */
	    case 'w':
		if (!strncmp(dp, "wsaerr@", 7)) {
		    dp += 7;
		    if (!wsaErr) {
			sprintf(wsaErrBuf, "%d", winsockerr);
			wsaErr = wsaErrBuf;
			wsaErrLen = strlen(wsaErr);
		    }
		    edir_sbuf_bincatn(sbuf, wsaErr, wsaErrLen);
		} else if (!strncmp(dp, "wsastr@", 7)) {
		    dp += 7;
		    if (!wsaStr) {
			wsaStr = edir_get_winsock_errstr(winsockerr,
					wsaStrBuf, _MAXERRLEN);
			wsaStrLen = strlen(wsaStr);
		    }
		    edir_sbuf_bincatn(sbuf, wsaStr, wsaStrLen);
		} else {
		    edir_sbuf_bincatn(sbuf, "@", 1);
		}
		break;
#endif /* WIN32 */
	    default:
		edir_sbuf_bincatn(sbuf, "@", 1);
		break;
	}
    }
    edir_sbuf_strcatn(sbuf, cp);
    cp = edir_sbuf_tostring(sbuf, NULL);
    edir_sbuf_free(sbuf);
    return cp;
}

#ifdef WIN32

/**
 * Obtain the WinSock error string for the given error code.
 * 
 * @param errorcode [in]     The error number to decode.
 * @param buf       [in,out] Pre-allocated buffer to write into.
 * @param bufsize   [in]     Size of pre-allocated buffer.
 * @return A pointer to the beginning of "buf" for convenience.
 */
static char *
edir_get_winsock_errstr(DWORD errorcode, char *buf, int bufsize)
{
    char *cp;
    switch(errorcode)
    {
	case WSAEACCES:
	    _snprintf(buf, bufsize, "Permission denied.");
	    break;
	case WSAEADDRINUSE:
	    _snprintf(buf, bufsize, "Address already in use.");
	    break;
	case WSAEADDRNOTAVAIL:
	    _snprintf(buf, bufsize, "Cannot assign requested address.");
	    break;
	case WSAEAFNOSUPPORT:
	    _snprintf(buf, bufsize, "Address family not supported by "\
				    "protocol family.");
	    break;
	case WSAEALREADY:
	    _snprintf(buf, bufsize, "Operation already in progress.");
	    break;
	case WSAECONNABORTED:
	    _snprintf(buf, bufsize, "Software caused connection abort.");
	    break;
	case WSAECONNREFUSED:
	    _snprintf(buf, bufsize, "Connection refused.");
	    break;
	case WSAECONNRESET:
	    _snprintf(buf, bufsize, "Connection reset by peer.");
	    break;
	case WSAEDESTADDRREQ:
	    _snprintf(buf, bufsize, "Destination address required.");
	    break;
	case WSAEFAULT:
	    _snprintf(buf, bufsize, "Bad address.");
	    break;
	case WSAEHOSTDOWN:
	    _snprintf(buf, bufsize, "Host is down.");
	    break;
	case WSAEHOSTUNREACH:
	    _snprintf(buf, bufsize, "No route to host.");
	    break;
	case WSAEINPROGRESS:
	    _snprintf(buf, bufsize, "Operation now in progress.");
	    break;
	case WSAEINTR:
	    _snprintf(buf, bufsize, "Interrupted function call.");
	    break;
	case WSAEINVAL:
	    _snprintf(buf, bufsize, "Invalid argument.");
	    break;
	case WSAEISCONN:
	    _snprintf(buf, bufsize, "Socket is already connected.");
	    break;
	case WSAEMFILE:
	    _snprintf(buf, bufsize, "Too many open files.");
	    break;
	case WSAEMSGSIZE:
	    _snprintf(buf, bufsize, "Message too long.");
	    break;
	case WSAENETDOWN:
	    _snprintf(buf, bufsize, "Network is down.");
	    break;
	case WSAENETRESET:
	    _snprintf(buf, bufsize, "Network dropped connection on reset.");
	    break;
	case WSAENETUNREACH:
	    _snprintf(buf, bufsize, "Network is unreachable.");
	    break;
	case WSAENOBUFS:
	    _snprintf(buf, bufsize, "No buffer space available.");
	    break;
	case WSAENOPROTOOPT:
	    _snprintf(buf, bufsize, "Bad protocol option.");
	    break;
	case WSAENOTCONN:
	    _snprintf(buf, bufsize, "Socket is not connected.");
	    break;
	case WSAENOTSOCK:
	    _snprintf(buf, bufsize, "Socket operation on non-socket.");
	    break;
	case WSAEOPNOTSUPP:
	    _snprintf(buf, bufsize, "Operation not supported.");
	    break;
	case WSAEPFNOSUPPORT:
	    _snprintf(buf, bufsize, "Protocol family not supported.");
	    break;
	case WSAEPROCLIM:
	    _snprintf(buf, bufsize, "Too many processes.");
	    break;
	case WSAEPROTONOSUPPORT:
	    _snprintf(buf, bufsize, "Protocol not supported.");
	    break;
	case WSAEPROTOTYPE:
	    _snprintf(buf, bufsize, "Protocol wrong type for socket.");
	    break;
	case WSAESHUTDOWN:
	    _snprintf(buf, bufsize, "Cannot send after socket shutdown.");
	    break;
	case WSAESOCKTNOSUPPORT:
	    _snprintf(buf, bufsize, "Socket type not supported.");
	    break;
	case WSAETIMEDOUT:
	    _snprintf(buf, bufsize, "Connection timed out.");
	    break;
	case WSATYPE_NOT_FOUND:
	    _snprintf(buf, bufsize, "Class type not found.");
	    break;
	case WSAEWOULDBLOCK:
	    _snprintf(buf, bufsize, "Operation on nonblocking socket " \
				    "would otherwise block.");
	    break;
	case WSAHOST_NOT_FOUND:
	    _snprintf(buf, bufsize, "Host not found.");
	    break;
	case WSA_INVALID_HANDLE:
	    _snprintf(buf, bufsize, "Specified event object handle is " \
				    "invalid.");
	    break;
	case WSA_INVALID_PARAMETER:
	    _snprintf(buf, bufsize, "One or more parameters are invalid.");
	    break;
#ifdef WSAINVALIDPROCTABLE
	case WSAINVALIDPROCTABLE:
	    _snprintf(buf, bufsize, "Invalid procedure table from service " \
				    "provider.");
	    break;
#endif
#ifdef WSAINVALIDPROVIDER
	case WSAINVALIDPROVIDER:
	    _snprintf(buf, bufsize, "Invalid service provider version number.");
	    break;
#endif
	case WSA_IO_INCOMPLETE:
	    _snprintf(buf, bufsize, "Overlapped I/O event object not in " \
				    "signaled state.");
	    break;
	case WSA_IO_PENDING:
	    _snprintf(buf, bufsize, "Overlapped operations will complete " \
				    "later.");
	    break;
	case WSA_NOT_ENOUGH_MEMORY:
	    _snprintf(buf, bufsize, "Insufficient memory available.");
	    break;
	case WSANOTINITIALISED:
	    _snprintf(buf, bufsize, "Successful WSAStartup not yet performed.");
	    break;
	case WSANO_DATA:
	    _snprintf(buf, bufsize, "Valid name, no data record of " \
				    "requested type.");
	    break;
	case WSANO_RECOVERY:
	    _snprintf(buf, bufsize, "This is a non-recoverable error.");
	    break;
#ifdef WSAPROVIDERFAILEDINIT
	case WSAPROVIDERFAILEDINIT:
	    _snprintf(buf, bufsize, "Unable to initialize a service provider.");
	    break;
#endif
	case WSASYSCALLFAILURE:
	    _snprintf(buf, bufsize, "System call failure.");
	    break;
	case WSASYSNOTREADY:
	    _snprintf(buf, bufsize, "Network subsystem is unavailable.");
	    break;
	case WSATRY_AGAIN:
	    _snprintf(buf, bufsize, "Non-authoritative host not found.");
	    break;
	case WSAVERNOTSUPPORTED:
	    _snprintf(buf, bufsize, "WINSOCK.DLL version out of range.");
	    break;
	case WSAEDISCON:
	    _snprintf(buf, bufsize, "Graceful shutdown in progress.");
	    break;
	case WSA_OPERATION_ABORTED:
	    _snprintf(buf, bufsize, "Overlapped operation aborted.");
	    break;
	default:
	    _snprintf(buf, bufsize, "(Socket error code %d)", errorcode);
	    break;
    }
    buf[bufsize-1] = '\0';
    /* Avoid dangerous printf format strings */
    for (cp=buf; *cp; cp++) if (*cp == '%') *cp = '#';
    return buf;
}

/**
 * Obtain the Win32 error string for the given Win32 error code.
 * <P>
 * This function also handles WinSock errors that have been 
 * adjusted to set bit 29 (The "application error" bit).
 * 
 * @param errorcode [in]     The error number to decode.
 * @param buf       [in,out] Pre-allocated buffer to write into.
 * @param bufsize   [in]     Size of pre-allocated buffer.
 * @return A pointer to the beginning of "buf" for convenience.
 */
static char *
edir_get_win32_errstr(DWORD errorcode, char *buf, int bufsize)
{
    int n;
    char *cp;
    n = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
		      FORMAT_MESSAGE_IGNORE_INSERTS, NULL, errorcode,
		      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		      (LPTSTR) buf, bufsize, NULL);
    if (n < 1) {
	_snprintf(buf, bufsize, "(Error code %d)", errorcode);
    } else {
	buf[n-1] = '\0';
    }
    /* Avoid dangerous printf format strings */
    for (cp=buf; *cp; cp++) if (*cp == '%') *cp = '#';
    return buf;
}

#endif /* Win32 */

