/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_mutex.h,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#ifndef __EDIR_MUTEX_H__
#define __EDIR_MUTEX_H__ 1
#ifdef __cplusplus
extern "C" {
#endif

/**
 * Interface definition for a generic Mutex for scoreboards.
 * <P>
 * System dependent mutex modules, such as edir_mutex_sysv.c,
 * implement functions fitting the prototypes defined below.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */
 
/**
 * Maximum # of mutexes per object.
 */
#define EDIR_MUTEX_MAX				(16)
 
/*
 * Tells edir_mutex_unlock to unlock every mutex in its array.
 */
#define EDIR_MUTEX_UNLOCK_ALL			(-1)

/**
 * Lock states.
 */
#define EDIR_MUTEX_STATE_UNLOCKED			(0)
#define EDIR_MUTEX_STATE_LOCKED_SHARE		(1)
#define EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE		(2)
 
/**
 * Header written to the beginning of files using fcntl or flock
 * locking.  this structure is the first part of the offsets
 * structure of the scoreboard, so that scoreboards can share
 * the file with their mutex object.
 */
typedef struct edir_mutex_file_header {
    uint32	magic;			 /* Space for magic # at head of file */
    uint32	numLocks;		 /* For recording size of lock array */
    octet	lockSlots[EDIR_MUTEX_MAX]; /* For range locking, if available */
} edir_mutex_file_header_t;

/**
 * Create a mutex in an initially disabled state.
 * <P>
 * Creates a brand new mutex.  For interprocess implementations, this
 * will be a system-wide mutex, lock, or semaphore.  The key may or may
 * not be used, depending upon the requirements of the implementation.
 * A NULL key means "use the system-dependent default".
 * Some implementations may leave system resources allocated,
 * even after the program has terminated, unless edir_mutex_destroy() is
 * called to clean up.
 * <P>
 * The new mutex is created in a state that will cause all threads or
 * processes attemting to obtain the mutex to block.  To begin normal
 * use of the mutex, the creator must at some point call edir_mutex_enable()
 * to initialize it.
 * 
 * @param key [in]  Handle to object to use for configuration data.
 * @param num [in]  Number of mutexes to create for this object.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return                An Handle to the mutex object.
 * @exception EDIR_EXCEPT_MUTEX If the mutex cannot be created.
 */
extern edir_ipc_handle_t *
edir_mutex_create(edir_ipc_handle_t *key, int num, edir_exception_t **exc);

/**
 * Destroy a mutex.
 * <P>
 * Deletes the specified mutex and removes its resources from the system.
 * If client threads or processes attempt to access the mutex after this
 * function is called, they will experience EDIR_EXCEPT_MUTEX_DELETED
 * exceptions.
 * 
 * @param handle [in]  Handle returned by edir_mutex_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If the mutex cannot be destroyed.
 */
extern void
edir_mutex_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc);

/**
 * Obtain an existing mutex.
 * <P>
 * A mutex must already exist for the given key (if keys are used).
 * If the mutex exists, this function returns a handle to a mutex object.
 * <P>
 * NOTE! This is NOT the same as "locking" the mutex.  This function
 * simply gets a reference to the mutex object from the system.  To lock
 * access to the resource controlled by this mutex, call edir_mutex_lock().
 * 
 * @param key [in]  Handle to object to use for configuration data.
 * @param num [out] The number of mutexes controlled by this object.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          An Handle to the mutex object.
 * @exception EDIR_EXCEPT_MUTEX If the specified mutex is not available.
 */
extern edir_ipc_handle_t *
edir_mutex_obtain(edir_ipc_handle_t *key, int *num, edir_exception_t **exc);

/**
 * Release a previously obtained mutex.
 * <P>
 * Releases this process's or thread's reference to the specified mutex,
 * but does not remove the underlying object.  This function is used to
 * "free" mutexes obtained using edir_mutex_obtain().
 * <P>
 * NOTE! This is NOT the same as "unlocking" the mutex.  To unlock
 * the mutex so other processes/threads can access it, use edir_mutex_unlock().
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while releasing the mutex.
 */
extern void
edir_mutex_release(edir_ipc_handle_t *handle, edir_exception_t **exc);

/**
 * Obtain an exclusive "Write" lock on a mutex.
 * <P>
 * An exclusive lock grants exclusive access to a resource.  In other
 * words, no other locks (whether shared or exclusive) will be granted
 * on the object until the holder of the current exclusive lock unlocks
 * the object.  An exclusive lock allows a critical area to be written
 * without danger of interfering with other writers OR with any readers.
 * This is the reason it is sometimes called a "Write" lock.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to lock, starting at 0.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             1 if the lock was obtained, otherwise 0.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while locking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern int
edir_mutex_lock_exclusive(edir_ipc_handle_t *handle, int num, edir_exception_t **exc);

/**
 * Obtain a shareable "read" lock on a mutex.
 * <P>
 * More than one shareable lock may be held by processes at any given
 * time.  However, not shareable (or exclusive) locks may be held if
 * one process has successfully obtained an exclusive lock.  If an
 * exclusive lock is held elsewhere then attempts to obtain a shareable
 * lock will block until the exclusive lock is unlocked.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to lock, starting at 0.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             1 if the lock was obtained, otherwise 0.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while locking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern int
edir_mutex_lock_shared(edir_ipc_handle_t *handle, int num, edir_exception_t **exc);

/**
 * Give up the currently held lock on a locked mutex.
 * <P>
 * Unlocks the mutex, allowing other threads or processes to access
 * the resource controlled by the given mutex.  Once this function
 * is called, the calling process or thread is no longer guaranteed
 * to have exclusive access to the controlled resource.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to unlock, starting at 0.
 * 		       A value of -1 means "unlock all".
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while unlocking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern void
edir_mutex_unlock(edir_ipc_handle_t *handle, int num, edir_exception_t **exc);

/**
 * Return zero if not locked, or nonzero state if locked.
 * <P>
 * If the mutex is not locked, this function returns zero, corresponding
 * to state EDIR_MUTEX_STATE_UNLOCKED, also corresponding to "false".
 * If the mutex is locked, the type of lock is returned as a positive
 * integer.  Possible return values are EDIR_MUTEX_STATE_LOCKED_SHARE if
 * a shareable lock is held, or EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE if an
 * exclusive lock is held.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the mutex to check.
 * @return             Zero(false) or held lock type.
 */
extern int
edir_mutex_islocked(edir_ipc_handle_t *handle, int num);

/**
 * Sets or unsets blocking mode on the indexed mutex.
 * <P>
 * If set (block == 1), then lock requests on the mutex will
 * block until the mutex becomes available.  If unset (block == 0),
 * then requests will return immediately without throwing an exception
 * if the lock is not available immediately.  If the lock is unavailable
 * in nonblocking mode, zero is returned by the lock functions.
 * <P>
 * This function does not throw exceptions.  Invalid values of "num"
 * are ignored.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the mutex for which to set/unset blocking.
 * @param block  [in]  1 for blocking, 0 for nonblocking.
 * @return             Zero(false) or held lock type.
 */
extern void
edir_mutex_setblocking(edir_ipc_handle_t *handle, int num, int block);

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * This pointer is a reference and MUST NOT be modified!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void * edir_mutex_handle_get_object(edir_ipc_handle_t *handle);

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char * edir_mutex_handle_get_object_name(edir_ipc_handle_t *handle);

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void edir_mutex_set_default_filename(char  *filename);

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern char * edir_mutex_default_name();

#ifdef __cplusplus
} /* Extern "C" */
#endif
#endif /* __EDIR_MUTEX_H__ */

