/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_mutex_filesys.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_mutex_filesys.h"

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef USE_FLOCK
# ifdef HAVE_SYS_FILE_H
# include <sys/file.h>
# endif
#endif
 
#ifdef USE_FLOCK
#error "FIXME!!!!! flock() code does not yet support multi-lock arrays"
#endif

/**
 * System-dependent implementation of the Enhydra Director mutex interface,
 * using Unix file locking calls.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/*
 * Internal contents pointed to by "handles".
 */
typedef struct edir_mutex_filesys_ctx {
    char	*pathname;
    int		fd;
    int		number;
    int		state[EDIR_MUTEX_MAX];
    int		wait[EDIR_MUTEX_MAX];
} edir_mutex_filesys_ctx_t;

#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_MUTEX)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

#define EDIR_OFFSET(T, M) ((int) (&(((T *)0)->M)))
#define HDR_OFFSET (EDIR_OFFSET(edir_mutex_file_header_t, lockSlots[0]))

static int DefaultSet = 0;
static char * DefaultFileName = EDIR_MUTEX_FILESYS_DEFAULT_FILE;

/**
 * Create a mutex in an initially disabled state.
 * <P>
 * Creates a brand new mutex.  For interprocess implementations, this
 * will be a system-wide mutex, lock, or semaphore.  The key may or may
 * not be used, depending upon the requirements of the implementation.
 * A NULL key means "use the system-dependent default".
 * Some implementations may leave system resources allocated,
 * even after the program has terminated, unless edir_mutex_destroy() is
 * called to clean up.
 * <P>
 * The new mutex is created in the locked state. 
 * 
 * @param key [in]  Handle to configuration info for this mutex.
 * @param num [in]  Number of mutexes to create for this object.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          An opaque handle to the mutex object.
 * @exception EDIR_EXCEPT_MUTEX If the mutex cannot be created.
 */
extern edir_ipc_handle_t *
edir_mutex_create(edir_ipc_handle_t *key, int num, edir_exception_t **exc)
{
    static char			*funcName = "edir_mutex_create";
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    edir_ipc_handle_t		*handle = NULL;
    int				i, nbytes, unlinkpath = 1;
#ifndef USE_FLOCK
    struct flock	 	lock;
#endif
    char		 	*tmpPath;
    edir_mutex_file_header_t	header;
    
    if ((num < 1) || (num > EDIR_MUTEX_MAX)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, 
		"Number of mutexes (%d) out of range (%d .. %d)",
		num, 0, EDIR_MUTEX_MAX);
	EDIR_THROW(*exc);
    }

    /*
     * Create a new handle for this mutex.  Externally this is a (void *).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_MUTEX;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_FILEDESC;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_mutex_filesys_ctx_t *) edir_malloc(sizeof(edir_mutex_filesys_ctx_t));
    handle->pData = (void *) ctx;
    if (key != NULL) {
	if (key->handle_type != EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, 
		"Mutexes only currently accept key handles of type %s (%d)",
		"EDIR_IPC_HANDLE_TYPE_CONFIG", EDIR_IPC_HANDLE_TYPE_CONFIG);
	    EDIR_THROW(*exc);
	}
	ctx->pathname = (char *) edir_ipc_config_handle_get_object_name(key);
	if (ctx->pathname == NULL)
	    ctx->pathname = edir_strdup(DefaultFileName);
    } else {
	ctx->pathname = edir_strdup(DefaultFileName);
    }
    ctx->fd = -1;
    ctx->number = num;
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->state[i] = -1;
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->wait[i] = 1;

    tmpPath = (char *)edir_malloc(strlen(ctx->pathname) + 30);
    sprintf(tmpPath, "%s.%d", ctx->pathname, getpid());
    unlink(tmpPath);

    /* Open the file privately, since it is opened unlocked. */
    ctx->fd = open(tmpPath, O_CREAT|O_EXCL|O_RDWR, 0600);
    if (ctx->fd < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't create initializing lock file \"%s\": @errstr@",
		tmpPath);
	EDIR_THROW(*exc);
    }
    
    memset((void *)(&header), 0, sizeof(header));
    header.numLocks = num;
    while ((nbytes = write(ctx->fd, (void *)(&header), sizeof(header))) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't write lock file header in \"%s\": @errstr@",
		    tmpPath);
	    EDIR_THROW(*exc);
	}
    }
    if (nbytes != sizeof(header)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't write lock file header in \"%s\": @errstr@",
		tmpPath);
	EDIR_THROW(*exc);
    }
    
#ifdef USE_FLOCK
    while (flock(ctx->fd, LOCK_EX) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't get initial lock on \"%s\": @errstr@", tmpPath);
	    EDIR_THROW(*exc);
	}
    }
    ctx->state[0] = EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE;
#else
    /* Obtain exclusive lock on all ranges before moving file */
    for (i=0; i<num; i++) {
	lock.l_whence = SEEK_SET;
	lock.l_start = HDR_OFFSET + i;
	lock.l_len = 1;
	lock.l_type = F_WRLCK;
	lock.l_pid = 0;
	while (fcntl(ctx->fd, F_SETLKW, &lock) < 0) {
	    if (errno != EINTR) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Can't get initial lock on \"%s\" [lock#%d]: @errstr@",
			tmpPath, i);
		EDIR_THROW(*exc);
	    }
	}
	ctx->state[i] = EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE;
    }
#endif

    /* Now, try to rename the file to the correct name. */
    while (link(tmpPath, ctx->pathname) < 0) {
	if (errno == EEXIST) {
	    /*
	     * Try to open existing lock and lock it.
	     * If the lock already exists AND is in use, as evidenced by
	     * the fact that we can't get an exclusive lock immediately,
	     * then throw an error.  If we get an exclusive lock, then
	     * it's probably an old leftover lock file.  Just use it
	     * instead of the one we just created.
	     */
	    int fd2;
	    while (((fd2 = open(ctx->pathname, O_RDWR))<0)&&(errno == EINTR));
	    if (fd2 >= 0) {
		int bad_file = 0;
		/* Attempt lock non-blocking */
#ifdef USE_FLOCK
		if (flock(fd2, LOCK_EX | LOCK_NB) < 0) {
		    close(fd2);
		    fd2 = -1;
		}
#else
		/* Get header from file */
		while ((nbytes =
			read(fd2, (void *)&header, sizeof(header))) < 0) {
		    if (errno != EINTR) {
			bad_file = 1;
#ifdef NOTDEF
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_EXISTS,
			    "Mutex '%s' already exists and is unavailable",
			    ctx->pathname);
			EDIR_THROW(*exc);
#endif
		    }
		}
		if (!bad_file && (nbytes != sizeof(header))) {
		    bad_file = 1;
#ifdef NOTDEF
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX_EXISTS,
			"Mutex '%s' already exists and is unreadable",
			ctx->pathname);
		    EDIR_THROW(*exc);
#endif
		}
		if (!bad_file &&
		    (header.numLocks > 0) && (header.numLocks != num)) {
		    bad_file = 1;
#ifdef NOTDEF
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX_EXISTS,
			"Mutex '%s' already exists and has different # locks",
			ctx->pathname);
		    EDIR_THROW(*exc);
#endif
		}
		if ((bad_file) && (unlink(ctx->pathname) < 0)) {
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX_EXISTS,
			"Could not replace corrupt mutex file '%s': @errstr@",
			ctx->pathname);
		    EDIR_THROW(*exc);
		}
		if (bad_file) continue; /* try moving file again */
		for (i=0; (fd2 >= 0) && (i < num); i++) {
		    lock.l_whence = SEEK_SET;
		    lock.l_start = HDR_OFFSET + i;
		    lock.l_len = 1;
		    lock.l_type = F_WRLCK;
		    lock.l_pid = 0;
		    while (fcntl(fd2, F_SETLK, &lock) < 0) {
			if (errno != EINTR) {
			    close(fd2);
			    fd2 = -1;
			    break;
			}
		    }
		}
#endif
	    }
	    if (fd2 < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_EXISTS,
		    "Mutex '%s' already exists and is in use", ctx->pathname);
		EDIR_THROW(*exc);
	    }
	    close(ctx->fd);
	    ctx->fd = fd2;
	    break;
	} 
	if (errno != EINTR) {
	    struct stat stbuf;
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't rename lockfile \"%s\" to \"%s\": @errstr@",
		tmpPath, ctx->pathname);
	    /* Could be on NFS and get fake bad link result. */
	    /* Try this bogon. */
	    if ((fstat(ctx->fd, &stbuf) >= 0) && (stbuf.st_nlink >= 2)) {
		edir_exception_free(exc);
		break;
	    }
	    EDIR_THROW(*exc);
	}
    }

    /* Link part of rename done, remove old path */
    unlink(tmpPath);
    edir_free(tmpPath);
    tmpPath = NULL;

    /* Done.  Have brand new lock which we currently hold. */
    return handle;

except:
    /* Last ditch exception cleanup. */
    if (tmpPath) {
	unlink(tmpPath);
	edir_free(tmpPath);
	tmpPath = NULL;
    }
    if (handle != NULL) {
	/* Ctx only ever valid if handle is valid */
	if (ctx != NULL) {
	    if (ctx->fd >= 0) close(ctx->fd);
	    ctx->fd = -1;
	    if (ctx->pathname != NULL) {
		/* If we get here, created file has not been moved yet */
		/* Don't unlink existing lock!!! */
		edir_free(ctx->pathname);
	    }
	    ctx->pathname = NULL;
	    edir_free(ctx);
	}
	edir_free(handle);
    }
    return NULL;
}

/**
 * Destroy a mutex.
 * <P>
 * Deletes the specified mutex and removes its resources from the system.
 * If client threads or processes attempt to access the mutex after this
 * function is called, they will experience EDIR_EXCEPT_MUTEX_DELETED
 * exceptions.
 * 
 * @param handle [in]  Handle returned by edir_mutex_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If the mutex cannot be destroyed.
 */
extern void
edir_mutex_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char			*funcName = "edir_mutex_destroy";
    edir_mutex_filesys_ctx_t	*ctx = NULL;

    if (!VALID_HANDLE(handle)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
	EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_DESTROY_SHARE(handle, exc);

    if (ctx->fd >= 0) {
	if (close(ctx->fd) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Can't close mutex \"%s\": @errstr@",
			ctx->pathname ? ctx->pathname : "???");
	    ctx->fd = -1;
	    EDIR_THROW(*exc);
	}
    }

    if (ctx->pathname != NULL) {
	if ((unlink(ctx->pathname) < 0) && (errno != ENOENT)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Can't delete mutex \"%s\": @errstr@",
			ctx->pathname ? ctx->pathname : "???");
	    edir_free(ctx->pathname);
	    ctx->pathname = NULL;
	    EDIR_THROW(*exc);
	}
	edir_free(ctx->pathname);
	ctx->pathname = NULL;
    }

    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Last ditch exception cleanup. At least free the handle memory. */
    if (ctx != NULL) {
	if (ctx->fd >= 0) {
	    close(ctx->fd);
	    ctx->fd = -1;
	}
	if (ctx->pathname != NULL) {
	    unlink(ctx->pathname);
	    edir_free(ctx->pathname);
	    ctx->pathname = NULL;
	}
	edir_free(ctx);
    }
    if (handle != NULL) edir_free(handle);
    return;
}

/**
 * Obtain an existing mutex.
 * <P>
 * A mutex must already exist for the given key (if keys are used).
 * If the mutex exists, this function returns a handle to a mutex object.
 * <P>
 * NOTE! This is NOT the same as "locking" the mutex.  This function
 * simply gets a reference to the mutex object from the system.  To lock
 * access to the resource controlled by this mutex, call edir_mutex_lock().
 * 
 * @param key [in]  Handle to configuration info for this mutex.
 * @param num [out] The number of mutexes controlled by this object.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          An opaque handle to the mutex object.
 * @exception EDIR_EXCEPT_MUTEX If the specified mutex is not available.
 */
extern edir_ipc_handle_t *
edir_mutex_obtain(edir_ipc_handle_t *key, int *num, edir_exception_t **exc)
{
    static char		 	*funcName = "edir_mutex_obtain";
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    edir_ipc_handle_t		*handle = NULL;
    edir_mutex_file_header_t	header;
    int				i, nbytes;

    /*
     * Create a new handle for this mutex.  Externally this is a (void *).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_MUTEX;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_FILEDESC;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_mutex_filesys_ctx_t *) edir_malloc(sizeof(edir_mutex_filesys_ctx_t));
    handle->pData = (void *)ctx;
    if (key != NULL) {
	if (key->handle_type != EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, 
		"Mutexes only currently accept key handles of type %s (%d)",
		"EDIR_IPC_HANDLE_TYPE_CONFIG", EDIR_IPC_HANDLE_TYPE_CONFIG);
	    EDIR_THROW(*exc);
	}
	ctx->pathname = (char *) edir_ipc_config_handle_get_object_name(key);
	if (ctx->pathname == NULL)
	    ctx->pathname = edir_strdup(DefaultFileName);
    } else {
	ctx->pathname = edir_strdup(DefaultFileName);
    }
    ctx->fd = -1;
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->state[i] = -1;
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->wait[i] = 1;

    while ((ctx->fd = open(ctx->pathname, O_RDWR)) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't open lock file \"%s\": @errstr@", ctx->pathname);
	    EDIR_THROW(*exc);
	}
    }
    
    while ((nbytes = read(ctx->fd, (void *)&header, sizeof(header))) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't read lock file header for \"%s\": @errstr@",
		    ctx->pathname);
	    EDIR_THROW(*exc);
	}
    }
    
    if (nbytes != sizeof(header)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Lock file header for \"%s\" is too small", ctx->pathname);
	EDIR_THROW(*exc);
    }
    
    if ((header.numLocks < 1) || (header.numLocks >= EDIR_MUTEX_MAX)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Lock file \"%s\" has %d locks, %s (0 .. %d)",
		ctx->pathname, header.numLocks, 
		"which is out of range", EDIR_MUTEX_MAX);
	EDIR_THROW(*exc);
    }
    
    *num = ctx->number = header.numLocks;

    for (i=0; i < *num; i++) ctx->state[i] = EDIR_MUTEX_STATE_UNLOCKED;

    /* Done.  Have brand new lock which we currently hold. */
    return handle;

except:
    /* Last ditch exception cleanup. */
    if (ctx != NULL) {
	if (ctx->fd >= 0) close(ctx->fd);
	ctx->fd = -1;
	if (ctx->pathname != NULL) edir_free(ctx->pathname);
	ctx->pathname = NULL;
	edir_free(ctx);
    }
    if (handle != NULL) edir_free(handle);
    return NULL;
}

/**
 * Release a previously obtained mutex.
 * <P>
 * Releases this process's or thread's reference to the specified mutex,
 * but does not remove the underlying object.  This function is used to
 * "free" mutexes obtained using edir_mutex_obtain().
 * <P>
 * NOTE! This is NOT the same as "unlocking" the mutex.  To unlock
 * the mutex so other processes/threads can access it, use edir_mutex_unlock().
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while releasing the mutex.
 */
extern void
edir_mutex_release(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		 	*funcName = "edir_mutex_release";
    edir_mutex_filesys_ctx_t	*ctx = NULL;

    if (!VALID_HANDLE(handle)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
	EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_RELEASE_SHARE(handle, exc);

    /* Just close the fd, it will automagically unlock. */
    if (ctx->fd >= 0) {
	if (close(ctx->fd) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't close lock file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    ctx->fd = -1;
	    EDIR_THROW(*exc);
	}
	ctx->fd = -1;
    }
    
    if (ctx->pathname != NULL) {
	edir_free(ctx->pathname);
	ctx->pathname = NULL;
    }

    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Last ditch exception cleanup. */
    if (ctx != NULL) {
	if (ctx->fd >= 0) close(ctx->fd);
	ctx->fd = -1;
	if (ctx->pathname != NULL) edir_free(ctx->pathname);
	ctx->pathname = NULL;
	edir_free(ctx);
    }
    if (handle != NULL) edir_free(handle);
    return;
}

/**
 * Obtain an exclusive "Write" lock on a mutex.
 * <P>
 * An exclusive lock grants exclusive access to a resource.  In other
 * words, no other locks (whether shared or exclusive) will be granted
 * on the object until the holder of the current exclusive lock unlocks
 * the object.  An exclusive lock allows a critical area to be written
 * without danger of interfering with other writers OR with any readers.
 * This is the reason it is sometimes called a "Write" lock.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to lock, starting at 0.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             1 if the lock was obtained, otherwise 0.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while locking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern int
edir_mutex_lock_exclusive(edir_ipc_handle_t *handle, int num, edir_exception_t **exc)
{
    static char			*funcName = "edir_mutex_lock_exclusive";
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    edir_mutex_file_header_t	header;
#ifndef USE_FLOCK
    struct flock	 	lock;
#endif

    if (!VALID_HANDLE(handle)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
	EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;
    
    if ((num < 0) || (num >= ctx->number)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Index %d out of range (%d .. %d)", num, 0, ctx->number - 1);
        EDIR_THROW(*exc);
    }

    /* Just return if we already have the requested type of lock. */
    if (ctx->state[num] == EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE) return 1;

#ifdef USE_FLOCK
    while (flock(ctx->fd, LOCK_EX) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Error while locking on file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    EDIR_THROW(*exc);
	}
    }
#else
    lock.l_whence = SEEK_SET;
    lock.l_start = HDR_OFFSET + num;
    lock.l_len = 1;
    lock.l_type = F_WRLCK;
    lock.l_pid = 0;
    while (fcntl(ctx->fd, ctx->wait[num] ? F_SETLKW : F_SETLK, &lock) < 0) {
	if (!(ctx->wait[num]) && ((errno == EAGAIN) || (errno == EACCES))) return 0;
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Error while locking on file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    EDIR_THROW(*exc);
	}
    }
#endif

    ctx->state[num] = EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE;

    return 1;

except:
    /* Nothing to clean up. */
    return 0;
}

/**
 * Obtain a shareable "read" lock on a mutex.
 * <P>
 * More than one shareable lock may be held by processes at any given
 * time.  However, not shareable (or exclusive) locks may be held if
 * one process has successfully obtained an exclusive lock.  If an
 * exclusive lock is held elsewhere then attempts to obtain a shareable
 * lock will block until the exclusive lock is unlocked.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to lock, starting at 0.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             1 if the lock was obtained, otherwise 0.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while locking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern int
edir_mutex_lock_shared(edir_ipc_handle_t *handle, int num, edir_exception_t **exc)
{
    static char		 	*funcName = "edir_mutex_lock_shared";
    edir_mutex_filesys_ctx_t	*ctx = NULL;
#ifndef USE_FLOCK
    struct flock	 	lock;
#endif

    if (!VALID_HANDLE(handle)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
	EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;
    
    if ((num < 0) || (num >= ctx->number)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Index %d out of range (%d .. %d)", num, 0, ctx->number - 1);
        EDIR_THROW(*exc);
    }

    /* Just return if we already have the requested type of lock. */
    if (ctx->state[num] == EDIR_MUTEX_STATE_LOCKED_SHARE) return 1;

#ifdef USE_FLOCK
    while (flock(ctx->fd, LOCK_SH) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Error while read-locking on file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    EDIR_THROW(*exc);
	}
    }
#else
    lock.l_whence = SEEK_SET;
    lock.l_start = HDR_OFFSET + num;
    lock.l_len = 1;
    lock.l_type = F_RDLCK;
    lock.l_pid = 0;
    while (fcntl(ctx->fd, ctx->wait[num] ? F_SETLKW : F_SETLK, &lock) < 0) {
	if (!(ctx->wait[num]) && ((errno == EAGAIN) || (errno == EACCES))) return 0;
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Error while read-locking on file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    EDIR_THROW(*exc);
	}
    }
#endif
    ctx->state[num] = EDIR_MUTEX_STATE_LOCKED_SHARE;

    return 1;

except:
    /* Nothing to clean up. */
    return 0;
}

/**
 * Give up the currently held lock on a locked mutex.
 * <P>
 * Unlocks the mutex, allowing other threads or processes to access
 * the resource controlled by the given mutex.  Once this function
 * is called, the calling process or thread is no longer guaranteed
 * to have exclusive access to the controlled resource.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to unlock, starting at 0.
 *                     A value of -1 means "unlock all".
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while unlocking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern void
edir_mutex_unlock(edir_ipc_handle_t *handle, int num, edir_exception_t **exc)
{
    static char		 	*funcName = "edir_mutex_unlock";
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    int				i;
#ifndef USE_FLOCK
    struct flock	 	lock;
#endif

    if (!VALID_HANDLE(handle)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
	EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;

    if (num == EDIR_MUTEX_UNLOCK_ALL) {
	for (i=0; i<ctx->number; i++) {
	    edir_mutex_unlock(handle, i, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	return;
    }
    
    if ((num < 0) || (num >= ctx->number)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Index %d out of range (%d .. %d)", num, 0, ctx->number - 1);
        EDIR_THROW(*exc);
    }

    /* Just return if already unlocked. */
    if (ctx->state[num] == EDIR_MUTEX_STATE_UNLOCKED) return;

#ifdef USE_FLOCK
    while (flock(ctx->fd, LOCK_UN) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Error while unlocking file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    EDIR_THROW(*exc);
	}
    }
#else
    lock.l_whence = SEEK_SET;
    lock.l_start = HDR_OFFSET + num;
    lock.l_len = 1;
    lock.l_type = F_UNLCK;
    lock.l_pid = 0;
    while (fcntl(ctx->fd, F_SETLKW, &lock) < 0) {
	if (errno != EINTR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Error while unlocking file \"%s\": @errstr@",
		ctx->pathname ? ctx->pathname : "???");
	    EDIR_THROW(*exc);
	}
    }
#endif

    ctx->state[num] = EDIR_MUTEX_STATE_UNLOCKED;

    return;

except:
    /* Nothing to clean up. */
    return;
}

/**
 * Return zero if not locked, or nonzero state if locked.
 * <P>
 * If the mutex is not locked, this function returns zero, corresponding
 * to state EDIR_MUTEX_STATE_UNLOCKED, also corresponding to "false".
 * If the mutex is locked, the type of lock is returned as a positive
 * integer.  Possible return values are EDIR_MUTEX_STATE_LOCKED_SHARE if
 * a shareable lock is held, or EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE if an
 * exclusive lock is held.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the mutex to check.
 * @return             Zero(false) or held lock type.
 */
extern int
edir_mutex_islocked(edir_ipc_handle_t *handle, int num)
{
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return EDIR_MUTEX_STATE_UNLOCKED;
    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;
    if ((num < 0) || (num >= ctx->number)) return EDIR_MUTEX_STATE_UNLOCKED;
    return ctx->state[num];
}

/**
 * Sets or unsets blocking mode on the indexed mutex.
 * <P>
 * If set (block == 1), then lock requests on the mutex will
 * block until the mutex becomes available.  If unset (block == 0),
 * then requests will return immediately without throwing an exception
 * if the lock is not available immediately.  If the lock is unavailable
 * in nonblocking mode, zero is returned by the lock functions.
 * <P>
 * This function does not throw exceptions.  Invalid values of "num"
 * are ignored.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the mutex for which to set/unset blocking.
 * @param block  [in]  1 for blocking, 0 for nonblocking.
 * @return             Zero(false) or held lock type.
 */
extern void
edir_mutex_setblocking(edir_ipc_handle_t *handle, int num, int block)
{
    edir_mutex_filesys_ctx_t *ctx;
    if (!VALID_HANDLE(handle)) return;
    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;
    if ((num < 0) || (num >= ctx->number)) return;
    ctx->wait[num] = block ? 1 : 0;
    return;
}

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * This pointer is a reference and MUST NOT be modified!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_mutex_handle_get_object(edir_ipc_handle_t *handle)
{
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;
    if (ctx->fd < 0) return NULL;
    return (void *)(&(ctx->fd));
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char *
edir_mutex_handle_get_object_name(edir_ipc_handle_t *handle)
{
    edir_mutex_filesys_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_mutex_filesys_ctx_t *) handle->pData;
    if (ctx->pathname == NULL) return NULL;
    return edir_strdup(ctx->pathname);
}

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void
edir_mutex_set_default_filename(char *filename)
{
    DefaultSet = 1;
    DefaultFileName = edir_strdup(filename);
}

/**
 * Get the default name/key for this class.
 * <P>
 * If no default has been set with edir_mutex_set_default_filename(),
 * then return NULL;
 * 
 * @return The default file name.
 */
extern char *
edir_mutex_default_name()
{
    if (!DefaultSet) return NULL;
    return edir_strdup(DefaultFileName);
}

