/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_mutex_sysv.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_mutex_sysv.h"
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

/**
 * System-dependent implement of the Enhydra Director mutex interface,
 * using System V IPC semaphores.
 * <P>
 * Uses System V semaphores via semget(), etc. to implement a simple
 * mutex that can be used to lock data in an interprocess shared
 * resource, such as a shared memory scoreboard.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

#define EDIR_SEM_LOCK(IDX)	((IDX) * 2)
#define EDIR_SEM_COUNT(IDX)	((IDX) * 2 + 1)

#define EDIR_SEMOP_WAITZERO	0
#define EDIR_SEMOP_DOWN		-1
#define EDIR_SEMOP_UP		1

/*
 * Duhhhh this is really lame.  Solaris does not define this anywhere!
 */
#ifdef EDIR_NEED_UNION_SEMUN
union semun {
    long val;
    struct semid_ds *buf;
    ushort *array;
};
#endif

/*
 * Internal contents pointed to by "handles".
 */
typedef struct edir_mutex_sysv_ctx {
    key_t	semkey;
    int		semid;
    int		number;
    int		state[EDIR_MUTEX_MAX];
    int		wait[EDIR_MUTEX_MAX];
} edir_mutex_sysv_ctx_t;

#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_MUTEX)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

/**
 * Create a mutex in an initially disabled state.
 * <P>
 * Creates a brand new mutex.  For interprocess implementations, this
 * will be a system-wide mutex, lock, or semaphore.  The key may or may
 * not be used, depending upon the requirements of the implementation.
 * A NULL key means "use the system-dependent default".
 * Some implementations may leave system resources allocated,
 * even after the program has terminated, unless edir_mutex_destroy() is
 * called to clean up.
 * <P>
 * The new mutex is created in the locked state. 
 * 
 * @param key [in]  Handle object to use as configuration data source.
 * @param num [in]  Number of mutexes to create for this object.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return                An Handle to the mutex object.
 * @exception EDIR_EXCEPT_MUTEX If the mutex cannot be created.
 */
extern edir_ipc_handle_t *
edir_mutex_create(edir_ipc_handle_t *key, int num, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_create";
    edir_mutex_sysv_ctx_t *ctx=NULL;
    edir_ipc_handle_t	*handle=NULL;
    union semun		semarg;
    struct sembuf	semopA[EDIR_MUTEX_MAX * 2];
    int			nopsA, i, j;
    
    if ((num > EDIR_MUTEX_MAX) || (num < 1)) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't create %d mutexes (max per object=%d)", 
		    num, EDIR_MUTEX_MAX);
	EDIR_THROW(*exc);
    }

    /*
     * Create a new handle for this mutex.  Externally this is a (void *).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_MUTEX;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_INTEGER;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_mutex_sysv_ctx_t *) edir_malloc(sizeof(edir_mutex_sysv_ctx_t));
    handle->pData = ctx;

    if (key != NULL) {
	char *endptr, *val;
	if (key->handle_type != EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, 
		"Mutexes only currently accept key handles of type %s (%d)",
		"EDIR_IPC_HANDLE_TYPE_CONFIG", EDIR_IPC_HANDLE_TYPE_CONFIG);
	    EDIR_THROW(*exc);
	}
	ctx->semkey = -1;
	val = (char *) edir_ipc_config_handle_get_object(key);
	if (val) {
	    errno = 0;
	    ctx->semkey = strtol(val, &endptr, 0);
	    if ((ctx->semkey <= 0) || (endptr == val)) errno = EINVAL;
	    if (endptr) while (*endptr) if (!isspace(*endptr++)) errno = EINVAL;
	    if (errno != 0) { 
		char *cp;
		int	hash = 0;
		for (cp=val; *cp; cp++) hash += (octet)(*cp);
		ctx->semkey = ftok("/dev/null", hash);
	    }
	}
	if (ctx->semkey <= 0) {
	    ctx->semkey = (key_t) EDIR_MUTEX_SYSV_DEFAULT_IPC_KEY;
	}
    } else {
	ctx->semkey = (key_t) EDIR_MUTEX_SYSV_DEFAULT_IPC_KEY;
    }
    ctx->semid = -1;
    ctx->number = num;

    /*
     * make sure unused slots are invalid.
     */
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->state[i] = -1; /* Invalid */
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->wait[i] = 1;

    /*
     * Create the global semaphore.  Semaphore is created with value
     * initialized to zero.  Any attempt by other processes to decrement
     * the semaphore count will block until we finish setting up the
     * shared memory area. 
     */
    ctx->semid = semget(ctx->semkey, 2 * num, IPC_CREAT | IPC_EXCL | 0600);
    if (ctx->semid < 0) {
	if (errno == EEXIST) {
	    ctx->semid = semget(ctx->semkey, 0, 0600);
	    if (ctx->semid < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		  "Could not get old semaphore: @errstr@ (errnum=@errnum@");
		EDIR_THROW(*exc);
	    }
	    semarg.val = 0;
	    if (semctl(ctx->semid, 0, IPC_RMID, semarg) < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		  "Could not destroy old semaphore: @errstr@ (errnum=@errnum@");
		EDIR_THROW(*exc);
	    }
	    ctx->semid = semget(ctx->semkey, 2 * num, IPC_CREAT|IPC_EXCL|0600);
	    if (ctx->semid < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		  "Could not create new semaphore: @errstr@ (errnum=@errnum@");
		EDIR_THROW(*exc);
	    }
	} else if (errno == ENOSYS) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"System V Semaphores must be enabled in kernel");
	    EDIR_THROW(*exc);
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Cannot create semaphore: @errstr@ (errnum=@errnum@)");
	    EDIR_THROW(*exc);
	}
    }
    
    /*
     * Mutex set is currently in exclusively locked state.
     * However we want SEM_UNDO semantics to release the mutexes
     * when/if we die.  To do this, do an atomic semop() call on
     * a vector of 2 * num that simply unlock and relock the main locks.
     * The relock operation adds the SEM_UNDO flag to ensure cleanup
     * on exit.
     */
    for (i=0; i<num; i++) {
	j = i * 2;
	/* Op #0 Unlocks the main lock. */
	semopA[j].sem_num = EDIR_SEM_LOCK(i);
	semopA[j].sem_op = EDIR_SEMOP_UP;
	semopA[j].sem_flg = 0;
	/* Op #1, in the same atomic action as #0, relocks with SEM_UNDO */
	semopA[j+1].sem_num = EDIR_SEM_LOCK(i);
	semopA[j+1].sem_op = EDIR_SEMOP_DOWN;
	semopA[j+1].sem_flg = SEM_UNDO;
    }
    nopsA = num * 2;

    while (semop(ctx->semid, semopA, nopsA) < 0) {
	if (errno != EINTR) {
	    if ((errno == EIDRM) || (errno == EINVAL)) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_DELETED,
		    "Semaphores deleted by creator");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Error relocking semaphores: @errstr@ (errnum=@errnum@)");
		EDIR_THROW(*exc);
	    }
	}
    }
    
    for (i=0; i<num; i++) ctx->state[i] = EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE;

    return handle;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	if (ctx->semid >= 0) {
	    semarg.val = 0;
	    semctl(ctx->semid, 0, IPC_RMID, semarg);
	}
	edir_free(ctx);
    }
    if (handle != NULL) edir_free(handle);
    return NULL;
}

/**
 * Destroy a mutex.
 * <P>
 * Deletes the specified mutex and removes its resources from the system.
 * If client threads or processes attempt to access the mutex after this
 * function is called, they will experience EDIR_EXCEPT_MUTEX_DELETED
 * exceptions.
 * 
 * @param handle [in]  Handle returned by edir_mutex_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If the mutex cannot be destroyed.
 */
extern void
edir_mutex_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_destroy";
    edir_mutex_sysv_ctx_t *ctx=NULL;
    union semun		semarg;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_DESTROY_SHARE(handle, exc);

    if (ctx->semid >= 0) {
	semarg.val = 0;
	if (semctl(ctx->semid, 0, IPC_RMID, semarg) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Cannot delete semaphore: @errstr@ (errnum=@errnum@)");
	    ctx->semid = -1;
	    EDIR_THROW(*exc);
	}
	ctx->semid = -1;
    }

    edir_free(ctx);

    return;

except:
    /* Last ditch exception cleanup. At least free the handle memory. */
    if (ctx) edir_free(ctx);
    if (handle) edir_free(handle);
    return;
}

/**
 * Obtain an existing mutex.
 * <P>
 * A mutex must already exist for the given key (if keys are used).
 * If the mutex exists, this function returns a handle to a mutex object.
 * <P>
 * NOTE! This is NOT the same as "locking" the mutex.  This function
 * simply gets a reference to the mutex object from the system.  To lock
 * access to the resource controlled by this mutex, call edir_mutex_lock().
 * 
 * @param key [in]  Handle object to use as configuration data source.
 * @param num [out] The number of mutexes controlled by this object.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          An Handle to the mutex object.
 * @exception EDIR_EXCEPT_MUTEX If the specified mutex is not available.
 */
extern edir_ipc_handle_t *
edir_mutex_obtain(edir_ipc_handle_t *key, int *num, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_obtain";
    edir_mutex_sysv_ctx_t *ctx=NULL;
    edir_ipc_handle_t	*handle=NULL;
    union semun		semarg;
    struct semid_ds	semstat;
    int			i;
    
    if (num == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Null 'num' pointer.");
	EDIR_THROW(*exc);
    }

    /*
     * Create a new handle for this mutex.  Externally this is a (void *).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_MUTEX;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_INTEGER;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_mutex_sysv_ctx_t *) edir_malloc(sizeof(edir_mutex_sysv_ctx_t));
    handle->pData = ctx;

    if (key != NULL) {
	char *endptr, *val;
	if (key->handle_type != EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, 
		"Mutexes only currently accept key handles of type %s (%d)",
		"EDIR_IPC_HANDLE_TYPE_CONFIG", EDIR_IPC_HANDLE_TYPE_CONFIG);
	    EDIR_THROW(*exc);
	}
	ctx->semkey = -1;
	val = (char *) edir_ipc_config_handle_get_object(key);
	if (val) {
	    errno = 0;
	    ctx->semkey = strtol(val, &endptr, 0);
	    if ((ctx->semkey <= 0) || (endptr == val)) errno = EINVAL;
	    if (endptr) while (*endptr) if (!isspace(*endptr++)) errno = EINVAL;
	    if (errno != 0) { 
		char *cp;
		int	hash = 0;
		for (cp=val; *cp; cp++) hash += (octet)(*cp);
		ctx->semkey = ftok("/dev/null", hash);
	    }
	}
	if (ctx->semkey <= 0) {
	    ctx->semkey = (key_t) EDIR_MUTEX_SYSV_DEFAULT_IPC_KEY;
	}
    } else {
	ctx->semkey = (key_t) EDIR_MUTEX_SYSV_DEFAULT_IPC_KEY;
    }

    ctx->semid = -1;
    /* Invalidate state vars until we fill in */
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->state[i] = -1;
    for (i=0; i<EDIR_MUTEX_MAX; i++) ctx->wait[i] = 1;

    /*
     * Find the specified semaphore.
     */
    ctx->semid = semget(ctx->semkey, 0, 0600);
    if (ctx->semid < 0) {
	if (errno == ENOENT) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Semaphore not found");
	    EDIR_THROW(*exc);
	} else if ((errno == EACCES) || (errno == EPERM)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Semaphore permission denied - are we running as right user?");
	    EDIR_THROW(*exc);
	} else if (errno == ENOSYS) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"System V Semaphores must be enabled in kernel");
	    EDIR_THROW(*exc);
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Cannot open semaphore: @errstr@ (errnum=@errnum@)");
	    EDIR_THROW(*exc);
	}
    }
    
    /*
     * Find out how many sems.
     */
    semarg.buf = &semstat;
    if (semctl(ctx->semid, 0, IPC_STAT, semarg) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't stat semaphore: @errstr@ (errnum=@errnum@)");
	EDIR_THROW(*exc);
    }
    ctx->number = semstat.sem_nsems / 2;
    
    if (ctx->number < 1) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Less than one lock (%d)", ctx->number);
	EDIR_THROW(*exc);
    }

    if (ctx->number > EDIR_MUTEX_MAX) {
	*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Obtained semaphore has too many locks (%d)", ctx->number);
	EDIR_THROW(*exc);
    }
    
    *num = ctx->number;
    
    for (i=0; i<ctx->number; i++) ctx->state[i] = EDIR_MUTEX_STATE_UNLOCKED;

    return handle;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    *num = 0;
    return NULL;
}

/**
 * Release a previously obtained mutex.
 * <P>
 * Releases this process's or thread's reference to the specified mutex,
 * but does not remove the underlying object.  This function is used to
 * "free" mutexes obtained using edir_mutex_obtain().
 * <P>
 * NOTE! This is NOT the same as "unlocking" the mutex.  To unlock
 * the mutex so other processes/threads can access it, use edir_mutex_unlock().
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while releasing the mutex.
 */
extern void
edir_mutex_release(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_release";
    edir_mutex_sysv_ctx_t *ctx=NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_RELEASE_SHARE(handle, exc);

    /* Oh well, we'll be "nice" and unlock if currently locked. */
    edir_mutex_unlock(handle, EDIR_MUTEX_UNLOCK_ALL, exc);
    if (*exc) EDIR_THROW(*exc);

    /* Nothing else to do but to free the handle memory. */
    ctx->semid = -1;
    ctx->semkey = (key_t) -1;
    edir_free(ctx);

    return;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	ctx->semid = -1;
	ctx->semkey = (key_t) -1;
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Obtain an exclusive "Write" lock on a mutex.
 * <P>
 * An exclusive lock grants exclusive access to a resource.  In other
 * words, no other locks (whether shared or exclusive) will be granted
 * on the object until the holder of the current exclusive lock unlocks
 * the object.  An exclusive lock allows a critical area to be written
 * without danger of interfering with other writers OR with any readers.
 * This is the reason it is sometimes called a "Write" lock.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to lock, starting at 0.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             1 if the lock was obtained, otherwise 0.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while locking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern int
edir_mutex_lock_exclusive(edir_ipc_handle_t *handle, int num, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_lock";
    edir_mutex_sysv_ctx_t *ctx=NULL;
    struct sembuf	semopA[3];
    int			nopsA;
    struct sembuf	semopB[1];
    int			nopsB;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;
    
    if ((num < 0) || (num >= ctx->number)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Index %d out of range (%d .. %d)", num, 0, ctx->number - 1);
        EDIR_THROW(*exc);
    }

    switch (ctx->state[num]) {
	case EDIR_MUTEX_STATE_LOCKED_SHARE:
	    if (ctx->wait[num]) {
		/* Promote to exclusive lock.  Like flock() this first
		 * releases the lock, then reobtains the write lock
		 * nonatomically.  In the midst of this, someone else
		 * could get an exclusive lock before us.  That's the
		 * way it goes.  Trying to do this atomically creates
		 * a severe deadlock hazard with other processes
		 * trying also to get an exclusive lock.
		 */
		edir_mutex_unlock(handle, num, exc);
		if (*exc) EDIR_THROW(*exc);
		edir_mutex_lock_exclusive(handle, num, exc);
		if (*exc) EDIR_THROW(*exc);
		return 1; /* Got the lock! */
	    } else {
		/*
		 * For nonblocking mode, we need to make sure
		 * state is restored if we fail.  This means
		 * an atomic action.  This is not a deadlock
		 * hazard, since we don't wait, thus breaking
		 * the hold and wait condition required for a deadlock.
		 */
		/* First, release share lock */
		semopA[0].sem_num = EDIR_SEM_COUNT(num);
		semopA[0].sem_op = EDIR_SEMOP_DOWN;
		semopA[0].sem_flg = SEM_UNDO | IPC_NOWAIT;
		/* Now, obtain main lock if possible */
		semopA[1].sem_num = EDIR_SEM_LOCK(num);
		semopA[1].sem_op = EDIR_SEMOP_DOWN;
		semopA[1].sem_flg = SEM_UNDO | IPC_NOWAIT;
		/* Check if there are still shared locks */
		semopA[2].sem_num = EDIR_SEM_COUNT(num);
		semopA[2].sem_op = EDIR_SEMOP_WAITZERO;
		semopA[2].sem_flg = SEM_UNDO | IPC_NOWAIT;
		nopsA = 3;
		nopsB = 0;
	    }
	    break;
	case EDIR_MUTEX_STATE_UNLOCKED:
	    if (ctx->wait[num]) {
		/* First gain main lock and keep it. */
		semopA[0].sem_num = EDIR_SEM_LOCK(num);
		semopA[0].sem_op = EDIR_SEMOP_DOWN;
		semopA[0].sem_flg = SEM_UNDO;
		nopsA = 1;
		/* Now, other processes can only unlock without blocking
		 * Also, deadlock is avoided by requiring that a holder
		 * of a share lock drop the lock non-atomically before
		 * "promoting" to an exclusive lock.
		 */
		semopB[0].sem_num = EDIR_SEM_COUNT(num);
		semopB[0].sem_op = EDIR_SEMOP_WAITZERO;
		semopB[0].sem_flg = SEM_UNDO;
		nopsB = 1;
	    } else {
		/*
		 * Try nonblocking operation in the same
		 * atomic attempt to ensure proper cleanup by the
		 * kernel if the lock is busy.
		 */
		semopA[0].sem_num = EDIR_SEM_LOCK(num);
		semopA[0].sem_op = EDIR_SEMOP_DOWN;
		semopA[0].sem_flg = SEM_UNDO | IPC_NOWAIT;
		semopA[1].sem_num = EDIR_SEM_COUNT(num);
		semopA[1].sem_op = EDIR_SEMOP_WAITZERO;
		semopA[1].sem_flg = SEM_UNDO | IPC_NOWAIT;
		nopsA = 2;
		nopsB = 0;
	    }
	    break;
	case EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE:
	    /* We're already locked as specified -- just return. */
	    return;
	default:
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Mutex in unknown state %d", ctx->state);
	    EDIR_THROW(*exc);
    }

    if (nopsA > 0) while (semop(ctx->semid, semopA, nopsA) < 0) {
	if (!(ctx->wait[num]) && errno == EAGAIN) return 0;
	if (errno != EINTR) {
	    if ((errno == EIDRM) || (errno == EINVAL)) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_DELETED,
			"Semaphore deleted by creator");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Error locking semaphore: @errstr@ (errnum=@errnum@)");
		EDIR_THROW(*exc);
	    }
	}
    }

    if (nopsB > 0) while (semop(ctx->semid, semopB, nopsB) < 0) {
	if (!(ctx->wait[num]) && errno == EAGAIN) return 0;
	if (errno != EINTR) {
	    if ((errno == EIDRM) || (errno == EINVAL)) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_DELETED,
			"Semaphore deleted by creator");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Error locking semaphore: @errstr@ (errnum=@errnum@)");
		EDIR_THROW(*exc);
	    }
	}
    }

    ctx->state[num] = EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE;

    return 1;

except:
    /* Nothing to clean up here. */
    return 0;
}

/**
 * Obtain a shareable "read" lock on a mutex.
 * <P>
 * More than one shareable lock may be held by processes at any given
 * time.  However, not shareable (or exclusive) locks may be held if
 * one process has successfully obtained an exclusive lock.  If an
 * exclusive lock is held elsewhere then attempts to obtain a shareable
 * lock will block until the exclusive lock is unlocked.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to lock, starting at 0.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             1 if the lock was obtained, otherwise 0.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while locking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern int
edir_mutex_lock_shared(edir_ipc_handle_t *handle, int num, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_lock_shared";
    edir_mutex_sysv_ctx_t *ctx=NULL;
    struct sembuf	semopA[3];
    int			nopsA;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;

    if ((num < 0) || (num >= ctx->number)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Index %d out of range (%d .. %d)", num, 0, ctx->number - 1);
        EDIR_THROW(*exc);
    }

    switch (ctx->state[num]) {
	case EDIR_MUTEX_STATE_UNLOCKED:
	    if (ctx->wait[num]) {
		/* Wait for shared lock */
		/* Op #0 requests main lock: wait until no exclusive locks */
		semopA[0].sem_num = EDIR_SEM_LOCK(num);
		semopA[0].sem_op = EDIR_SEMOP_DOWN;
		semopA[0].sem_flg = SEM_UNDO;
		/* Op #1 ups the share lock count, blocking exclusive locks */
		semopA[1].sem_num = EDIR_SEM_COUNT(num);
		semopA[1].sem_op = EDIR_SEMOP_UP;
		semopA[1].sem_flg = SEM_UNDO;
		/* Op #2 releases the main lock to allow other shared locks. */
		semopA[2].sem_num = EDIR_SEM_LOCK(num);
		semopA[2].sem_op = EDIR_SEMOP_UP;
		semopA[2].sem_flg = SEM_UNDO;
		nopsA = 3;
	    } else {
		/* Test and get shared lock without blocking, if possible */
		/* Op #0 requests main lock: wait until no exclusive locks */
		semopA[0].sem_num = EDIR_SEM_LOCK(num);
		semopA[0].sem_op = EDIR_SEMOP_DOWN;
		semopA[0].sem_flg = SEM_UNDO | IPC_NOWAIT;
		/* Op #1 ups the share lock count, blocking exclusive locks */
		semopA[1].sem_num = EDIR_SEM_COUNT(num);
		semopA[1].sem_op = EDIR_SEMOP_UP;
		semopA[1].sem_flg = SEM_UNDO | IPC_NOWAIT;
		/* Op #2 releases the main lock to allow other shared locks. */
		semopA[2].sem_num = EDIR_SEM_LOCK(num);
		semopA[2].sem_op = EDIR_SEMOP_UP;
		semopA[2].sem_flg = SEM_UNDO | IPC_NOWAIT;
		nopsA = 3;
	    }
	    break;
	case EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE:
	    if (ctx->wait[num]) {
		/* "Demote" atomically to shared lock */
		/* Note, we already have the main lock */
		/* Op #0 ups the share lock count, blocking exclusive locks */
		semopA[0].sem_num = EDIR_SEM_COUNT(num);
		semopA[0].sem_op = EDIR_SEMOP_UP;
		semopA[0].sem_flg = SEM_UNDO;
		/* Op #1 releases the main lock to allow other shared locks. */
		semopA[1].sem_num = EDIR_SEM_LOCK(num);
		semopA[1].sem_op = EDIR_SEMOP_UP;
		semopA[1].sem_flg = SEM_UNDO;
		nopsA = 2;
	    } else {
		/* "Demote" atomically to shared lock without blocking
		 * Note that this is somewhat anal retentive, since
		 * we are doing two unlocks, which will never block 
		 * anyway.
		 */
		semopA[0].sem_num = EDIR_SEM_COUNT(num);
		semopA[0].sem_op = EDIR_SEMOP_UP;
		semopA[0].sem_flg = SEM_UNDO | IPC_NOWAIT;
		/* Op #1 releases the main lock to allow other shared locks. */
		semopA[1].sem_num = EDIR_SEM_LOCK(num);
		semopA[1].sem_op = EDIR_SEMOP_UP;
		semopA[1].sem_flg = SEM_UNDO | IPC_NOWAIT;
		nopsA = 2;
	    }
	    break;
	case EDIR_MUTEX_STATE_LOCKED_SHARE:
	    /* We're already locked as specified -- just return. */
	    return 1;
	default:
	    /* Unexpected state - pitch a fit */
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Mutex in unexpected state %d", ctx->state);
	    EDIR_THROW(*exc);
    }

    while (semop(ctx->semid, semopA, nopsA) < 0) {
	if (!(ctx->wait[num]) && (errno == EAGAIN)) return 0;
	if (errno != EINTR) {
	    if ((errno == EIDRM) || (errno == EINVAL)) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_DELETED,
			"Semaphore deleted by creator (A)");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Error locking semaphore (A): @errstr@ (errnum=@errnum@)");
		EDIR_THROW(*exc);
	    }
	}
    }

    ctx->state[num] = EDIR_MUTEX_STATE_LOCKED_SHARE;

    return 1;

except:
    /* Nothing to clean up here. */
    return 0;
}   

/**
 * Give up the currently held lock on a locked mutex.
 * <P>
 * Unlocks the mutex, allowing other threads or processes to access
 * the resource controlled by the given mutex.  Once this function
 * is called, the calling process or thread is no longer guaranteed
 * to have exclusive access to the controlled resource.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the particular mutex to unlock, starting at 0.
 *                     A value of -1 means "unlock all".
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_MUTEX If a failure occurs while unlocking the mutex.
 * @exception EDIR_EXCEPT_MUTEX_DELETED If the mutex was deleted by its creator.
 */
extern void
edir_mutex_unlock(edir_ipc_handle_t *handle, int num, edir_exception_t **exc)
{
    static char		*funcName = "edir_mutex_unlock";
    edir_mutex_sysv_ctx_t *ctx=NULL;
    struct sembuf	semopA[1];
    int			nopsA, i;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX, "Invalid mutex handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;
    
    if (num == EDIR_MUTEX_UNLOCK_ALL) {
	for (i=0; i<ctx->number; i++) {
	    edir_mutex_unlock(handle, i, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	return;
    }
    
    if ((num < 0) || (num >= ctx->number)) {
        *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Index %d out of range (%d .. %d)", num, 0, ctx->number - 1);
        EDIR_THROW(*exc);
    }

    switch (ctx->state[num]) {
	case EDIR_MUTEX_STATE_LOCKED_SHARE:
	    /* Unlock a shareable lock */
	    semopA[0].sem_num = EDIR_SEM_COUNT(num);
	    semopA[0].sem_op = EDIR_SEMOP_DOWN;
	    semopA[0].sem_flg = SEM_UNDO;
	    nopsA = 1;
	    break;
	case EDIR_MUTEX_STATE_LOCKED_EXCLUSIVE:
	    /* Unlock an exclusive lock */
	    semopA[0].sem_num = EDIR_SEM_LOCK(num);
	    semopA[0].sem_op = EDIR_SEMOP_UP;
	    semopA[0].sem_flg = SEM_UNDO;
	    nopsA = 1;
	    break;
	case EDIR_MUTEX_STATE_UNLOCKED:
	    /* We're already unlocked -- just return. */
	    return;
	default:
	    /* Unexpected state - pitch a fit */
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Mutex in unexpected state %d", ctx->state);
	    EDIR_THROW(*exc);
    }

    while (semop(ctx->semid, semopA, nopsA) < 0) {
	if (errno != EINTR) {
	    if ((errno == EIDRM) || (errno == EINVAL)) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX_DELETED,
			"Semaphore deleted by creator (A)");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Error unlocking semaphore: @errstr@ (errnum=@errnum@)");
		EDIR_THROW(*exc);
	    }
	}
    }

    ctx->state[num] = EDIR_MUTEX_STATE_UNLOCKED;

    return;

except:
    /* Nothing to clean up here. */
    return;
}

/**
 * Return 1 (true) if the mutex is locked, otherwise return 0 (false).
 * <P>
 * This is mostly a convenience for exception cleanup code to ensure
 * that locks are released.  It allows code such as:
 * 	except:
 * 		if (edir_mutex_islocked(mutex)) edir_mutex_unlock(mutex);
 * 		return NULL;
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the mutex to check.
 * @return             1 (true) if locked, otherwise 0 (false).
 */
extern int
edir_mutex_islocked(edir_ipc_handle_t *handle, int num)
{
    edir_mutex_sysv_ctx_t *ctx;
    if (!VALID_HANDLE(handle)) return EDIR_MUTEX_STATE_UNLOCKED;
    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;
    if ((num < 0) || (num >= ctx->number)) return EDIR_MUTEX_STATE_UNLOCKED;
    return ctx->state[num];
}

/**
 * Sets or unsets blocking mode on the indexed mutex.
 * <P>
 * If set (block == 1), then lock requests on the mutex will
 * block until the mutex becomes available.  If unset (block == 0),
 * then requests will return immediately without throwing an exception
 * if the lock is not available immediately.  If the lock is unavailable
 * in nonblocking mode, zero is returned by the lock functions.
 * <P>
 * This function does not throw exceptions.  Invalid values of "num"
 * are ignored.
 * 
 * @param handle [in]  Handle returned by edir_mutex_obtain().
 * @param num    [in]  Index of the mutex for which to set/unset blocking.
 * @param block  [in]  1 for blocking, 0 for nonblocking.
 * @return             Zero(false) or held lock type.
 */
extern void
edir_mutex_setblocking(edir_ipc_handle_t *handle, int num, int block)
{
    edir_mutex_sysv_ctx_t *ctx;
    if (!VALID_HANDLE(handle)) return;
    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;
    if ((num < 0) || (num >= ctx->number)) return;
    ctx->wait[num] = block ? 1 : 0;
    return;
}

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * This pointer is a reference and MUST NOT be modified!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_mutex_handle_get_object(edir_ipc_handle_t *handle)
{
    edir_mutex_sysv_ctx_t		*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;
    if (ctx->semkey == -1) return NULL;
    return (void *)(&(ctx->semkey));
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char *
edir_mutex_handle_get_object_name(edir_ipc_handle_t *handle)
{
    edir_mutex_sysv_ctx_t	*ctx = NULL;
    char		buf[64];
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_mutex_sysv_ctx_t *) handle->pData;
    if (ctx->semkey == -1) return NULL;
    sprintf(buf, "%08x", ctx->semkey);
    return edir_strdup(buf);
}

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void
edir_mutex_set_default_filename(char  *filename)
{
    ; /* Do nothing.  SysV IPC uses keys, not files. */
}

/**
 * Get the default name/key for this class.
 * <P>
 * If no default has been set with edir_mutex_set_default_filename(),
 * then return NULL;
 * 
 * @return The default file name.
 */
extern char *
edir_mutex_default_name()
{
    /* No filename can ever be set for sysv */
    return NULL;
}


