/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_proto.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"

/**
 * fills in a 4 octet array with an encoded protocol header for
 * a PKT_TYPE_HEADER packet.
 * <P>
 * The actual length of packet data is (nameLen) + (valueLen).
 * 
 * @param pkt      [in,out] Buffer to write packet to.
 * @param type     [in]     Type of this packet.
 * @param subtype  [in]     Subtype of this packet.
 * @param length   [in]     Length of the data portion of this packet.
 */
extern void
edir_proto_make_header(octet *pkt, int16 type, int16 subtype, int32 length)
{
    pkt[0] = (octet) ((type << 4) | (subtype & 0xf));
    pkt[1] = (octet) ((length >> 16) & 0xff);
    pkt[2] = (octet) ((length >> 8) & 0xff);
    pkt[3] = (octet) (length & 0xff);
}

/**
 * fills in a 4 octet array with an encoded protocol header for
 * a PKT_TYPE_HEADER packet.
 * <P>
 * The actual length of packet data is (nameLen) + (valueLen).
 * 
 * @param pkt      [in,out] Buffer to write packet to.
 * @param subtype  [in]     Subtype of this packet.
 * @param nameLen  [in]     Length in octets of the name of this CGI header.
 * @param valueLen [in]     Length in octets of the value of this CGI header.
 */
extern void
edir_proto_make_header_header(octet *pkt, int16 subtype,
			    int32 nameLen, int32 valueLen)
{
    int32 length;
    static int32 type = ( EDIR_PKT_TYPE_HEADER << 4 );
    length = ((nameLen & 0x3ff) << 14) | (valueLen & 0x3fff);
    pkt[0] = (octet) type | (subtype & 0xf);
    pkt[1] = (octet) ((length >> 16) & 0xff);
    pkt[2] = (octet) ((length >> 8) & 0xff);
    pkt[3] = (octet) (length & 0xff);
}

/**
 * Sends a generic non-cgi-header packet over the connection.
 * <P>
 * Encodes an appropriate protocol header, then sends the header and
 * any packet data over the connection to the server.
 * <P>
 * This function does not check for reentry due to non-blocking I/O.
 * It expects the state machine to go to a state where 
 * 
 * @param conn     [in,out] Per-connection context block.
 * @param type     [in,out] Type of this packet.
 * @param subtype  [in]     Subtype of this packet.
 * @param nameLen  [in]     Length in octets of the name of this CGI header.
 * @param valueLen [in]     Length in octets of the value of this CGI header.
 */
extern void
edir_proto_send_packet(edir_connection_t *conn, int type, int subtype,
		     edir_exception_t **exc )
{
    static char *funcName = "edir_proto_send_packet";

    edir_proto_make_header(conn->headerBuf, (int16)type, (int16)subtype, conn->sendDataLen);
    if (conn->sbuf == NULL)
	conn->sbuf = edir_sbuf_create();
    else
	edir_sbuf_clear(conn->sbuf);
    edir_sbuf_bincatn(conn->sbuf, conn->headerBuf, 4);
    edir_sbuf_bincatn(conn->sbuf, conn->sendData, conn->sendDataLen);
    edir_socket_write_sbuf(conn, exc, conn->sbuf);
    if (*exc) EDIR_THROW(*exc); /* Add to exception stack trace. */
    return;
except:
    /* Nothing to clean up. */
    return;
}

/**
 * Sends a generic non-cgi-header packet over the connection.
 * <P>
 * Encodes an appropriate protocol header, then sends the header and
 * any packet data over the connection to the server.
 * <P>
 * This function does not check for reentry due to non-blocking I/O.
 * It expects the state machine to go to a state where 
 * 
 * @param conn     [in,out] Per-connection context block.
 * @param major	   [in] Major version number to send.
 * @param minor	   [in] Minor version number to send.
 * @param exc	   [out] Filled in if an error occurs.
 */
extern void
edir_proto_send_version(edir_connection_t *conn, int major, int minor,
		      edir_exception_t **exc)
{
    static char *funcName = "edir_proto_send_version";
    uint32 val;
    val = (major << 12) & 0xfff000;
    val |= minor & 0xfff;
    edir_proto_make_header(conn->headerBuf, EDIR_PKT_TYPE_PROTO,
			 EDIR_PKT_SUBTYPE_PROTO_VERSION, val);
    edir_socket_write(conn, exc, conn->headerBuf, 4);
    if (*exc) EDIR_THROW(*exc); /* Add to exception stack trace. */
    return;
except:
    /* Nothing to clean up. */
    return;
}

/**
 * Sends a packet of type EDIR_PKT_TYPE_HEADER to the server.
 * <P>
 * Encodes the packet header and sends the header, (conn->headerName),
 * and (conn->headerValue) to the server.
 * 
 * @param conn     [in,out] Per-connection context block.
 * @param type     [in,out] Type of this packet.
 * @param exc	   [out]    Exception context.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR If a error is encountered while writing
 * 	to the socket.
 */
extern void
edir_proto_send_header_packet(edir_connection_t *conn, edir_exception_t **exc )
{
    static char *funcName = "edir_proto_send_header_packet";

    edir_proto_make_header_header(conn->headerBuf, conn->packetSubType,
			    conn->headerNameLen, conn->headerValueLen);
    if (conn->sbuf == NULL)
	conn->sbuf = edir_sbuf_create();
    else
	edir_sbuf_clear(conn->sbuf);
    edir_sbuf_bincatn(conn->sbuf, conn->headerBuf, 4);
    edir_sbuf_bincatn(conn->sbuf, conn->headerName, conn->headerNameLen);
    edir_sbuf_bincatn(conn->sbuf, conn->headerValue, conn->headerValueLen);
    edir_socket_write_sbuf(conn, exc, conn->sbuf);
    if (*exc) EDIR_THROW(*exc); /* Add to exception stack trace. */
    return;
except:
    /* Nothing to clean up. */
    return;
}

/**
 * Finish sending data when non-blocking write becomes possible.
 * <P>
 * This function is only called by non-blocking implementations.
 * We get to this point because select() or some other event driver
 * has determined that we can now write at least one byte of data to
 * the socket without blocking.  
 * <P>
 * The call to this method may be repeated numerous times until the
 * output buffer has been entirely sent.  For this reason, this function
 * should be called from its own state in the state machine so that
 * "writeable" event is dispatched back here.
 * <P>
 * In the previous edir_socket_write*() call, the output buffer and
 * buffer position were set up so that the state of the write
 * operation is always accurately reflected in the connection
 * context.
 * 
 * @param conn     [in,out] Per-connection context block.
 * @param exc	   [out]    Exception context.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR If a error is encountered while writing
 * 	to the socket.
 */
extern void
edir_proto_finish_send(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_proto_finish_send";
    edir_socket_finish_send(conn, exc);
    if (*exc) EDIR_THROW(*exc); /* Add to exception stack trace. */
    return;
except:
    /* Nothing to clean up. */
    return;
}

/**
 * Receive a single 4 byte packet header into (conn->headerBuf).
 * <P>
 * Also decodes the contents and sets (conn->packetType),
 * (conn->packetSubType), (conn->packetLength), and (conn->packetRemainder)
 * to their correct values for the new packet.
 * <P>
 * Sets (conn->packetOctetsIn) to 0 to indicate that no octets from the
 * packet have been read yet.
 * <P>
 * In cases of non-blocking I/O, this function may be repeatedly entered
 * by a state function until EDIR_IO_PENDING(conn) returns false.
 * 
 * @param conn	[in,out]   Connection context in which we read data.
 * @param exc	[in,out]   Exception context. (NULL OK)
 */
extern void
edir_proto_receive_packet_header(edir_connection_t *conn, edir_exception_t **exc)
{
    octet *rawHdr;
    int32 length;
    static char *funcName = "edir_proto_receive_packet_header";
    if (EDIR_IO_PENDING(conn)) {
	edir_socket_finish_receive(conn, exc);
    } else {
	conn->packetType = 0;
	conn->packetSubType = 0;
	conn->packetLength = 0;
	conn->packetRemainder = 0;
	conn->dataExpected = 4;
	edir_socket_receive_all(conn, exc, conn->headerBuf, 4);
    }
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) return; /* Still must wait for data */

    if (conn->dataActual < conn->dataExpected) {
	if (conn->dataActual <= 0)
	    *exc = edir_exception_create(EDIR_EXCEPT_PROTO,
			"Unexpected end-of-data (expected packet)");
	else
	    *exc = edir_exception_create(EDIR_EXCEPT_PROTO,
			"Unexpected end-of-data (mid-packet-header)");
	EDIR_THROW(*exc);
    }

    rawHdr = &(conn->headerBuf[0]);
    conn->packetType = rawHdr[0] >> 4;
    conn->packetSubType = rawHdr[0] & 0xf;
    length = ((int32) rawHdr[1]) << 16;
    length |= ((int32) rawHdr[2]) << 8;
    length |= ((int32) rawHdr[3]);
    /* Important! Decode header packets correctly. */
    switch (conn->packetType) {
	case EDIR_PKT_TYPE_HEADER:
	    conn->headerNameLen = (length >> 14) & 0x3ff;
	    conn->headerValueLen = length & 0x3fff;
	    conn->packetLength = conn->headerNameLen + conn->headerValueLen;
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    /* Version header always encodes major:12 minor:12 */
	    if (conn->packetSubType == EDIR_PKT_SUBTYPE_PROTO_VERSION) {
		conn->serverMajorVersion = (length >> 12) & 0xfff;
		conn->serverMinorVersion = length & 0xfff;
		conn->packetLength = 0;
	    } else conn->packetLength = length;
	    break;
	default:
	    conn->packetLength = length;
	    break;
    }
    conn->packetRemainder = conn->packetLength;
    return;
except:
    return;
}

/**
 * Receive a whole packet.
 * <P>
 * fills it with the contents of an entire packet, up to a specified
 * maximum length.  Any data octets beyond the maximum length are
 * read and discarded.
 * <PRE>
 * Special situations:
 * 	1.  Recognize incoming data packet and only receive header.
 * 		Data packets can be very large, so receipt of them 
 * 		must take place in a distinct connection state so that
 * 		reasonable buffered chunks can be processed.
 * 	2.  Recognize headers and use correct header length encoding.
 * 		Headers use a different encoding in order to encode
 * 		the length of both the header name and value in 
 * 		the 24 bit length field of the header.
 * 		This is taken care of in edir_proto_receive_packet_header().
 * </PRE>
 * 		
 * @param conn [in,out]  Connection context in which to read.
 * @param exc  [out]     Exception context.
 */
extern void
edir_proto_receive_packet(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_proto_receive_packet";
    int datalen;
    int len;

    len = conn->receiveDataMax;

    if (!conn->haveHeader) {
	edir_proto_receive_packet_header(conn, exc);
	if (*exc) EDIR_THROW(*exc);
	if (EDIR_IO_PENDING(conn)) return; /* Still must wait for data */
	conn->haveHeader = 1;
    }
    /* Situation #1 (above) -- Incoming data packet */
    if (conn->packetType == EDIR_PKT_TYPE_RESP) {
	if (conn->packetSubType == EDIR_PKT_SUBTYPE_RESP_DATA) {
	    conn->haveHeader = conn->haveData = 0;
	    return; /* Only get header for data packets. */
	}
    }
    if (!conn->haveData) {
	if (EDIR_IO_PENDING(conn)) {
	    edir_socket_finish_receive(conn, exc);
	} else {
	    if (len < conn->packetLength) datalen = len;
	    else datalen = conn->packetLength;
	    conn->dataExpected = datalen; /* Expect exactly this many. */
	    edir_socket_receive_all(conn, exc, conn->receiveData, datalen);
	}
	if (*exc) EDIR_THROW(*exc);
	if (EDIR_IO_PENDING(conn)) return; /* Still must wait for data */
	conn->packetRemainder -= conn->dataActual;

	/* Check for short packet */
	if (conn->dataActual < conn->dataExpected) {
	    *exc = edir_exception_create(EDIR_EXCEPT_PROTO,
			"Unexpected end-of-data (mid-packet)");
	    EDIR_THROW(*exc);
	}
	conn->haveData = 1;
	conn->receiveDataLen = conn->dataActual;
    }
    /* Now, discard any extra data. */
    if (EDIR_IO_PENDING(conn)) {
	edir_socket_finish_skip(conn, exc);
    } else {
	if (conn->packetRemainder > 0) {
	    edir_socket_skip(conn, exc, conn->packetRemainder);
	    if (*exc) EDIR_THROW(*exc);
	    if (EDIR_IO_PENDING(conn)) return; /* Still must wait for data */
	}
    }
    /* Finally, we have the data -- clear these flags for later calls. */
    conn->haveHeader = conn->haveData = 0;
    /* Important! Set header name and value ptrs if needed! */
    if (conn->packetType == EDIR_PKT_TYPE_HEADER) {
	conn->headerName = conn->receiveData;
	conn->headerValue = conn->receiveData + conn->headerNameLen;
    }
    return;
except:
    /* Exception cleanup. */
    conn->haveHeader = conn->haveData = 0;
    return;
}

/**
 * Receive a response data fragment from the current incoming packet.
 * <P>
 * This function expects the state machine to have already set up
 * the (conn->receiveData) buffer to the default buffering length.
 * 
 * @param conn  [in,out] Connection context in which to read data.
 * @param exc   [out]    Exception context.
 */
extern void
edir_proto_receive_packet_data(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_proto_receive_packet_data";
    if (EDIR_IO_PENDING(conn)) {
	edir_socket_finish_receive(conn, exc);
    } else {
	edir_socket_receive(conn, exc, conn->receiveData,
		 EDIR_MIN(conn->packetRemainder, conn->receiveDataMax));
    }
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) return; /* Still must wait for data */

    /* Check for unexpected end-of-data. */
    if (conn->dataActual < 1) {
	*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
		    "Unexpected end-of-data (mid-data-packet)");
	EDIR_THROW(*exc);
    }

    conn->packetRemainder -= conn->dataActual;

    /* Return any data received immediately. */
    conn->receiveDataLen = conn->dataActual;
    return;
    
except:
    /* Exception cleanup. */
    return;
}


