/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_refcount.h,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#ifndef __EDIR_REFCOUNT_H__
#define __EDIR_REFCOUNT_H__ 1
#ifdef __cplusplus
extern "C" {
#endif

/**
 * Interface definition for a generic reference count objects.
 * <P>
 * For interprocess scoreboards, we need reliable reference counting.
 * Currently, this means System V IPC (Semaphores) or the like.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/**
 * Create a new reference count array.
 * 
 * @param key [in]  Handle to object from which to obtain configuration.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          A handle to the new reference count object.
 * @exception EDIR_EXCEPT_REFCOUNT If the array cannot be created.
 */
extern edir_ipc_handle_t *
edir_refcount_create(edir_ipc_handle_t *key, int count, edir_exception_t **exc);

/**
 * Destroy a reference count array.
 * 
 * @param handle [in]  Handle returned by edir_refcount_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If the array cannot be destroyed.
 */
extern void
edir_refcount_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc);

/**
 * Obtain an existing reference count array.
 * 
 * @param key [in]  Handle to object from which to obtain configuration.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          A handle to the reference count object.
 * @exception EDIR_EXCEPT_REFCOUNT If the specified array is not available.
 */
extern edir_ipc_handle_t *
edir_refcount_obtain(edir_ipc_handle_t *key, edir_exception_t **exc);

/**
 * Release a previously obtained reference count array.
 * 
 * @param handle [in]  Handle returned by edir_refcount_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while releasing the array.
 */
extern void
edir_refcount_release(edir_ipc_handle_t *handle, edir_exception_t **exc);

/**
 * Increment a reference count within a refcount array.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param index  [in]  Reference count within the array to increment.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while incrementing.
 */
extern void
edir_refcount_increment(edir_ipc_handle_t *handle, int index, edir_exception_t **exc);

/**
 * Decrement a reference count within a refcount array.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param index  [in]  Reference count within the array to decrement.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while decrementing.
 */
extern void
edir_refcount_decrement(edir_ipc_handle_t *handle, int index, edir_exception_t **exc);

/**
 * Return a pointer to the current array of reference counts.
 * <P>
 * This array is maintained by the object and should not be modified
 * or freed.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param count  [out] Filled in with the number of entries returned.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             Integer array containing all refcount values.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while checking.
 */
extern int32 *
edir_refcount_values(edir_ipc_handle_t *handle, int32 *count, edir_exception_t **exc);

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * This pointer is a reference and MUST NOT be modified!
 * 
 * @param handle [in]  Handle to a reference count object.
 * @return             Reference to the underlying object, or NULL if not valid.
 */
extern void * edir_refcount_handle_get_object(edir_ipc_handle_t *handle);

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in]  Handle to a reference count object.
 * @return             The name of the underlying object, or NULL if not valid.
 */
extern char * edir_refcount_handle_get_object_name(edir_ipc_handle_t *handle);

#ifdef __cplusplus
} /* Extern "C" */
#endif
#endif /* __EDIR_REFCOUNT_H__ */

