/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_refcount_simple.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"

/**
 * System-dependent implementation of the Enhydra Director refcount interface,
 * using the refCount fields of the scoreboard's server slots.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/*
 * Internal contents pointed to by "handles".
 */
typedef struct edir_refcount_simple_ctx {
    int32			numServers;
    edir_load_sb_server_t	*servers;   /* Must NOT be freed. */
    int32			*refCounts; /* Must be freed. */
} edir_refcount_simple_ctx_t;

#define _NUMSV(H) (((edir_refcount_simple_ctx_t *)(H))->numServers >= 0)
#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_REFCOUNT)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H) && _NUMSV(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

/**
 * Create a new reference count array.
 * 
 * @param key [in]  Handle to object from which to obtain configuration.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          A handle to the new reference count object.
 * @exception EDIR_EXCEPT_REFCOUNT If the array cannot be created.
 */
extern edir_ipc_handle_t *
edir_refcount_create(edir_ipc_handle_t *key, int count, edir_exception_t **exc)
{
    static char		     *funcName = "edir_refcount_create";
    edir_refcount_simple_ctx_t *ctx = NULL;
    edir_ipc_handle_t	     *handle = NULL;
    edir_load_sb_offsets_t     *sb = NULL;

    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_REFCOUNT;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_SCOREBOARD;
    handle->nrefs = 1;
    handle->share = NULL;
    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_SHMEM) {
	    /*
	     * Shmem handles always return a pointer to their attached 
	     * segment as the "object".
	     */
	    sb = (edir_load_sb_offsets_t *) edir_shmem_handle_get_object(key);
	    handle->share = key;
	    handle->share->nrefs++;
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
		"Simple refcount requires SHMEM handle, not %d",
		key->handle_type);
	    EDIR_THROW(*exc);
	}
    } else {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
	EDIR_THROW(*exc);
    }

    if (!EDIR_SCOREBOARD_VALID(sb)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid scoreboard");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_simple_ctx_t *)
	    edir_malloc(sizeof(edir_refcount_simple_ctx_t));
    handle->pData = (void *)ctx;
    ctx->servers = EDIR_SB_SERVERS(sb);
    ctx->numServers = sb->numServers;
    ctx->refCounts = (int32 *) edir_malloc(ctx->numServers * sizeof(int32));

    return handle;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	if (ctx->refCounts) edir_free(ctx->refCounts);
	ctx->refCounts = NULL;
	edir_free(ctx);
    }
    if (handle) {
	if (handle->share) handle->share->nrefs--;
	handle->share = NULL;
	edir_free(handle);
    }
    return NULL;
}

/**
 * Destroy a reference count array.
 * 
 * @param handle [in]  Handle returned by edir_refcount_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If the array cannot be destroyed.
 */
extern void
edir_refcount_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		     *funcName = "edir_refcount_destroy";
    edir_refcount_simple_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_simple_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_DESTROY_SHARE(handle, exc);

    if (ctx->refCounts) edir_free(ctx->refCounts);
    ctx->numServers = -1;
    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Only way we get here is if ctx is invalid, so don't free */
    return;
}

/**
 * Obtain an existing reference count array.
 * 
 * @param key [in]  Handle to object from which to obtain configuration.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          A handle to the reference count object.
 * @exception EDIR_EXCEPT_REFCOUNT If the specified array is not available.
 */
extern edir_ipc_handle_t *
edir_refcount_obtain(edir_ipc_handle_t *key, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_obtain";
    void		   *ret;

    /* For this implementation, _create() does the same thing */
    ret = edir_refcount_create(key, 0, exc);
    if (*exc) EDIR_THROW(*exc);

    return ret;

except:
    return NULL;
}

/**
 * Release a previously obtained reference count array.
 * 
 * @param handle [in]  Handle returned by edir_refcount_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while releasing the array.
 */
extern void
edir_refcount_release(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		     *funcName = "edir_refcount_release";
    edir_refcount_simple_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_simple_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_RELEASE_SHARE(handle, exc);

    if (ctx->refCounts) edir_free(ctx->refCounts);
    ctx->numServers = -1;
    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Only way we get here is if ctx is invalid, so don't free */
    return;
}

/**
 * Increment a reference count within a refcount array.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param index  [in]  Reference count within the array to increment.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while incrementing.
 */
extern void
edir_refcount_increment(edir_ipc_handle_t *handle, int index, edir_exception_t **exc)
{
    static char		     *funcName = "edir_refcount_increment";
    edir_refcount_simple_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_simple_ctx_t *) handle->pData;

    if ((index < 0) || (index >= ctx->numServers)) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
				    "increment: index out of range");
	EDIR_THROW(*exc);
    }

    ctx->servers[index].refCount++;

    return;

except:
    /* Nothing to clean up here. */
    return;
}

/**
 * Decrement a reference count within a refcount array.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param index  [in]  Reference count within the array to decrement.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while decrementing.
 */
extern void
edir_refcount_decrement(edir_ipc_handle_t *handle, int index, edir_exception_t **exc)
{
    static char		     *funcName = "edir_refcount_decrement";
    edir_refcount_simple_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_simple_ctx_t *) handle->pData;

    if ((index < 0) || (index >= ctx->numServers)) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
				    "decrement: index out of range");
	EDIR_THROW(*exc);
    }

    ctx->servers[index].refCount--;

    return;

except:
    /* Nothing to clean up here. */
    return;
}

/**
 * Return a pointer to the current array of reference counts.
 * <P>
 * This array is maintained by the object and should not be modified
 * or freed.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param count  [out] Filled in with the number of entries returned.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             Integer array containing all refcount values.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while checking.
 */
extern int32 *
edir_refcount_values(edir_ipc_handle_t *handle, int32 *count, edir_exception_t **exc)
{
    static char		     *funcName = "edir_refcount_values";
    edir_refcount_simple_ctx_t *ctx = NULL;
    int32		     i;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_simple_ctx_t *) handle->pData;

    if (count == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
				    "Invalid 'count' pointer");
	EDIR_THROW(*exc);
    }

    for (i=0; i<ctx->numServers; i++)
	ctx->refCounts[i] = ctx->servers[i].refCount;

    *count = ctx->numServers;
    return ctx->refCounts;

except:
    *count = -1;
    return NULL;
}

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * This pointer is a reference and MUST NOT be modified!
 * 
 * @param handle [in]  Handle to a reference count object.
 * @return             Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_refcount_handle_get_object(edir_ipc_handle_t *handle)
{
    edir_refcount_simple_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_refcount_simple_ctx_t *) handle->pData;
    return (void *)(ctx->servers);
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in]  Handle to a reference count object.
 * @return             The name of the underlying object, or NULL if not valid.
 */
extern char *
edir_refcount_handle_get_object_name(edir_ipc_handle_t *handle)
{
    if (!VALID_HANDLE(handle)) return NULL;
    return "scoreboard"; /* Scoreboard does not have a "name" */
}

