/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_refcount_sysv.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_refcount_sysv.h"
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

/**
 * System-dependent implement of the Enhydra Director refcount interface,
 * using System V IPC semaphores.
 * <P>
 * Uses System V semaphores via semget(), etc. to implement a reliable
 * interprocess reference count in which references are automatically
 * decremented by the system when a process dies unexpectedly.
 * <P>
 * This is needed for CGI because some web servers (like apache) do
 * nasty things like sending SIGTERM or closing the output pipe.
 * Catching these signals can be flakey.  Semaphores provide a much more
 * leak resistant method of keeping track of the number of connections
 * to the servers.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/*
 * Duhhhh this is really lame.  Solaris does not define this anywhere!
 */
#ifdef EDIR_NEED_UNION_SEMUN
union semun {
    long val;
    struct semid_ds *buf;
    ushort *array;
};
#endif

/*
 * Internal contents pointed to by "handles".
 */
typedef struct edir_refcount_sysv_ctx {
    key_t	   semkey;  /* Global ID used by semget() to find the set. */
    int		   semid;   /* Unique system ID for this set */
    int		   numSem;  /* Number of semaphores in this set (array) */
    unsigned short *values;
    int32	   *refCounts;
} edir_refcount_sysv_ctx_t;

#define _NUMSV(H) (((edir_refcount_sysv_ctx_t *)(H))->numSem >= 0)
#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_REFCOUNT)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H) && _NUMSV(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

/**
 * Create a new reference count array.
 * 
 * @param key [in]  Handle to object from which to obtain configuration.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          A handle to the new reference count object.
 * @exception EDIR_EXCEPT_REFCOUNT If the array cannot be created.
 */
extern edir_ipc_handle_t *
edir_refcount_create(edir_ipc_handle_t *key, int count, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_create";
    edir_refcount_sysv_ctx_t *ctx = NULL;
    edir_ipc_handle_t	   *handle = NULL;
    char		   *stringName = NULL, *endptr = NULL;
    union semun		   semarg;

    /*
     * Create a new handle for this refcount.  Externally this is a (void *).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_REFCOUNT;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_INTEGER;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_refcount_sysv_ctx_t *) edir_malloc(sizeof(edir_refcount_sysv_ctx_t));
    handle->pData = (void *)ctx;
    ctx->semkey = -1;

    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    stringName = edir_ipc_config_handle_get_object_name(key);
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Unexpected object type %d from CONFIG",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else if (key->handle_type == EDIR_IPC_HANDLE_TYPE_SHMEM) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_INTEGER:
		    /* Sysv IPC passes name as 8 hex digits of key value */
		    stringName = edir_shmem_handle_get_object_name(key);
		    errno = 0;
		    ctx->semkey = strtol(stringName, &endptr, 16);
		    if (errno || (endptr && *endptr)) ctx->semkey = -1;
		    if (ctx->semkey > 0) {
			ctx->semkey += EDIR_REFCOUNT_SYSV_IPC_OFFSET;
			edir_free(stringName);
			stringName = NULL;
		    } else ctx->semkey = -1;
		    break;
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    /* For shmem, the object is the shared memory itself */
		    /* We want the shmem key */
		    stringName = edir_shmem_handle_get_object_name(key);
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Unexpected object type %d from SHMEM",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
		"Unexpected handle type %d", key->handle_type);
	    EDIR_THROW(*exc);
	}
    }
    if (ctx->semkey <= 0) {
	if (stringName != NULL) {
	    errno = 0;
            ctx->semkey = strtol(stringName, &endptr, 0);
            if ((ctx->semkey <= 0) || (endptr == stringName)) errno = EINVAL;
            if (endptr) while (*endptr) if (!isspace(*endptr++)) errno = EINVAL;
            if (errno != 0) {
                char *cp;
                int     hash = 0;
                for (cp=stringName; *cp; cp++) hash += (octet)(*cp);
                ctx->semkey = ftok("/dev/null", hash);
            }
	}
    }
    if (ctx->semkey <= 0) ctx->semkey = EDIR_REFCOUNT_SYSV_DEFAULT_IPC_KEY;

    if (stringName) edir_free(stringName);
    stringName = NULL;

    ctx->semid = -1;
    ctx->numSem = count;
    ctx->values = (unsigned short *) edir_malloc(count * sizeof(unsigned short));
    ctx->refCounts = (int32 *) edir_malloc(count * sizeof(int32));

    /*
     * Create the ref count.  Destroy the old one if necessary.
     */
    ctx->semid = semget(ctx->semkey, ctx->numSem, IPC_CREAT | IPC_EXCL | 0600);
    if (ctx->semid < 0) {
	if (errno == EEXIST) {
	    ctx->semid = semget(ctx->semkey, 1, 0600);
	    if (ctx->semid < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Could not get old semaphore to destroy it");
		EDIR_THROW(*exc);
	    }
	    semarg.val = 0;
	    if (semctl(ctx->semid, 0, IPC_RMID, semarg) < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Could not destroy old semaphore");
		EDIR_THROW(*exc);
	    }
	    ctx->semid = semget(ctx->semkey, 1, IPC_CREAT|IPC_EXCL|0600);
	    if (ctx->semid < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Could not create new semaphore");
		EDIR_THROW(*exc);
	    }
	} else if (errno == ENOSYS) {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"System V Semaphores must be enabled in kernel");
	    EDIR_THROW(*exc);
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Cannot create semaphore");
	    EDIR_THROW(*exc);
	}
    }

    return handle;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	if (ctx->semid >= 0) {
	    semarg.val = 0;
	    semctl(ctx->semid, 0, IPC_RMID, semarg);
	}
	if (ctx->values) edir_free(ctx->values);
	ctx->values = NULL;
	if (ctx->refCounts) edir_free(ctx->refCounts);
	ctx->refCounts = NULL;
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    if (stringName) edir_free(stringName);
    return NULL;
}

/**
 * Destroy a reference count array.
 * 
 * @param handle [in]  Handle returned by edir_refcount_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If the array cannot be destroyed.
 */
extern void
edir_refcount_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_destroy";
    edir_refcount_sysv_ctx_t *ctx = NULL;
    union semun		   semarg;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;

    EDIR_IPC_DESTROY_SHARE(handle, exc);

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    if (ctx->semid >= 0) {
	semarg.val = 0;
	if (semctl(ctx->semid, 0, IPC_RMID, semarg) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Cannot delete semaphore");
	    ctx->semid = -1;
	    EDIR_THROW(*exc);
	}
	ctx->semid = -1;
    }

    if (ctx->values) edir_free(ctx->values);
    ctx->values = NULL;
    if (ctx->refCounts) edir_free(ctx->refCounts);
    ctx->refCounts = NULL;

    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Last ditch exception cleanup. At least free the handle memory. */
    if (ctx) {
	if (ctx->values) edir_free(ctx->values);
	ctx->values = NULL;
	if (ctx->refCounts) edir_free(ctx->refCounts);
	ctx->refCounts = NULL;
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Obtain an existing reference count array.
 * 
 * @param key [in]  Handle to object from which to obtain configuration.
 * @param exc [out] Exception pointer filled in if an exception occurs.
 * @return          A handle to the reference count object.
 * @exception EDIR_EXCEPT_REFCOUNT If the specified array is not available.
 */
extern edir_ipc_handle_t *
edir_refcount_obtain(edir_ipc_handle_t *key, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_obtain";
    edir_refcount_sysv_ctx_t *ctx = NULL;
    edir_ipc_handle_t	   *handle = NULL;
    char		   *stringName = NULL, *endptr = NULL;
    union semun		   semarg;
    struct semid_ds	   semstat;

    /*
     * Create a new handle for this refcount.  Externally this is a (void *).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_REFCOUNT;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_INTEGER;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_refcount_sysv_ctx_t *) edir_malloc(sizeof(edir_refcount_sysv_ctx_t));
    handle->pData = (void *)ctx;
    ctx->semkey = -1;

    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    stringName = edir_ipc_config_handle_get_object_name(key);
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Unexpected object type %d from CONFIG",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else if (key->handle_type == EDIR_IPC_HANDLE_TYPE_SHMEM) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_INTEGER:
		    /* Sysv IPC passes name as 8 hex digits of key value */
		    stringName = edir_shmem_handle_get_object_name(key);
		    errno = 0;
		    ctx->semkey = strtol(stringName, &endptr, 16);
		    if (errno || (endptr && *endptr)) ctx->semkey = -1;
		    if (ctx->semkey > 0) {
			ctx->semkey += EDIR_REFCOUNT_SYSV_IPC_OFFSET;
			edir_free(stringName);
			stringName = NULL;
		    } else ctx->semkey = -1;
		    break;
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    /* For shmem, the object is the shared memory itself */
		    /* We want the shmem key */
		    stringName = edir_shmem_handle_get_object_name(key);
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Unexpected object type %d from SHMEM",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
		"Unexpected handle type %d", key->handle_type);
	    EDIR_THROW(*exc);
	}
    }
    if (ctx->semkey <= 0) {
	if (stringName != NULL) {
	    errno = 0;
            ctx->semkey = strtol(stringName, &endptr, 0);
            if ((ctx->semkey <= 0) || (endptr == stringName)) errno = EINVAL;
            if (endptr) while (*endptr) if (!isspace(*endptr++)) errno = EINVAL;
            if (errno != 0) {
                char *cp;
                int     hash = 0;
                for (cp=stringName; *cp; cp++) hash += (octet)(*cp);
                ctx->semkey = ftok("/dev/null", hash);
            }
	}
    }
    if (ctx->semkey <= 0) ctx->semkey = EDIR_REFCOUNT_SYSV_DEFAULT_IPC_KEY;

    if (stringName) edir_free(stringName);
    stringName = NULL;

    ctx->semid = -1;
    ctx->numSem = 0;

    /*
     * Get the semaphore array used to implement the refcount array.
     */
    ctx->semid = semget(ctx->semkey, 0, 0600);
    if (ctx->semid < 0) {
	if (errno == ENOENT) {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
					"Semaphore not found");
	    EDIR_THROW(*exc);
	} else if ((errno == EACCES) || (errno == EPERM)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
		"Semaphore permission denied - are we running as right user?");
	    EDIR_THROW(*exc);
	} else if (errno == ENOSYS) {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
		"System V Semaphores must be enabled in kernel");
	    EDIR_THROW(*exc);
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Cannot create semaphore");
	    EDIR_THROW(*exc);
	}
    }

    /*
     * Get info about the semaphore.  Namely, how meny sems are in
     * this set.
     */
    semarg.buf = &semstat;
    if (semctl(ctx->semid, 0, IPC_STAT, semarg) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Cannot get semaphore set size");
	EDIR_THROW(*exc);
    }
    ctx->numSem = semstat.sem_nsems;
    ctx->values = (unsigned short *)
		    edir_malloc(ctx->numSem * sizeof(unsigned short));
    ctx->refCounts = (int32 *) edir_malloc(ctx->numSem * sizeof(int32));

    return handle;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	if (ctx->semid >= 0) {
	    semarg.val = 0;
	    semctl(ctx->semid, 0, IPC_RMID, semarg);
	}
	if (ctx->values) edir_free(ctx->values);
	ctx->values = NULL;
	if (ctx->refCounts) edir_free(ctx->refCounts);
	ctx->refCounts = NULL;
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    if (stringName) edir_free(stringName);
    return NULL;
}

/**
 * Release a previously obtained reference count array.
 * 
 * @param handle [in]  Handle returned by edir_refcount_create().
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while releasing the array.
 */
extern void
edir_refcount_release(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_release";
    edir_refcount_sysv_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;

    EDIR_IPC_RELEASE_SHARE(handle, exc);

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    /* Nothing else to do but to free the handle memory. */
    ctx->semid = -1;
    ctx->semkey = (key_t) -1;
    ctx->numSem = 0;
    if (ctx->values) edir_free(ctx->values);
    ctx->values = NULL;
    if (ctx->refCounts) edir_free(ctx->refCounts);
    ctx->refCounts = NULL;
    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Last ditch exception cleanup. */
    if (ctx) {
	ctx->semid = -1;
	ctx->semkey = (key_t) -1;
	ctx->numSem = 0;
	if (ctx->values) edir_free(ctx->values);
	ctx->values = NULL;
	if (ctx->refCounts) edir_free(ctx->refCounts);
	ctx->refCounts = NULL;
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Increment a reference count within a refcount array.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param index  [in]  Reference count within the array to increment.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while incrementing.
 */
extern void
edir_refcount_increment(edir_ipc_handle_t *handle, int index, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_increment";
    edir_refcount_sysv_ctx_t *ctx = NULL;
    struct sembuf	   semcmd;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;

    if ((index < 0) || (index >= ctx->numSem)) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
				    "increment: index out of range");
	EDIR_THROW(*exc);
    }

    semcmd.sem_num  = index;
    semcmd.sem_op   = 1;	  /* Increment refcount by one. */
    semcmd.sem_flg  = SEM_UNDO;   /* System automatically decrements on exit */
    semcmd.sem_flg |= IPC_NOWAIT; /* Never block.  Blocking is an error */

    while (semop(ctx->semid, &semcmd, 1) < 0) {
	if (errno != EINTR) {
	    if (errno == EAGAIN) {
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			    "Semaphore unexpectedly blocks on increment!");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			    "Error incrementing semaphore");
		EDIR_THROW(*exc);
	    }
	}
    }

    return;

except:
    /* Nothing to clean up here. */
    return;
}

/**
 * Decrement a reference count within a refcount array.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param index  [in]  Reference count within the array to decrement.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while decrementing.
 */
extern void
edir_refcount_decrement(edir_ipc_handle_t *handle, int index, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_decrement";
    edir_refcount_sysv_ctx_t *ctx = NULL;
    struct sembuf	   semcmd;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;

    if ((index < 0) || (index >= ctx->numSem)) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
				    "decrement: index out of range");
	EDIR_THROW(*exc);
    }

    semcmd.sem_num  = index;
    semcmd.sem_op   = -1;	  /* Decrement refcount by one. */
    semcmd.sem_flg  = SEM_UNDO;   /* Cancel out SEM_UNDO from previous incr. */
    semcmd.sem_flg |= IPC_NOWAIT; /* Never block.  Blocking is an error */

    while (semop(ctx->semid, &semcmd, 1) < 0) {
	if (errno != EINTR) {
	    if (errno == EAGAIN) {
		/* Semaphore fails with EAGAIN if it would go below zero. */
		/* (Assuming IPC_NOWAIT, of course) */
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			    "Value less than zero (unbalanced decrement?)");
		EDIR_THROW(*exc);
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			    "Error decrementing semaphore");
		EDIR_THROW(*exc);
	    }
	}
    }

    return;

except:
    /* Nothing to clean up here. */
    return;
}

/**
 * Return a pointer to the current array of reference counts.
 * <P>
 * This array is maintained by the object and should not be modified
 * or freed.
 * 
 * @param handle [in]  Handle to a reference count object.
 * @param count  [out] Filled in with the number of entries returned.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @return             Integer array containing all refcount values.
 * @exception EDIR_EXCEPT_REFCOUNT If a failure occurs while checking.
 */
extern int32 *
edir_refcount_values(edir_ipc_handle_t *handle, int32 *count, edir_exception_t **exc)
{
    static char		   *funcName = "edir_refcount_values";
    edir_refcount_sysv_ctx_t *ctx = NULL;
    union semun		   semarg;
    struct semid_ds	   semstat;
    int32		   i;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT, "Invalid handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;

    if (count == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
				    "Invalid 'count' pointer");
	EDIR_THROW(*exc);
    }

    /*
     * Get the value of every semaphore in the set into a ushort array.
     */
    semarg.array = ctx->values;
    if (semctl(ctx->semid, 0, GETALL, semarg) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_REFCOUNT,
			"Cannot get semaphore values");
	EDIR_THROW(*exc);
    }
    for (i=0; i<ctx->numSem; i++) {
	ctx->refCounts[i] = ctx->values[i];
    }

    *count = ctx->numSem;
    return ctx->refCounts;

except:
    *count = -1;
    return NULL;
}

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * This pointer is a reference and MUST NOT be modified!
 * 
 * @param handle [in]  Handle to a reference count object.
 * @return             Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_refcount_handle_get_object(edir_ipc_handle_t *handle)
{
    edir_refcount_sysv_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;
    if (ctx->semkey == -1) return NULL;
    return (void *)(&(ctx->semkey));
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in]  Handle to a reference count object.
 * @return             The name of the underlying object, or NULL if not valid.
 */
extern char *
edir_refcount_handle_get_object_name(edir_ipc_handle_t *handle)
{
    edir_refcount_sysv_ctx_t	*ctx = NULL;
    char			buf[64];
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_refcount_sysv_ctx_t *) handle->pData;
    if (ctx->semkey == -1) return NULL;
    sprintf(buf, "%08x", ctx->semkey);
    return edir_strdup(buf);
}

