/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_shmem.h,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#ifndef __EDIR_SHMEM_H__
#define __EDIR_SHMEM_H__ 1
#ifdef __cplusplus
extern "C" {
#endif

/**
 * Interface definition for all Enhydra Director shared memory access.
 * <P>
 * System dependent shared memory access modules, such as edir_shmem_sysv.c,
 * implement functions fitting the prototypes defined below.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/**
 * Kind of a hack, but necessary to prevent some nasty cleanup
 * race conditions with scoreboards.  Basically, we need
 * to mark a scoreboard as "dead" AFTER its systemwide resource
 * has been unlinked.  This callback type is provided to allow
 * that sort of thing when edir_shmem_destroy is called.
 * 
 * @param ptr [in]  Pointer the the mapped memory area.
 * @param exc [out] Can be filled in with an exception that is then
 * 		    rethrown by edir_shmem_destroy.
 */
typedef void (*edir_shmem_callback_t)(void *ptr, edir_exception_t **exc);

/**
 * Create a shared memory region.
 * <P>
 * Creates a brand new shared memory region of the specified size.  The
 * key may or may not be used, depending on the system-dependent
 * implementation.  A NULL key means "use the system-dependent default".
 * Some implementations may leave shared memory resources allocated,
 * even after the program has terminated, unless edir_shmem_destroy() is
 * called to clean up.
 * 
 * @param key     [in]  Handle to object to use as configuration source.
 * @param size    [in]  The desired size of the shared memory region.
 * @param func    [in]  Optional destroy callback.
 * @param exc     [out] Exception pointer filled in if an exception occurs.
 * @return		A shared memory handle object.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be created.
 */
extern edir_ipc_handle_t *
edir_shmem_create(edir_ipc_handle_t *key, int32 size,
		edir_shmem_callback_t func, edir_exception_t **exc);

/**
 * Destroy a shared memory region.
 * <P>
 * Detaches the given shared memory region from the data segment and
 * removes its key and ID from the system shared memory tables, if
 * applicable.
 * 
 * @param handle [in]  Handle to the shared memory object.
 * @param func   [in]  Function to call after unlink (may be NULL).
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be destroyed.
 */
extern void
edir_shmem_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc);

/**
 * Obtain an existing shared memory region.
 * <P>
 * A shared memory region must already be mapped to the specified key.
 * Otherwise an exception is generated.  This function returns a pointer
 * to the mapped shared memory area if it exists and can be mapped.
 * 
 * @param key     [in]  Handle to object to use as configuration source.
 * @param exc     [out] Exception pointer filled in if an exception occurs.
 * @return		A shared memory handle object.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be mapped.
 */
extern edir_ipc_handle_t *
edir_shmem_obtain(edir_ipc_handle_t *key, edir_exception_t **exc);

/**
 * Detach a shared memory region.
 * <P>
 * Detaches the given shared memory region from the data segment, but does not
 * remove its key and ID from the system shared memory tables.
 * 
 * @param handle [in]  Handle to the shared memory object.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be detached.
 */
extern void
edir_shmem_release(edir_ipc_handle_t *handle, edir_exception_t **exc);

/**
 * Sets the destruction callback function for the specified instance.
 * <P>
 * The destroy callback performs actions such as setting flags and
 * the like, just after the shared memory was unlinked from its system
 * wide name and just before it is unmapped from this process.
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @param func   [in] Function to call upon destroy.
 * @return The old destroy function
 */
extern edir_shmem_callback_t
edir_shmem_set_destroy_func(edir_ipc_handle_t *handle, edir_shmem_callback_t func);

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * For shared memory objects, this pointer always points to the beginning
 * of the mapped memory area.
 * <P>
 * This pointer is a reference MUST NOT be freed!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void * edir_shmem_handle_get_object(edir_ipc_handle_t *handle);

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char * edir_shmem_handle_get_object_name(edir_ipc_handle_t *handle);

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void
edir_shmem_set_default_filename(char  *filename);

/**
 * Get the default name/key for this class.
 * <P>
 * If no default has been set with edir_mutex_set_default_filename(),
 * then return NULL;
 * 
 * @return The default file name.
 */
extern char *
edir_shmem_default_name();

#ifdef __cplusplus
} /* Extern "C" */
#endif
#endif /* __EDIR_SHMEM_H__ */

