/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_shmem_mmap.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_shmem_mmap.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

/**
 * Shared memory access functions for mmap() based shared memory.
 * <P>
 * This module implements the Enhydra Director shared memory interface
 * using mmap() on a well known file path.  It should be more stable
 * and clean that System V shared memory (shmget()), since it does
 * not depend nearly as much on kernel tunables and does not leave
 * resources in wierd places that must be cleaned using
 * "ipcs" and "ipcrm".
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/**
 * System-dependent context info for a mmap() based shared memory.
 */
typedef struct edir_shmem_mmap_ctx {
    char		*pathname;
    void		*attach;
    int32		size;
    edir_shmem_callback_t destroyFunc;
} edir_shmem_mmap_ctx_t;

#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_SHMEM)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

static int DefaultSet = 0;
static char * DefaultFileName = EDIR_SHMEM_MMAP_DEFAULT_FILE;

/**
 * Create a shared memory region.
 * <P>
 * Creates a brand new shared memory region of the specified size.  The
 * key may or may not be used, depending on the system-dependent
 * implementation.  A NULL key means "use the system-dependent default".
 * Some implementations may leave shared memory resources allocated,
 * even after the program has terminated, unless edir_shmem_destroy() is
 * called to clean up.
 * 
 * @param key     [in]  Handle to object to use as configuration source.
 * @param size    [in]  The desired size of the shared memory region.
 * @param func    [in]  Optional destroy callback.
 * @param exc     [out] Exception pointer filled in if an exception occurs.
 * @return		A shared memory handle object.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be created.
 */
extern edir_ipc_handle_t *
edir_shmem_create(edir_ipc_handle_t *key, int32 size,
		edir_shmem_callback_t func, edir_exception_t **exc)
{
    static char		*funcName = "edir_shmem_create";
    static char		zchar = '\0';
    edir_shmem_mmap_ctx_t	*ctx = NULL;
    edir_ipc_handle_t	*handle = NULL;
    int			fd = -1;
    int			*intp = NULL;

    /*
     * Set up a context structure for this new region.
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_SHMEM;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_STRING;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_shmem_mmap_ctx_t *) edir_malloc(sizeof(edir_shmem_mmap_ctx_t));
    ctx->destroyFunc = func;
    handle->pData = (void *) ctx;
    fd = -1;

    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    ctx->pathname = edir_ipc_config_handle_get_object_name(key);
		    fd = -1;
		    handle->share = NULL;
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from CONFIG",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else if (key->handle_type == EDIR_IPC_HANDLE_TYPE_MUTEX) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    ctx->pathname = edir_mutex_handle_get_object_name(key);
		    fd = -1;
		    handle->share = NULL;
		    break;
		case EDIR_IPC_OBJECT_TYPE_INTEGER:
		    intp = (int *) edir_mutex_handle_get_object(key);
		    if (intp == NULL) {
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			    "Can't get integer from shared mutex object");
			EDIR_THROW(*exc);
		    } else {
			char numbuf[20], *def, *ext, *cp;
			def = DefaultFileName;
			ext = strrchr(def, '.');
			if (!ext) ext = "";
			ctx->pathname = (char *) edir_malloc(strlen(def) + 40);
			strcpy(ctx->pathname, def);
			cp = strrchr(ctx->pathname, '.');
			if (cp) *cp = '\0';
			sprintf(numbuf, "-%08x", *intp);
			strcat(ctx->pathname, numbuf);
			strcat(ctx->pathname, ext);
			fd = -1;
			handle->share = NULL;
		    }
		    break;
		case EDIR_IPC_OBJECT_TYPE_FILEDESC:
		    intp = (int *) edir_mutex_handle_get_object(key);
		    if ((intp == NULL) || (*intp < 0)) {
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			    "Invalid fd from shared mutex object");
			EDIR_THROW(*exc);
		    }
		    fd = *intp;
		    ctx->pathname = edir_mutex_handle_get_object_name(key);
		    handle->share = key;
		    handle->share->nrefs++;
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from MUTEX",
		    key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Unexpected handle type %d", key->handle_type);
	    EDIR_THROW(*exc);
	}
    }
    if (handle->share == NULL) {
	if (ctx->pathname == NULL)
	    ctx->pathname = edir_strdup(DefaultFileName);
    }

    ctx->attach = NULL;
    ctx->size = size;

    /* Don't allow huge file to cause huge process. */
    if ((ctx->size < 0) || (ctx->size > EDIR_SHMEM_MMAP_MAX_SIZE)) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"Shared memory size %d too big.  Reconfigure %s and %s",
		ctx->size, "EDIR_SHMEM_MMAP_MAX_SIZE", "rebuild Enhydra Director");
	EDIR_THROW(*exc);
    }

    if (handle->share == NULL) {
	fd = open(ctx->pathname, O_CREAT|O_RDWR, 0600);
	if (fd < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		    "Can't create mmap() shared memory file \"%s\": @errstr@",
		    ctx->pathname);
	    EDIR_THROW(*exc);
	}
    }

    /* Seek out to maximum configured file size */
    if (lseek(fd, ctx->size, SEEK_SET) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"lseek failed to resize scoreboard file %s to max %d: @errstr@",
		ctx->pathname, ctx->size);
	EDIR_THROW(*exc);
    }

    /* Write one byte after end of max size to create a "sparse" file. */
    if (write(fd, &zchar, 1) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"write failed to resize scoreboard file %s to max %d: @errstr@",
		ctx->pathname, ctx->size);
	EDIR_THROW(*exc);
    }

    ctx->attach = mmap((caddr_t)0, ctx->size, 
			PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (ctx->attach == (caddr_t)-1) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"Could not mmap file %s: @errstr@", ctx->pathname);
	ctx->attach = NULL;
	EDIR_THROW(*exc);
    }

    /* No longer need fd open.  We're mmap()-ed! */
    if (handle->share == NULL) {
	/* Only close if we're not sharing the fd. */
	if (close(fd) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		    "Could not close fd for %s after mmap(): @errstr@",
		    ctx->pathname);
	    EDIR_THROW(*exc);
	}
    }

    return handle;

except:
    /* Exception cleanup */
    if (ctx != NULL) {
	/* Detach if attached */
	if (ctx->attach != NULL) {
	    munmap(ctx->attach, ctx->size);
	    ctx->attach = NULL;
	}
	/* Kill file and close fd to it. */
	if ((fd >= 0) && (ctx->pathname != NULL)) {
	    if (handle->share == NULL) {
		/* Only close if not sharing */
		unlink(ctx->pathname);
		close(fd);
	    }
	}
	if (ctx->pathname != NULL) edir_free(ctx->pathname);
	edir_free(ctx);
    }
    if (handle) {
	if (handle->share != NULL) handle->share->nrefs--;
	handle->share = NULL;
	edir_free(handle);
    }
    return NULL;
}

/**
 * Destroy a shared memory region.
 * <P>
 * Detaches the given shared memory region from the data segment and
 * removes its key and ID from the system shared memory tables, if
 * applicable.
 * 
 * @param handle [in]  Handle to the shared memory object.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be destroyed.
 */
extern void
edir_shmem_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char 	*funcName = "edir_shmem_destroy";
    edir_shmem_mmap_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, "Invalid shmem handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_shmem_mmap_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_DESTROY_SHARE(handle, exc);

    /* Kill file and close fd to it. */
    if (ctx->pathname != NULL) {
	if ((unlink(ctx->pathname) < 0) && (errno != ENOENT)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, 
		"Can't unlink mmap()-ed scoreboard file %s: @errstr@",
		ctx->pathname ? ctx->pathname : "(null)");
	    EDIR_THROW(*exc);
	}
	edir_free(ctx->pathname);
	ctx->pathname = NULL;
    }
    /* Detach if attached */
    if (ctx->attach != NULL) {
	if (ctx->destroyFunc != NULL) {
	    (ctx->destroyFunc)(ctx->attach, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	if (munmap(ctx->attach, ctx->size) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, 
		"Can't munmap() region "EDIR_PTR": @errstr@", ctx->attach);
	    ctx->attach = NULL; /* Already tried to munmap() */
	    EDIR_THROW(*exc);
	}
	ctx->attach = NULL;
    }

    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Last-ditch exception cleanup */
    if (ctx != NULL) {
	if (ctx->attach != NULL) {
	    munmap(ctx->attach, ctx->size);
	    ctx->attach = NULL;
	}
	if (ctx->pathname != NULL) {
	    unlink(ctx->pathname);
	    edir_free(ctx->pathname);
	    ctx->pathname = NULL;
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Obtain an existing shared memory region.
 * <P>
 * A shared memory region must already be mapped to the specified key.
 * Otherwise an exception is generated.  This function returns a pointer
 * to the mapped shared memory area if it exists and can be mapped.
 * 
 * @param key     [in]  Handle to object to use as configuration source.
 * @param exc     [out] Exception pointer filled in if an exception occurs.
 * @return		A shared memory handle object.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be mapped.
 */
extern edir_ipc_handle_t *
edir_shmem_obtain(edir_ipc_handle_t *key, edir_exception_t **exc)
{
    static char		*funcName = "edir_shmem_obtain";
    edir_shmem_mmap_ctx_t	*ctx = NULL;
    edir_ipc_handle_t	*handle = NULL;
    int			fd;
    int			*intp = NULL;
    struct stat		st;

    /*
     * Set up a context structure for this new region.
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_SHMEM;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_STRING;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_shmem_mmap_ctx_t *) edir_malloc(sizeof(edir_shmem_mmap_ctx_t));
    ctx->destroyFunc = NULL;
    handle->pData = (void *) ctx;
    fd = -1;

    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    ctx->pathname = edir_ipc_config_handle_get_object_name(key);
		    fd = -1;
		    handle->share = NULL;
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from CONFIG",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else if (key->handle_type == EDIR_IPC_HANDLE_TYPE_MUTEX) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    ctx->pathname = edir_mutex_handle_get_object_name(key);
		    fd = -1;
		    handle->share = NULL;
		    break;
		case EDIR_IPC_OBJECT_TYPE_INTEGER:
		    intp = (int *) edir_mutex_handle_get_object(key);
		    if (intp == NULL) {
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			    "Can't get integer from shared mutex object");
			EDIR_THROW(*exc);
		    } else {
			char numbuf[20], *def, *ext, *cp;
			def = DefaultFileName;
			ext = strrchr(def, '.');
			if (!ext) ext = "";
			ctx->pathname = (char *) edir_malloc(strlen(def) + 40);
			strcpy(ctx->pathname, def);
			cp = strrchr(ctx->pathname, '.');
			if (cp) *cp = '\0';
			sprintf(numbuf, "-%08x", *intp);
			strcat(ctx->pathname, numbuf);
			strcat(ctx->pathname, ext);
			fd = -1;
			handle->share = NULL;
		    }
		    break;
		case EDIR_IPC_OBJECT_TYPE_FILEDESC:
		    intp = (int *) edir_mutex_handle_get_object(key);
		    if ((intp == NULL) || (*intp < 0)) {
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			    "Invalid fd from shared mutex object");
			EDIR_THROW(*exc);
		    }
		    fd = *intp;
		    ctx->pathname = edir_mutex_handle_get_object_name(key);
		    handle->share = key;
		    handle->share->nrefs++;
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from MUTEX",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Unexpected handle type %d", key->handle_type);
	    EDIR_THROW(*exc);
	}
    }
    if (handle->share == NULL) {
	if (ctx->pathname == NULL)
	    ctx->pathname = edir_strdup(DefaultFileName);
    }

    ctx->attach = NULL;
    ctx->size = 0;

    if (handle->share == NULL) {
	fd = open(ctx->pathname, O_RDWR);
	if (fd < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		    "Can't open mmap() shared memory file \"%s\": @errstr@",
		    ctx->pathname);
	    EDIR_THROW(*exc);
	}
    }

    /* Determine shared memory file size. */
    if (fstat(fd, &st) < 0) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"Can't stat mmap() shared memory file \"%s\": @errstr@",
		ctx->pathname);
	EDIR_THROW(*exc);
    }
    ctx->size = st.st_size;

    /* Don't allow huge file to cause huge process. */
    if ((ctx->size < 0) || (ctx->size > (EDIR_SHMEM_MMAP_MAX_SIZE + 1))) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"Shared memory size %d too big.  Reconfigure %s and %s",
		ctx->size, "EDIR_SHMEM_MMAP_MAX_SIZE", "rebuild Enhydra Director");
	EDIR_THROW(*exc);
    }

    ctx->attach = mmap((caddr_t)0, ctx->size,
			PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (ctx->attach == (caddr_t)-1) {
	ctx->attach = NULL;
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		"Could not mmap file %s: @errstr@", ctx->pathname);
	EDIR_THROW(*exc);
    }

    /* No longer need fd open.  We're mmap()-ed! */
    if (handle->share == NULL) {
	if (close(fd) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
		    "Could not close fd for %s after mmap(): @errstr@",
		    ctx->pathname);
	    fd = -1; /* Already tried to close. */
	    EDIR_THROW(*exc);
	}
    }

    return handle;

except:
    /* Exception cleanup */
    if (ctx != NULL) {
	/* Detach if attached */
	if (ctx->attach != NULL) {
	    munmap(ctx->attach, ctx->size);
	    ctx->attach = NULL;
	}
	/* Close open file */
	if (handle && (handle->share == NULL) && (fd >= 0)) close(fd);
	if (ctx->pathname != NULL) edir_free(ctx->pathname);
	ctx->pathname = NULL;
	edir_free(ctx);
    }
    if (handle) {
	if (handle->share != NULL) handle->share->nrefs--;
	handle->share = NULL;
	edir_free(handle);
    }
    return NULL;
}

/**
 * Detach a shared memory region.
 * <P>
 * Detaches the given shared memory region from the data segment, but does not
 * remove its key and ID from the system shared memory tables.
 * 
 * @param handle [in]  Handle to the shared memory object.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be detached.
 */
extern void
edir_shmem_release(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char 	*funcName = "edir_shmem_destroy";
    edir_shmem_mmap_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, "Invalid shmem handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_shmem_mmap_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_RELEASE_SHARE(handle, exc);

    /* Detach if attached */
    if (ctx->attach != NULL) {
	if (munmap(ctx->attach, ctx->size) < 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, 
		"Can't munmap() region "EDIR_PTR" for %s: @errstr@",
		ctx->attach, ctx->pathname ? ctx->pathname : "(null)");
	    ctx->attach = NULL; /* Already tried to munmap() */
	    EDIR_THROW(*exc);
	}
	ctx->attach = NULL;
    }

    if (ctx->pathname != NULL) edir_free(ctx->pathname);
    ctx->pathname = NULL;

    edir_free(ctx);
    edir_free(handle);

    return;

except:
    /* Last-ditch exception cleanup */
    if (ctx != NULL) {
	if (ctx->attach != NULL) {
	    munmap(ctx->attach, ctx->size);
	    ctx->attach = NULL;
	}
	if (ctx->pathname != NULL) {
	    edir_free(ctx->pathname);
	    ctx->pathname = NULL;
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Sets the destruction callback function for the specified instance.
 * <P>
 * The destroy callback performs actions such as setting flags and
 * the like, just after the shared memory was unlinked from its system
 * wide name and just before it is unmapped from this process.
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @param func   [in] Function to call upon destroy.
 * @return The old destroy function
 */
extern edir_shmem_callback_t
edir_shmem_set_destroy_func(edir_ipc_handle_t *handle, edir_shmem_callback_t func)
{
    edir_shmem_mmap_ctx_t *ctx = NULL;
    edir_shmem_callback_t ret;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_shmem_mmap_ctx_t *) handle->pData;
    ret = ctx->destroyFunc;
    ctx->destroyFunc = func;
    return ret;
}

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * For shared memory objects, this pointer always points to the beginning
 * of the mapped memory area.
 * <P>
 * This pointer is a reference MUST NOT be freed!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_shmem_handle_get_object(edir_ipc_handle_t *handle)
{
    if (!VALID_HANDLE(handle)) return NULL;
    return ((edir_shmem_mmap_ctx_t *)(handle->pData))->attach;
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char *
edir_shmem_handle_get_object_name(edir_ipc_handle_t *handle)
{
    edir_shmem_mmap_ctx_t	*ctx = NULL;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_shmem_mmap_ctx_t *) handle->pData;
    if (ctx->pathname == NULL) return NULL;
    return edir_strdup(ctx->pathname);
}

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void
edir_shmem_set_default_filename(char  *filename)
{
    DefaultSet = 1;
    DefaultFileName = edir_strdup(filename);
}

/**
 * Get the default name/key for this class.
 * <P>
 * If no default has been set with edir_mutex_set_default_filename(),
 * then return NULL;
 * 
 * @return The default file name.
 */
extern char *
edir_shmem_default_name()
{
    if (!DefaultSet) return NULL;
    return edir_strdup(DefaultFileName);
}
