/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_shmem_sysv.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_shmem_sysv.h"
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>

/**
 * Shared memory access functions for System V IPC.
 * <P>
 * The functions in this file implement the generic Enhydra Director
 * shared memory interface using the System V shmget() and shmat() calls.
 * Some systems, such as FreeBSD and Linux must be configured to enable
 * System V IPC compatibility in their kernels in order to use this
 * shared memory method.
 * 
 * @author John Marco
 * @since  Enhydra 2.4
 */

/**
 * System-dependent context info for a System V IPC shared memory region.
 */
typedef struct edir_shmem_sysv_ctx {
    key_t		shmkey;
    int			shmid;
    void		*attach;
    edir_shmem_callback_t	destroyFunc;
} edir_shmem_sysv_ctx_t;

#define _VALID_TYPE(H) ((H)->handle_type == EDIR_IPC_HANDLE_TYPE_SHMEM)
#define _VALID_HANDLE(H) ((H) && ((H)->pData) && _VALID_TYPE(H))
#define VALID_HANDLE(V) _VALID_HANDLE(((edir_ipc_handle_t *)(V)))

/**
 * Create a shared memory region.
 * <P>
 * Creates a brand new shared memory region of the specified size.  The
 * key may or may not be used, depending on the system-dependent
 * implementation.  A NULL key means "use the system-dependent default".
 * Some implementations may leave shared memory resources allocated,
 * even after the program has terminated, unless edir_shmem_destroy() is
 * called to clean up.
 * 
 * @param key     [in]  Handle to object to use as configuration source.
 * @param size    [in]  The desired size of the shared memory region.
 * @param func    [in]  Optional destroy function.  (May be NULL)
 * @param exc     [out] Exception pointer filled in if an exception occurs.
 * @return		A shared memory handle object.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be created.
 */
extern edir_ipc_handle_t *
edir_shmem_create(edir_ipc_handle_t *key, int32 size,
		edir_shmem_callback_t func, edir_exception_t **exc)
{
    static char		*funcName = "edir_shmem_create";
    edir_shmem_sysv_ctx_t	*ctx = NULL;
    edir_ipc_handle_t	*handle = NULL;
    char		*stringName = NULL, *endptr = NULL;
    int			*intp;

    /*
     * Set up a context structure for this new region.
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_SHMEM;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_INTEGER;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_shmem_sysv_ctx_t *) edir_malloc(sizeof(edir_shmem_sysv_ctx_t));
    ctx->destroyFunc = func;
    handle->pData = (void *)ctx;
    ctx->shmkey = -1;

    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    stringName = edir_ipc_config_handle_get_object_name(key);
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from CONFIG",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else if (key->handle_type == EDIR_IPC_HANDLE_TYPE_MUTEX) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_FILEDESC:
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    stringName = edir_mutex_handle_get_object_name(key);
		    break;
		case EDIR_IPC_OBJECT_TYPE_INTEGER:
		    intp = (int *) edir_mutex_handle_get_object(key);
		    if (intp == NULL) {
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			    "Can't get integer from shared mutex object");
			EDIR_THROW(*exc);
		    } else {
			ctx->shmkey = *intp;
		    }
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from MUTEX",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Unexpected handle type %d", key->handle_type);
	    EDIR_THROW(*exc);
	}
    }
    if (ctx->shmkey <= 0) {
	if (stringName != NULL) {
	    errno = 0;
            ctx->shmkey = strtol(stringName, &endptr, 0);
            if ((ctx->shmkey <= 0) || (endptr == stringName)) errno = EINVAL;
            if (endptr) while (*endptr) if (!isspace(*endptr++)) errno = EINVAL;
            if (errno != 0) {
                char *cp;
                int     hash = 0;
                for (cp=stringName; *cp; cp++) hash += (octet)(*cp);
                ctx->shmkey = ftok("/dev/null", hash);
            }
	}
    }
    if (ctx->shmkey <= 0) ctx->shmkey = EDIR_SHMEM_SYSV_DEFAULT_IPC_KEY;

    if (stringName) edir_free(stringName);
    stringName = NULL;

    ctx->shmid = -1;
    ctx->attach = NULL;

    /*
     * Obtain the region using the given key.
     */
    ctx->shmid = shmget(ctx->shmkey, size, IPC_CREAT|IPC_EXCL|0600);
    if (ctx->shmid < 0) {
	if (errno == EEXIST) {
	    ctx->shmid = shmget(ctx->shmkey, 1, 0);
	    if (ctx->shmid < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Could not get old shared memory to destroy it");
		EDIR_THROW(*exc);
	    }
	    if (shmctl(ctx->shmid, IPC_RMID, NULL) < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Could not destroy old shared memory");
		EDIR_THROW(*exc);
	    }
	    ctx->shmid = shmget(ctx->shmkey, size, IPC_CREAT|IPC_EXCL|0600);
	    if (ctx->shmid < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Could not create new shared memory");
		EDIR_THROW(*exc);
	    }
	} else if (errno == ENOSYS) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"System V IPC must be enabled in the kernel");
	    EDIR_THROW(*exc);
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Cannot create shared memory region");
	    EDIR_THROW(*exc);
	}
    }

    /*
     * Attach the region to our address space.
     */
    ctx->attach = (char *) shmat(ctx->shmid, 0, 0);
    if (ctx->attach == ((void *)(-1))) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			    "Cannot attach shared memory region");
	ctx->attach = NULL;
	EDIR_THROW(*exc);
    }

    return handle;

except:
    /* Exception cleanup */
    if (ctx) {
	if (ctx->attach) {
	    shmdt(ctx->attach);
	}
	if (ctx->shmid >= 0) {
	    shmctl(ctx->shmid, IPC_RMID, NULL);
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    if (stringName) edir_free(stringName);
    return NULL;
}

/**
 * Destroy a shared memory region.
 * <P>
 * Detaches the given shared memory region from the data segment and
 * removes its key and ID from the system shared memory tables, if
 * applicable.
 * 
 * @param handle [in]  Handle to the shared memory object.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be destroyed.
 */
extern void
edir_shmem_destroy(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char 	*funcName = "edir_shmem_destroy";
    edir_shmem_sysv_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, "Invalid shmem handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_shmem_sysv_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_DESTROY_SHARE(handle, exc);

    if (ctx) {
	if (ctx->shmid >= 0) {
	    if (shmctl(ctx->shmid, IPC_RMID, NULL) < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Could not remove shared memory region");
		ctx->shmid = -1;
		EDIR_THROW(*exc);
	    }
	    ctx->shmid = -1;
	}
	if (ctx->attach) {
	    if (ctx->destroyFunc != NULL) {
		(ctx->destroyFunc)(ctx->attach, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    if (shmdt(ctx->attach) < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Could not detach shared memory");
		ctx->attach = NULL;
		EDIR_THROW(*exc);
	    }
	    ctx->attach = NULL;
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;

except:
    /* Last-ditch exception cleanup */
    if (ctx) {
	if (ctx->shmid >= 0) {
	    shmctl(ctx->shmid, IPC_RMID, NULL);
	}
	if (ctx->attach) {
	    shmdt(ctx->attach);
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Obtain an existing shared memory region.
 * <P>
 * A shared memory region must already be mapped to the specified key.
 * Otherwise an exception is generated.  This function returns a pointer
 * to the mapped shared memory area if it exists and can be mapped.
 * 
 * @param key     [in]  Handle to object to use as configuration source.
 * @param exc     [out] Exception pointer filled in if an exception occurs.
 * @return		A shared memory handle object.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be mapped.
 */
extern edir_ipc_handle_t *
edir_shmem_obtain(edir_ipc_handle_t *key, edir_exception_t **exc)
{
    static char		*funcName = "edir_shmem_obtain";
    edir_shmem_sysv_ctx_t	*ctx = NULL;
    edir_ipc_handle_t	*handle = NULL;
    char		*stringName = NULL, *endptr = NULL;
    int			*intp = NULL;

    /*
     * Set up a context structure for the region (if found).
     */
    handle = (edir_ipc_handle_t *) edir_malloc(sizeof(edir_ipc_handle_t));
    handle->handle_type = EDIR_IPC_HANDLE_TYPE_SHMEM;
    handle->object_type = EDIR_IPC_OBJECT_TYPE_INTEGER;
    handle->nrefs = 1;
    handle->share = NULL;
    ctx = (edir_shmem_sysv_ctx_t *) edir_malloc(sizeof(edir_shmem_sysv_ctx_t));
    ctx->destroyFunc = NULL;
    handle->pData = (void *)ctx;
    ctx->shmkey = -1;

    if (key != NULL) {
	if (key->handle_type == EDIR_IPC_HANDLE_TYPE_CONFIG) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    stringName = edir_ipc_config_handle_get_object_name(key);
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from CONFIG",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else if (key->handle_type == EDIR_IPC_HANDLE_TYPE_MUTEX) {
	    switch (key->object_type) {
		case EDIR_IPC_OBJECT_TYPE_FILEDESC:
		case EDIR_IPC_OBJECT_TYPE_STRING:
		    stringName = edir_mutex_handle_get_object_name(key);
		    break;
		case EDIR_IPC_OBJECT_TYPE_INTEGER:
		    intp = (int *) edir_mutex_handle_get_object(key);
		    if (intp == NULL) {
			*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			    "Can't get integer from shared mutex object");
			EDIR_THROW(*exc);
		    } else {
			ctx->shmkey = *intp;
		    }
		    break;
		default:
		    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
			"Unexpected object type %d from MUTEX",
			key->object_type);
		    EDIR_THROW(*exc);
	    }
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Unexpected handle type %d", key->handle_type);
	    EDIR_THROW(*exc);
	}
    }
    if (ctx->shmkey <= 0) {
	if (stringName != NULL) {
	    errno = 0;
            ctx->shmkey = strtol(stringName, &endptr, 0);
            if ((ctx->shmkey <= 0) || (endptr == stringName)) errno = EINVAL;
            if (endptr) while (*endptr) if (!isspace(*endptr++)) errno = EINVAL;
            if (errno != 0) {
                char *cp;
                int     hash = 0;
                for (cp=stringName; *cp; cp++) hash += (octet)(*cp);
                ctx->shmkey = ftok("/dev/null", hash);
            }
	}
    }
    if (ctx->shmkey <= 0) ctx->shmkey = EDIR_SHMEM_SYSV_DEFAULT_IPC_KEY;

    if (stringName) edir_free(stringName);
    stringName = NULL;

    ctx->shmid = -1;
    ctx->attach = NULL;

    /*
     * Obtain the region using the given key.
     */
    ctx->shmid = shmget(ctx->shmkey, 1, 0600);
    if (ctx->shmid < 0) {
	if (errno == ENOSYS) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"System V IPC must be enabled in the kernel");
	    EDIR_THROW(*exc);
	} else {
	    *exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Cannot find shared memory region");
	    EDIR_THROW(*exc);
	}
    }

    /*
     * Attach the region to our address space.
     */
    ctx->attach = (char *) shmat(ctx->shmid, 0, 0);
    if (ctx->attach == ((void *)(-1))) {
	*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			    "Cannot attach shared memory region");
	ctx->attach = NULL;
	EDIR_THROW(*exc);
    }

    return handle;

except:
    /* Exception cleanup */
    if (ctx) {
	if (ctx->attach) {
	    shmdt(ctx->attach);
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    if (stringName) edir_free(stringName);
    return NULL;
}

/**
 * Detach a shared memory region.
 * <P>
 * Detaches the given shared memory region from the data segment, but does not
 * remove its key and ID from the system shared memory tables.
 * 
 * @param handle [in]  Handle to the shared memory object.
 * @param exc    [out] Exception pointer filled in if an exception occurs.
 * @exception EDIR_EXCEPT_SHMEM If the shared memory region cannot be detached.
 */
extern void
edir_shmem_release(edir_ipc_handle_t *handle, edir_exception_t **exc)
{
    static char 	*funcName = "edir_shmem_release";
    edir_shmem_sysv_ctx_t *ctx = NULL;

    if (!VALID_HANDLE(handle)) {
        *exc = edir_exception_create(EDIR_EXCEPT_SHMEM, "Invalid shmem handle");
        EDIR_THROW(*exc);
    }

    ctx = (edir_shmem_sysv_ctx_t *) handle->pData;

    handle->nrefs--;
    if (handle->nrefs > 0) return; /* Someone still using this */

    EDIR_IPC_RELEASE_SHARE(handle, exc);

    if (ctx) {
	if (ctx->attach) {
	    if (shmdt(ctx->attach) < 0) {
		*exc = edir_exception_create(EDIR_EXCEPT_SHMEM,
			"Could not detach shared memory");
		ctx->attach = NULL;
		EDIR_THROW(*exc);
	    }
	    ctx->attach = NULL;
	}
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;

except:
    /* Last-ditch exception cleanup */
    if (ctx) {
	edir_free(ctx);
    }
    if (handle) edir_free(handle);
    return;
}

/**
 * Sets the destruction callback function for the specified instance.
 * <P>
 * The destroy callback performs actions such as setting flags and
 * the like, just after the shared memory was unlinked from its system
 * wide name and just before it is unmapped from this process.
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @param func   [in] Function to call upon destroy.
 * @return The old destroy function
 */
extern edir_shmem_callback_t
edir_shmem_set_destroy_func(edir_ipc_handle_t *handle, edir_shmem_callback_t func)
{
    edir_shmem_sysv_ctx_t *ctx = NULL;
    edir_shmem_callback_t ret;
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_shmem_sysv_ctx_t *) handle->pData;
    ret = ctx->destroyFunc;
    ctx->destroyFunc = func;
    return ret;
}

/**
 * Return a pointer to the underlying handle object.
 * <P>
 * For shared memory objects, this pointer always points to the beginning
 * of the mapped memory area.
 * <P>
 * This pointer is a reference MUST NOT be freed!
 * 
 * @param handle [in] The handle from which to obtain the pointer.
 * @return Reference to the underlying object, or NULL if not valid.
 */
extern void *
edir_shmem_handle_get_object(edir_ipc_handle_t *handle)
{
    if (!VALID_HANDLE(handle)) return NULL;
    return ((edir_shmem_sysv_ctx_t *)(handle->pData))->attach;
}

/**
 * Return a string representation of the underlying object's name.  
 * <P>
 * This string must be freed using edir_free().
 * 
 * @param handle [in] The handle from which to obtain the name.
 * @return The name of the underlying object, or NULL if not valid.
 */
extern char * edir_shmem_handle_get_object_name(edir_ipc_handle_t *handle)
{
    edir_shmem_sysv_ctx_t	*ctx = NULL;
    char		buf[64];
    if (!VALID_HANDLE(handle)) return NULL;
    ctx = (edir_shmem_sysv_ctx_t *) handle->pData;
    if (ctx->shmkey == -1) return NULL;
    sprintf(buf, "%08x", ctx->shmkey);
    return edir_strdup(buf);
}

/**
 * Allow specific modules to set their own default file name.
 * <P>
 * This file name may be ignored, if the underlying mechanism does
 * not use files.
 * 
 * @param filename [in] The new default file name.
 */
extern void
edir_shmem_set_default_filename(char  *filename)
{
    ; /* do nothing -- files not used here */
}

/**
 * Get the default name/key for this class.
 * <P>
 * If no default has been set with edir_mutex_set_default_filename(),
 * then return NULL;
 * 
 * @return The default file name.
 */
extern char *
edir_mutex_default_name()
{
    return NULL;
}

