/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_socket.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"

/**
 * Convert a valid dotted-quad-decimal IP address into an in_addr
 * structure suitable for use in socket calls.
 * 
 * @param hostname [in] Pointer to string containing a valid dotted-decimal
 * 	IP address to convert.
 * @param ap       [out] Pointer to an in_addr variable to receive the converted
 * 	address value.
 * @param exc      [out] Exception pointer to receive any errors.
 * @return True(1) if the address is valid, False(0) otherwise.
 */
extern int
edir_ip_addr(char *hostname, struct in_addr *ap)
{
    char buf[20]; /* Max length is 15 -- 5 more to simplify parser */
    char *cp1, *cp2, *cp3;
    octet *ucp;
    int32 d[4]; /* atoi returns signed int. */
    int i, len;
    
    /*
     * Be anal about length checking.
     */
    if (strlen(hostname) > 15) goto fail;
    memset(buf, 0, 20);
    strncpy(buf, hostname, 19);

    cp1 = buf;
    for (i=0; i<4; i++) {
	for (cp2=cp1; *cp2 && (*cp2 != '.'); cp2++) ;
	*cp2 = '\0';
	len = strlen(cp1);
	if ((len < 1) || (len > 3)) goto fail;
	for (cp3=cp1; *cp3; cp3++) if (!isdigit(*cp3)) goto fail;
	d[i] = atoi(cp1);
	if ((d[i] < 0) || (d[i] > 255)) goto fail;
	cp1 = cp2 + 1;
    }
    if (i < 4) goto fail;
    /*
     * At this point, we have 4 valid dotted decimals.
     * Write them out, MSB first. (Net byte order)
     */
    ucp = (octet *) &(ap->s_addr);
    *ucp++ = (octet) d[0];
    *ucp++ = (octet) d[1];
    *ucp++ = (octet) d[2];
    *ucp   = (octet) d[3];
    return 1;
fail:
    return 0;
}

/**
 * Create an Internet Domain "sockaddr_in" structure from a given
 * host name and port number.  This function only handles the TCP
 * protocol for Internet Domain socket addresses.
 * 
 * @param hostname [in]  Host name (dotted IP, host, or host.domain) 
 * @param port     [in]  Port to associate this address with.  If
 * 	port is EDIR_PORT_ANY, then associate with default (all/any) port.
 * @param sin      [out] Pointer to sockaddr_in struct variable to receive
 * 	the address data.
 * @param exc      [out] Exception pointer to receive any error.
 * @exception EDIR_EXCEPT_IO_ERROR If the host name cannot be converted.
 */
extern void
edir_sockaddr_in(char *hostname, int port,
	       struct sockaddr_in *sin, edir_exception_t **exc)
{
    static char *funcName = "edir_sockaddr_in";
    struct hostent *hp;
    
    sin->sin_family = AF_INET;
    if (port == EDIR_PORT_ANY) sin->sin_port = 0;
    else sin->sin_port = htons((uint16)port);
    if ((hostname == NULL) || (strlen(hostname)==0)) {
	sin->sin_addr.s_addr = INADDR_ANY;
	return;
    }
    /* If quad dotted decimal, don't go DNS lookup. */
    if (edir_ip_addr(hostname, &(sin->sin_addr))) {
	return;
    }
    if ((hp = gethostbyname(hostname)) == NULL) {
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
				   "%s: gethostbyname() failed", hostname);
	EDIR_THROW(*exc);
    }
    if ((hp->h_addr_list == NULL) || (hp->h_addr_list[0] == NULL)) {
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
				   "%s: Unknown host", hostname);
	EDIR_THROW(*exc);
    }
    if (hp->h_addrtype != AF_INET) {
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
				   "%s: gethostbyname() returned %s",
				   hostname, " unknown address type.");
	EDIR_THROW(*exc);
    }
    memcpy((char *)&sin->sin_addr, (char *)hp->h_addr, sizeof(sin->sin_addr));
    return;

except:
    return;
}

/**
 * Convenience function to convert the IP address of a sockaddr_in struct
 * into an ASCII string suitable for printing.
 * <P>
 * Note that this function expects to be passed a character buffer
 * with at least 16 characters in it.  If a smaller buffer is passed,
 * the likely result will be a crash of the application due to a 
 * buffer overrun.
 * 
 * @param sin [in] Pointer to a sockaddr_in struct to convert.
 * @param buf [in,out] Pointer to character array with room for at least
 * 		16 characters.  Will be filled in with a null-terminated
 * 		string representing the IP address.
 */
extern void 
edir_sockaddr_to_str(struct sockaddr_in *sin, char *buf)
{
    uint32 ipaddr = (uint32) ntohl(sin->sin_addr.s_addr);
    sprintf(buf, "%d.%d.%d.%d", (ipaddr >> 24) & 0xff,
				(ipaddr >> 16) & 0xff, 
				(ipaddr >>  8) & 0xff, 
				 ipaddr        & 0xff);
}

