/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_socket_win32.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"

#ifdef DEBUG
static void hexdump(char *msg, octet *buf, int len);
static void hexdumpv(char *msg, struct iovec *iov, int niovs);
#endif

/**
 * Create a new socket descriptor for a connection.
 * <P>
 * Creates a new AF_INET socket descriptor and places it in the 
 * connection context referred to by <code>conn</code>.
 *
 * @param conn [in,out] Connection context to contain the new socket.
 * @param exc  [out]	Exception pointer to receive any exceptions.
 * 
 * @exception	EDIR_EXCEPT_IO_ERROR
 * 	If creation of the new socket fails due to system resource
 * 	shortages or initialization errors.
 */
extern void
edir_socket(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_socket";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    conn->ioPending = 0; /* This is a blocking implementation */
    conn->socket = socket(AF_INET, SOCK_STREAM, 0);
    if (EDIR_SOCKET_ISNULL(conn->socket)) {
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
		    "Error creating socket: @wsastr@ (errnum=@wsaerr@)");
	EDIR_THROW(*exc);
    }
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;

except:
    if (conn) conn->socket = EDIR_NULL_SOCKET;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Bind a connections socket to its bind host and address.
 * <P>
 * A socket <code>bind()</code> is attempted to the local bind address
 * listed in the connection context referred to by <code>conn</code>
 * <P>
 * If the bind host string is null or of zero length, we bind to
 * the default (any) address.  If the port is EDIR_PORT_ANY, bind to
 * any port the system chooses.
 * 
 * @param conn [in,out] Connection context handle.
 * @param exc  [out]    Exception handle filled in upon errors.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If the socket cannot be bound to the local bind address contained
 * 	in the connection context referenced by <code>conn</code>.
 */
extern void
edir_socket_bind(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_socket_bind";
    int status, tries=3;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    conn->ioPending = 0; /* This is a blocking implementation */

    conn->localAddr.sin_family = AF_INET;
    status = bind(conn->socket, (struct sockaddr FAR *) &(conn->localAddr),
		  sizeof(struct sockaddr_in));
    if (status == SOCKET_ERROR) {
	char host[16];
	int port, wsastatus;
	wsastatus = WSAGetLastError();
	edir_sockaddr_to_str(&(conn->localAddr), host);
	port = ntohs(conn->localAddr.sin_port);
#ifdef WSAEADDRNOTAVAIL
	while (wsastatus == WSAEADDRNOTAVAIL) {
	    if (tries-- > 0) {
		Sleep(500);
		continue;
	    }
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR, "%s%s%s%s",
		    "Cannot bind to INADDR_ANY:\n",
		    "There may be too many connections in TIME_WAIT.\n",
		    "Consider tuning your system to reduce the ",
		    "TIME_WAIT interval");
	    EDIR_THROW(*exc);
	}
#endif
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
		    "Error binding to %s:%d: @wsastr@ (errnum=@wsaerr@)",
		    host, port);
	EDIR_THROW(*exc);
    }

except:
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Connect a connection's socket to its server IP address and port.
 * <P>
 * A socket <code>connect()</code> is attempted to the remote address
 * listed in the connection context referred to by <code>conn</code>
 * 
 * @param conn [in,out] Connection context handle.
 * @param exc  [out]    Exception handle filled in upon errors.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If the socket cannot be connected to the remote address contained
 * 	in the connection context referenced by <code>conn</code>.
 */
extern void
edir_socket_connect(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_socket_connect";
    int status;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    conn->ioPending = 0; /* This is a blocking implementation */
    conn->serverAddr.sin_family = AF_INET;

    do {
	status = connect(conn->socket,
			 (struct sockaddr FAR *) &(conn->serverAddr),
			 sizeof(struct sockaddr_in));
    } while ((status == SOCKET_ERROR) && (WSAGetLastError() == WSAEINTR));
    if (status == SOCKET_ERROR) {
	char host[16];
	int port;
	edir_sockaddr_to_str(&(conn->serverAddr), host);
	port = ntohs(conn->serverAddr.sin_port);
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
		    "Error connecting to %s:%d: @wsastr@ (errnum=@wsaerr@)",
		    host, port);
	EDIR_THROW(*exc);
    }

except:
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Close a connection's socket.
 * <P>
 * A socket <code>close()</code> is attempted to the possibly open socket
 * descriptor in the connection context referred to by <code>conn</code>.
 * 
 * @param conn [in,out] Connection context handle.
 * @param exc  [out]    Exception handle filled in upon errors.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If a socket error occurs while closing the socket.
 */
extern void
edir_socket_close(edir_connection_t *conn, edir_exception_t **exc) 
{
    static char *funcName = "edir_socket_connect";
    int status;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    conn->ioPending = 0; /* This is a blocking implementation */
    do {
	status = closesocket(conn->socket);
    } while ((status == SOCKET_ERROR) && (WSAGetLastError() == WSAEINTR));
    if (status == SOCKET_ERROR) {
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
		    "Error closing socket: @wsastr@ (errnum=@wsaerr@)");
	EDIR_THROW(*exc);
    }

except:
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Writes an entire buffer to the connection's socket.
 * <P>
 * Blocks as necessary until all bytes have been written, or an
 * exception occurs.
 * <P>
 * This is a blocking implementation.  It never returns until
 * I/O is complete.  conn->ioPending is always 0 after this function
 * is called.
 * 
 * @param conn [in,out] Connection context handle.
 * @param exc  [out]    Exception handle filled in upon errors.
 * @param buf  [in]     Buffer of data to send on the connection.
 * @param len  [in]     Number of bytes to send.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If a socket error occurs while sending the data.
 */
extern void
edir_socket_write(edir_connection_t *conn, edir_exception_t **exc,
		void *buf, int32 len)
{
    static char *funcName = "edir_socket_write";
    int32 act, actual=0, remainder;
    octet *bp;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", "EDIR_PTR", %d)", \
		      funcName, conn, *exc, buf, len));

    EDIR_IF_DEBUG( if (edir_debug & EDIR_DEBUG_MASK_SOCK_SEND) \
		 hexdump("Socket Write:", buf, len); )
   
    remainder = len;
    bp = (octet *) buf;
    while (remainder > 0) {
	do {
	    act = send(conn->socket, (char *) bp, remainder, 0);
	} while ((act == SOCKET_ERROR) && (WSAGetLastError() == WSAEINTR));
	if (act <= 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
		    "send() failed: @wsastr@ (errnum=@wsaerr@)");
	    EDIR_THROW(*exc);
	}
	actual += act;
	remainder -= act;
	bp += act;
    }

except:
    conn->dataActual += actual;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Scatter-gather writes the contents of a string/data buffer object to the
 * connection's socket.
 * <P>
 * For the Win32 implementation, we call WSASend() to get scatter-gather
 * I/O.  The observant among us will note that I macroed struct iovec to
 * be the same as WSABUF.  Isn't that conveeenient?
 * <P>
 * This function still has 'write entire buffer' semantics, but it should
 * be easier to implement a non-blocking version.
 * <P>
 * Note that this function acts destructively on the supplied sbuf.  It
 * discards any buffered data successfully sent over the connection.  This
 * will allow a nonblocking version to call multiple times.
 * 
 * @param conn [in,out] Connection context handle.
 * @param exc  [out]    Exception handle filled in upon errors.
 * @param sbuf [in]     Opaque 'edir_sbuf' handle.
 * 
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If a socket error occurs while sending the data.
 */
extern void
edir_socket_write_sbuf(edir_connection_t *conn, edir_exception_t **exc,
		       void *sbuf)
{
    static char		*funcName = "edir_socket_write_sbuf";
    struct iovec	*iovp;
    int32		niovs = 0, iovskip = 0;
    int32		act, actual=0, remainder;
    int			status;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", "EDIR_PTR")",
		        funcName, conn, *exc, sbuf));

    /*
     * Get the length of the data.
     */
    remainder = edir_sbuf_length(sbuf);
    if (remainder < 1) {
	conn->dataActual = 0;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
        return;
    }

    /*
     * Get a pointer to the internal iovecs.
     */
    iovp = edir_sbuf_get_iov(sbuf, &niovs);
    EDIR_IF_DEBUG( if (edir_debug & EDIR_DEBUG_MASK_SOCK_SEND) \
		 hexdumpv("Socket WriteV:", iovp, niovs); )

    while (remainder > 0) {
	do {
	    status = WSASend(conn->socket, iovp, niovs, &act, 0, NULL, NULL);
	} while ((status != 0) && (WSAGetLastError() == WSAEINTR));
	if (status != 0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
		    "WSASend() failed: @wsastr@ (errnum=@wsaerr@)");
	    EDIR_THROW(*exc);
	}
	actual += act;
	remainder -= act;
	/* See if we need to advance the iovec pointer. */
	if (remainder > 0) {
	    while (act >= (int32) iovp->iov_len) {
		act -= iovp->iov_len;
		iovp++;
		niovs--;
		iovskip++;
	    }
	    /* Adjust len,base in new first iov */
	    iovp->iov_base += act;
	    iovp->iov_len -= act;
	}
    }

except:
    conn->dataActual = actual;
    edir_sbuf_skip(sbuf, iovskip);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Receive at least one byte of data on a connection.
 * <P>
 * Reads the connection's socket and receives at least one byte
 * of data if an error does not occur.  If data are available,
 * returns immediately, even if the actual number read is less
 * than the "len" parameter.
 * 
 * @param conn	[in,out] Connection to receive on.
 * @param exc	[out]    Exception pointer to receive errors.
 * @param buf	[in]     Incominig data buffer.
 * @param len	[in]     Length of buffer.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while reading.
 */
extern void
edir_socket_receive(edir_connection_t *conn, edir_exception_t **exc,
		void *buf, int32 len)
{
    static char *funcName = "edir_socket_receive";
    int32 act = 0;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", "EDIR_PTR", %d)", \
		        funcName, conn, *exc, buf, len));
    do {
        act = recv(conn->socket, buf, len, 0);
    } while ((act == SOCKET_ERROR) && (WSAGetLastError() == WSAEINTR));
    if (act == SOCKET_ERROR) {
	edir_exception_create(EDIR_EXCEPT_IO_ERROR,
			"Socket read error: @wsastr@ (errnum=@wsaerr@)");
	EDIR_THROW(*exc);
    }
	if (act > 0) {
    EDIR_IF_DEBUG( if (edir_debug & EDIR_DEBUG_MASK_SOCK_RECV) \
		 hexdump("Socket Receive:", buf, act); )
	}
	conn->dataActual = (act > 0) ? act : 0;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
	return;

except:
	conn->dataActual = (act > 0) ? act : 0;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Receive an entire buffer of data.
 * <P>
 * Reads the connection's socket and receives bytes until the
 * buffer is filled, or an error occurs.
 * 
 * @param conn	[in,out] Connection to receive on.
 * @param exc	[out]    Exception pointer to receive errors.
 * @param buf	[in]     Incominig data buffer.
 * @param len	[in]     Length of buffer.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while reading.
 */
extern void
edir_socket_receive_all(edir_connection_t *conn, edir_exception_t **exc,
		void *buf, int32 len)
{
    static char *funcName = "edir_socket_receive";
    int32 act, actual=0, remainder;
    octet *bp;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", "EDIR_PTR", %d)", \
		      funcName, conn, *exc, buf, len));
    remainder = len;
    bp = (octet *) buf;
    while (remainder > 0) {
	do {
	    act = recv(conn->socket, (char *) bp, remainder, 0);
	} while ((act == SOCKET_ERROR) && (WSAGetLastError() == WSAEINTR));
	if (act == SOCKET_ERROR) {
	    edir_exception_create(EDIR_EXCEPT_IO_ERROR,
			"Socket read error: @wsastr@ (errnum=@wsaerr@)");
	    EDIR_THROW(*exc);
	}
	if (act == 0) break;
	actual += act;
	remainder -= act;
	bp += act;
    }
except:
    EDIR_IF_DEBUG( if (edir_debug & EDIR_DEBUG_MASK_SOCK_RECV) \
		 hexdump("Socket Receive All:", buf, actual); )
    conn->dataActual = actual;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Receive and discard data from the connection's receive socket.
 * <P>
 * Receives and discards <code>len</code> bytes of data from the connection's
 * receiver socket.  If end-of-data is reached, then fewer than len
 * bytes may be read.
 * 
 * @param conn	[in,out] Connection to discard from
 * @param exc	[out]    Exception pointer to receive errors.
 * @param len	[in]     Number of bytes to discard.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while receiving data to discard.
 */
extern void
edir_socket_skip(edir_connection_t *conn, edir_exception_t **exc, int32 len)
{
    static char *funcName = "edir_socket_skip";
    int32 act, actual=0, remainder;
    octet buf[512];
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", %d)", \
		      funcName, conn, *exc, len));
    remainder = len;
    while (remainder > 0) {
	do {
	    act = recv(conn->socket, (char *) buf,
		       (remainder<512) ? remainder : 512, 0);
	} while ((act == SOCKET_ERROR) && (WSAGetLastError() == WSAEINTR));
	if (act == SOCKET_ERROR) {
	    edir_exception_create(EDIR_EXCEPT_IO_ERROR,
			"Socket discard error: @wsastr@ (errnum=@wsaerr@)");
	    EDIR_THROW(*exc);
	}
	if (act == 0) break;
	actual += act;
	remainder -= act;
    }
except:
    conn->dataActual = actual;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Nonblocking I/O completion poll.
 * <P>
 * This implementation is for blocking sockets, so this function
 * always sets conn->ioPending to false(0) and returns immediately.
 * 
 * @param conn	[in,out] Connection to discard from.
 * @param exc	[out]    Exception pointer to receive errors.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while receiving data to discard.
 */
extern void
edir_socket_finish_skip(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_socket_finish_skip("EDIR_PTR", "EDIR_PTR")", \
		      conn, *exc));
    conn->ioPending = 0;
    EDIR_DEBUG_FUNCS_OUT(("func: edir_socket_finish_skip() returns void"));
}

/**
 * Nonblocking I/O completion poll.
 * <P>
 * This implementation is for blocking sockets, so this function
 * always sets conn->ioPending to false(0) and returns immediately.  
 * 
 * @param conn	[in,out] Connection to send to.
 * @param exc	[out]    Exception pointer to receive errors.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while sending data.
 */
extern void
edir_socket_finish_send(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_socket_finish_send("EDIR_PTR", "EDIR_PTR")", \
		      conn, *exc));
    conn->ioPending = 0;
    EDIR_DEBUG_FUNCS_OUT(("func: edir_socket_finish_send() returns void"));
}

/**
 * Nonblocking I/O completion poll.
 * <P>
 * This implementation is for blocking sockets, so this function
 * always sets conn->ioPending to false(0) and returns immediately.  
 * 
 * @param conn	[in,out] Connection to receive from.
 * @param exc	[out]    Exception pointer to receive errors.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while sending data.
 */
extern void
edir_socket_finish_receive(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_socket_finish_receive("EDIR_PTR", "EDIR_PTR")", \
		      conn, *exc));
    conn->ioPending = 0;
    EDIR_DEBUG_FUNCS_OUT(("func: edir_socket_finish_receive() returns void"));
}

/**
 * Nonblocking I/O completion poll.
 * <P>
 * This implementation is for blocking sockets, so this function
 * always sets conn->ioPending to false(0) and returns immediately.
 * 
 * @param conn	[in,out] Connection awaiting establishment.
 * @param exc	[out]    Exception pointer to receive errors.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 	If an error occurs while establishing the socket connection.
 */
extern void
edir_socket_finish_connect(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_socket_finish_connect("EDIR_PTR", "EDIR_PTR")", \
		      conn, *exc));
    conn->ioPending = 0;
    EDIR_DEBUG_FUNCS_OUT(("func: edir_socket_finish_connect() returns void"));
}

#ifdef DEBUG
/**
 * Debugging function to hexdump a byte buffer to the debug log.
 * <P>
 * Dumps the buffer, 16 bytes per line, in the format of SCO's 'hd',
 * or BSD's 'hexdump' command.  Very useful for debugging protocol
 * errors.
 * 
 * @param msg [in] Message to prepend to the log output.
 * @param buf [in] The buffer to dump.
 * @param len [in] The length of the buffer.
 */
static void
hexdump(char *msg, octet *buf, int len)
{
    octet xbuf[16];
    char  pbuf[128];
    char  lineBuf[512];
    int ch, i, j, k, n=0;
    if (msg) edir_log_debug("%s", msg);
    sprintf(lineBuf, "00000000:");
    for (j=0; j<len; j++, n++) {
	ch = ((unsigned char *)buf)[j] & 0xff;
	sprintf(pbuf, " %02x", ch);
	strcat(lineBuf, pbuf);
	xbuf[n%16] = (octet) ch;
	if ((n % 16) == 15) {
	    strcat(lineBuf, "   ");
	    for (k=0; k<16; k++) {
		if ((xbuf[k] < ' ') || (xbuf[k] > '~') ||
			(xbuf[k] == '<') || (xbuf[k] == '>')) 
		    strcat(lineBuf, ".");
		else
		    strncat(lineBuf, (char *)(&(xbuf[k])), 1);
	    }
	    edir_log_debug("%s", lineBuf);
	    sprintf(lineBuf, "%08x:", n + 1);
	}
    }
    j = n % 16;
    k = 16 - j;
    if (j) {
	for (i=0; i<k; i++) strcat(lineBuf, "   ");
	strcat(lineBuf, "   ");
	for (i=0; i<j; i++) {
	    if ((xbuf[i] < ' ') || (xbuf[i] > '~') ||
		    (xbuf[i] == '<') || (xbuf[i] == '>'))
		strcat(lineBuf, ".");
	    else
		strncat(lineBuf, (char *)(&(xbuf[i])), 1);
	}
    }
    edir_log_debug("%s", lineBuf);
}

/**
 * Debugging function to hexdump a scatter-gather buffer to the debug log.
 * <P>
 * Like <code>hexdump()</code>, except that it takes a pointer to an
 * array of 'struct iovec', such as would be passed to sendmsg().
 * Dumps the buffer, 16 bytes per line, in the format of SCO's 'hd',
 * or BSD's 'hexdump' command.  Very useful for debugging protocol
 * errors.  Adds a '/' character between hex-printed bytes at the 
 * boundary of each iovec fragment.
 * 
 * @param msg   [in] Message to prepend to the log output.
 * @param iov   [in] Array of <code>struct iovec</code>.
 * @param niovs [in] Number of <code>struct iovec</code> elements.
 */
static void
hexdumpv(char *msg, struct iovec *iov, int niovs)
{
    octet xbuf[16];
    char  pbuf[128];
    char  lineBuf[512];
    int ch, i, j, k, n=0;
    char *spc = " ";
    edir_log_debug("%s", msg);
    sprintf(lineBuf, "00000000:");
    for (i=0; i<niovs; i++) {
	for (j=0; j<(int)(iov[i].iov_len); j++, n++) {
	    ch = ((unsigned char *)((iov[i]).iov_base))[j] & 0xff;
	    sprintf(pbuf, "%02x", ch);
	    strcat(lineBuf, spc);
	    strcat(lineBuf, pbuf);
	    spc = " ";
	    xbuf[n%16] = (octet) ch;
	    if ((n % 16) == 15) {
		strcat(lineBuf, "   ");
		for (k=0; k<16; k++) {
		    if ((xbuf[k] < ' ') || (xbuf[k] > '~') || (xbuf[k] == '<') || (xbuf[k] == '>'))
			strcat(lineBuf, ".");
		    else
			strncat(lineBuf, (char *)(&(xbuf[k])), 1);
		}
		edir_log_debug("%s", lineBuf);
		sprintf(lineBuf, "%08x:", n + 1);
	    }
	}
	spc="/";
    }
    j = n % 16;
    k = 16 - j;
    if (j) {
	for (i=0; i<k; i++) strcat(lineBuf, "   ");
	strcat(lineBuf, "   ");
	for (i=0; i<j; i++) {
	    if ((xbuf[i] < ' ') || (xbuf[i] > '~') || (xbuf[i] == '<') || (xbuf[i] == '>'))
		strcat(lineBuf, ".");
	    else
		strncat(lineBuf, (char *)(&(xbuf[i])), 1);
	}
    }
    edir_log_debug("%s", lineBuf);
}
#endif
