/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_state.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"

static void edir_internal_fatal(edir_connection_t *, edir_exception_t **);
static void edir_internal_close(edir_connection_t *, edir_exception_t **);
static void edir_internal_proto_exception(edir_connection_t *, edir_exception_t **);
static void edir_internal_badsync_exception(edir_connection_t *,edir_exception_t **);

/**
 * State machine driver function for the Enhydra Director Protocol Version 1.x
 * <P>
 * This is a simple state machine that loops until one of the state 
 * functions indicates that we should break out and allow the client
 * module code to do some processing.
 * <P>
 * This state machine has been designed for use in both blocking and
 * nonblocking/event-driven environments.  However, only blocking socket
 * environments have been tested so far, so it remains to be seen how
 * well this would work in an event driver environment.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_PROTO_AUTH
 * 		If authentication with the Enhydra server was not successful.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern void
edir_state_machine(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_machine";
    EDIR_IF_DEBUG( int prevstate; char *trace = NULL; )
    int keep_going = 1;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    while (keep_going) {
	EDIR_IF_DEBUG( prevstate = conn->state; trace = NULL; )
	EDIR_DEBUG_STATE_IN(("Entering state \"%s\"\n", \
			  edir_statename(conn->state)));
	switch (conn->state) {
	    case EDIR_STATE_CLOSED:
		keep_going = edir_state_closed(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_WAIT_CONN:
		keep_going = edir_state_wait_conn(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_WAIT_VERSION:
		keep_going = edir_state_wait_version(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_WAIT_CHALLENGE:
		keep_going = edir_state_wait_challenge(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_VERSION:
		keep_going = edir_state_send_version(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_AUTH:
		keep_going = edir_state_send_auth(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_WAIT_AUTH_REPLY:
		keep_going = edir_state_wait_auth_reply(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_OPENING:
		keep_going = edir_state_opening(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_OPEN:
		keep_going = edir_state_open(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_HEADERS:
		keep_going = edir_state_send_headers(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_HEADER:
		keep_going = edir_state_send_header(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_HEADERS_SENT:
		keep_going = edir_state_headers_sent(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_ENTITY:
		keep_going = edir_state_send_entity(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_DATA:
		keep_going = edir_state_send_data(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_RECEIVE_HEADERS:
		keep_going = edir_state_receive_headers(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_RECEIVE_RESPONSE:
		keep_going = edir_state_receive_response(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_RECEIVE_RESPONSE_PACKET:
		keep_going = edir_state_receive_response_packet(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_SEND_FUNC:
		keep_going = edir_state_send_func(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_WAIT_FUNC_RETURN:
		keep_going = edir_state_wait_func_return(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_PROCESS_CB:
		keep_going = edir_state_process_cb(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_FINISH_SEND:
		keep_going = edir_state_finish_send(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_CLOSE:
		keep_going = edir_state_close(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_CLOSING:
		keep_going = edir_state_closing(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    case EDIR_STATE_ABORT:
		keep_going = edir_state_abort(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	    default:
		keep_going = edir_state_bad(conn, exc);
		if (*exc) EDIR_THROW(*exc);
		break;
	}
	EDIR_DEBUG_STATE_OUT(("Leaving state  \"%s\" Next: \"%s\"\n", \
			edir_statename(prevstate), edir_statename(conn->state)));
	continue;
except:
	EDIR_DEBUG_NORMAL(("State machine caught exception. (%s)", \
		    keep_going ? "Continuing" : "Returning"));
	EDIR_DEBUG_VERBOSE(("%s", ((*exc) && \
			 (trace = edir_exception_stack_trace(*exc))) ? \
		         trace : "(No exception trace)"));
	EDIR_DEBUG_STATE_OUT(("Leaving state  \"%s\" Next: \"%s\"\n", \
			edir_statename(prevstate), edir_statename(conn->state)));
	if (!keep_going) {
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
	    return;
	}
    }
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
}

/**
 * State function for the CLOSED state.
 * <P>
 * This is the state of a connection for which the socket has not
 * yet been opened, or for which an existing connection has been
 * completely closed and reset.
 * <P>
 * Normal Transitions: WAIT_VERSION, WAIT_CONN
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_closed(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_closed";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    
    conn->needAuth = 0;

    if (conn->receiveData == NULL) {
	/* Doh! Receive buffer missing. Required in this state. */
	*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				   "Input buffer required");
	EDIR_THROW(*exc);
	/* NOTREACHED */
    }
    if (conn->receiveDataMax < EDIR_PROTO_MAX_CHALLENGE_LEN) {
	/* Doh! Receive buffer too small. */
	*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				   "Input buffer too small");
	EDIR_THROW(*exc);
	/* NOTREACHED */
    }
    if (conn->sendData == NULL) {
	/* Doh! Send buffer missing. Required in this state. */
	*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				   "Output buffer required");
	EDIR_THROW(*exc);
	/* NOTREACHED */
    }
    if (conn->sendDataMax < EDIR_PROTO_MAX_AUTH_REPLY_LEN) {
	/* Doh! Send buffer too small. */
	*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				   "Output buffer too small");
	EDIR_THROW(*exc);
	/* NOTREACHED */
    }

    /* Create the socket descriptor */
    edir_socket(conn, exc);
    if (*exc) EDIR_THROW(*exc);

    /* Bind to desired address */
    edir_socket_bind(conn, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * Connect to the Servlet engine.  In some implementations, this
     * could be a non-blocking call that completes in the
     * STATE_WAIT_CONN state.
     */
    edir_socket_connect(conn, exc);
    if (*exc) EDIR_THROW(*exc);

    conn->returnState = EDIR_STATE_WAIT_VERSION;

    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_WAIT_CONN;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_close(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the WAIT_CONN state.
 * <P>
 * This is the state of a connection that is waiting in nonblocking
 * mode for a connect() to complete to the Enhydra server.  This 
 * state will only be entered in a nonblocking environment in which
 * EDIR_IO_PENDING(conn) returns true.
 * <P>
 * Normal Transitions: WAIT_VERSION
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_wait_conn(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_wait_conn";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_socket_finish_connect(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }
    
    conn->state = EDIR_STATE_WAIT_VERSION;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_close(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the WAIT_VERSION state.
 * <P>
 * In this state, the connection is waiting for the version packet
 * [15,15] to be sent from the server.  The version packet is always
 * the first packet received by either side of the connection.  It
 * contains no data, and simply encodes the major and minor protocol
 * that its sender is using in the length field.
 * <P>
 * Normal Transitions: WAIT_CHALLENGE
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_wait_version(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_wait_version";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_proto_receive_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Breakout: Nonblocking I/O pending. */
    }

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_PROTO:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_PROTO_VERSION:
		    if (conn->serverMajorVersion != EDIR_PROTO_MAJOR_VERSION) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Incompatible protocol version %d.%d",
				conn->serverMajorVersion,
				conn->serverMinorVersion);
			EDIR_THROW(*exc);
		    }
		    /* Auth challenge or auth-OK immediately follows */
		    conn->state = EDIR_STATE_WAIT_CHALLENGE;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going. */
		default:
		    break;
	    }
	    break;
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the WAIT_CHALLENGE state.
 * <P>
 * In this state, the connection is waiting for either an
 * [AUTH,SVR] or [AUTH,OK] packet from the server.  If [AUTH,SVR], then
 * we must send a valid auth response to the server in order to establish
 * the connection.  If [AUTH,OK] then the server allows unauthenticated
 * connections and we go directly to an open connection.
 * <P>
 * However, we must send the our version to the server before any [AUTH,CLI]
 * or [HEADER] packets.  Therefore we take not of whether a reply is needed,
 * and then go to the SEND_VERSION state.
 * <P>
 * Normal Transitions: SEND_VERSION
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_PROTO_AUTH
 * 		If authentication with the Enhydra server was not successful.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_wait_challenge(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_wait_challenge";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_proto_receive_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Breakout: Nonblocking I/O pending. */
    }

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_AUTH:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_AUTH_SVR:
		    conn->state = EDIR_STATE_SEND_VERSION;
		    conn->needAuth = 1;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going. */
		case EDIR_PKT_SUBTYPE_AUTH_OK:
		    if (conn->receiveDataLen > 0) {
			conn->openFlags = conn->receiveData[0];
		    } else {
			conn->openFlags = 0;
		    }
		    conn->state = EDIR_STATE_SEND_VERSION;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going. */
		case EDIR_PKT_SUBTYPE_AUTH_FAIL:
		    *exc = edir_exception_create(EDIR_EXCEPT_PROTO_AUTH,
			    "Connection to server denied.");
		    EDIR_THROWTO(*exc, authexcept);
		    /* NOTREACHED */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
authexcept:
    edir_internal_close(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the SEND_VERSION state.
 * <P>
 * In this state we send the version packet to the server.  It gets
 * a separate state mainly for the benefit of nonblocking versions,
 * which may need to return to this state later after the send() of
 * the packet completes.
 * <P>
 * After the packet has been sent, we look at the "auth reqired" flag
 * we recorded in the WAIT_CHALLENGE state, and go to SEND_AUTH or
 * OPENING, as appropriate.
 * <P>
 * In nonblocking implementations, we may go to the FINISH_SEND state
 * to wait for the send() of the packet to complete.
 * <P>
 * Normal Transitions: SEND_AUTH, OPENING, FINISH_SEND
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_send_version(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_send_version";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_proto_send_version(conn, EDIR_PROTO_MAJOR_VERSION,
			  EDIR_PROTO_MINOR_VERSION, exc);
    if (*exc) EDIR_THROW(*exc);

    if (conn->needAuth)
	conn->returnState = EDIR_STATE_SEND_AUTH;
    else
	conn->returnState = EDIR_STATE_OPENING;

    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_FINISH_SEND;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the SEND_AUTH state.
 * <P>
 * We reach this state if we must send a valid authentication reply
 * to the server.  For the 1.x protocol, we take the authentication
 * challenge we received from the server, concatenate it with the
 * locally stored authentication key, MD5 hash the resultant data,
 * and send it back.  If the hashed data matches on the server, we
 * should get an [AUTH,OK] packet back in the WAIT_AUTH_REPLY state,
 * otherwise we get [AUTH,FAIL];
 * <P>
 * In nonblocking implementations, we may go to the FINISH_SEND state
 * to wait for the send() of the packet to complete.
 * <P>
 * Normal Transitions: WAIT_AUTH_REPLY, FINISH_SEND
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_PROTO_AUTH
 * 		If authentication with the Enhydra server was not successful.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_send_auth(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_send_auth";
    edir_MD5_CTX mdctx;
    int len;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    if (conn->authData == NULL) {
	/* Doh! No auth data -- we fail. */
	*exc = edir_exception_create(EDIR_EXCEPT_PROTO_AUTH,
				   "Auth secret required");
	EDIR_THROW(*exc);
	/* NOTREACHED */
    }

    /* Create the MD5 digester context.  */
    edir_MD5Init(&mdctx);

    /* Digest the challenge data */
    if (conn->receiveDataLen > 0) {
	edir_MD5Update(&mdctx, conn->receiveData, conn->receiveDataLen);
    }

    /* Digest the authentication secret.  */
    if ((len = strlen(conn->authData)) > 0) {
	edir_MD5Update(&mdctx, (octet *) conn->authData, len);
    }

    /* Finish the digest.  */
    edir_MD5Final(conn->sendData, &mdctx);

    conn->sendDataLen = EDIR_PROTO_MD5_HASH_SIZE;

    /* Send the auth repsonse packet.  This may be non-blocking. */
    edir_proto_send_packet(conn, EDIR_PKT_TYPE_AUTH, EDIR_PKT_SUBTYPE_AUTH_CLI, exc);
    if (*exc) EDIR_THROW(*exc);

    conn->returnState = EDIR_STATE_WAIT_AUTH_REPLY;

    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_FINISH_SEND;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the WAIT_AUTH_REPLY state.
 * <P>
 * In this state we are waiting for an authentication reply from the
 * server.  If we receive [AUTH,OK] then we continue to the OPENING
 * state.  If, however, we get [AUTH,FAIL], then an authentication
 * exception is raised, and we transition to the CLOSING state to
 * close the connection and free resources.  Eventually we return
 * to the client module and report the error.
 * <P>
 * Normal Transitions: OPENING, CLOSING
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_PROTO_AUTH
 * 		If authentication with the Enhydra server was not successful.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_wait_auth_reply(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_wait_auth_reply";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_proto_receive_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking I/O pending. */
    }

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_AUTH:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_AUTH_OK:
		    if (conn->receiveDataLen > 0) {
			conn->openFlags = conn->receiveData[0];
		    } else {
			conn->openFlags = 0;
		    }
		    conn->state = EDIR_STATE_OPENING;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Continue */
		case EDIR_PKT_SUBTYPE_AUTH_FAIL:
		    *exc = edir_exception_create(EDIR_EXCEPT_PROTO_AUTH,
			    "Authentication to server failed. (Incorrect authdata)");
		    EDIR_THROWTO(*exc, authexcept);
		    /* NOTREACHED */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
authexcept:
    edir_internal_close(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the OPENING state.
 * <P>
 * The OPENING state exists to allow optional callbacks from the server
 * immediately after authentication, but before handling a request.
 * Future versions of the protocol will allow a connection to be kept open
 * through more than one request, and this allows general per-connection
 * settings to be established.  For example, we could negotiate to not
 * send certain constant headers on each request.
 * <P>
 * If the server did not specify callbacks in the [AUTH,OK] packet, then
 * we just go straight to the OPEN state.  If callbacks are requested,
 * we go to the PROCESS_CB state.
 * <P>
 * Normal Transitions: OPEN, PROCESS_CB
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_opening(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_opening";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    EDIR_DEBUG_EVERYTHING(("openFlags = 0x%08x", conn->openFlags));

    /*
     * If server did not specify pre-open callbacks, then just
     * go to next state.  Otherwise wait for the callbacks.
     */
    if (!(conn->openFlags & EDIR_PKT_OPENFLAGS_OPENCB)) {
	EDIR_DEBUG_EVERYTHING(("Open callbacks NOT requested."));
	conn->state = EDIR_STATE_OPEN;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	return 1;
    }

    EDIR_DEBUG_EVERYTHING(("Open callbacks requested."));

    edir_proto_receive_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_FUNC:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_FUNC_CALL:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in pre-open %s",
				"callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 0;
		    conn->returnState = EDIR_STATE_OPENING;
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_ONEWAY:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in pre-open %s",
				"one way callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 1;
		    conn->returnState = EDIR_STATE_OPENING;
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_END:
		    conn->state = EDIR_STATE_OPEN;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* We're finally open and established! */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the OPEN state.
 * <P>
 * The OPEN state is the point at which <code>edir_client_connect()</code>
 * returns to the client module code.  This allows the client module
 * to gather request headers and data and prepare to begin a new request.
 * <P>
 * Normal Transitions: Always returns to client module code.
 * <BR>
 * Note that the client module causes a transition to SEND_HEADERS by
 * sending the first request header.  It could also cause a transition
 * to CLOSING or ABORTING by closing or aborting the connection via
 * the client API.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 */
extern int
edir_state_open(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_state_open("EDIR_PTR", "EDIR_PTR")", conn, *exc));
    EDIR_DEBUG_FUNCS_OUT(("func: edir_state_open() returns 0"));
    /* Nothing to do in this state. */
    return 0;
}

/**
 * State function for the SEND_HEADERS state.
 * <P>
 * In this state, the client module code controls the flow of 
 * execution, and is in the process of sending request headers to
 * the server via the client API.
 * <P>
 * Normal Transitions: Always returns to client module code.
 * <BR>
 * Note that the client module causes a transition to HEADERS_SENT by
 * sending the terminating request header.  It could also cause a transition
 * to CLOSING or ABORTING by closing or aborting the connection via
 * the client API.
 * <P>
 * Each time the client code sends a header, we do a transition to
 * SEND_HEADER to send the header to the server, then return to 
 * SEND_HEADERS until all headers have been sent.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 */
extern int
edir_state_send_headers(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_state_send_headers("EDIR_PTR", "EDIR_PTR")", \
		     conn, *exc));
    EDIR_DEBUG_FUNCS_OUT(("func: edir_state_send_headers() returns 0"));
    /* Nothing to do in this state. */
    return 0;
}

/**
 * State function for the SEND_HEADER state.
 * <P>
 * In this state we send a single request header to the server.
 * (XXX FIX  A performance optimization would be to gather the
 * headers and send them all at once to save trips into the kernel)
 * <P>
 * If this header is flagged as the terminating header, we transition
 * to HEADERS_SENT, otherwise we transition back to SEND_HEADERS to 
 * allow the client to send more headers.
 * <P>
 * In either case, in nonblocking mode we may return with a transition
 * to FINISH_SEND if we have to wait for the send() to complete.
 * <P>
 * Normal Transitions: SEND_HEADERS, HEADERS_SENT, FINISH_SEND
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_send_header(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_send_header";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    edir_proto_send_header_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);

    if (conn->isLast) 
	conn->returnState = EDIR_STATE_HEADERS_SENT;
    else
	conn->returnState = EDIR_STATE_SEND_HEADERS;

    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_FINISH_SEND;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the HEADERS_SENT state.
 * <P>
 * The HEADERS_SENT state is yet another "callback" state.  Here we allow
 * the server to make callbacks after receiving request headers but 
 * before receiving the request entity.
 * <P>
 * To enable these callbacks, the server must specify a desire to
 * make these callbacks in the [AUTH,OK] packet, and if this is done,
 * EVERY request must send an "end of callbacks" indication for the
 * ENTIRE lifetime of the connection.  The usefulness of having this
 * functionality is questionable, except in the case of checking 
 * validity of large form uploads before beginning receipt of entity
 * data.  Therefore, this functionality is off by default.  It is
 * not currently used anywhere.
 * <P>
 * Enabling these callbacks will result in the performance hit of 
 * having to do a full two way synchronous handshake between the
 * server and client.
 * <P>
 * If no callbacks are expected, we go to SEND_ENTITY, otherwise we
 * go to PROCESS_CB
 * <P>
 * Normal Transitions: SEND_ENTITY, PROCESS_CB
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_headers_sent(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_headers_sent";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    /*
     * If server did not specify post-header callbacks, then just
     * go to next state.  Otherwise wait for the callbacks.
     */
    if (!(conn->openFlags & EDIR_PKT_OPENFLAGS_OPENCB)) {
	conn->state = EDIR_STATE_SEND_ENTITY;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	return 1;
    }

    edir_proto_receive_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_FUNC:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_FUNC_CALL:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in post-header %s",
				"callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 0;
		    conn->returnState = EDIR_STATE_HEADERS_SENT;
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_ONEWAY:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in post-header %s",
				"one way callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 1;
		    conn->returnState = EDIR_STATE_HEADERS_SENT;
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_END:
		    conn->state = EDIR_STATE_SEND_ENTITY;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the SEND_ENTITY state.
 * <P>
 * In this state, execution is taking place in client module code, which
 * is receiving request entity data from the browser client and is
 * the repeatedly sending chunks of data to the Enhydra server via the
 * client API.
 * <P>
 * The client module code will cause repeated transitions to the SEND_DATA
 * state as it sends blocks of data via the client API.  The SEND_DATA
 * state will transition to the RECEIVE_HEADERS state when the last block 
 * is sent.
 * <P>
 * Normal Transitions: SEND_DATA, RECEIVE_HEADERS
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_send_entity(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_state_send_entity("EDIR_PTR", "EDIR_PTR")",conn,*exc));
    EDIR_DEBUG_FUNCS_OUT(("func: edir_state_send_entity() returns 0"));
    /* Nothing to do here.  Waiting for module to call us. */
    return 0;
}

/**
 * State function for the SEND_DATA state.
 * <P>
 * In this state, we are sending a block of data to the server.
 * If this is the last block, we then transition to the RECEIVE_HEADERS
 * state.  Otherwise we transition back to SEND_ENTITY.
 * <P>
 * In nonblocking implementations, we could also transition to FINISH_SEND
 * while the client module waits for the send() to complete.
 * <P>
 * Normal Transitions: SEND_ENTITY, RECEIVE_HEADERS
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_send_data(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_send_data";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    
    if (conn->isLast)
	edir_proto_send_packet(conn, EDIR_PKT_TYPE_REQ,
			    EDIR_PKT_SUBTYPE_REQ_LASTDATA, exc);
    else
	edir_proto_send_packet(conn, EDIR_PKT_TYPE_REQ,
			    EDIR_PKT_SUBTYPE_REQ_DATA, exc);
    if (*exc) EDIR_THROW(*exc);

    if (conn->isLast)
	conn->returnState = EDIR_STATE_RECEIVE_HEADERS;
    else
	conn->returnState = EDIR_STATE_SEND_ENTITY;

    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_FINISH_SEND;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the RECEIVE_HEADERS state.
 * <P>
 * In this state, we are receiving response headers from the server.
 * The client module code calls the client API to receive each header,
 * and we go into this state to wait for the header to come from
 * the server.  When the last header is received, we transition
 * to RECEIVE_RESPONSE.  If a normal callback is requested by the
 * server, we go to PROCESS_CB.
 * <P>
 * XXX FIXME Performance can probably be improved by using a buffered
 * socket for the socket implementation.  This would reduce trips
 * into the kernel to receive the headers.
 * <P>
 * Normal Transitions: RECEIVE_HEADERS, RECEIVE_RESPONSE, PROCESS_CB
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_receive_headers(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_receive_headers";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    conn->headerName = conn->headerValue = NULL;
    conn->headerNameLen = conn->headerValueLen = 0;

    /* edir_proto_receive_packet() has the brains to decode headers. */
    edir_proto_receive_packet(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_HEADER:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_HEADER_HEADER:
		case EDIR_PKT_SUBTYPE_HEADER_COOKIE:
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
		    return 0; /* Break out to allow header to be processed */
		case EDIR_PKT_SUBTYPE_HEADER_LAST:
		    conn->state = EDIR_STATE_RECEIVE_RESPONSE;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
		    return 0; /* Break out to process last header */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_FUNC:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_FUNC_CALL:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in %s",
				"callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 0;
		    conn->returnState = conn->state; /* Return to this state*/
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_ONEWAY:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in %s",
				"one way callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 1;
		    conn->returnState = conn->state; /* Return to this state*/
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the RECEIVE_RESPONSE state.
 * <P>
 * At this point, all response headers have been read and processed
 * by the client module code and we are receiving response data blocks.
 * <P>
 * The edir_proto_receive_packet() call normally receives the entire
 * packet, EXCEPT in the case of a response data packet.  In that
 * case it only receives the header.  When we get data packets,
 * we go into the RECEIVE_RESPONSE_PACKET state to allow the client
 * module to obtain the incoming data and send it to the client
 * browser.
 * <P>
 * Upon completion of processing of the last data packet, we transition
 * back to the OPEN state because the request is now complete.
 * <P>
 * The server can make callbacks at any time while sending response data,
 * in which case we transition to PROCESS_CB to handle callbacks.
 * <P>
 * Normal Transitions: RECEIVE_RESPONSE_PACKET, OPEN, PROCESS_CB
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_receive_response(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_receive_response";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    conn->headerName = conn->headerValue = NULL;
    conn->headerNameLen = conn->headerValueLen = 0;

    /* edir_proto_receive_packet only receives the _header_ of a data frag. */
    edir_proto_receive_packet(conn, exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }
    if (*exc) EDIR_THROW(*exc);

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_RESP:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_RESP_DATA:
		    conn->state = EDIR_STATE_RECEIVE_RESPONSE_PACKET;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Break out to allow header to be processed */
		case EDIR_PKT_SUBTYPE_RESP_END:
		    conn->state = EDIR_STATE_OPEN;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Done with this request. */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_FUNC:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_FUNC_CALL:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in %s",
				"callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 0;
		    conn->returnState = conn->state; /* Return to this state*/
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_ONEWAY:
		    if (conn->receiveDataLen < 1) {
			*exc = edir_exception_create(EDIR_EXCEPT_PROTO,
				"Function type missing in %s",
				"one way callback request.");
			EDIR_THROW(*exc);
			/* NOTREACHED */
		    }
		    conn->funcType = conn->receiveData[0];
		    conn->funcStatus = 0;
		    conn->funcOneWay = 1;
		    conn->returnState = conn->state; /* Return to this state*/
		    conn->state = EDIR_STATE_PROCESS_CB;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the RECEIVE_RESPONSE_PACKET state.
 * <P>
 * We are receiving a response data packet, which may be quite large.
 * In this state we repeatedly return to the client module to allow
 * the client module to get the incoming data via the client API.
 * <P>
 * At the end of the packet, we go back to the RECEIVE_RESPONSE state.
 * <P>
 * Normal Transitions: RECEIVE_RESPONSE
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_receive_response_packet(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_receive_response_packet";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_proto_receive_packet_data(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }

    if (conn->packetRemainder < 1) {
	conn->state = EDIR_STATE_RECEIVE_RESPONSE;
    }
    /* Return any data received immediately. */
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
    return 0; /* Break out. */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the SEND_FUNC state.
 * <P>
 * If the client module code requests a client-side function call to
 * the server, we go to this state to send the function call packet.
 * Once the packet is sent, we go to the WAIT_FUNC_RETURN state (if 
 * not a one-way function) to wait for the function results to come back.
 * <P>
 * Normal Transitions: WAIT_FUNC_RETURN
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_send_func(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_send_func";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    
    if (EDIR_IO_PENDING(conn)) {
	edir_proto_finish_send(conn, exc);
	if (*exc) EDIR_THROW(*exc);
	if (EDIR_IO_PENDING(conn)) {
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	    return 0; /* Must wait yet again. */
	}
    }

    conn->state = EDIR_STATE_WAIT_FUNC_RETURN;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the WAIT_FUNC_RETURN state.
 * <P>
 * In this state we are waiting for a function call to return from the
 * server.  This is a generic state that can be called from one of
 * several other states in which client-side functions are allowable.
 * When we get the function results we return them to the client and
 * transition back to the stored return state.
 * <P>
 * Normal Transitions: State immediately prior to this function request.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_wait_func_return(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_wait_func_return";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    edir_proto_receive_packet(conn, exc);
    if (EDIR_IO_PENDING(conn)) return 0; /* Break out nonblocking io */
    if (*exc) EDIR_THROW(*exc);

    conn->funcStatus = 0;

    switch (conn->packetType) {
	case EDIR_PKT_TYPE_FUNC:
	    switch (conn->packetSubType) {
		case EDIR_PKT_SUBTYPE_FUNC_OK:
		    conn->funcStatus = 1;
		    conn->state = conn->returnState;
		    conn->returnState = -1;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		case EDIR_PKT_SUBTYPE_FUNC_FAIL:
		    conn->funcStatus = 0;
		    conn->state = conn->returnState;
		    conn->returnState = -1;
		    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
		    return 1; /* Keep going */
		default:
		    break;
	    }
	    break;
	case EDIR_PKT_TYPE_PROTO:
	    edir_internal_proto_exception(conn, exc);
	    edir_internal_close(conn, exc);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	    return 1; /* Continue closing */
	default:
	    break;
    }
    /* Illegal packet! */
    edir_internal_badsync_exception(conn, exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the generic PROCESS_CB state.
 * <P>
 * This is a generic state for processing callbacks.  It can be entered
 * from several other states, and always returns to the state from
 * which it entered.
 * <P>
 * This function determines whether a requested function ID is known 
 * and enabled in this implementation, and if so, runs the function with
 * the server-supplied byte array as data.  If the function is not known
 * or enabled then return an error packet to the server stating so.
 * <P>
 * For one-way functions, NEVER send a response, not even an ERROR
 * response, because the sender of a one-way function does not block
 * for the results.
 * <P>
 * Normal Transitions: FINISH_SEND, The state we were in before this one.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_process_cb(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_process_cb";
    octet subtype;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    /* We allow callbacks in non-blocking mode. */
    edir_process_callback(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }

    /*
     * If this was a one-way function, Ignore return value
     * and return without sending reply
     */
    if (conn->funcOneWay) {
	conn->state = conn->returnState;
	conn->returnState = -1;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
	return 1; /* Keep going. */
    }

    /* Send reply */
    if (conn->funcStatus)
	subtype = EDIR_PKT_SUBTYPE_FUNC_OK;
    else
	subtype = EDIR_PKT_SUBTYPE_FUNC_FAIL;
    edir_proto_send_packet(conn, EDIR_PKT_TYPE_FUNC, subtype, exc);
    if (*exc) EDIR_THROW(*exc);

    /* If I/O is pending in nonblocking mode, must wait for completion. */
    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_FINISH_SEND;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    /* If not I/O pending, just return to state machine. */
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the generic FINISH_SEND state.
 * <P>
 * This state exists for nonblocking socket implementations that need
 * to make repeated event-driven calls to reliably send a block of data.
 * We split this waiting out into its own state so that it is easy to
 * return to the same place when, for instance, select() reports that
 * our socket is again writeable without blocking.
 * <P>
 * This state normally transitions back to the state from which it was
 * entered, unless a fatal error occurs.
 * <P>
 * Normal Transitions: The state we were in before this one.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_finish_send(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_finish_send";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    
    if (EDIR_IO_PENDING(conn)) {
	edir_proto_finish_send(conn, exc);
	if (*exc) EDIR_THROW(*exc);
	if (EDIR_IO_PENDING(conn)) {
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	    return 0; /* Must wait yet again. */
	}
    }

    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * State function for the CLOSE state.
 * <P>
 * This is the initial connection closing state for a "graceful" close.
 * In previous revisions of this function it sent [PROTO,CLOSE] to the
 * server.  This is completely redundant because closing the socket
 * will be detected and interpreted as exactly the same thing.
 * <P>
 * Normal Transitions: CLOSING
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 */
extern int
edir_state_close(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_close";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    conn->state = EDIR_STATE_CLOSING;

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */
}

/**
 * State function for the ABORT state.
 * <P>
 * This is the initial connection closing state for a "nasty" close.
 * This happens when something fatal happens and we want to try an
 * emergency shutdown of the socket, and send some error text to the
 * server for server-side logging (if possible).  We do this with
 * the [PROTO,ABORT] packet, followed by a transition to CLOSING.
 * <P>
 * Normal Transitions: CLOSING, FINISH_SEND
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_abort(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_abort";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));

    /* If we have a send buffer, try to send a message and trace. */
    if ((conn->sendData) && (conn->sendDataMax > 0)) {
	char *msg = NULL;
	char *trace = NULL;
	if (conn->fatalException) {
	    msg = conn->fatalException->message;
	    trace = edir_exception_stack_trace(conn->fatalException);
	}
	edir_snprintf((char *) conn->sendData, conn->sendDataMax,
		    "Client Exception: %s\n%s%s",
		    (msg ? msg : "Unknown error"),
		    (trace ? msg : ""), (trace ? "\n" : ""));
	if (trace) edir_free(trace);
	trace = NULL;
	conn->sendData[conn->sendDataMax - 1] = '\0';
	conn->sendDataLen = strlen((char *) conn->sendData);
    } else conn->sendDataLen = 0;

    /* Send a PROTO_FATAL request to the server. */
    edir_proto_send_packet(conn, EDIR_PKT_TYPE_PROTO,
			 EDIR_PKT_SUBTYPE_PROTO_FATAL, exc);
    if (*exc) EDIR_THROW(*exc);

    conn->returnState = EDIR_STATE_CLOSING;

    if (EDIR_IO_PENDING(conn)) {
	conn->state = EDIR_STATE_FINISH_SEND;
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Nonblocking: Waiting for send */
    }
    conn->state = conn->returnState;
    conn->returnState = -1;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */

except:
    /* Oh well.  If we can't send a PROTO message, just close socket. */
    conn->state = EDIR_STATE_CLOSING;
    if (*exc) edir_exception_free(exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Keep going */
}

/**
 * State function for the CLOSING state.
 * <P>
 * This state handles the act of closing the socket to the server.
 * It is separated into its own state to allow nonblocking closure
 * of a socket to guarantee completion before going into the CLOSED state.
 * <P>
 * Normal Transitions: CLOSED
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_closing(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_closing";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    /* Close the socket. */
    edir_socket_close(conn, exc);
    if (*exc) EDIR_THROW(*exc);
    if (EDIR_IO_PENDING(conn)) {
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
	return 0; /* Break out: Nonblocking I/O pending. */
    }
except:
    /* Clear all state information, ignore exceptions */
    if (*exc) edir_exception_free(exc);
    /* Restore any saved fatal exception */
    *exc = conn->fatalException;
    if (conn->sbuf != NULL) {
	edir_sbuf_free(conn->sbuf);
	conn->sbuf = NULL;
    }
    /* Clear the connection. */
    memset(conn, 0, sizeof(edir_connection_t));
    /* Go back to closed state. */
    conn->state = EDIR_STATE_CLOSED;
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 0", funcName));
    return 0; /* Break out.  Socket now closed. */
}

/**
 * Pseudo-state function for unknown state numbers.
 * <P>
 * Throws a bad state connection and causes the state machine to
 * abort the connection.
 *
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
extern int
edir_state_bad(edir_connection_t *conn, edir_exception_t **exc)
{
    static char *funcName = "edir_state_bad";
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, conn, *exc));
    *exc = edir_exception_create(EDIR_EXCEPT_PROTO,
	    "Unknown state: %d", conn->state);
    EDIR_THROW(*exc);
except:
    edir_internal_fatal(conn, exc);
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns 1", funcName));
    return 1; /* Continue closing */
}

/**
 * Convenience function to cause an abort to occur if a fatal error occurs.
 * <P>
 * Stores away the exception to allow an ABORT of the connection to take
 * place.  Any exeptions raised by the ABORT are later ignored and the
 * original saved exception is reported to the client module code.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
static void
edir_internal_fatal(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_internal_fatal("EDIR_PTR", "EDIR_PTR")", conn, *exc));
    conn->fatalException = *exc;
    conn->state = EDIR_STATE_ABORT;
    *exc = NULL;
    EDIR_DEBUG_FUNCS_OUT(("func: edir_internal_fatal() returns void"));
}

/**
 * Convenience function to cause an abort to occur if a fatal error occurs.
 * <P>
 * Similar to edir_internal_fatal, but is called in cases where we know
 * the server is not connected yet.  We just do a close in this case.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 * @exception EDIR_EXCEPT_IO_ERROR
 * 		If a socket-level or system-level I/O error is detected while
 * 		attempting to communicate with the Enhydra server.
 */
static void
edir_internal_close(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_internal_close("EDIR_PTR", "EDIR_PTR")", conn, *exc));
    conn->fatalException = *exc;
    conn->state = EDIR_STATE_CLOSING;
    *exc = NULL;
    EDIR_DEBUG_FUNCS_OUT(("func: edir_internal_close() returns void"));
}

/**
 * Generic convenience function to handle incoming fatal protocol 
 * error reports from the server.
 * <P>
 * Generates an EDIR_EXCEPT_PROTO exception to report the server 
 * error.  This exception is then raised by the calling state function.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 */ 
static void
edir_internal_proto_exception(edir_connection_t *conn, edir_exception_t **exc)
{
    char *msg = NULL;
    EDIR_DEBUG_FUNCS_IN(("func: edir_internal_proto_exception("EDIR_PTR", "EDIR_PTR")", \
		     conn, *exc));
    if (conn->receiveData && (conn->receiveDataLen > 0)) {
	msg = edir_malloc(conn->receiveDataLen + 1);
	memcpy(msg, conn->receiveData, conn->receiveDataLen);
	msg[conn->receiveDataLen] = '\0';
    } else {
	msg = edir_strdup("Unknown server error.");
    }
    *exc = edir_exception_create(EDIR_EXCEPT_PROTO,
			      "Server Exception: %s", msg);
    edir_free(msg);
    EDIR_DEBUG_FUNCS_OUT(("func: edir_internal_proto_exception() returns void"));
}

/**
 * Generic convenience function to handle protocol state errors.
 * <P>
 * Generates an EDIR_EXCEPT_PROTO exception to report that the protocol
 * state for this connection is out of sync with the server and/or reality.
 * The calling state function will then raise this exception and send an
 * ABORT to the server if possible.
 * 
 * @param conn [in,out] Connection handle to maintain connection state.
 * @param exc  [out]    Exception handle filled in if an exception is raised.
 * @return Integer value.  If zero, state machine driver returns to client
 * 	module.  If non-zero, state machine driver keeps processing.
 * @exception EDIR_EXCEPT_PROTO
 * 		If an error related to the Enhydra Director protocol occurs.
 */ 
static void
edir_internal_badsync_exception(edir_connection_t *conn, edir_exception_t **exc)
{
    EDIR_DEBUG_FUNCS_IN(("func: edir_internal_badsync_exception("EDIR_PTR", "EDIR_PTR")", \
		     conn, *exc));
    *exc = edir_exception_create(EDIR_EXCEPT_PROTO,
	    "Connection out of sync: Unexpected packet [%d,%d] in state %s.",
	    (int)(conn->packetType), (int)(conn->packetSubType),
	    edir_statename(conn->state));
    EDIR_DEBUG_FUNCS_OUT(("func: edir_internal_badsync_exception() returns void"));
}

/**
 * Returns a text name for the specified state.
 * <P>
 * This character pointer points to a statically allocated string that
 * should not be modified or freed.
 * 
 * @param state  [in] The state number to convert.
 * @return A pointer to a static character string containg the state's name.
 */
extern char *
edir_statename(int state)
{
    switch (state) {
	case EDIR_STATE_CLOSED:
	    return "CLOSED";
	    break;
	case EDIR_STATE_WAIT_CONN:
	    return "WAIT_CONN";
	    break;
	case EDIR_STATE_WAIT_VERSION:
	    return "WAIT_VERSION";
	    break;
	case EDIR_STATE_WAIT_CHALLENGE:
	    return "WAIT_CHALLENGE";
	    break;
	case EDIR_STATE_SEND_VERSION:
	    return "SEND_VERSION";
	    break;
	case EDIR_STATE_SEND_AUTH:
	    return "SEND_AUTH";
	    break;
	case EDIR_STATE_WAIT_AUTH_REPLY:
	    return "WAIT_AUTH_REPLY";
	    break;
	case EDIR_STATE_OPENING:
	    return "OPENING";
	    break;
	case EDIR_STATE_OPEN:
	    return "OPEN";
	    break;
	case EDIR_STATE_SEND_HEADERS:
	    return "SEND_HEADERS";
	    break;
	case EDIR_STATE_SEND_HEADER:
	    return "SEND_HEADER";
	    break;
	case EDIR_STATE_HEADERS_SENT:
	    return "HEADERS_SENT";
	    break;
	case EDIR_STATE_SEND_ENTITY:
	    return "SEND_ENTITY";
	    break;
	case EDIR_STATE_SEND_DATA:
	    return "SEND_DATA";
	    break;
	case EDIR_STATE_RECEIVE_HEADERS:
	    return "RECEIVE_HEADERS";
	    break;
	case EDIR_STATE_RECEIVE_RESPONSE:
	    return "RECEIVE_RESPONSE";
	    break;
	case EDIR_STATE_RECEIVE_RESPONSE_PACKET:
	    return "RECEIVE_RESPONSE_PACKET";
	    break;
	case EDIR_STATE_SEND_FUNC:
	    return "SEND_FUNC";
	    break;
	case EDIR_STATE_WAIT_FUNC_RETURN:
	    return "WAIT_FUNC_RETURN";
	    break;
	case EDIR_STATE_PROCESS_CB:
	    return "PROCESS_CB";
	    break;
	case EDIR_STATE_FINISH_SEND:
	    return "FINISH_SEND";
	    break;
	case EDIR_STATE_CLOSE:
	    return "CLOSE";
	    break;
	case EDIR_STATE_CLOSING:
	    return "CLOSING";
	    break;
	case EDIR_STATE_ABORT:
	    return "ABORT";
	    break;
	default:
	    return "Unknown";
	    break;
    }
}

