/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_state.h,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#ifndef __EDIR_CONNECTION_H__
#define __EDIR_CONNECTION_H__ 1
#ifdef __cplusplus
extern "C" {
#endif

#define EDIR_STATE_CLOSED				0
#define EDIR_STATE_WAIT_CONN			1
#define EDIR_STATE_WAIT_VERSION			2
#define EDIR_STATE_WAIT_CHALLENGE			3
#define EDIR_STATE_SEND_VERSION			4
#define EDIR_STATE_SEND_AUTH			5
#define EDIR_STATE_WAIT_AUTH_REPLY		6
#define EDIR_STATE_OPENING			7
#define EDIR_STATE_OPEN				8
#define EDIR_STATE_SEND_HEADERS			9
#define EDIR_STATE_SEND_HEADER			10
#define EDIR_STATE_HEADERS_SENT			11
#define EDIR_STATE_SEND_ENTITY			12
#define EDIR_STATE_SEND_DATA			13
#define EDIR_STATE_RECEIVE_HEADERS		14
#define EDIR_STATE_RECEIVE_RESPONSE		15
#define EDIR_STATE_RECEIVE_RESPONSE_PACKET	16
#define EDIR_STATE_SEND_FUNC			17
#define EDIR_STATE_WAIT_FUNC_RETURN		18
#define EDIR_STATE_PROCESS_CB			19
#define EDIR_STATE_FINISH_SEND			20
#define EDIR_STATE_CLOSE				253
#define EDIR_STATE_ABORT				254
#define EDIR_STATE_CLOSING			255

#define EDIR_STATE_MAX_HOSTNAME_LEN		256

/**
 * Per-connection state for a connection to a Servlet Engine.
 */
typedef struct edir_connection {
    /**
     * Optional module-specific data pointer for per-connection state
     * maintained by module-specific code.
     */
    void		*moduleData;

    /**
     * Major version of server side protocol engine.
     */
    int			serverMajorVersion;

    /**
     * Minor version of server side protocol engine.
     */
    int			serverMinorVersion;

    /*===================================================================
     * General connection state.
     * The socket, host name, port, flags, etc.
     *===================================================================*/
    /**
     * The actual socket connection to the server.
     */
    EDIR_SOCKET		socket;

    /**
     * The current state of this connection's state machine.
     */
    octet		state;

    /**
     * State to transition to when finished with a "common" state
     * function.  This allows elimination of much duplicate code.
     */
    octet		returnState;

    octet		ioPending;

    /**
     * Local internet address and port to bind to.
     */
    struct sockaddr_in	localAddr;

    /**
     * Server internet address to connect to.
     */
    struct sockaddr_in	serverAddr;

    /**
     * Options requested by the server in the AUTH_OK packet.
     */
    octet		openFlags;

    /**
     * Decoded packet type of the last received packet.
     */
    octet		packetType;

    /**
     * Decoded packet subtype of the last received packet.
     */
    octet		packetSubType;

    /**
     * Length of the last received packet's data section.
     */
    int32		packetLength;

    /**
     * Number of packet bytes that have not yet been received from
     * the socket connection.
     */
    int32		packetRemainder;

    /**
     * Temporary buffer to hold an incoming packet protocol header.
     * This requires 4 octets.
     */
    octet		headerBuf[4];

    /*===================================================================
     * Bookkeeping fields for edir_proto_receive_packet, and the edir_socket
     * receive functions.
     *===================================================================*/
    octet		haveHeader;
    octet		haveData;
    int32		dataActual;
    int32		dataExpected;

    /**
     * Whether this is the last function, fragment, or header.
     */
    octet		isLast;

    /*===================================================================
     * Authentication Data Fields.
     * This field is always freed once authentication is done.
     *===================================================================*/
    /**
     * Authentication secret data from the client configuration.
     */
    char		authData[EDIR_AUTH_MAX+1];

    /**
     * Whether we need to send an auth reply
     */
    int			needAuth;

    /*===================================================================
     * Incoming packet buffer fields.
     * This points to a buffer for incoming packet data.
     *===================================================================*/
    /**
     * Buffer to receive incoming packet data.
     */
    octet		*receiveData;

    /**
     * Buffer to receive incoming packet data.
     */
    octet		*receivePtr;

    /**
     * Current number of packet data bytes in "packetData".
     */
    int32		receiveDataLen;

    /**
     * Maximum number of data bytes in "packetData".
     */
    int32		receiveDataMax;

    /*===================================================================
     * Outgoing packet buffer fields.
     * This points to outgoing packet data.
     *===================================================================*/
    /**
     * Buffer containing outgoing packet data.
     */
    octet		*sendData;

    /**
     * Buffer containing outgoing packet data.
     */
    octet		*sendPtr;

    /**
     * Current number of packet data bytes in "sendData".
     */
    int32		sendDataLen;

    /**
     * Maximum number of packet data bytes in "sendData".
     */
    int32		sendDataMax;

    /**
     * Scatter-gather buffer handle allocated upon connection.
     * Used at various points to interface with edir_socket_write_sbuf().
     */
    void		*sbuf;

    /*===================================================================
     * Callback and function call fields.
     * Function argument data is specified by "sendData", return data
     * comes in as "receiveData".
     *===================================================================*/
    octet		funcStatus;

    octet		funcType;

    octet		funcOneWay;

    /*===================================================================
     * Header name and value fields.
     * These fields hold the name and value of the last HTTP header
     * received from the server.
     * They are also used to pass outgoing header names and values
     * to the state machine to send them to the server.
     *===================================================================*/
    octet		*headerName;
    int32		headerNameLen;

    octet		*headerValue;
    int32		headerValueLen;

    edir_exception_t	*fatalException;
} edir_connection_t;

extern void
edir_state_machine(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_closed(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_wait_conn(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_wait_version(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_wait_challenge(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_version(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_auth(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_wait_auth_reply(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_opening(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_open(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_headers(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_header(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_headers_sent(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_entity(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_data(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_receive_headers(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_receive_response(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_receive_response_packet(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_send_func(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_wait_func_return(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_process_cb(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_finish_send(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_close(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_abort(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_closing(edir_connection_t *conn, edir_exception_t **exc);

extern int
edir_state_bad(edir_connection_t *conn, edir_exception_t **exc);

extern char *
edir_statename(int state);

extern void
edir_process_callback(edir_connection_t *conn, edir_exception_t **exc);

#ifdef __cplusplus
} /* Extern "C" */
#endif
#endif /* __EDIR_CONNECTION_H__ */
