/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_string.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"

/**
 * Replacement for <code>strdup()</code> that uses <code>edir_malloc()</code>.
 * <P>
 * The returned string is an identical copy of <code>s</code> that should
 * be freed using <code>edir_free()</code>.
 * 
 * @param s [in] The string to duplicate.
 * @return A newly allocated character array containing the copied string.
 */
extern char *
edir_strdup(const char *s)
{
    char *cp;
    cp = edir_malloc(strlen((char *)s)+1);
    strcpy(cp, (char *)s);
    return cp;
}

/**
 * Format a string into a newly allocated character array using
 * a "stdarg.h" <code>va_list</code> handle for the arguments.
 * <P>
 * Formats a string in the style of <code>sprintf()</code> and
 * places the resulting string into a newly allocated character array
 * that should be freed using <code>edir_free()</code>.
 * 
 * @param fmt [in]  Format string as used with <code>printf()</code>.
 * @param ap  [in]  Argument list handle ala "stdarg.h".
 * @return A newly allocated character array containing the formatted string.
 */
extern char *
edir_vasprintf(const char *fmt, va_list ap)
{
    char buf[512];
    edir_vsnprintf(buf, 512, (char *)fmt, ap);
    buf[511] = '\0';
    return edir_strdup(buf);
}

/**
 * Format a string into a newly allocated character array.
 * <P>
 * Formats a string in the style of <code>sprintf()</code> and
 * places the resulting string into a newly allocated character array
 * that should be freed using <code>edir_free()</code>.
 * 
 * @param fmt [in]  Format string as used with <code>printf()</code>.
 * @param ... [in]  Variadic argument list as used with <code>printf()</code>.
 * @return A newly allocated character array containing the formatted string.
 */
extern char *
edir_asprintf(const char *fmt, ...)
{
    va_list ap;
    char *ret = NULL;
    va_start(ap, fmt);
    ret = edir_vasprintf((char *)fmt, ap);
    va_end(ap);
    return ret;
}

/**
 * Compares two strings without sensitivity to case.
 * <P>
 * If both strings match in a case-insensitive manner, then
 * <code>0</code> is returned to indicate a match.
 * <P>
 * If <code>s1</code> is greater than s2, then <code>1</code>
 * is returned.
 * <P>
 * If <code>s1</code> is lexically less than s2 then <code>-1</code>
 * is returned.
 * 
 * @param s1  [in] The first string to match.
 * @param s2  [in] The second string to match.
 * @return Integer value of <code>-1</code>, <code>0</code>, or <code>1</code>
 * 	if <code>(s1 < s2)</code>, <code>(s1 == s2)</code>, or 
 * 	<code>(s1 > s2)</code>, respectively.
 */
extern int
edir_strcasecmp(const char *s1, const char *s2)
{
#ifdef HAVE_STRCASECMP
    return strcasecmp((char *)s1, (char *)s2);
#elif HAVE_STRCMPI
    return strcmpi((char *)s1, (char *)s2);
#elif HAVE_STRICMP
    return stricmp((char *)s1, (char *)s2);
#else /* Must implement */
    unsigned char c1, c2;
    while (*s1 && *s2) {
	c1 = (unsigned char) tolower(*s1);
	c2 = (unsigned char) tolower(*s2);
	if (c1 > c2) return 1;
	if (c1 < c2) return -1;
	s1++;
	s2++;
    }
    if (*s1) return 1; /* s1 greater, contains s2 plus suffix */
    if (*s2) return -1; /* s2 greater, contains s1 plus suffix */
    return 0; /* Strings match. */
#endif
}

/**
 * Compares a fixed-length prefix of two strings without sensitivity to case.
 * <P>
 * If the first <code>len</code> characters of both strings
 * match in a case-insensitive manner, then <code>0</code> is
 * returned to indicate a match.
 * <P>
 * If <code>s1</code> is greater than s2, then <code>1</code>
 * is returned.
 * <P>
 * If <code>s1</code> is lexically less than s2 then <code>-1</code>
 * is returned.
 * 
 * @param s1  [in] The first string to match.
 * @param s2  [in] The second string to match.
 * @param len [in] Maximum number of characters to compare.
 * @return Integer value of <code>-1</code>, <code>0</code>, or <code>1</code>
 * 	if <code>(s1 < s2)</code>, <code>(s1 == s2)</code>, or 
 * 	<code>(s1 > s2)</code>, respectively.
 */
extern int
edir_strncasecmp(const char *s1, const char *s2, int32 len)
{
#ifdef WIN32
    return _strnicmp(s1, s2, len);
#elif HAVE_STRNCASECMP
    return strncasecmp((char *)s1, (char *)s2, len);
#elif HAVE_STRNCMPI
    return strncmpi((char *)s1, (char *)s2, len);
#elif HAVE_STRINCMP
    return strincmp((char *)s1, (char *)s2, len);
#else /* Must implement */
    unsigned char c1, c2;
    while ((len > 0) && *s1 && *s2) {
	c1 = (unsigned char) tolower(*s1);
	c2 = (unsigned char) tolower(*s2);
	if (c1 > c2) return 1;
	if (c1 < c2) return -1;
	s1++;
	s2++;
	len--;
    }
    if (len <= 0) return 0; /* First <len> chars match. */
    if (*s1) return 1; /* s1 greater, contains s2 plus suffix */
    if (*s2) return -1; /* s2 greater, contains s1 plus suffix */
    return 0; /* Whole strings match. */
#endif
}

/**
 * Finds a substring within a larger string using case insensitive matching.
 * <P>
 * Returns the first substring of <code>s</code> that matches <code>find</code>
 * in a case-insensitive fashion, or <code>NULL</code> if no match is found.
 * <P>
 * Note that the returned character pointer does NOT point to a newly
 * allocated string, so it should NOT be passed to <code>free()</code>.
 * 
 * @param s    [in] The string in which to match.
 * @param find [in] The substring to search for.
 * @return A pointer to the position within <code>s</code> where <code>find
 * 	</code> was found, or <code>NULL</code> if no match.
 */
extern char *
edir_strcasestr(register const char *s, register const char *find)
{
    register char c, sc;
    register int32 len;
    if ((c = tolower(*find++)) != 0) {
	len = strlen(find);
	do {
	    do {
		if ((sc = tolower(*s++)) == 0)
		    return (NULL);
	    } while (sc != c);
	} while (edir_strncasecmp(s, find, len) != 0);
	s--;
    }
    return ((char *)s);
}

/**
 * Remove leading and trailing whitespace characters.
 * 
 * @param s [in] The string to trim.
 * @return A newly allocated character array containing the trimmed string.
 * 	   This array should be freed using <code>edir_free()</code>.
 */
extern char *
edir_strtrim(const char *s)
{
    char *start, *end, *trim;
    int32 len;
    for (start=(char *)s; *start && (*start <= ' '); start++) ;
    if ((len = strlen(start)) < 1) return edir_strdup("");
    for (end = start + len - 1; end >= start; end--) if (*end > ' ') break;
    if (end < start) return edir_strdup("");
    len = end - start + 1;
    trim = (char *) edir_malloc(len + 1);
    strncpy(trim, start, len);
    trim[len] = '\0';
    return trim;
}

/**
 * Remove leading and trailing whitespace characters and path separators.
 * <P>
 * This function handles both the Unix '/' path separator and the DOS
 * '\' path separator.
 * 
 * @param s [in] The string to trim.
 * @return A newly allocated character array containing the trimmed string.
 * 	   This array should be freed using <code>edir_free()</code>.
 */
extern char *
edir_strtrimpath(const char *s)
{
    char *start, *end, *trim;
    int32 len;
    for (start=(char *)s;
	 *start && ((*start <= ' ') || (*start == '/') || (*start == '\\'));
	 start++) ;
    if ((len = strlen(start)) < 1) return edir_strdup("");
    for (end = start + len - 1; end >= start; end--)
	if ((*end > ' ') && (*end != '/') && (*end != '\\')) break;
    if (end < start) return edir_strdup("");
    len = end - start + 1;
    trim = (char *) edir_malloc(len + 1);
    strncpy(trim, start, len);
    trim[len] = '\0';
    return trim;
}

/**
 * Error-checking <code>atoi</code> replacement.
 * <P>
 * Uses <code>strtol</code> in order to check <code>errno</code>
 * for formatting or out-of-range errors in the input string. 
 * Returns <code>1</code> (true) if the string was successfully converted,
 * <code>0</code> if not.
 * 
 * @param s    [in]  The string to convert.
 * @param iptr [out] Pointer to 32 bit integer to be filled in upon 
 * 		     successful conversion.
 * @return True (1) if conversion successful, otherwise False (0).
 */
extern int
edir_atoi(const char *s, int32 *iptr)
{
    char *endptr;
    int32 ival;
    *iptr = 0;
    errno = 0;
    if (s == NULL) return 0;
    ival = (int32) strtol(s, &endptr, 0);
    if ((errno == ERANGE) || (errno == EINVAL)) return 0;
    if (endptr) {
	while (*endptr && (*endptr <= ' ')) endptr++;
	if (*endptr) return 0; /* trailing garbage. */
    }
    *iptr = ival;
    return 1;
}

/**
 * Internal string/data buffer header pointed to by the externally
 * visible opaque handles.
 * 
 * @member num_iovs	Current number of iovec entries in use.
 * @member alloc_iovs	Number of iovec slots currently allocated.
 * @member iov		Array of iovec slots.
 * @member free		Array of pointers to free if not NULL.
 */
typedef struct edir_sbuf {
    int32		num_iovs;
    int32		alloc_iovs;
    struct iovec	*iov;
    char		**free;
} edir_sbuf_t;

#define EDIR_SBUF_INCR	50

/**
 * Create an iovec-based string/data buffer.
 * <P>
 * This is an append-only data buffer that allows string or binary
 * data to be written to the end.  Data bytes may be appended as
 * copies, or as references (with functions ending in 'n'), which
 * are stored internally as an array of struct 'iovec'.
 * <P>
 * This allows some debugging routines to build up stack traces or
 * printouts in memory.  It also can allow modules to gather data
 * to be written using scatter-gather system calls such as writev()
 * or sendmsg().
 * 
 * @return A new string buffer handle.
 */
extern void *
edir_sbuf_create()
{
    edir_sbuf_t *sb;
    sb = (edir_sbuf_t *) edir_malloc(sizeof(edir_sbuf_t));
    sb->num_iovs = 0;
    sb->alloc_iovs = EDIR_SBUF_INCR;
    sb->iov = (struct iovec *) edir_malloc(EDIR_SBUF_INCR * sizeof(struct iovec));
    memset((char *)sb->iov, 0, EDIR_SBUF_INCR * sizeof(struct iovec));
    sb->free = (char **) edir_malloc(EDIR_SBUF_INCR * sizeof(char *));
    memset((char *)sb->free, 0, EDIR_SBUF_INCR * sizeof(char *));
    return (void *) sb;
}

/**
 * Internally resize a string buffer object upon overflow.
 * <P>
 * Increases the length of the internal iovec and free arrays
 * by EDIR_SBUF_INCR elements.  It does so by reallocating, copying,
 * and then freeing the old arrays.
 * 
 * @param sb [in,out] The string buffer object to resize.
 */
static void
edir_sbuf_realloc(edir_sbuf_t *sb)
{
    int32		newnum;
    struct iovec	*newiov;
    char		**newfree;
    newnum = sb->num_iovs + EDIR_SBUF_INCR;
    newiov = (struct iovec *) edir_malloc(newnum * sizeof(struct iovec));
    memcpy((char *)newiov, (char *)sb->iov,
	    sb->num_iovs * sizeof(struct iovec));
    memset((char *)&(newiov[sb->num_iovs]), 0,
	    EDIR_SBUF_INCR * sizeof(struct iovec));
    edir_free(sb->iov);
    sb->iov = newiov;
    newfree = (char **) edir_malloc(newnum * sizeof(char *));
    memcpy(newfree, sb->free, sb->num_iovs * sizeof(char *));
    memset((char *)&(newfree[sb->num_iovs]), 0, EDIR_SBUF_INCR * sizeof(char *));
    edir_free(sb->free);
    sb->free = newfree;
    sb->alloc_iovs = newnum;
}

/**
 * Append a formatted string to the end of this string/data buffer.
 * <P>
 * The trailing null character is not included, since these functions
 * keep track of the length of each data extent internally.  This
 * allows binary data to be added.
 * 
 * @param sbuf [in,out] The string/data buffer to append to.
 * @param fmt  [in]     Printf()-style format string.
 * @param ...  [in]     Printf()-style variadic argument list.
 * @return The number of bytes actually added to the buffer.
 */
extern int32
edir_sbuf_appendf(void *sbuf, char *fmt, ...)
{
    va_list		ap;
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		len;
    char		*cpy;
    va_start(ap, fmt);
    cpy = edir_vasprintf((char *)fmt, ap);
    va_end(ap);
    len = strlen(cpy);
    if (sb->num_iovs >= sb->alloc_iovs) edir_sbuf_realloc(sb);
    sb->iov[sb->num_iovs].iov_base = cpy;
    sb->iov[sb->num_iovs].iov_len = len;
    sb->free[sb->num_iovs] = cpy; /* Want to free this eventually. */
    sb->num_iovs++;
    return len;
}

/**
 * Append a null-terminated character string to a string/data buffer.
 * <P>
 * A newly allocated copy of the passed-in string is added to
 * the string/data buffer.
 * 
 * @param sbuf [in,out] The buffer to append to.
 * @param str  [in]     The string to append.
 * @return The number of characters appended.
 */
extern int32
edir_sbuf_strcat(void *sbuf, char *str)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		len;
    char		*cpy;
    if (sb->num_iovs >= sb->alloc_iovs) edir_sbuf_realloc(sb);
    cpy = edir_strdup(str);
    len = strlen(cpy);
    sb->iov[sb->num_iovs].iov_base = cpy;
    sb->iov[sb->num_iovs].iov_len = len;
    sb->free[sb->num_iovs] = cpy; /* Want to free this eventually. */
    sb->num_iovs++;
    return len;
}

/**
 * Append a null-terminated character string to a string/data buffer.
 * <P>
 * A reference to the passed in string is saved in the string/data buffer.
 * This reference is not freed when edir_sbuf_free() is eventually called.
 * However, care must be taken not to free this string before finished
 * with this buffer, or heap corruption and horrible program death may
 * occur.
 * 
 * @param sbuf [in,out] The buffer to append to.
 * @param str  [in]     The string to append.
 * @return The number of characters appended.
 */
extern int32
edir_sbuf_strcatn(void *sbuf, char *str)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		len;
    if (sb->num_iovs >= sb->alloc_iovs) edir_sbuf_realloc(sb);
    len = strlen(str);
    sb->iov[sb->num_iovs].iov_base = str;
    sb->iov[sb->num_iovs].iov_len = len;
    sb->free[sb->num_iovs] = NULL; /* Don't free this element */
    sb->num_iovs++;
    return len;
}

/**
 * Append a generic byte array to the string/data buffer.
 * <P>
 * A newly allocated copy of the passed-in array is added to
 * the string/data buffer.
 * 
 * @param sbuf [in,out] The buffer to append to.
 * @param data [in]     The byte array to append.
 * @param len  [in]     The number of bytes to append.
 * @return The number of bytes appended. (len)
 */
extern int32
edir_sbuf_bincat(void *sbuf, void *data, int32 len)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    char		*cpy;
    if (sb->num_iovs >= sb->alloc_iovs) edir_sbuf_realloc(sb);
    cpy = (char *) edir_malloc(len);
    memcpy(cpy, data, len);
    sb->iov[sb->num_iovs].iov_base = cpy;
    sb->iov[sb->num_iovs].iov_len = len;
    sb->free[sb->num_iovs] = cpy; /* Want to free this eventually. */
    sb->num_iovs++;
    return len;
}

/**
 * Append a generic byte array to the string/data buffer.
 * <P>
 * A reference to the passed in array is saved in the string/data buffer.
 * This reference is not freed when edir_sbuf_free() is eventually called.
 * However, care must be taken not to free this string before finished
 * with this buffer, or heap corruption and horrible program death may
 * occur.
 * 
 * @param sbuf [in,out] The buffer to append to.
 * @param data [in]     The byte array to append.
 * @param len  [in]     The number of bytes to append.
 * @return The number of bytes appended. (len)
 */
extern int32
edir_sbuf_bincatn(void *sbuf, void *data, int32 len)
{
    edir_sbuf_t 	*sb = (edir_sbuf_t *)sbuf;
    if (sb->num_iovs >= sb->alloc_iovs) edir_sbuf_realloc(sb);
    sb->iov[sb->num_iovs].iov_base = data;
    sb->iov[sb->num_iovs].iov_len = len;
    sb->free[sb->num_iovs] = NULL; /* Don't free this element */
    sb->num_iovs++;
    return len;
}

/**
 * Copies the entire contents of the string/data buffer to a string.
 * <P>
 * The bytes are copied with no regard to null characters.  (That is,
 * a binary copy is performed).  If binary data is expected, a pointer
 * to an integer may be passed as 'lenp' to receive the actual number
 * of bytes in the new array.
 * <P>
 * The returned array is allocated using edir_malloc() and should
 * eventually be freed using edir_free().
 * 
 * @param sbuf [in] The buffer to convert to string.
 * @param lenp [in] Optional (may be NULL) int32 pointer to receive length.
 * @return A pointer to the newly copied data string.
 */
extern char *
edir_sbuf_tostring(void *sbuf, int32 *lenp)
{
    edir_sbuf_t 	*sb = (edir_sbuf_t *)sbuf;
    int32		len, niovs, i;
    char		*cp, *cpy;
    niovs = sb->num_iovs;
    for (i=0, len=0; i < niovs; i++) len += sb->iov[i].iov_len;
    cp = cpy = (char *) edir_malloc(len + 1);
    for (i=0; i < niovs; i++) {
	memcpy(cp, sb->iov[i].iov_base, sb->iov[i].iov_len);
	cp += sb->iov[i].iov_len;
    }
    *cp++ = '\0';
    if (lenp) *lenp = len;
    return cpy;
}

/**
 * Returns the iovec array containing this string/data buffer's data.
 * <P>
 * This is useful for writing out data using writev() or sendmsg().
 * The iovec array may be modified, within bounaries of not overrunning
 * any of the internal arrays.  The pointers to free upon the call
 * to edir_sbuf_free() are maintained separately.  In particular, DO
 * NOT free ANY of the 'iov_base' pointers!  This will cause horrible
 * death due to heap corruption.  Let edir_sbuf_free() do its job.
 * 
 * @param sbuf [in]	The string/data buffer from which to obtain iovecs.
 * @param lenp [in,out] Pointer to int32 to receive the number of iovecs.
 * @return A pointer to an array of iovecs.
 */
extern struct iovec *
edir_sbuf_get_iov(void *sbuf, int32 *lenp)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    *lenp = sb->num_iovs;
    return sb->iov;
}

/**
 * Copy data from a string/data buffer into an existing contiguous
 * data buffer.
 * <P>
 * This function copies no more than 'buflen' bytes from 'sbuf'
 * into the contiguous buffer referenced by 'buf'.  No null 
 * terminating character is appended to the copied data.  Instead,
 * the number of bytes actually copied is returned.
 * 
 * @param sbuf   [in]  The string/data buffer to copy from.
 * @param buf    [out] The data buffer to copy into.
 * @param buflen [in]  Maximum number of bytes to copy to 'buf'.
 * @return The actual number of bytes copied.
 */
extern int32
edir_sbuf_copy_into(void *sbuf, void *buf, int32 buflen)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		len, niovs, i, tocpy;
    size_t		left;
    char		*cp;
    niovs = sb->num_iovs;
    for (i=0, len=0; i < niovs; i++) len += sb->iov[i].iov_len;
    left = len = EDIR_MIN(len, buflen);
    for (cp=buf,i=0; (left > 0) && (i < niovs); i++) {
	tocpy = EDIR_MIN(sb->iov[i].iov_len, left);
	memcpy(cp, sb->iov[i].iov_base, tocpy);
	left -= tocpy;
	cp += tocpy;
    }
    return len;
}

/**
 * Returns the total length of data contained by a string/data buffer.
 * <P>
 * The returned length does not include a terminating null character.
 * 
 * @param sbuf [in] The string/data buffer to measure.
 * @return The length in bytes of the buffer's contained data.
 */
extern int32
edir_sbuf_length(void *sbuf)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		len, niovs, i;
    niovs = sb->num_iovs;
    for (i=0, len=0; i < niovs; i++) len += sb->iov[i].iov_len;
    return len;
}

/**
 * Free any allocated strings contained within the string/data buffer,
 * but don't destroy the buffer.
 * <P>
 * After this function is called, the buffer referenced by 'sbuf'
 * will be in the same state as a newly created (empty) buffer.
 * 
 * @param sbuf [in,out] The buffer to clear.
 */
extern void
edir_sbuf_clear(void *sbuf)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		i, niovs;
    niovs = sb->num_iovs;
    for (i=0; i < niovs; i++) if (sb->free[i]) edir_free(sb->free[i]);
    memset((char *)sb->iov,  0, niovs * sizeof(struct iovec));
    memset((char *)sb->free, 0, niovs * sizeof(char *));
    sb->num_iovs = 0;
}

/**
 * Remove iovecs from the beginning of the internal array.
 * <P>
 * This is useful for code that 'eats' data from the beginning
 * of the sbuf and then wishes to discard it.  Code that handles
 * partial and/or nonblocking writev() or sendmsg() calls can
 * make use of this functionality.
 * 
 * @param sbuf  [in,out] The sbuf to trim.
 * @param niovs [in]     The number of iovecs to skip and discard.
 */
extern void
edir_sbuf_skip(void *sbuf, int32 niovs)
{
    edir_sbuf_t		*sb = (edir_sbuf_t *)sbuf;

    int			i, remain;
    if ((remain = sb->num_iovs - niovs) < 0) {
	remain = 0;
	niovs = sb->num_iovs;
    }
    if (niovs <= 0) return;
    for (i=0; i < niovs; i++) if (sb->free[i]) edir_free(sb->free[i]);
    for (i=0; i < remain; i++) sb->iov[i] = sb->iov[i+niovs];
    memset((char *)&(sb->iov[i]), 0, niovs * sizeof(struct iovec));
    memset((char *)&(sb->free[i]), 0, niovs * sizeof(char *));
}

/**
 * Destroy a buffer and free all of its allocated data.
 * <P>
 * Frees all copied data arrays, the internal free and iovec
 * arrays, and the buffer header.  The passed in handle should 
 * not be used after calling this function on it.
 * 
 * @param sbuf [in,out] Handle to the buffer to destroy.
 */ 
extern void
edir_sbuf_free(void *sbuf)
{
    edir_sbuf_t 		*sb = (edir_sbuf_t *)sbuf;
    int32		i, niovs;
    niovs = sb->num_iovs;
    if (sb->free) {
	for (i=0; i < niovs; i++) if (sb->free[i]) edir_free(sb->free[i]);
	edir_free(sb->free);
    }
    if (sb->iov) edir_free(sb->iov);
    edir_free(sb);
}

/**
 * Destructively split up a URI into Path-Info, Path-Parameters, 
 * and Query-String, and pass each portion to a callback for
 * processing.
 * <P>
 * This allows Enhydra Director client modules to easily obtain the servlet
 * prefix and URL-encoded session IDs in a module-specific way.
 * 
 * @param uri [in,out] The uri buffer to destructively parse.  This function
 * 		modifies the buffer pointed to by 'uri', so it should only
 * 		contain copied data.
 * @param funcData [in,out] Module specific data passed by the calling
 * 		module.  The edir_parse_uri() function does not dereference
 * 		this pointer, so it may be NULL.
 * @param func [in] Pointer to function that will be passed the <code>
 * 		funcData</code> argument, an integer token type, a
 * 		'left hand side' string, and a 'right hand side' string.
 */
extern void
edir_parse_uri(char *uri, void *funcData, 
		   void (*func)(void *, int, char *, char *))
{
    char	*pinfo = NULL, *pparam = NULL, *qs = NULL, *cp, *dp;
    int		proceed, have_pparam=0, have_qs=0;

    if (!uri) return;
    while (isspace(*uri)) uri++;
    while (*uri == '/') uri++;
    pinfo = uri;
    for (proceed=1; proceed;) {
	switch (*uri) {
	    case '\0': case '#':
		*uri = '\0';
		pparam = "";
		qs = "";
		proceed = 0;
		break;
	    case ';': 
		*uri++ = '\0';
		have_pparam=1;
		proceed = 0;
		break;
	    case '?': 
		pparam = "";
		*uri++ = '\0';
		have_qs=1;
		proceed = 0;
		break;
	    default:
		uri++;
		break;
	}
    }
    /* Strip trailing and multiple slashes */
    for (cp = dp = pinfo; *cp; ) {
	if ((*dp++ = *cp++) == '/') while (*cp == '/') cp++;
    }
    *dp-- = '\0';
    /* XXX Actually, we want to LEAVE trailing slash! */
    /* while ((dp >= pinfo) && (*dp == '/')) *dp-- = '\0'; */
    
    /* Get any "path" parameters */
    if (have_pparam) for (pparam = uri, proceed=1; proceed; ) {
	switch (*uri) {
	    case '\0': case '#':
		*uri = '\0';
		qs = "";
		proceed = 0;
		break;
	    case '?': 
		*uri++ = '\0';
		have_qs=1;
		proceed = 0;
		break;
	    default:
		uri++;
		break;
	}
    }

    /* Query string, if any, is the last part of the URI */
    if (have_qs) for (qs = uri, proceed=1; proceed; ) {
	switch (*uri) {
	    case '\0': case '#': 
		*uri = '\0';
		proceed = 0;
		break;
	    default:
		uri++;
		break;
	}
    }

    (func)(funcData, EDIR_PARSE_URI_PATH_INFO, pinfo, NULL);

    if (have_pparam && pparam) {
	(func)(funcData, EDIR_PARSE_URI_PATH_PARAMS, pparam, NULL);
	while (dp = strchr(pparam, ';')) {
	    *dp++ = '\0';
	    if (cp = strchr(pparam, '=')) {
		*cp++ = '\0';
		(func)(funcData, EDIR_PARSE_URI_PATH_PARAM_2, pparam, cp);
	    } else {
		(func)(funcData, EDIR_PARSE_URI_PATH_PARAM_1, pparam, NULL);
	    }
	    pparam = dp;
	}
	if (cp = strchr(pparam, '=')) {
	    *cp++ = '\0';
	    (func)(funcData, EDIR_PARSE_URI_PATH_PARAM_2, pparam, cp);
	} else {
	    (func)(funcData, EDIR_PARSE_URI_PATH_PARAM_1, pparam, NULL);
	}
    }

    if (have_qs) (func)(funcData, EDIR_PARSE_URI_QUERY_STRING, qs, NULL);
}

