/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_isapi_log_line.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_isapi_log.h"
#include "edir_isapi_log_line.h"

static LPCTSTR evtKey="SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application";
static char * edir_log_isapi_quote_nl(char *str, char *buf, int32 bufsize);
static TCHAR eventLogName[256];

/**
 * Log a single line of text to the correct log channel.
 * <P>
 * This function should only be called by the general logging functions
 * in 'edir_log.c'.  The common line logging function has been split out
 * into a separate module to allow specific client implementations to
 * integrate logging into the native web server's mechanism.
 * 
 * @param level [in]  The logging level, as per syslog.
 * @param line  [in]  Single line of pre-formatted log text.
 */
extern void
edir_log_line(int level, char *line)
{
     switch (level) {
	case LOG_ERR:
	    EDIR_IF_DEBUG( NTDebugLog(line); )
	    NTEventLog(line, 1);
	    break;
	case LOG_DEBUG:
	    NTDebugLog(line);
	    break;
	case LOG_NOTICE:
	    EDIR_IF_DEBUG( NTDebugLog(line); )
	    NTEventLog(line, 0);
	    break;
	default:
	    NTDebugLog(line);
	    break;
    }
}

/**
 * Set whether the log system is in 'daemon' mode.
 * <P>
 * In the standard log module, this means "use syslog instead of printf".
 * It may or may not be used by other modules, such as Apache.
 * 
 * @param progName [in]  The name of the program doing the logging.
 * @param isDaemon [in]  True for daemon mode, false for "normal".
 */
extern void
edir_log_setdaemon(char *progName, int isDaemon)
{
    /* No-Op on Win32 */
}

/**
 * List of message ids for messages with between zero and 50 lines.
 * If a message has more than 50 lines, the last line is the rest of
 * the lines glommed together.
 */
static DWORD messageIds[] = {
    EVTLOG_MULTILINE_0, 
    EVTLOG_MULTILINE_1, 
    EVTLOG_MULTILINE_2, 
    EVTLOG_MULTILINE_3, 
    EVTLOG_MULTILINE_4, 
    EVTLOG_MULTILINE_5, 
    EVTLOG_MULTILINE_6, 
    EVTLOG_MULTILINE_7, 
    EVTLOG_MULTILINE_8, 
    EVTLOG_MULTILINE_9, 
    EVTLOG_MULTILINE_10, 
    EVTLOG_MULTILINE_11, 
    EVTLOG_MULTILINE_12, 
    EVTLOG_MULTILINE_13, 
    EVTLOG_MULTILINE_14, 
    EVTLOG_MULTILINE_15, 
    EVTLOG_MULTILINE_16, 
    EVTLOG_MULTILINE_17, 
    EVTLOG_MULTILINE_18, 
    EVTLOG_MULTILINE_19, 
    EVTLOG_MULTILINE_20, 
    EVTLOG_MULTILINE_21, 
    EVTLOG_MULTILINE_22, 
    EVTLOG_MULTILINE_23, 
    EVTLOG_MULTILINE_24, 
    EVTLOG_MULTILINE_25, 
    EVTLOG_MULTILINE_26, 
    EVTLOG_MULTILINE_27, 
    EVTLOG_MULTILINE_28, 
    EVTLOG_MULTILINE_29, 
    EVTLOG_MULTILINE_30, 
    EVTLOG_MULTILINE_31, 
    EVTLOG_MULTILINE_32, 
    EVTLOG_MULTILINE_33, 
    EVTLOG_MULTILINE_34, 
    EVTLOG_MULTILINE_35, 
    EVTLOG_MULTILINE_36, 
    EVTLOG_MULTILINE_37, 
    EVTLOG_MULTILINE_38, 
    EVTLOG_MULTILINE_39, 
    EVTLOG_MULTILINE_40, 
    EVTLOG_MULTILINE_41, 
    EVTLOG_MULTILINE_42, 
    EVTLOG_MULTILINE_43, 
    EVTLOG_MULTILINE_44, 
    EVTLOG_MULTILINE_45, 
    EVTLOG_MULTILINE_46, 
    EVTLOG_MULTILINE_47, 
    EVTLOG_MULTILINE_48, 
    EVTLOG_MULTILINE_49, 
    EVTLOG_MULTILINE_50
};
#define MAX_LINES 50

/**
 * Add an event to the Windows NT event log.
 * <P>
 * This implementation correctly handles embedded newlines and 
 * "does it right" as far as using message strings stored in
 * the DLL and registering with the event sources in the registry.
 * <P>
 * Due to limitations in the NT event log system, there is a hard
 * limit of 50 separate lines of text.  If there are more than 50
 * lines in <code>message</code>, the lines past the 50th are 
 * appended to the 50th without line breaks.  You really shouldn't
 * be firing off NT log events that big anyhow.
 *
 * @param message [in] The message to log.
 * @param type    [in] True if this is an error event.
 */
extern void
NTEventLog(LPTSTR message, int type)
{
    LPTSTR  tmpbuf, cp, dp;
    DWORD   nlines;
    LPTSTR  lines[MAX_LINES + 1];
    HANDLE  hEventSource;

    tmpbuf = (char *)_alloca(strlen(message)+1);
    strcpy(tmpbuf, message);
    nlines = 0;
    for (cp = tmpbuf; *cp && (dp = strchr(cp, '\n')); cp = dp + 1) {
	if (nlines >= (MAX_LINES - 1)) break;
	if ((dp > cp) && (*(dp-1) == '\r')) {
	    *(dp - 1) = '\0';
	} else {
	    *dp = '\0';
	}
	lines[nlines++] = cp;
    }
    if (*cp) lines[nlines++] = cp;

    if (lines > 0) {
	hEventSource = RegisterEventSource(NULL, eventLogName);
	if (hEventSource != NULL) {
	    ReportEvent(
		hEventSource,	     // handle of event source
		(WORD) (type ? EVENTLOG_ERROR_TYPE : EVENTLOG_INFORMATION_TYPE), 
		(WORD) 0,	     // No category.
		messageIds[nlines],  // Standard message template ('%1')
		(PSID) NULL,	     // No SID.
		(WORD) nlines,	     // One string in the array.
		0,		     // Zero supplemental data bytes.
		lines,	   	     // The strings to output.
		NULL);		     // No supplemental data
	    DeregisterEventSource(hEventSource);
	}
    }
}

/**
 * Write a message to the Windows NT Debug logging system.
 * <P>
 * To view these debug messages, use the 'DBMON.EXE' example
 * program from the Microsoft MSDN CD.
 *
 * @param message [in] The message to log.
 */
extern void
NTDebugLog(LPTSTR message)
{
    LPTSTR  cp, dp, ep;
    TCHAR   saveChar;
    for (cp = message; *cp && (dp = strchr(cp, '\n')); cp = ep) {
	ep = dp + 1;
	saveChar = *ep;
	*ep = '\0';
	OutputDebugString(cp);
	*ep = saveChar;
    }
    if (*cp) {
	OutputDebugString(cp);
	OutputDebugString("\n");
    }
}

/**
 * Register with the NT event log system.
 * <P>
 * Allows full text messages to be logged to the system event log.
 * 
 * @param dllName [in]     The name of the currently running DLL.
 * @param displayName [in] The name to display in log message headings.
 * @return TRUE if successful, otherwise FALSE.
 */
extern BOOL
edir_isapi_event_log_register(LPCTSTR dllName, LPCTSTR displayName)
{
    LONG ret;
    HKEY keyHandle;
    DWORD disposition;
    DWORD eventTypes;
    TCHAR keybuf[1024];
    
    _snprintf(keybuf, 1023, "%s\\%s", evtKey, displayName);
    keybuf[1023] = '\0';

    strncpy(eventLogName, displayName, sizeof(eventLogName) - 1);
    eventLogName[sizeof(eventLogName) - 1] = '\0';

    ret = RegCreateKeyEx(HKEY_LOCAL_MACHINE, keybuf, 0, NULL,
			 REG_OPTION_NON_VOLATILE,
			 KEY_SET_VALUE, NULL, &keyHandle, &disposition);
    if (ret != ERROR_SUCCESS) {
	EDIR_DEBUG_SPARSE(("Can't create Event Logging keys."));
	return FALSE;
    }

    if (disposition == REG_OPENED_EXISTING_KEY) {
	EDIR_DEBUG_VERBOSE(("Updating existing Event Logging keys."));
    }

    ret = RegSetValueEx(keyHandle, "EventMessageFile", 0, REG_EXPAND_SZ,
			(LPBYTE) dllName, strlen(dllName) + 1);
    if (ret != ERROR_SUCCESS) {
	RegCloseKey(keyHandle);
	EDIR_DEBUG_SPARSE(("Can't create 'EventMessageFile' registry key."));
	return FALSE;
    }

    eventTypes = EVENTLOG_ERROR_TYPE | EVENTLOG_INFORMATION_TYPE;
    ret = RegSetValueEx(keyHandle, "TypesSupported", 0, REG_DWORD, 
			(LPBYTE) &eventTypes, sizeof(DWORD));
    if (ret != ERROR_SUCCESS) {
	RegCloseKey(keyHandle);
	EDIR_DEBUG_SPARSE(("Can't create 'TypesSupported' registry key."));
	return FALSE;
    }

    RegCloseKey(keyHandle);
    EDIR_DEBUG_VERBOSE(("Registered to send messages to Event Log."));

    return TRUE;
}
 
