/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_isapi_memory.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_isapi_memory.h"

#undef EDIR_DEBUG_MALLOC

static DWORD	edir_malloc_tlsindex = 0xffffffff;
static DWORD	edir_mallopt_tlsindex = 0xffffffff;

#ifdef EDIR_DEBUG_MALLOC
#define MEMHDR_MAGIC 0xdeadbeef
#define ALIGNBYTES 8
#define _ALIGNSZ(S)  ((((S) + ALIGNBYTES - 1) / ALIGNBYTES) * ALIGNBYTES)
typedef struct memhdr {
    int32	magic;
    void	*heap;
    int32	size;
} memhdr_t;
static int memhdr_size; /* Must initialize */
static void edir_isapi_memory_die(char *);
#endif

/**
 * Set up the thread local storage identifier used by threads to
 * store their local heaps.
 * <P>
 * This implementation uses the Win32 TlsAlloc() function to allocate
 * a TLS index and store it in a global variable accessible to all threads.
 * All threads use this index to get at their thread-private heap handle.
 * <P>
 * This function should be called in the ISAPI GetFilterVersion function
 * or method to prepare for per-thread heaps.
 */
extern BOOL
edir_isapi_memory_init()
{
#ifdef EDIR_DEBUG_MALLOC
    memhdr_size = _ALIGNSZ(sizeof(memhdr_t));
#endif    
    edir_malloc_tlsindex = TlsAlloc();
    edir_mallopt_tlsindex = TlsAlloc();
    if ((edir_malloc_tlsindex == -1) ||
	(edir_mallopt_tlsindex == -1)) {
	return FALSE;
    }
    return TRUE;
}

/**
 * Creates EnhydraDirector's per-thread heap for the current thread.
 * <P>
 * This function MUST be called prior to any <code>edir_malloc()</code>
 * calls in the current request.  At the end of the request,
 * the <code>edir_malloc_delete_heap()</code> should be used to free
 * the heap, and possibly any leaked memory.
 * 
 * @return A handle for the newly created heap.  This should be stored
 * 	in the per-request context structure.
 */
extern HANDLE
edir_isapi_memory_create_heap(DWORD options)
{
    HANDLE heap;
    heap = HeapCreate(options, 0, 0);
    TlsSetValue(edir_malloc_tlsindex, (LPVOID) heap);
    TlsSetValue(edir_mallopt_tlsindex, (LPVOID) options);
    return heap;
}

/**
 * Sets EnhydraDirector's per-thread heap for the current thread.
 * <P>
 * This function should be called at the beginning of each ISAPI class
 * member to set this thread's heap to the one stored in its private
 * context.  This is done because it is not clearly in writing that
 * IIS won't do something extremely stupid like switch threads on
 * a request between calls to OnReadRaw() for the same request.
 * Maybe (probably) it doesn't, but doing this is cheap and make sure
 * we don't get heap corruption if IIS does switch threads on us.
 * 
 * @param heap [in] The heap to place in thread local data.
 */
extern void
edir_isapi_memory_set_heap(HANDLE heap, DWORD options)
{
    TlsSetValue(edir_malloc_tlsindex, (LPVOID) heap);
    TlsSetValue(edir_mallopt_tlsindex, (LPVOID) options);
}

/**
 * Deletes this thread's per-request heap and frees any memory
 * still allocated from it.
 * <P>
 * Once this function has been called, <code>edir_malloc()</code> will no
 * longer work, and ALL pointers previously allocated by
 * <code>edir_malloc()</code> become invalid.
 */
extern void
edir_isapi_memory_delete_heap()
{
    HANDLE heap = (HANDLE)TlsGetValue(edir_malloc_tlsindex);
    HeapDestroy(heap);
    TlsSetValue(edir_malloc_tlsindex, NULL);
}

/**
 * Allocate memory from the current request's heap.
 * <P>
 * @param sz [in] The number of bytes to allocate.
 * @return A pointer to the allocated memory region.
 */
extern void *
edir_malloc(int32 sz)
{
#ifndef EDIR_DEBUG_MALLOC
    return HeapAlloc((HANDLE)TlsGetValue(edir_malloc_tlsindex),
		  (DWORD)TlsGetValue(edir_mallopt_tlsindex),		      
		  (DWORD) sz);
#else
    int32 fullSize = memhdr_size + _ALIGNSZ(sz) + ALIGNBYTES;
    HANDLE heap;
    DWORD opts;
    memhdr_t *hdr;
    heap = (HANDLE) TlsGetValue(edir_malloc_tlsindex);
    opts = (DWORD) TlsGetValue(edir_mallopt_tlsindex);
    hdr = (memhdr_t *) HeapAlloc(heap, opts, (DWORD)fullSize);
    hdr->magic = MEMHDR_MAGIC;
    hdr->size = sz;
    hdr->heap = heap;
    return (void *)(((char *)hdr) + memhdr_size);
#endif
}

/**
 * Deallocate a memory block which was previously allocated using
 * <code>edir_malloc()</code>
 * 
 * @param p [in] Pointer to the memory region to free.
 */
extern void
edir_free(void *p)
{
#ifndef EDIR_DEBUG_MALLOC
    HeapFree((HANDLE)TlsGetValue(edir_malloc_tlsindex),
	     (DWORD)TlsGetValue(edir_mallopt_tlsindex),		      
	     (LPVOID) p);
#else
    HANDLE heap;
    DWORD opts;
    memhdr_t *hdr;
    hdr = (memhdr_t *)(((char *)p) - memhdr_size);
    heap = (HANDLE) TlsGetValue(edir_malloc_tlsindex);
    opts = (DWORD) TlsGetValue(edir_mallopt_tlsindex);
    if (hdr->magic != MEMHDR_MAGIC) {
	edir_isapi_memory_die("Unrecognized heap header");
	return; /* NOTREACHED */
    }
    if (hdr->heap != heap) {
	edir_isapi_memory_die("Wrong heap!");
	return; /* NOTREACHED */
    }
    hdr->magic = 0xfeedbeef;
    HeapFree(heap, opts, (LPVOID) hdr);
#endif
}

/**
 * Reallocate a memory block which has previously been allocated
 * using <code>edir_malloc()</code>.
 *
 * @param p [in] Pointer to the block to resize.
 * @param sz [in] The desired new block size.
 * @return A pointer to the reallocated block.
 */
extern void *
edir_realloc(void *p, int32 sz)
{
#ifndef EDIR_DEBUG_MALLOC
    return HeapReAlloc((HANDLE)TlsGetValue(edir_malloc_tlsindex),
		       (DWORD)TlsGetValue(edir_mallopt_tlsindex),
	    	       p, (DWORD)sz);
#else
    int32 fullSize = memhdr_size + _ALIGNSZ(sz) + ALIGNBYTES;
    HANDLE heap;
    DWORD opts;
    memhdr_t *hdr, *newhdr;
    hdr = (memhdr_t *)(((char *)p) - memhdr_size);
    heap = (HANDLE) TlsGetValue(edir_malloc_tlsindex);
    opts = (DWORD) TlsGetValue(edir_mallopt_tlsindex);
    if (hdr->magic != MEMHDR_MAGIC) {
	edir_isapi_memory_die("Unrecognized heap header (realloc)");
	return NULL; /* NOTREACHED */
    }
    if (hdr->heap != heap) {
	edir_isapi_memory_die("Wrong heap! (realloc)");
	return NULL; /* NOTREACHED */
    }
    hdr->magic = 0xdededede;
    newhdr = (memhdr_t *) HeapReAlloc(heap, opts, (LPVOID) hdr, fullSize);
    newhdr->magic = MEMHDR_MAGIC;
    newhdr->size = sz;
    newhdr->heap = heap;
    return (void *)(((char *)newhdr) + memhdr_size);
#endif
}

#ifdef EDIR_DEBUG_MALLOC
static void
edir_isapi_memory_die(char *s)
{
    char *p = NULL;
    OutputDebugString(s);
    while(1) {
	Sleep(1000);
	strcpy(p, s); /* This intentionally crashes the prog. */
    }
}
#endif
