/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_isapi_filter.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

/* 
 * ENHYDRADIRECTOR.CPP - ISAPI filter for EnhydraDirector.
 * 
 * This module implements the EnhydraDirector client for IIS HTTP servers.
 * It is implemented as an ISAPI filter so that we can switch on the
 * URL based on the enhydra_director.conf file, and so we can get at the
 * raw incoming data at a low level.
 * <P>
 * ISAPI filters are a bit messy, since the incoming data from a browser
 * POST or upload comes as separate calls to the OnRawRead() method,
 * and we don't have direct access to a stream object we can read from.
 * <P>
 * Another annoyance is the fact that we can't get the parsed headers
 * until later, at which point we no longer have access to the POST 
 * upload data (unless we buffer it and we could get multi GB uploaded
 * movies and the like so that's not going to happen).
 */

#include "edir.h"
#include "edir_isapi.h"
#include <httpfilt.h>

#define EDIR_ISAPI_MAX_URL_LEN 1024
#define EDIR_ISAPI_MAX_HOST_NAME_LEN 256
#define DEFAULT_ENHYDRA_HANDLER_URL "/EnhydraDirector/EnhydraHandler.dll"

/*
 * Externs not defined in header files.
 * XXX should define in header file.
 */
extern int edir_log_isapi_register(LPCTSTR dllName);

/**
 * Internal copy of this DLL's instance handle, needed by ISAPI.
 */
static HINSTANCE myInstanceHandle;

/*
 * Local 'am I detached?' flag.
 */
static int Dead = 1;

/**
 * The global memory heap.
 * <P>
 * Bytes allocated from this heap are not destroyed en masse until
 * this DLL is detached.
 */
static HANDLE globalHeap = NULL;

/**
 * Local copy of this DLL's full path name.
 */
static LPTSTR myDllName = NULL;

/**
 * Local copy of the full path of the config file we're using.
 */
static LPTSTR myConfig = NULL;

/**
 * Handler URL to pass EnhydraDirector requests to.
 */
static LPTSTR handlerURL = DEFAULT_ENHYDRA_HANDLER_URL;

/**
 * Pointer to the global scoreboard.  This scoreboard follows the 
 * same format as the NSAPI and Apache modules.  It is position 
 * independent, even though it does not need to be.  This allows for
 * maintenance of the scoreboard access code in only one place, which
 * should facilitate increased robustness and reliablility across all
 * platforms.
 */
static edir_load_sb_offsets_t *scoreboard = NULL;

/**
 * Number of bytes allocated for the scoreboard.
 */
static int32 sbSize = 0;

/**
 * Global reference count array to use instead of scoreboard itself.
 * <P>  
 * The scoreboard itself has a reference count for each server entry,
 * but the actual scoreboard access code takes an array of references
 * in order to allow flexible handling of reference counts in multi-
 * process environments such as Apache.
 * <P>
 * It turns out that it is more efficient and less complex to just keep
 * a separate array of reference counts in parallel to the scoreboard.
 * XXX FIX A better solution will be to rearrange the scoreboard to
 * store the reference counts in a contiguous array, as that would also
 * allow non-systemV refcount methods to implement simpler code.
 */
static int32 *refCounts = NULL;

/**
 * The number of servers (and thus reference counts).
 */
static int32 numServers = 0;

/**
 * Handle filter-wide initializations when the DLL is attached.
 * <P>
 * This function is called by <code>DllMain()</code> when the DLL is
 * attached by ISAPI.  This is a good place to perform global, filter-
 * wide initializations that are not freed until the DLL is unloaded.
 * 
 * @param handle [in] The instance handle for this DLL.
 * @return True if the DLL should load, False if it should fail to load.
 */
static int
edir_isapi_filter_init(HINSTANCE handle)
{
    static char		    *funcName = "edir_isapi_filter_init";
    edir_exception_t	    *_exc=NULL, **exc=&_exc;
    char		    *trace = NULL, *cp;
#define                     MODNAMESIZE 4096
    char		    moduleName[MODNAMESIZE];
    void		    *sbuf;
    HANDLE		    tmpHeap = NULL;
    DWORD		    debugMask = 0;
    edir_conf_t		    *appCfg = NULL;
    int			    i;

    // First real thing we ever do is register with the event log
    // so we can complain if anything else fails.
    GetModuleFileName(handle, moduleName, MODNAMESIZE);
    moduleName[MODNAMESIZE-1] = '\0';

    /*
     * Use unwrappered malloc here -- this is NEVER freed!
     * I have seen tlsAlloc() fail before, so I want to spew
     * a complaint to the NT event log if we fail for that reason.
     */
    myDllName = malloc(strlen(moduleName)-1);
    strcpy(myDllName, moduleName);
    edir_isapi_event_log_register(myDllName, "EnhydraFilter");

    // Initialize the EnhydraDirector heap system.
    if (!edir_isapi_memory_init()) {
	NTEventLog("Fatal error: Thread Local storage not available", 1);
	return FALSE; /* Fail to load the filter. */
    }

    // Initialize a serialized global heap.
    // This heap is not freed until the DLL is unloaded.
    globalHeap = edir_isapi_memory_create_heap(0);

    /*
     * Compute the default config file name from the DLL name
     * in case we need it.
     */
    sbuf = edir_sbuf_create();
    for (cp=moduleName; cp = strchr(cp, '/');) *cp++ = '\\';
    if (cp = strrchr(moduleName, '\\')) *cp = '\0';
    edir_sbuf_strcatn(sbuf, moduleName);
    edir_sbuf_strcatn(sbuf, "\\enhydra_director.conf");
    myConfig = edir_sbuf_tostring(sbuf, NULL);
    edir_sbuf_free(sbuf);
 
    EDIR_DEBUG_VERBOSE(("Using Config File: '%s'", myConfig));

    /*
     * Create a temporary heap to store the config file data.
     * This heap gets deleted right after we fill in the scoreboard.
     * This ensures that no memory is wasted on config info that
     * we no longer need.
     */
    tmpHeap = HeapCreate(HEAP_NO_SERIALIZE, 0, 0);
    edir_isapi_memory_set_heap(tmpHeap, HEAP_NO_SERIALIZE);

    /*
     * Try to open and process the enhydra_director.conf file.
     */
    appCfg = edir_conf_process(myConfig, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * Set debugging mask if needed.
     */
    if (appCfg->debug) edir_debug |= appCfg->debugMask;

    /*
     * Allocate space for the scoreboard from the global memory
     * memory heap.  This memory is never freed until the DLL is
     * unloaded.  Note that we don't use edir_malloc because it's
     * set to the _temporary_ heap.
     */
    sbSize = edir_load_sb_compute_size(appCfg);
    scoreboard = (edir_load_sb_offsets_t *) HeapAlloc(globalHeap, 0, sbSize);

    /*
     * Convert the config data to a position-independent scoreboard.
     * We don't really need position-independent data here, but we
     * use it anyhow so that we can use the same scoreboard code
     * for all module implementations, including Apache.
     */
    edir_load_sb_from_config(appCfg, (void *)scoreboard);

    /*
     * Delete the temporary heap.
     */
    edir_isapi_memory_set_heap(globalHeap, 0);
    HeapDestroy(tmpHeap);
    tmpHeap = (HANDLE) NULL;

    /*
     * Create a global reference count array.  For the filter, we
     * never touch the refcounts anyway so they will always be zero.
     */
    numServers = edir_load_sb_num_servers((void *)scoreboard);
    refCounts = (int32 *) edir_malloc(numServers * sizeof(int32));
    for (i=0; i<numServers; i++) refCounts[i] = 0;

    /* debug: print out the scoreboard contents */
    trace = edir_load_sb_debug_string((void *)scoreboard,
				refCounts, numServers);
    edir_log("Enhydra Director filter started:\nScoreboard:\n%s", trace);
    edir_free(trace);

    return TRUE;

except:
    trace = edir_exception_stack_trace(*exc);
    edir_log_error("Enhydra Director filter could not start:\n%s", trace);
    edir_free(trace);
    edir_exception_free(exc);
    if (tmpHeap != NULL) HeapDestroy(tmpHeap);
    if (globalHeap != NULL) HeapDestroy(globalHeap);
    return FALSE;
}

/**
 * Handle URL translation for the EnhydraDirector ISAPI client.
 * <P>
 * This function handles the PREPROC_HEADERS notification and 
 * checks the URL to see if it is one handled by EnhydraDirector.
 * If so, the value of the URL is replaced with a URL to the
 * EnhydraDirector handler module, which does the actual work.
 * The original URL, and the name of the config file, are placed in
 * the headers to for use by the handler.
 * 
 * @param req_info [in,out] Pointer to a per-request context structure.
 * @param data     [in]     Header data for this request.
 * @return		    Integer status code.
 */
static DWORD 
edir_isapi_filter_translator(HTTP_FILTER_CONTEXT *req_info, void *data)
{
    static char			*funcName = "edir_isapi_filter_translator";
    edir_exception_t	    	*_exc=NULL, **exc=&_exc;
    HTTP_FILTER_PREPROC_HEADERS *hdr_info;
    TCHAR			url[EDIR_ISAPI_MAX_URL_LEN+1];
    TCHAR			buf[EDIR_ISAPI_MAX_HOST_NAME_LEN+1];
    TCHAR			dbgBuf[20];
    DWORD			nbytes;
    int32			iv, do_status = 0;
    struct sockaddr_in		remote_addr, local_addr;
	LPVOID			pszURL;
 
	DWORD			alloc_dummy; 
	alloc_dummy = 0;

    hdr_info = (HTTP_FILTER_PREPROC_HEADERS *) data;

    /*
     * Security: This prevents some bozo from manually
     * setting these control headers to try to get at EnhydraHandler.dll.
     */
    (hdr_info->SetHeader)(req_info, "edir-url:", NULL);
    (hdr_info->SetHeader)(req_info, "edir-conf:", NULL);
    (hdr_info->SetHeader)(req_info, "edir-debug:", NULL);
    (hdr_info->SetHeader)(req_info, "edir-status:", NULL);

    /*
     * Get the URL and do a tentative check to quickly filter
     * out requests we're not interested in.
     */
    url[0] = '\0';
    nbytes = sizeof(url) - 1;
    (hdr_info->GetHeader)(req_info, "url", url, &nbytes);
    url[nbytes] = '\0';
    if ((!edir_load_status_check_path(scoreboard, url, NULL, NULL, NULL)) && 
	(!edir_load_check_path(scoreboard, url, NULL, NULL, NULL))) {
	return SF_STATUS_REQ_NEXT_NOTIFICATION;
    }

    /* XXX Only spew debug if we're remotely interested in the request */
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR")", funcName, \
			req_info, data));

    /*
     * Get the remote and local connection parameters so we
     * can do a more careful check, taking server and client
     * restrictions into account.
     */
    /* REMOTE_ADDR (Port is zero - we don't care about it) */
    nbytes = sizeof(buf) - 1;
    if ((req_info->GetServerVariable)(req_info, "REMOTE_ADDR",
				      buf, &nbytes) ) {
	buf[nbytes] = '\0';
	if (!edir_ip_addr(buf, &(remote_addr.sin_addr))) {
	    edir_log_error("Can't translate remote IP: '%s'", buf);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
				 "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	    return SF_STATUS_REQ_NEXT_NOTIFICATION;
	}
    } else {
	edir_log_error("Can't obtain remote IP.");
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
			     "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	return SF_STATUS_REQ_NEXT_NOTIFICATION;
    }
    remote_addr.sin_port = 0;
    remote_addr.sin_family = AF_INET;
    
    /* SERVER_ADDR */
    nbytes = sizeof(buf) - 1;
    if ((req_info->GetServerVariable)(req_info, "LOCAL_ADDR",
				      buf, &nbytes) ) {
	buf[nbytes] = '\0';
	if (!edir_ip_addr(buf, &(local_addr.sin_addr))) {
	    edir_log_error("Can't translate local IP: '%s'", buf);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
				 "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	    return SF_STATUS_REQ_NEXT_NOTIFICATION;
	}
    } else {
	edir_log_error("Can't obtain server address.");
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
			     "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	return SF_STATUS_REQ_NEXT_NOTIFICATION;
    }
    
    /* SERVER_PORT */
    nbytes = sizeof(buf) - 1;
    if ((req_info->GetServerVariable)(req_info, "SERVER_PORT",
				      buf, &nbytes) ) {
	buf[nbytes] = '\0';
	if (!edir_atoi(buf, &iv)) {
	    edir_log_error("Can't translate server port: '%s'", buf);
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
				 "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	    return SF_STATUS_REQ_NEXT_NOTIFICATION;
	}
	local_addr.sin_port = htons((unsigned short) iv);
    } else {
	edir_log_error("Can't obtain server port.");
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
			     "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	return SF_STATUS_REQ_NEXT_NOTIFICATION;
    }
    local_addr.sin_family = AF_INET;
    
    /* SERVER_NAME */
    nbytes = sizeof(buf) - 1;
    if ((req_info->GetServerVariable)(req_info, "SERVER_NAME",
				      buf, &nbytes) ) {
	char *cp;
	buf[nbytes] = '\0';
	if (cp = strchr(buf, ':')) *cp = '\0';
    } else {
	edir_sockaddr_to_str(&local_addr, buf);
    }

    if (!edir_load_status_check_path(scoreboard, url, 
	&local_addr, &remote_addr, buf)) {
	if (!edir_load_check_path(scoreboard, url, 
			       &local_addr, &remote_addr, buf)) {
	    EDIR_DEBUG_VERBOSE(("URL '%s' did not match anything.", url));
	    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
				 "SF_STATUS_REQ_NEXT_NOTIFICATION"));
	    return SF_STATUS_REQ_NEXT_NOTIFICATION;
	}
    } else {
	do_status = 1;
    }

	pszURL = req_info->AllocMem(req_info, EDIR_ISAPI_MAX_URL_LEN, alloc_dummy);
    sprintf(pszURL, "%s", url);
	req_info->pFilterContext = pszURL;

    EDIR_DEBUG_VERBOSE(("Translating '%s' --> '%s'", url, handlerURL));
    (hdr_info->SetHeader)(req_info, "url", handlerURL);
    (hdr_info->SetHeader)(req_info, "edir-url:", url);
    (hdr_info->SetHeader)(req_info, "edir-conf:", myConfig);
    sprintf(dbgBuf, "0x%08x", edir_debug);
    (hdr_info->SetHeader)(req_info, "edir-debug:", dbgBuf);
    if (do_status) {
	(hdr_info->SetHeader)(req_info, "edir-status:", "true");
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, \
				 "SF_STATUS_REQ_NEXT_NOTIFICATION"));
    return SF_STATUS_REQ_NEXT_NOTIFICATION;
}


/*
 * Repair IIS Logfile
 * written by Christian Swoboda
 */
static DWORD 
edir_isapi_filter_repair_log(HTTP_FILTER_CONTEXT *req_info, void *data)

{
    HTTP_FILTER_LOG *log_info;

 	log_info = (HTTP_FILTER_LOG *) data;

    if (strcmp(log_info->pszTarget, DEFAULT_ENHYDRA_HANDLER_URL)==0)
	{
	 log_info->pszTarget = req_info->pFilterContext;
	}

	return SF_STATUS_REQ_NEXT_NOTIFICATION;
}


/*
 * Functions that follow just do ISAPI bookkeeping.
 */

/**
 * Get EnhydraDirector filter version and options.
 * <P>
 * In the MFC implementation, this is a reasonable place to do filter-wide
 * initialization.  We initialize the thread-local storage index for
 * the per-request heap here, and try to read the configuration file
 * in order to set up the scoreboard.
 * 
 * @param pVer [in,out] Version information block in which to set version
 * 			information and option bits.
 * @return True if this filter should continue loading, or False if the
 * 	load should fail.
 */
BOOL WINAPI
GetFilterVersion(HTTP_FILTER_VERSION *vinfo)
{
    BOOL ret;
    ret = (BOOL) edir_isapi_filter_init(myInstanceHandle);
    if (ret) {
	/*
	 * Do some bookkeeping needed by IIS.
	 */
	vinfo->dwFilterVersion = HTTP_FILTER_REVISION;	
	strncpy(vinfo->lpszFilterDesc, "Enhydra Director Filter",
		SF_MAX_FILTER_DESC_LEN);
	// Clear the flags set by base class
	vinfo->dwFlags &= ~SF_NOTIFY_ORDER_MASK;
	// Set the flags we are interested in
	vinfo->dwFlags |=     SF_NOTIFY_ORDER_LOW
			    | SF_NOTIFY_SECURE_PORT
			    | SF_NOTIFY_NONSECURE_PORT
			    | SF_NOTIFY_PREPROC_HEADERS
				| SF_NOTIFY_LOG;
    }
    return ret;
}

/**
 * Event dispatcher for the EnhydraDirector ISAPI module.
 * <P>
 * The current implementation handles the SF_NOTIFY_READ_RAW_DATA and
 * SF_NOTIFY_END_OF_NET_SESSION events.
 * 
 * @param req_info [in,out] Pointer to a per-request context structure.
 * @param event    [in]     The event to handle.
 * @param data     [in]     Event-specific data.
 * @return	Integer status code.
 */
DWORD WINAPI
HttpFilterProc(HTTP_FILTER_CONTEXT *req_info, DWORD event, void *data)
{
    DWORD retcode = SF_STATUS_REQ_NEXT_NOTIFICATION;
    if (!Dead) {
	switch (event) {
	    case SF_NOTIFY_PREPROC_HEADERS:
		    edir_isapi_memory_create_heap(HEAP_NO_SERIALIZE);
			retcode = edir_isapi_filter_translator(req_info, data);
			edir_isapi_memory_delete_heap();
			break;
		case SF_NOTIFY_LOG:
			retcode = edir_isapi_filter_repair_log(req_info, data);
			break;

	}
    }
    return retcode;
}

/**
 * Function needed by ISAPI to get the instance handle for this DLL.
 * <P>
 * Does nothing specific to Enhydra Director.
 * 
 * @return The instance handle for this DLL.
 */
HINSTANCE __stdcall
AfxGetResourceHandle()
{
    return myInstanceHandle;
}

/**
 * Entry point for this DLL.
 * <P>
 * This is called by the ISAPI server to initialize the DLL.  It is the
 * first code executed when the DLL is loaded, and is a good place to
 * perform filter-wide initializations such as setting up the 
 * thread-local storage index for the per-request heap.
 * 
 * @param handle   [in] The instance handle for this DLL.
 * @param reason   [in] DLL access event code.  (i.e. "attach")
 * @param reserved [in] Reserved data pointer.
 * @return True if the DLL load should succeed, False if it should fail.
 */
BOOL WINAPI
DllMain(HINSTANCE handle, unsigned long reason, void *reserved)
{
    switch (reason) {
	case DLL_PROCESS_ATTACH:
	    myInstanceHandle = handle;
	    Dead = 0;
	    break;
	case DLL_PROCESS_DETACH:
	    if (globalHeap) HeapDestroy(globalHeap);
	    globalHeap = NULL;
	    refCounts = NULL;
	    numServers = 0;
	    Dead = 1;
	    break;
	default:
	    break;
    }
    return TRUE;
}

