/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_isapi_handler.c,v 1.1 2003/04/14 12:25:59 sinisa Exp $
 */

#include "edir.h"
#include "edir_isapi.h"
#include <httpext.h>

static void
edir_isapi_handle_header(edir_kvtable_t *httpVars,
			 edir_kvtable_t *cookies,
		         char *line, int32 len);

#define EDIR_ISAPI_MAX_URL_LEN	1024
#define EDIR_INBUFSIZE		4096
#define EDIR_OUTBUFSIZE		4096

/**
 * URI parse info structure.
 *
 * @member basePath     The path portion without any params or query string.
 * @member params       Url 'path parameters'.
 * @member sessionID    The value of the 'jsessionid' path parameter or NULL.
 * @member queryString  The value of the query string if it exists.
 */
typedef struct edir_isapi_uridata {
    char    *basePath;
    char    *params;
    char    *sessionID;
    char    *queryString;
} edir_isapi_uridata_t;


/**
 * Internal copy of this DLL's instance handle, needed by ISAPI.
 */
static HINSTANCE myInstanceHandle = NULL;

/*
 * Local 'am I detached?' flag.
 */
static int Dead = 1;

/**
 * The global memory heap.
 * <P>
 * Bytes allocated from this heap are not destroyed en masse until
 * this DLL is detached.
 */
static HANDLE globalHeap = NULL;

/**
 * Local copy of this DLL's full path name.
 */
static LPTSTR myDllName = NULL;

/**
 * Local copy of the full path of the config file we're using.
 */
static LPTSTR myConfig = NULL;

/**
 * Pointer to the global scoreboard.  This scoreboard follows the 
 * same format as the NSAPI and Apache modules.  It is position 
 * independent, even though it does not need to be.  This allows for
 * maintenance of the scoreboard access code in only one place, which
 * should facilitate increased robustness and reliablility across all
 * platforms.
 */
static edir_load_sb_offsets_t *scoreboard = NULL;

/**
 * Scoreboard mutex.
 */
static HANDLE scoreboardMutex = (HANDLE)0;

/**
 * Number of bytes allocated for the scoreboard.
 */
static int32 sbSize = 0;

/**
 * Whether debugging has been set from the filter.
 */
static BOOL debugSet = FALSE;

/**
 * Global reference count array to use instead of scoreboard itself.
 * <P>  
 * The scoreboard itself has a reference count for each server entry,
 * but the actual scoreboard access code takes an array of references
 * in order to allow flexible handling of reference counts in multi-
 * process environments such as Apache.
 * <P>
 * It turns out that it is more efficient and less complex to just keep
 * a separate array of reference counts in parallel to the scoreboard.
 * XXX FIX A better solution will be to rearrange the scoreboard to
 * store the reference counts in a contiguous array, as that would also
 * allow non-systemV refcount methods to implement simpler code.
 */
static int32 *refCounts = NULL;

/**
 * The number of servers (and thus reference counts).
 */
static int32 numServers = 0;

/**
 * Handle filter-wide initializations when the DLL is attached.
 * <P>
 * This function is called by <code>DllMain()</code> when the DLL is
 * attached by ISAPI.  This is a good place to perform global, filter-
 * wide initializations that are not freed until the DLL is unloaded.
 * 
 * @param handle [in] The instance handle for this DLL.
 * @return True if the DLL should load, False if it should fail to load.
 */
static int
edir_isapi_handler_init(HINSTANCE handle)
{
    static char		    *funcName = "edir_isapi_filter_init";
    edir_exception_t	    *_exc=NULL, **exc=&_exc;
    char		    *trace = NULL;
#define                     MODNAMESIZE 4096
    char		    moduleName[MODNAMESIZE];
    HANDLE		    tmpHeap = NULL;
    DWORD		    debugMask = 0;
    edir_conf_t		    *appCfg = NULL;

    // First real thing we ever do is register with the event log
    // so we can complain if anything else fails.
    GetModuleFileName(handle, moduleName, MODNAMESIZE);
    moduleName[MODNAMESIZE-1] = '\0';
    /*
     * Use unwrappered malloc here -- this is NEVER freed!
     * I have seen tlsAlloc() fail before, so I want to spew
     * a complaint to the NT event log if we fail for that reason.
     */
    myDllName = malloc(strlen(moduleName)+1);
    strcpy(myDllName, moduleName);
    edir_isapi_event_log_register(myDllName, "EnhydraHandler");

    // Initialize the EnhydraDirector heap system.
    if (!edir_isapi_memory_init()) {
	NTEventLog("Fatal error: Thread Local storage not available", 1);
	return FALSE; /* Fail to load DLL */
    }

    // Initialize a serialized global heap.
    // This heap is not freed until the DLL is unloaded.
    globalHeap = edir_isapi_memory_create_heap(0);
    if (globalHeap == NULL) {
	NTEventLog("Fatal error: Could not create global heap.", 1);
	return FALSE; /* Fail to load DLL */
    }

    /*
     * Create the scoreboard mutex in an unlocked state.
     * Initialization code uses this to ensure the scoreboard is
     * only loaded once.  Once initialized, this mutex provides
     * serialization for the reference counts.
     */
    if ((scoreboardMutex = CreateMutex(NULL, FALSE, NULL)) == NULL) {
	NTEventLog("Fatal error: Could not create scoreboard mutex.", 1);
	return FALSE;
    }
 
    return TRUE;
}

/**
 * Initialize the scoreboard during the first request.
 * <P>
 * Initialization of the scoreboard is done during a request so that
 * the filter can pass the name of the config file.  This allows
 * the config file to exist in the directory of the filter, while
 * the handler DLLs can just be placed where needed.
 *
 * @param exc [in,out] Exception handle filled in if an error occurs.
 * @exception EDIR_EXCEPT_CONFIG If an configuration error is detected.
 */
static void
edir_isapi_handler_init_scoreboard(edir_exception_t **exc)
{
    static char		    *funcName = "edir_isapi_handler_init_scoreboard";
    char		    *trace = NULL;
    DWORD		    debugMask = 0;
    edir_conf_t		    *appCfg = NULL;
    int			    i;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR")", funcName, *exc));

    /*
     * Try to open and process the enhydra_director.conf file.
     */
    appCfg = edir_conf_process(myConfig, exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * Set debugging mask if needed.
     */
    if (appCfg->debug) edir_debug |= appCfg->debugMask;

    /*
     * Allocate space for the scoreboard from the global memory
     * memory heap.  This memory is never freed until the DLL is
     * unloaded.  Note that we don't use edir_malloc because it's
     * set to the _temporary_ heap.
     */
    sbSize = edir_load_sb_compute_size(appCfg);
    scoreboard = (edir_load_sb_offsets_t *) HeapAlloc(globalHeap, 0, sbSize);

    /*
     * Convert the config data to a position-independent scoreboard.
     * We don't really need position-independent data here, but we
     * use it anyhow so that we can use the same scoreboard code
     * for all module implementations, including Apache.
     */
    edir_load_sb_from_config(appCfg, (void *)scoreboard);

    /*
     * Create a global reference count array.  For the filter, we
     * never touch the refcounts anyway so they will always be zero.
     */
    numServers = edir_load_sb_num_servers((void *)scoreboard);
    refCounts = (int32 *) HeapAlloc(globalHeap, 0, numServers * sizeof(int32));
    for (i=0; i<numServers; i++) refCounts[i] = 0;

    /* debug: print out the scoreboard contents */
    trace = edir_load_sb_debug_string((void *)scoreboard,
				refCounts, numServers);
    edir_log("%s\n%s:%s\nScoreboard:\n%s", 
	     "Enhydra Director Handler Initialized.", 
	     "Config File", myConfig, trace);
    edir_free(trace);

    return;

except:
    if (scoreboard) HeapFree(globalHeap, 0, scoreboard);
    scoreboard = NULL;
    if (refCounts) HeapFree(globalHeap, 0, refCounts);
    refCounts = NULL;
    numServers = 0;
    return;
}

/**
 * Function called by edir_parse_uri() to handle uri components
 * <P>
 * The generic URI parser calls this function when it has each of
 * various parts of the URI it is parsing.  This function places
 * the parsed components into the context structure for later
 * processing into standard CGI and HTTP variables.
 *
 * @param ptr  [in,out] Opaque pointer to module context structure.
 * @param type [in]     The type of component that was just parsed.
 * @param lhs  [in]     Left hand side of a 'key=value' URL parameter,
 *                      or the complete value of other URI parts.
 * @param rhs  [in]     Right hand side of a 'key=value' URI parameter
 *                      or NULL for other parts.
 */
static void
edir_isapi_uri_parse_func(void *ptr, int type, char *lhs, char *rhs)
{
    edir_isapi_uridata_t *uriData = (edir_isapi_uridata_t *) ptr;
    switch (type) {
	case EDIR_PARSE_URI_PATH_INFO:
	    if (uriData->basePath) edir_free(uriData->basePath);
	    uriData->basePath = edir_strdup(lhs);
	    break;
	case EDIR_PARSE_URI_PATH_PARAMS:
	    if (uriData->params) edir_free(uriData->params);
	    uriData->params = edir_strdup(lhs);
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_1:
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_2:
	    if (!edir_strcasecmp(lhs, "jsessionid")) {
		uriData->sessionID = edir_strdup(rhs);
	    }
	    break;
	case EDIR_PARSE_URI_QUERY_STRING:
	    uriData->queryString = edir_strdup(lhs);
	    break;
    }
}

/**
 * Send an error response to the client
 * <P>
 * Sends an HTTP header to the client with the specified status
 * code and status message.
 * 
 * @param req_info [in,out] The request context for the current request.
 * @param codeval  [in]     Numeric HTTP status code in string form.
 * @param msg      [in]     The text message to send along with the status.
 */
static void
SendClientError(EXTENSION_CONTROL_BLOCK *req_info,
		int32 code, char *msg)
{
    char	*header, *body, *sline;
    void	*sbuf;
    int32	bodyLen, headerLen, bytesSent, nbytes;
    sbuf = edir_sbuf_create();
    edir_sbuf_appendf(sbuf, "%03d %s", code, msg);
    sline = edir_sbuf_tostring(sbuf, NULL);
    edir_sbuf_clear(sbuf);
    edir_sbuf_appendf(sbuf, "<HTML>\r\n");
    edir_sbuf_appendf(sbuf, "<HEAD>\r\n");
    edir_sbuf_appendf(sbuf, "<TITLE>Error: %03d %s</TITLE>\r\n", code, msg);
    edir_sbuf_appendf(sbuf, "</HEAD>\r\n");
    edir_sbuf_appendf(sbuf, "<BODY>\r\n");
    edir_sbuf_appendf(sbuf, "<CENTER><H2>Error: %03d %s</CENTER></H2>\r\n", code, msg);
    edir_sbuf_appendf(sbuf, "</BODY>\r\n");
    edir_sbuf_appendf(sbuf, "</HTML>\r\n");
    body = edir_sbuf_tostring(sbuf, &bodyLen);
    edir_sbuf_clear(sbuf);
    edir_sbuf_appendf(sbuf, "Content-Type: text/html\r\n");
    edir_sbuf_appendf(sbuf, "Content-Length: %d\r\n", bodyLen);
    edir_sbuf_appendf(sbuf, "\r\n");
    header = edir_sbuf_tostring(sbuf, &headerLen);
    edir_sbuf_free(sbuf);
    (req_info->ServerSupportFunction)(req_info->ConnID,
		HSE_REQ_SEND_RESPONSE_HEADER,
		sline, (LPDWORD)headerLen, (LPDWORD)header);

    for (bytesSent = 0; bytesSent < bodyLen; bytesSent += nbytes) {
	nbytes = bodyLen - bytesSent;
	if (!(req_info->WriteClient)(req_info->ConnID,
			    body + bytesSent, &nbytes, 0)) break;
	if (nbytes < 1) break;
    }
    
    edir_free(header);
    edir_free(body);
    edir_free(sline);
}

/**
 * Handler function for the Enhydra Director ISAPI client.
 *
 * @param req_info [in,out] Extension control block containing the
 *	    information for the current request.
 * @return An integer status code indicating the success of failure
 *	    of the current requet, and whether to honor keepalives.
 */
static DWORD
edir_isapi_handler(EXTENSION_CONTROL_BLOCK *req_info)
{
    static char		    *funcName = "edir_isapi_handler";
    edir_exception_t	    *_exc=NULL, **exc=&_exc;
    char		    inBuf[EDIR_INBUFSIZE], outBuf[EDIR_OUTBUFSIZE];
    char		    *cp, *dp, *ep, *key, *val, **vals;
    char		    *url = NULL, *stickyTag = NULL;
    char		    *virt_host = NULL, *doc_root = NULL;
    char		    *path_info = NULL, *request_uri = NULL;
    DWORD		    nbytes;
    int32		    len, nvals, i, iv;
    struct sockaddr_in	    local_addr, remote_addr;
    edir_kvtable_t	    *cookies, *httpVars, *cgiVars;
    edir_isapi_uridata_t    uriData;
    edir_load_sb_app_t	    *app = NULL;
    edir_load_sb_server_t   *server = NULL;
    edir_connection_t	    _conn, *conn=NULL;
    time_t		    now, startTime, lastTime, elapsedTime;
    int32		    stickyRetries = 0, refIndex = -1, do_status=0;
    int32		    contentLeft, bytesSent, remain;
    int32		    statusCode = 500, headerSent = 0;
    int32		    retStatus = HSE_STATUS_ERROR;
    int32		    keepAlive = FALSE;
    char		    *statusMessage = "Server Error";
    char		    *statusline = NULL, *headers = NULL;
    void		    *sbuf = NULL;
    BOOL		    connAborted = FALSE;
    memset(&uriData, 0, sizeof(edir_isapi_uridata_t));

    cookies = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
    httpVars = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
    cgiVars = edir_kvtable_create(EDIR_KVTABLE_TO_UPPER);

    /*
     * Get HTTP headers
     */
    nbytes = sizeof(inBuf);
    (req_info->GetServerVariable)(req_info->ConnID, "ALL_HTTP", inBuf, &nbytes);
    cp = inBuf;
    for (ep = cp + nbytes; *cp && (cp < ep); cp = dp) {
	for (dp=cp; *dp && (*dp != '\n'); dp++) ;
	len = dp - cp;
	if ((dp > cp) && (*(dp-1) == '\r')) { len--; *(dp-1) = '\0'; }
	if (*dp) *dp++ = '\0';
	edir_isapi_handle_header(httpVars, cookies, cp, len);
    }

    if (!debugSet) {
	if (val = edir_kvtable_find(httpVars, "edir-debug")) {
	    if (edir_atoi(val, &iv)) {
		edir_debug |= (uint32) iv;
	    }
	}
	debugSet = TRUE;
    }

    /*
     * Check for URL and config file name passed to us by the filter.
     */
    url = edir_kvtable_find(httpVars, "edir-url");
    if (!url) {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET,
		    "Filter 'Edir-Url' hint missing.");
	EDIR_THROW(*exc);
    }
    cp = (char *) _alloca(strlen(url) + 1);
    strcpy(cp, url);
    url = cp;

    /*
     * See if we matched a 'status' URL in the filter.
     */
    do_status = FALSE;
    if (val = edir_kvtable_find(httpVars, "edir-status")) {
	if (!edir_strcasecmp(val, "true")) do_status = TRUE;
    }

    /*
     * Initialize the scoreboard if it has not already been set up.
     */
    if (scoreboard == NULL) {
	if (WaitForSingleObject(scoreboardMutex, INFINITE)
				!= WAIT_OBJECT_0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't lock scoreboard mutex: @laststr@ (errnum=@lasterr@)");
	    EDIR_THROW(*exc);
	}

	myConfig = edir_kvtable_find(httpVars, "edir-conf");
	if (!myConfig) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SERVLET,
		    "Filter 'Edir-Conf' hint missing.");
	    ReleaseMutex(scoreboardMutex);
	    EDIR_THROW(*exc);
	}

	/* This is global, so allocate it from the global heap */
	cp = HeapAlloc(globalHeap, 0, strlen(myConfig) + 1);
	strcpy(cp, myConfig);
	myConfig = cp;

	/* Check again in case another thread beat us to the draw */
	if (scoreboard == NULL) {
	    edir_isapi_handler_init_scoreboard(exc);
	    if (*exc) {
		ReleaseMutex(scoreboardMutex);
		EDIR_THROW(*exc);
	    }
	}
	ReleaseMutex(scoreboardMutex);
    }

    edir_kvtable_remove(httpVars, "edir-conf");
    edir_kvtable_remove(httpVars, "edir-url");
    edir_kvtable_remove(httpVars, "edir-debug");
    edir_kvtable_remove(httpVars, "edir-status");

    /*
     * Get the remote and local connection parameters so we
     * can do a more careful check, taking server and client
     * restrictions into account.
     */
    /* REMOTE_ADDR (Port is zero - we don't care about it) */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "REMOTE_ADDR",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (!edir_ip_addr(inBuf, &(remote_addr.sin_addr))) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
		    "Can't translate remote address: '%s'", inBuf);
	    EDIR_THROW(*exc);
	}
    } else {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
		    "Can't obtain remote address.");
	EDIR_THROW(*exc);
    }
    remote_addr.sin_port = 0;
    remote_addr.sin_family = AF_INET;
    edir_kvtable_add(cgiVars, "REMOTE_ADDR", inBuf);

    /* REMOTE_PORT - optional */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "REMOTE_PORT",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (edir_atoi(inBuf, &iv)) {
	    remote_addr.sin_port = htons((unsigned short) iv);
	    edir_kvtable_add(cgiVars, "REMOTE_PORT", inBuf);
	}
    }
    local_addr.sin_family = AF_INET;

    /* REMOTE_HOST */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "REMOTE_HOST",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
    } else {
	edir_sockaddr_to_str(&remote_addr, inBuf);
    }
    edir_kvtable_add(cgiVars, "REMOTE_HOST", inBuf);
    
    /* SERVER_ADDR */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "LOCAL_ADDR",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (!edir_ip_addr(inBuf, &(local_addr.sin_addr))) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
		    "Can't translate server address: '%s'", inBuf);
	    EDIR_THROW(*exc);
	}
    } else {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
		    "Can't obtain server address.");
	EDIR_THROW(*exc);
    }
    edir_kvtable_add(cgiVars, "SERVER_ADDR", inBuf);
    
    /* SERVER_PORT */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "SERVER_PORT",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (!edir_atoi(inBuf, &iv)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
		    "Can't translate server port: '%s'", inBuf);
	    EDIR_THROW(*exc);
	}
	local_addr.sin_port = htons((unsigned short) iv);
    } else {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
		    "Can't obtain server port.");
	EDIR_THROW(*exc);
    }
    edir_kvtable_add(cgiVars, "SERVER_PORT", inBuf);
    local_addr.sin_family = AF_INET;
    
    /* SERVER_NAME */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "SERVER_NAME",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
    } else {
	edir_sockaddr_to_str(&local_addr, inBuf);
    }
    edir_kvtable_add(cgiVars, "SERVER_NAME", inBuf);
    if (val = edir_kvtable_find(httpVars, "host")) {
	virt_host = edir_strdup(val);
    } else {
	virt_host = edir_strdup(inBuf);
    }
    if (cp = strchr(virt_host, ':')) *cp = '\0';

    /* AUTH_TYPE */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "AUTH_TYPE",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (inBuf[0]) {
	    edir_kvtable_add(cgiVars, "AUTH_TYPE", inBuf);
	}
    }

    /* AUTH_USER */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "AUTH_USER",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (inBuf[0]) {
	    edir_kvtable_add(cgiVars, "AUTH_USER", inBuf);
	}
    }

    /* REMOTE_USER */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "REMOTE_USER",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (inBuf[0]) {
	    edir_kvtable_add(cgiVars, "REMOTE_USER", inBuf);
	}
    }

    /* REQUEST_METHOD */
    edir_kvtable_add(cgiVars, "REQUEST_METHOD", req_info->lpszMethod);

    /* SERVER_PROTOCOL */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "SERVER_PROTOCOL",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (inBuf[0]) {
	    edir_kvtable_add(cgiVars, "SERVER_PROTOCOL", inBuf);
	}
    }

    /* SERVER_SCHEME */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "SERVER_PORT_SECURE",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	if (edir_atoi(inBuf, &iv) && (iv)) {
	    edir_kvtable_add(cgiVars, "SERVER_SCHEME", "https");
	} else {
	    edir_kvtable_add(cgiVars, "SERVER_SCHEME", "http");
	}
    }

    /* DOCUMENT_ROOT - needed for PATH_TRANSLATED */
    nbytes = sizeof(inBuf) - 1;
    if ((req_info->GetServerVariable)(req_info, "DOCUMENT_ROOT",
				      inBuf, &nbytes) ) {
	inBuf[nbytes] = '\0';
	doc_root = edir_strdup(inBuf);
	cp = strrchr(doc_root, '\\');
	if (cp && (*(cp + 1) == '\0')) *cp = '\0';
    } else doc_root = NULL;

    /*
     * Now, parse the URL to get the base URL, query string,
     * session ID and params, etc.
     */
    /* Add the "REQUEST_URI" parameter */
    edir_parse_uri(url, &uriData, edir_isapi_uri_parse_func);
    if (uriData.basePath) {
	if (uriData.params) {
	    edir_snprintf(inBuf, sizeof(inBuf)-1, "/%s;%s", 
			  uriData.basePath, uriData.params);
	} else {
	    edir_snprintf(inBuf, sizeof(inBuf)-1, "/%s",
			  uriData.basePath);
	}
    } else {
	if (uriData.params) {
	    edir_snprintf(inBuf, sizeof(inBuf)-1, "/;%s", 
			  uriData.params);
	} else {
	    edir_snprintf(inBuf, sizeof(inBuf)-1, "/");
	}
    }
    edir_kvtable_set(cgiVars, "REQUEST_URI", inBuf);
    request_uri = edir_kvtable_find(cgiVars, "REQUEST_URI");

    /* Add the "QUERY_STRING" parameter */
    if (uriData.queryString) {
	edir_kvtable_set(cgiVars, "QUERY_STRING", uriData.queryString);
    }

    /*
     * Set up the session ID.  URL rewriting takes precedent
     * over cookies.
     */
    if (uriData.sessionID == NULL) {
	val = edir_kvtable_find(cookies, "JSESSIONID");
	if (val) {
	    uriData.sessionID = edir_strdup(val);
	}
    }
    
    /*
     * See if we have a 'sticky session' tag.
     */
    if (uriData.sessionID) {
	if (cp = strchr(uriData.sessionID, '(')) {
	    cp++;
	    stickyTag = (char *)_alloca(9);
	    strncpy(stickyTag, cp, 8);
	    stickyTag[8] = '\0';
	    stickyRetries = EDIR_CONNECT_STICKY_SESSION_RETRIES;
	}
    }

#ifdef DEBUG

    edir_log_debug("HTTP Headers:");
    vals = edir_kvtable_first(httpVars, &key, &nvals);
    while (vals) {
	for (i=0; i<nvals; i++) {
	    edir_log_debug("%s='%s'", key, vals[i]);
	}
	vals = edir_kvtable_next(httpVars, &key, &nvals);
    }

    edir_log_debug("CGI Headers:");
    vals = edir_kvtable_first(cgiVars, &key, &nvals);
    while (vals) {
	for (i=0; i<nvals; i++) {
	    edir_log_debug("%s='%s'", key, vals[i]);
	}
	vals = edir_kvtable_next(cgiVars, &key, &nvals);
    }

    edir_log_debug("Cookies:");
    vals = edir_kvtable_first(cookies, &key, &nvals);
    while (vals) {
	for (i=0; i<nvals; i++) {
	    edir_log_debug("%s='%s'", key, vals[i]);
	}
	vals = edir_kvtable_next(cookies, &key, &nvals);
    }

    edir_log_debug("Session Tag: %s", stickyTag ? stickyTag : "(null)");
#endif
    
    
    /*
     * Check for basic request sanity.  We don't support chunked content
     * encoding, so if the request is a POST, we require Content-Length
     * to be specified.
     * NOTE We handle chunked output to the browser just fine -- we
     * don't even know it's there.  We just don't parse input (yet).
     * XXX - FIX - Need to support this.
     */
    if (!edir_strcasecmp(req_info->lpszMethod, "POST")) {
	val = edir_kvtable_find(httpVars, "Content-Length");
	if ((val == NULL) || (!edir_atoi(val, &iv))) {
	    statusCode = 411;
	    statusMessage = "Content Length Required";
	    *exc = NULL; /* Not an exception - but abort anyway */
	    EDIR_THROW(*exc);
	}
	contentLeft = iv;
    } else contentLeft = 0;


    if (!do_status) {
	/*
	 * Now, lock the global scoreboard mutex
	 */
/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	if (WaitForSingleObject(scoreboardMutex, INFINITE) != WAIT_OBJECT_0) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't lock scoreboard mutex: @laststr@ (errnum=@lasterr@)");
	    EDIR_THROW(*exc);
	}

	/*
	 * Select a server.
	 */
	edir_load_choose(scoreboard, request_uri, stickyTag, 
			 &app, &server, refCounts, numServers, 
			 &local_addr, &remote_addr, virt_host, exc);
	if (*exc) {
	    ReleaseMutex(scoreboardMutex);
	    EDIR_THROW(*exc);
	}
	EDIR_DEBUG_NORMAL(("Got server: %s:%d", server->hostName, \
			    server->port));
	server->useCount++;
    
	/*
	 * Bump the reference count.
	 */
	refIndex = server->refIndex;
	refCounts[refIndex]++;

	/*
	 * Unlock the scoreboard.
	 */
	if (!ReleaseMutex(scoreboardMutex)) {
	    *exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		"Can't unlock scoreboard mutex: @laststr@ (errnum=@lasterr@)");
	    EDIR_THROW(*exc);
	}
/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
	path_info = request_uri + app->appPrefixLen + 1;
	if (*path_info) {
	    edir_kvtable_set(cgiVars, "PATH_INFO", path_info);
	} else {
	    edir_kvtable_remove(cgiVars, "PATH_INFO");
	}
	if (doc_root && *path_info) {
	    _snprintf(outBuf, EDIR_OUTBUFSIZE-1, "%s\\%s",
		      doc_root, path_info);
	    for (cp=outBuf; *cp; cp++) if (*cp == '/') *cp = '\\';
	    edir_kvtable_set(cgiVars, "PATH_TRANSLATED", outBuf);
	} else {
	    edir_kvtable_remove(cgiVars, "PATH_TRANSLATED");
	}
    }

    /*
     * Connect to the server.
     */
    if (!(edir_debug & EDIR_DEBUG_MASK_FAKE_IT) && !do_status) {
	time(&startTime);
	lastTime = startTime;
	elapsedTime = 0;
	while (elapsedTime <= EDIR_CONNECT_TIMEOUT) {
	    if (*exc) edir_exception_free(exc);
	    conn = &_conn;
	    memset(conn, 0, sizeof(edir_connection_t));
	    edir_client_connect(conn, NULL, &(server->sin),
		    server->authData, outBuf, EDIR_OUTBUFSIZE,
		    inBuf, EDIR_INBUFSIZE, exc);
	    if (!(*exc)) break; /* Connection established! */
	    if ((*exc)->id != EDIR_EXCEPT_IO_ERROR) EDIR_THROW(*exc);

	    /* XXX Should have 'logmask' */
	    if (edir_debug & EDIR_DEBUG_MASK_NORMAL) {
		char *trace = NULL;
		if (*exc) trace = edir_exception_stack_trace(*exc);
		edir_log_error("Warning: Error connecting to Enhydra.");
		edir_log_error("%s", trace ? trace : "No details.");
		if (trace) edir_free(trace);
	    }
	    if (*exc) edir_exception_free(exc);

	    /* Mark this server as 'failing' */
/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	    if (WaitForSingleObject(scoreboardMutex, INFINITE) 
					    != WAIT_OBJECT_0) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't lock scoreboard mutex: @laststr@ (errnum=@lasterr@)");
		EDIR_THROW(*exc);
	    }
	    
	    if (!(server->failing)) {
		server->failing = TRUE;
		time(&(server->failTime));
	    }
	    server->failCount++;

	    /*
	     * If this request has a 'sticky session' token in the session
	     * ID then we want to try several times to go to the desired
	     * server.  After a number of attempts, though, we give up and
	     * fall back to round robin.  
	     */
	    if (stickyRetries <= 0) stickyTag = NULL;
	    else stickyRetries--;
	    refCounts[refIndex]--;
	    refIndex = -1;
	    server->useCount--;

	    /*
	     * Select a server.
	     */
	    edir_load_choose(scoreboard, request_uri, stickyTag, 
			    &app, &server, refCounts, numServers, 
			    &local_addr, &remote_addr, virt_host, exc);
	    if (*exc) {
		ReleaseMutex(scoreboardMutex);
		EDIR_THROW(*exc);
	    }
	    EDIR_DEBUG_NORMAL(("Got alternate server: %s:%d", \
			    server->hostName, server->port));
	    server->useCount++;
    
	    /*
	     * Bump the reference count.
	     */
	    refIndex = server->refIndex;
	    refCounts[refIndex]++;

	    /*
	     * Unlock the scoreboard.
	     */
	    if (!ReleaseMutex(scoreboardMutex)) {
		*exc = edir_exception_create(EDIR_EXCEPT_MUTEX,
		    "Can't unlock scoreboard mutex: @laststr@ (errnum=@lasterr@)");
		EDIR_THROW(*exc);
	    }
/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
	    path_info = request_uri + app->appPrefixLen + 1;
	    if (*path_info) {
		edir_kvtable_set(cgiVars, "PATH_INFO", path_info);
	    } else {
		edir_kvtable_remove(cgiVars, "PATH_INFO");
	    }
	    if (doc_root && *path_info) {
		_snprintf(outBuf, EDIR_OUTBUFSIZE-1, "%s\\%s",
		      doc_root, path_info);
		for (cp=outBuf; *cp; cp++) if (*cp == '/') *cp = '\\';
		edir_kvtable_set(cgiVars, "PATH_TRANSLATED", outBuf);
	    } else {
		edir_kvtable_remove(cgiVars, "PATH_TRANSLATED");
	    }
	    
	    /* Avoid spinning wildly on errors */
	    time(&now);
	    if (EDIR_TIMEDIFF(now, lastTime) < 1) {
		Sleep(1000);
		time(&now);
	    }

	    lastTime = now;
	    elapsedTime = EDIR_TIMEDIFF(now, startTime);
	} /* While not timed out */

	/*
	 * If we couldn't get a connection in time, fail.
	 */
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Check for case of timing out after freeing exception.
	 * XXX There's got to be a better way.
	 */
	if (conn->state != EDIR_STATE_OPEN) {
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR, 
		    "Can't connect to any Enhydra server for %d %s",
		    EDIR_CONNECT_TIMEOUT, "seconds -- giving up.");
	    EDIR_THROW(*exc);
	}

	/* Send standard CGI headers */
	vals = edir_kvtable_first(cgiVars, &key, &nvals);
	while (vals != NULL) {
	    for (i=0; i<nvals; i++) {
		len = strlen(key);
		edir_client_send_header(conn, key, len, vals[i],
		    strlen(vals[i]), EDIR_PKT_SUBTYPE_HEADER_HEADER,
		    outBuf, EDIR_OUTBUFSIZE, inBuf, EDIR_INBUFSIZE, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(cgiVars, &key, &nvals);
	}

	/* Send standard HTTP headers */
	vals = edir_kvtable_first(httpVars, &key, &nvals);
	while (vals != NULL) {
	    for (i=0; i<nvals; i++) {
		len = strlen(key);
		edir_client_send_header(conn, key, len, vals[i],
		    strlen(vals[i]), EDIR_PKT_SUBTYPE_HEADER_HTTP,
		    outBuf, EDIR_OUTBUFSIZE, inBuf, EDIR_INBUFSIZE, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(httpVars, &key, &nvals);
	}

	/* Send cookies */
	vals = edir_kvtable_first(cookies, &key, &nvals);
	while (vals != NULL) {
	    for (i=0; i<nvals; i++) {
		len = strlen(key);
		edir_client_send_header(conn, key, len, vals[i],
		    strlen(vals[i]), EDIR_PKT_SUBTYPE_HEADER_COOKIE,
		    outBuf, EDIR_OUTBUFSIZE, inBuf, EDIR_INBUFSIZE, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(cookies, &key, &nvals);
	}

	/* Send 'last header' indication */
	edir_client_send_header(conn,"EDIR_CLIENT",sizeof("EDIR_CLIENT")-1,
				"ISAPI", sizeof("ISAPI") - 1, 
				EDIR_PKT_SUBTYPE_HEADER_LAST,
				outBuf, EDIR_OUTBUFSIZE,
				inBuf, EDIR_INBUFSIZE, exc); 
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Now send the request entity, if any 
	 */
	if (contentLeft > 0) {
	    char *data;
	    /* Send the first buffer */
	    len = (int32) req_info->cbAvailable;
	    data = (char *) req_info->lpbData;
	    for (bytesSent=0; bytesSent < len; bytesSent += nbytes) {
		remain = len - bytesSent;
		nbytes = EDIR_MIN(remain, EDIR_OUTBUFSIZE);
		memcpy(outBuf, data + bytesSent, nbytes);
		edir_client_send_data(conn, nbytes, FALSE, outBuf, 
			EDIR_OUTBUFSIZE, inBuf, EDIR_INBUFSIZE, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    contentLeft -= req_info->cbAvailable;
	    /* Send any remaining data from the input stream */
	    while (contentLeft > 0) {
		nbytes = EDIR_MIN(contentLeft, EDIR_OUTBUFSIZE);
		if (!(req_info->ReadClient)(req_info->ConnID,
		    outBuf, &nbytes)) {
		    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
				"Error reading from client: @laststr@");
		    EDIR_THROW(*exc);
		}
		edir_client_send_data(conn, nbytes, FALSE, outBuf, 
			EDIR_OUTBUFSIZE, inBuf, EDIR_INBUFSIZE, exc);
		if (*exc) EDIR_THROW(*exc);
		contentLeft -= nbytes;
		if (nbytes < 1) break; /* Client closed - not error */
	    }
	}

	/*
	 * Send the 'end of data' packet.
	 */
	edir_client_send_data(conn, 0, TRUE, 
			      outBuf, EDIR_OUTBUFSIZE,
			      inBuf, EDIR_INBUFSIZE, exc);
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Receive response headers.
	 */
	sbuf = edir_sbuf_create();
	while (conn->state == EDIR_STATE_RECEIVE_HEADERS) {
	    if (conn->packetSubType == EDIR_PKT_SUBTYPE_HEADER_COOKIE) {
		edir_sbuf_bincatn(sbuf, "Set-Cookie: ", 12);
		edir_sbuf_bincat(sbuf, conn->headerName, 
			         conn->headerNameLen);
		edir_sbuf_bincatn(sbuf, "=", 1);
		edir_sbuf_bincat(sbuf, conn->headerValue, 
			         conn->headerValueLen);
		edir_sbuf_bincatn(sbuf, "\r\n", 2);
	    } else {
		key = conn->headerName;
		nbytes = conn->headerNameLen;
		if (!edir_strncasecmp(key, "Status", nbytes)) {
		    if (statusline) edir_free(statusline);
		    statusline = edir_malloc(conn->headerValueLen + 1);

		    strncpy(statusline, conn->headerValue, 
			    conn->headerValueLen);
		    statusline[conn->headerValueLen] = '\0';
		} else if (!edir_strncasecmp(key, "Connection", nbytes)) {
		    /* See if we want keepalive */
		    len = EDIR_MIN(sizeof("keep-alive"), conn->headerValueLen);
		    if (len >= sizeof("keep-alive")) {
			if (!edir_strncasecmp(conn->headerValue,
			    "keep-alive", len)) {
			    keepAlive = TRUE;
			}
		    }
		    edir_sbuf_bincat(sbuf, key, nbytes); 
		    edir_sbuf_bincatn(sbuf, ": ", 2);
		    edir_sbuf_bincat(sbuf, conn->headerValue, 
			             conn->headerValueLen);
		    edir_sbuf_bincatn(sbuf, "\r\n", 2);
		} else {
		    edir_sbuf_bincat(sbuf, key, nbytes); 
		    edir_sbuf_bincatn(sbuf, ": ", 2);
		    edir_sbuf_bincat(sbuf, conn->headerValue, 
			             conn->headerValueLen);
		    edir_sbuf_bincatn(sbuf, "\r\n", 2);
		}
	    }
	    edir_client_next_header(conn, outBuf, EDIR_OUTBUFSIZE,
		                    inBuf, EDIR_INBUFSIZE, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	headers = NULL;
	len = 0;
	if (edir_sbuf_length(sbuf) > 0) {
	    edir_sbuf_bincatn(sbuf, "\r\n", 2);
	    headers = edir_sbuf_tostring(sbuf, &len);
	    edir_sbuf_free(sbuf);
	    sbuf = NULL;
	} else {
	    edir_sbuf_free(sbuf);
	    sbuf = NULL;
	    len = 0;
	}

	/*
	 * Send the response header using the appropriate ISAPI CB.
	 */
	headerSent = 1;
	EDIR_DEBUG_VERBOSE(("Sending status: %s", \
			    statusline ? statusline : "200 OK"));
	EDIR_DEBUG_VERBOSE(("Sending headers: %s", headers));
	if (!(req_info->ServerSupportFunction)(req_info->ConnID, 
		    HSE_REQ_SEND_RESPONSE_HEADER, statusline, 
		    (LPDWORD)&len, (LPDWORD)headers)) {
	    if ((WSAGetLastError() == WSAECONNABORTED) ||
		(WSAGetLastError() == WSAECONNRESET)) {
		connAborted = TRUE;
	    } else {
		*exc = edir_exception_create(EDIR_EXCEPT_SERVLET,
		    "Error sending response header: @laststr@ (errnum=@lasterr@)");
		EDIR_THROW(*exc);
	    }
	}
	if (headers) edir_free(headers);
	headers = NULL;
	if (statusline) edir_free(statusline);
	statusline = NULL;

	/*
	 * Send the response entity.
	 */
	while ((conn->state == EDIR_STATE_RECEIVE_RESPONSE) ||
	       (conn->state == EDIR_STATE_RECEIVE_RESPONSE_PACKET)) {
	    if (connAborted) break;
	    len = conn->receiveDataLen;
	    if (!(req_info->WriteClient)(req_info->ConnID, 
					 conn->receiveData,
				         &len, 0)) {
		/* Don't complain about write fails because they
		 * happen all the time when the user aborts requests 
		 */
		EDIR_DEBUG_VERBOSE(("%s: @laststr@ (errnum=@lasterr@)", \
				    "Error Sending Response Entity"));
		connAborted = TRUE;
	    }
	    edir_client_next_data(conn, outBuf, EDIR_OUTBUFSIZE,
			          inBuf, EDIR_INBUFSIZE, exc);
	    if (*exc) EDIR_THROW(*exc);
	}

	/*
	 * Close the connection to Enhydra
	 */
	if (conn->state != EDIR_STATE_OPEN) {
	    EDIR_DEBUG_NORMAL(("Connection not in OPEN state after %s", \
			  "successful request."));
	    edir_client_abort(conn, outBuf, EDIR_OUTBUFSIZE, exc);
	} else {
	    EDIR_DEBUG_VERBOSE(("Request finished.  Closing connection."));
	    edir_client_close(conn, outBuf, EDIR_OUTBUFSIZE, exc);
	}
	conn = NULL;
	if (*exc) EDIR_THROW(*exc);

	/*
	 * Decrement the reference count on the Enhydra server entry
	 */
	if (refIndex >= 0) {
	    WaitForSingleObject(scoreboardMutex, INFINITE);
	    refCounts[refIndex]--;
	    refIndex = -1;
	    ReleaseMutex(scoreboardMutex);
	}

	/*
	 * Set the return status.
	 */
	if (keepAlive) retStatus = HSE_STATUS_SUCCESS_AND_KEEP_CONN;
	else retStatus = HSE_STATUS_SUCCESS;


	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", \
			funcName, "TRUE"));
	return TRUE;
    } else { /* edir_debug & EDIR_DEBUG_MASK_FAKE_IT */
	void *sbuf;
	edir_kvtable_t *respHeaders;
	char *sline, *header, *body, *conType, **vals;
	int32 headerLen, bodyLen, nvals;
	respHeaders = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	sbuf = edir_load_status(scoreboard, request_uri, refCounts, 
				numServers, respHeaders, exc);
	if (*exc) EDIR_THROW(*exc);
	headerSent = TRUE;

	sline = edir_kvtable_find(respHeaders, "status");
	if (!val) sline = "200 OK";
	edir_kvtable_remove(respHeaders, "status");

	body = edir_sbuf_tostring(sbuf, &bodyLen);
	bodyLen = strlen(body);

	edir_sbuf_clear(sbuf);
	conType = edir_kvtable_find(respHeaders, "content-type");
	edir_sbuf_appendf(sbuf, "Content-Type: %s\r\n",
			  conType ? conType : "text/html");
	edir_kvtable_remove(respHeaders, "content-type");
	edir_kvtable_remove(respHeaders, "content-length");
	edir_sbuf_appendf(sbuf, "Content-Length: %d\r\n", bodyLen);
	
	vals = edir_kvtable_first(respHeaders, &key, &nvals);
	while (vals) {
	    edir_sbuf_appendf(sbuf, "%s: %s\r\n", key, vals[nvals-1]);
	    vals = edir_kvtable_next(respHeaders, &key, &nvals);
	}
	edir_sbuf_appendf(sbuf, "\r\n");
	edir_kvtable_destroy(respHeaders);

	header = edir_sbuf_tostring(sbuf, &headerLen);
	edir_sbuf_free(sbuf);

	EDIR_DEBUG_EVERYTHING(("Before sending header"));
	(req_info->ServerSupportFunction)(req_info->ConnID, 
		    HSE_REQ_SEND_RESPONSE_HEADER,
		    sline, (LPDWORD)&headerLen, (LPDWORD)header);

	EDIR_DEBUG_EVERYTHING(("Before sending body"));
	for (bytesSent = 0; bytesSent < bodyLen; bytesSent += nbytes) {
	    nbytes = bodyLen - bytesSent;
	    if (!(req_info->WriteClient)(req_info->ConnID,
				    body + bytesSent, &nbytes, 0)) break;
	    
	    EDIR_DEBUG_EVERYTHING(("Sent %d: %d of %d bytes", nbytes, bytesSent, bodyLen));
	    if (nbytes < 1) break;
	}
	EDIR_DEBUG_EVERYTHING(("Done sending body."));

	edir_free(header);
	edir_free(body);

	return HSE_STATUS_SUCCESS;
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns %s", funcName, retStatus));
    
except:
    if (*exc) {
	char *trace = edir_exception_stack_trace(*exc);
	edir_log_error("%s", trace);
	edir_free(trace);
	trace = NULL;
	edir_exception_free(exc);
	if (conn) {
	    edir_client_abort(conn, outBuf, EDIR_OUTBUFSIZE, exc);
	    if (*exc) edir_exception_free(exc);
	    conn = NULL;
	}
	if (refIndex >= 0) {
	    WaitForSingleObject(scoreboardMutex, INFINITE);
	    refCounts[refIndex]--;
	    ReleaseMutex(scoreboardMutex);
	}
	if (!headerSent) {
	    SendClientError(req_info, statusCode, statusMessage);
	}
	retStatus = HSE_STATUS_ERROR;
    }
    if (cookies) edir_kvtable_destroy(cookies);
    if (httpVars) edir_kvtable_destroy(httpVars);
    if (cgiVars) edir_kvtable_destroy(cgiVars);
    if (uriData.basePath) edir_free(uriData.basePath);
    if (uriData.params) edir_free(uriData.params);
    if (uriData.sessionID) edir_free(uriData.sessionID);
    if (uriData.queryString) edir_free(uriData.queryString);
    if (sbuf) edir_sbuf_free(sbuf);
    if (headers) edir_free(headers);
    if (statusline) edir_free(statusline);
    return retStatus;
}

/**
 * Handle a single request header line in the incoming HTTP request header.
 * <P>
 * Parse the request header line into a key and a value, and stuff it
 * into the appropriate key-value table.  Look out for cookies and
 * Accept headers and perform further parsing.  Also look for Content-Length
 * and Content-Encoding to help determine whether to receive POST data.
 * 
 * @param req_info [in,out] Pointer to a per-request context structure.
 * @param data     [in]     Event-specific data.
 * @param ctx      [in,out] EnhydraDirector private per-request context.
 * @return <code>TRUE</code> if successful, or <code>FALSE</code>
 *	    if there is an error.
 */
static void
edir_isapi_handle_header(edir_kvtable_t *httpVars,
			 edir_kvtable_t *cookies,
		         char *line, int32 len)
{
    static char		*funcName = "edir_isapi_handle_header";
    char		*cp, *dp, *ep, *name, *value, *cname, *cval;
    char		*upname;
    int			done;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR", "EDIR_PTR", \"%s\", %d)", \
	funcName, httpVars, cookies, line ? line : "(null)", len));
    ep = line + len;
    for (cp=line; (cp < ep) && isspace(*cp); ) cp++; /* Leading WS */
    name = cp;
    while ((cp < ep) && (*cp != ':')) cp++;           /* Header name */
    if (cp < ep) {
	*cp++ = '\0';
	while ((cp < ep) && isspace(*cp)) cp++;          /* WS again. */
	value = cp;
	for (cp = ep - 1; (cp >= value) && isspace(*cp); ) *cp-- = '\0';
	/* Convert CGI header names to 'normal' HTTP header names */
	upname = name;
	if (*upname) {
	    *upname = toupper(*upname);
	    upname++;
	}
	while (*upname) {
	    *upname = tolower(*upname);
	    if (*upname == '_') {
		*upname++ = '-';
		if (*upname) *upname = toupper(*upname);
	    }
	    if (*upname) upname++;
	}
	if (!edir_strncasecmp(name, "http-", 5)) name += 5;
	upname = name;
	
	if (!edir_strcasecmp(name, "Cookie")) {
	    for (cp=value, done=0; !done; cp = ep) {
		ep = strchr(cp, ';');
		if (ep == NULL) {
		    for (ep=cp; *ep; ) ep++;
		    done=1;
		}
    		for (dp=ep - 1; (dp >= cp) && isspace(*dp); ) *dp-- = '\0';
		*ep++ = '\0';
		while (*cp && isspace(*cp)) cp++;
		cname = cp;
		cval = strchr(cname, '=');
		if (cval) {
		    for (dp=cval - 1; (dp >= cname) && isspace(*dp); ) {
			*dp-- = '\0';
		    }
		    *cval++ = '\0';
		    while (isspace(*cval)) cval++;
		} else cval = "";
		edir_kvtable_add(cookies, cname, cval);
	    }
	} else {
	    edir_kvtable_add(httpVars, upname, value);
	}
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/*
 * Functions that follow just do ISAPI bookkeeping.
 */

/**
 * Get EnhydraDirector extension version and options.
 * <P>
 * This is where we set up the scoreboard initialization mutex and the
 * global memory heap.
 * 
 * @param vinfo [in,out] Version information block in which to set version
 * 			 information and extension name.
 * @return True if this extension should continue loading, or False if the
 * 	load should fail.
 */
BOOL WINAPI
GetExtensionVersion(HSE_VERSION_INFO *vinfo)
{
    BOOL ret;
    ret = (BOOL) edir_isapi_handler_init(myInstanceHandle);
    if (ret) {
	/*
	 * Do some bookkeeping needed by IIS.
	 */
	vinfo->dwExtensionVersion = 
		MAKELONG(HSE_VERSION_MINOR,HSE_VERSION_MAJOR);	
	strncpy(vinfo->lpszExtensionDesc, "Enhydra Director Handler",
		HSE_MAX_EXT_DLL_NAME_LEN);
    }
    return ret;
}

/**
 * Request entry point for the Enhydra Director ISAPI handler.
 * <P>
 * This function is called by IIS when a request is translated 
 * to this module by the Enhydra Director filter.  It creates
 * a per-request heap and then calls <code>edir_isapi_handler()</code>.
 *
 * @param req_info [in,out] Pointer to the request information.
 * @return	Integer status code.
 */
DWORD WINAPI
HttpExtensionProc(EXTENSION_CONTROL_BLOCK *req_info)
{
    DWORD retcode = HSE_STATUS_ERROR;
    if (!Dead) {
	if (edir_isapi_memory_create_heap(HEAP_NO_SERIALIZE) == NULL) {
	    NTEventLog("Can't allocate memory. Request aborted.", 1);
	    retcode = HSE_STATUS_ERROR;
	} else {
	    retcode = edir_isapi_handler(req_info);
	    edir_isapi_memory_delete_heap();
	}
    }
    return retcode;
}

/**
 * ISAPI Extension termination handler.
 * <P>
 * This handler is called when IIS is about to unload the extenstion
 * DLL from its address space.  This is a good place to put cleanup
 * of global resources.
 * <P>
 * Since this extension keeps track of load across many requests, it
 * is best if this DLL is NOT unloaded unless the server is shuttnig
 * down.  In that case it will send HSE_TERM_MUST_UNLOAD.
 * 
 * @param flags [in] If <code>HSE_TERM_MUST_UNLOAD</code> this
 *                   extension is about to be unconditionally terminated.
 *		     If <code>HSE_TERM_ADVISORY_UNLOAD</code> this 
 *		     extension is being asked to voluntarily unload.  It
 *		     can refuse by returning FALSE.
 * @return Returns TRUE if being unconditionally unloaded (as if it really
 *		     has any choice).  Normally it returns FALSE in order
 *		     to maintain scoreboard persistence.
 */
BOOL WINAPI
TerminateExtension(DWORD flags)
{
    if (flags == HSE_TERM_MUST_UNLOAD) {
	/*
	 * Server insists on unloading NOW.  
	 * Most likely because of pending shutdown of IIS.
	 */
	if (myDllName) free(myDllName); /* This was allocated with malloc*/
	myDllName = NULL;
	if (scoreboardMutex) CloseHandle(scoreboardMutex);
	scoreboardMutex = NULL;
	if (globalHeap) HeapDestroy(globalHeap);
	globalHeap = NULL;
	refCounts = NULL;
	numServers = 0;
	return TRUE;
    }
    return FALSE; /* NEVER willingly unload */
}

/**
 * Entry point for this DLL.
 * <P>
 * This is called by the IIS server to initialize the DLL.  It is the
 * first code executed when the DLL is loaded, and is a good place to
 * perform global initializations such as setting up the 
 * thread-local storage index for the per-request heap.
 * 
 * @param handle   [in] The instance handle for this DLL.
 * @param reason   [in] DLL access event code.  (i.e. "attach")
 * @param reserved [in] Reserved data pointer.
 * @return True if the DLL load should succeed, False if it should fail.
 */
BOOL WINAPI
DllMain(HINSTANCE handle, unsigned long reason, void *reserved)
{
    switch (reason) {
	case DLL_PROCESS_ATTACH:
	    myInstanceHandle = handle;
	    Dead = 0;
	    break;
	case DLL_PROCESS_DETACH:
	    Dead = 1;
	    break;
	default:
	    break;
    }
    return TRUE;
}

