/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_nsapi_handle.c,v 1.1 2003/04/14 12:26:02 sinisa Exp $
 */

#ifdef XP_WIN32
#define NSAPI_PUBLIC __declspec(dllexport)
#include <winsock2.h>
#else /* !XP_WIN32 */
#define NSAPI_PUBLIC
#endif /* !XP_WIN32 */ 

#include "nsapi.h"
#include "edir.h"
#include "edir_kvtable.h"
/* #include "nsapi.h" */
#include "edir_nsapi.h"

/**
 * Replacement for <code>strndup()</code> that uses <code>edir_malloc()</code>.
 * <P>
 * The returned string is an identical copy of <code>s</code> for the first
 * <code>n</code> characters, and should be freed using 
 * <code>edir_free()</code>.
 * 
 * @param s [in] The string to duplicate.
 * @param n [in] the limit to the ammount of characters duplicated
 * @return A newly allocated character array containing the copied string.
 */
extern char *
edir_strndup(const char *s, int n)
{
    char *cp;
    cp = edir_malloc(n+1);
    strncpy(cp, (char *)s, n);
    cp[n] = '\0';
    return cp;
}

/**
 * Convenience function to do in-place removal of leading and trailing 
 * whitespace characters.
 * <P>
 * Side effects:<BR>
 * This function modifies the string pointed to by "s".
 * 
 * @param s [in,out] String to trim in place.
 * @return A pointer to the trimmed portion of the string.
 */
static char *
trim_inplace(char *s)
{
    char *cp;
    EDIR_DEBUG_FUNCS_IN(("func: trim_inplace(\"%s\")", s ? s : "(null)"));
    while (*s && isspace(*s)) s++;
    for (cp=s; *cp; cp++) ;
    for (--cp; (cp >= s) && isspace(*cp); cp--) *cp = '\0';
    EDIR_DEBUG_FUNCS_OUT(("func: trim_inplace() returns \"%d\"",s?s:"(null)"));
    return s;
}

/**
 * Function to handle components of the request URI as it is parsed.
 * <P>
 * The edir_parse_uri() function calls this function for each part
 * of the url that it parses.  The four possible values of the <code>
 * type</code> parameter are described below. <BR>
 * <code>EDIR_PARSE_URI_PATH_INFO</code><BR>
 * The value of the <code>lhs</code> parameter will point to a character
 * string containing the stripped path portion of the URI, with leading
 * and trailing slashes removed, and with interior runs of more than
 * one slash reduced to one slash.
 * <code>EDIR_PARSE_URI_PATH_PARAM_1</code>
 * Returns a pointer to a string in <code>lhs</code> that contains a URL
 * "path parameter" that is not of the form "key=value".  The <code>rhs</code>
 * parameter is always <code>NULL</code>.
 * <code>EDIR_PARSE_URI_PATH_PARAM_2</code>
 * Returns strings comprising "path parameter" in the form of a "key=value"
 * pair, with the key in <code>lhs</code>, and the value in <code>rhs</code>.
 * Neither parameters are ever <code>NULL</code> for this type.
 * <code>EDIR_PARSE_URI_QUERY_STRING</code>
 * Returns a pointer to the beginning of the "query string" in <code>lhs</code>.
 * <P>
 * NOTE: The strings passed to this function are part of a larger
 * string buffer that is being destructively parsed by edir_parse_uri(). They
 * should not be freed, and they should only be stored in conjunction with
 * the larger URI buffer that contains them so the buffer does not
 * get freed while the strings are still in use.
 * 
 * @param ptr  [in,out] Arbitrary data structure to store things in.
 * @param type [in]     What type of component has been parsed.
 * @param lhs  [in]     "Left hand side" of "key=value" pair, or full
 * 			String of "value" item.
 * @param rhs  [in]     "Right hand side" of "key=value" pair, or NULL.
 */
static void
edir_nsapi_uri_parse_func(void *ptr, int type, char *lhs, char *rhs)
{
    edir_uri_data_t *uriData = (edir_uri_data_t *) ptr;
    switch (type) {
	case EDIR_PARSE_URI_PATH_INFO:
	    uriData->pathInfo = lhs;
	    break;
	case EDIR_PARSE_URI_PATH_PARAMS:
	    uriData->pathParams = edir_strdup(lhs);
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_1:
	    break;
	case EDIR_PARSE_URI_PATH_PARAM_2:
	    if (lhs && !strcasecmp(lhs, "jsessionid")) {
	        EDIR_DEBUG_NORMAL(("Found session id in uri"));
		uriData->sessionID = rhs;
	    }
	    break;
	case EDIR_PARSE_URI_QUERY_STRING:
	    uriData->queryString = lhs;
	    break;
    }
}

#ifdef DEBUG
static void
dump_pblock_info(char *blockName, pblock *pblk)
{
    int i;

    for (i = 0; i<pblk->hsize; i++) {
        pb_param *param = NULL;
	struct pb_entry *next = NULL;

	if (pblk->ht[i] == NULL) {
	    continue;
	}

	param = pblk->ht[i]->param;
	next = pblk->ht[i]->next;

	if (param) {
	    fprintf(stdout, "%s #%d: ", blockName, i);
	    fflush(stdout);

	    if (param->name) {
	        fprintf(stdout, "%s", param->name);
		fflush(stdout);
	    }

	    if (param->value) {
	        fprintf(stdout, "=%s", param->value);
		fflush(stdout);
	    }

	    if (next != NULL) {
	        fprintf(stdout, "  (NEXT->PARAM->NAME=%s)", next->param->name);
		fflush(stdout);
	    }

	    fprintf(stdout, "\n");
	    fflush(stdout);
	}
    }
}
#endif

/**
 * Function to parse a cookie header.
 * <P>
 * This function takes a Cookie header, which is a semicolon-delimited
 * list of cookie key-value pairs.  Each key-value pair is placed
 * separately into the key-value table passed through the "tablePtr" pointer.
 * The table will be created if it doesn't already exist.
 * 
 * @param tablePtr     [in,out] The key-value table to store the cookies in
 * @param value        [in]	 The value found by cookie fetcher
 * @param sessionIdPtr [in,out] If given, but currently empty, the session ID
 *                              will be written in.
 */
static void
parse_cookie(edir_kvtable_t **tablePtr,
	     const char      *value,
	     char           **sessionIdPtr)
{
    char		*start, *end, *name, *val, *fp = NULL;
    EDIR_DEBUG_FUNCS_IN(("func: parse_cookie("EDIR_PTR", \"%s\", "EDIR_PTR")",\
			 tablePtr, value ? value : "(null)", sessionIdPtr));

    if (value == NULL) {
	EDIR_DEBUG_FUNCS_OUT(("func: parse_cookie() - no cookie found"));
	return;
    }

    if (tablePtr != NULL && *tablePtr == NULL)
	*tablePtr = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE); 

    fp = start = edir_strdup(value);
    end = strchr(start, ';');
    while (end) {
	if (strncasecmp(end, ";path=", 6) == 0) {
	    end++;
	    end = strchr(end, ';');
	    if (end == NULL)
		break;
	}
	*end++ = '\0';
        name = start;
        val = strchr(name, '=');
        if (val) *val++ = '\0';
        else val = "";
	name = trim_inplace(name);
	val = trim_inplace(val);
	if (!strcasecmp(name, "jsessionid")) {
	    char *l = name;
	    while (*l) {
		*l++ = toupper(*l);
	    }
	    if ((sessionIdPtr != NULL) && (*sessionIdPtr == NULL)) {
		*sessionIdPtr = edir_strdup(val);
	    }
	}
	edir_kvtable_add(*tablePtr, name, val);
        start = end;
        end = strchr(start, ';');
    }
    name = start;
    val = strchr(name, '=');
    if (val)
        *val++ = '\0';
    else
        val = "";
    name = trim_inplace(name);
    val = trim_inplace(val);
    if (!strcasecmp(name, "jsessionid")) {
	char *l = name;
	while (*l) {
	    *l++ = toupper(*l);
	}
	if ((sessionIdPtr != NULL) && (*sessionIdPtr == NULL)) {
	    *sessionIdPtr = edir_strdup(val);
	}
    }
    edir_kvtable_add(*tablePtr, name, val);
    edir_free(fp);
    EDIR_DEBUG_FUNCS_OUT(("func: parse_cookie() found cookies"));
    return;
}

static void
headers_to_kvtable(pblock *pb, edir_kvtable_t *kvt)
{
    static char		*funcName = "pblock_to_kvtable";
    pb_param		*param;
    char		*name, *namePtr, *value, *valuePtr, *cp;
    int          	i;
    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR")", funcName, pb));

    for (i = 0; i<pb->hsize; i++) {
	if (pb->ht[i] == NULL) continue;
	param = pb->ht[i]->param;
	if (param && param->name) {
	    namePtr = edir_strdup(param->name);
	    name = trim_inplace(namePtr);
	    valuePtr = edir_strdup(param->value ? param->value : "");
	    value = trim_inplace(valuePtr);
	    if (*name) *name = toupper(*name);
	    for (cp = name; *cp; cp++) {
		if (*cp == '-') {
		    cp++;
		    if (*cp) *cp = toupper(*cp);
		    else break;
		}
	    }
	    edir_kvtable_add(kvt, name, value);
	    edir_free(namePtr);
	    edir_free(valuePtr);
	}
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns", funcName));
    return;
}

static void
edir_connect_to_enhydra(edir_load_sb_server_t  *server,
			Request                *pRequest,
			Session                *pSession,
			edir_nsapi_cfg_t       *cfg,
			edir_uri_data_t	       *uriData,
			edir_load_sb_app_t     *app,
			char		       *stickyTag,
			int                    *releaseConnPtr,
			CRITICAL                vLock,
			struct sockaddr_in	*local_addr,
			struct sockaddr_in	*remote_addr,
			edir_exception_t      **exc)
{
    static char                 *funcName = "edir_connect_to_enhydra";
    edir_connection_t		*conn = NULL;
    time_t			 startTime, elapsedTime, lastTime, now;
    octet			*inBuf, *outBuf;
    int				 inBufSize, outBufSize;
    int			         stickyRetries = 0;
    int				 cgi_status = PROTOCOL_SERVER_ERROR;
    void			*sbuf = NULL;
    char                        *path_info = NULL, *request_uri = NULL;
    edir_kvtable_t		*cgi_vars = NULL;  /* Request vars */
    edir_kvtable_t		*http_vars = NULL; /* Request headers */
    edir_kvtable_t              *cookies = NULL;   /* Response cookies */
    int                          content_length = 0;
    int32			 nbytes, bytesToRead, remain;
    int                          content_type_given_flag = 0;
    char			*remAddr = NULL;
    char			*key, *val, **vals, *sep;
    char			ipBuf[20], portBuf[20];
    int32			i, nvals;
    EDIR_IF_DEBUG(char		*trace = NULL; )

    EDIR_DEBUG_FUNCS_IN(("func: %s()", funcName));

    if (stickyTag) stickyRetries = EDIR_CONNECT_STICKY_SESSION_RETRIES;
    path_info = uriData->pathInfo + app->appPrefixLen;

    request_uri = uriData->pathInfo ? uriData->pathInfo : "";

    /* 
     * Get config and set up buffers
     */
    inBufSize = cfg->inBufSize;
    outBufSize = cfg->outBufSize;
    inBuf = edir_malloc(inBufSize);
    outBuf = edir_malloc(outBufSize);

    conn = (edir_connection_t *) edir_malloc(sizeof(edir_connection_t));
    memset(conn, 0, sizeof(edir_connection_t));
    time(&startTime);
    lastTime = startTime;
    elapsedTime = 0;
    while (elapsedTime <= EDIR_CONNECT_TIMEOUT) {
        if (*exc) edir_exception_free(exc);
	edir_client_connect(conn, NULL, &(server->sin), server->authData,
			    outBuf, outBufSize, inBuf, inBufSize, exc);
	if (!(*exc)) break; /* Connection established! */
	if ((*exc)->id != EDIR_EXCEPT_IO_ERROR) EDIR_THROW(*exc);

	/* XXX Should have 'logmask' */
	if (edir_debug & EDIR_DEBUG_MASK_NORMAL) {
	    char *trace = NULL;
	    if (*exc) trace = edir_exception_stack_trace(*exc);
	    edir_log_error("Warning: Error connecting to Enhydra.");
	    edir_log_error("%s", trace ? trace : "No details.");
	    if (trace) edir_free(trace);
	}
	if (*exc) edir_exception_free(exc);

	/* Mark this server as 'failing' */
	/* ###### ENTER CRITICAL SECTION - scoreboard ###### */

	crit_enter(vLock);

	if (!(server->failing)) {

	    /**
	     * XXX FIX  - Idea: Run notification script.
	     * Especially on Unix, where it's easy, run a script if we
	     * find a server to be 'failing'.  Probably would want to
	     * only run the script occasionally to avoid spamming,
	     * and would want to run it outside any critical section
	     * to avoid blocking for prolonged periods while holding
	     * the global scoreboard lock.
	     */
	    server->failing = 1;
	    time(&(server->failTime));
	}
	server->failCount++;
	/*
	 * If this request has a 'sticky session' token in the session
	 * cookie, then we want to try several times to go to the desired
	 * server.  After a number of attempts, though, we give up and
	 * fall back to round robin.  If getterData.retries is greater than
	 * zero, that means 'keep trying this server'.
	 */
	if (stickyRetries <= 0) stickyTag = NULL;
	else stickyRetries--;
	*releaseConnPtr = 0;
	if (refCounts[server->refIndex] > 0)
	    refCounts[server->refIndex]--;
	server->useCount--; /* Oops, undo rejected server use count. */

	EDIR_DEBUG_NORMAL(("Failed to get %sserver: %s:%d\n", (stickyTag) ? "sticky " : "", server->hostName, server->port));

	/* Get new server */
	edir_load_choose(scoreboard, uriData->pathInfo, stickyTag, 
			 &app, &server, refCounts, numServers, 
			 local_addr, remote_addr, uriData->virtHost, exc);
	if (*exc) {
	    crit_exit(vLock);
	    EDIR_THROW(*exc);
	}
	EDIR_DEBUG_NORMAL(("Got alternate server: %s:%d", \
			   server->hostName, server->port));
	server->useCount++; /* Now, up the use count for new server */
	path_info = uriData->pathInfo + app->appPrefixLen;
	*releaseConnPtr = 1;

	refCounts[server->refIndex]++;

	/* Release lock so we don't hold it while waiting for conn */
	crit_exit(vLock);
	/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */

	time(&now);
	/* Avoid spinning wildly on errors */
	if (EDIR_TIMEDIFF(now, lastTime) < 1) {
#ifdef XP_WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
	    time(&now);
	}
	lastTime = now;
	elapsedTime = EDIR_TIMEDIFF(now, startTime);
    } /* While not timed out */

    /**
     * If we couldn't get a connection in time, just fail.
     */
    if (*exc) EDIR_THROW(*exc);

    /*
     * Check for case of timing out after freeing exception.
     * XXX There's got to be a better way.
     */
    if (conn->state != EDIR_STATE_OPEN) {
	*exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR, 
		"Can't connect to any Enhydra server for %d %s",
		EDIR_CONNECT_TIMEOUT, "seconds -- giving up.");
	EDIR_THROW(*exc);
    }

    /**
     * Get some header values we need for our own processing
     */
    if (val = pblock_findval("content-length", pRequest->headers)) {
	content_length = atoi(val);
    }

    /**
     * Use NSAPI's built-in functions for obtaining the
     * standard request headers. 
     * 
     * XXX - Probably want to add a header packet type for 
     * numbered standard headers, and to pare down the list
     * of headers sent.  Eventually, with connection pooling,
     * we can cache immutable headers.
     */
    http_vars = edir_kvtable_create(EDIR_KVTABLE_TO_UPPER_PRESERVE);
    cgi_vars = edir_kvtable_create(EDIR_KVTABLE_TO_UPPER);

    /* 
     * Gather specific headers to send.
     */
    if (val = pblock_findval("auth-type", pRequest->vars)) {
	edir_kvtable_add(cgi_vars, "AUTH_TYPE", val);
    }

    if (val = pblock_findval("auth-user", pRequest->vars)) {
	edir_kvtable_add(cgi_vars, "AUTH_USER", val);
    }

    if (val = pblock_findval("path-translated", pRequest->vars)) {
	edir_kvtable_add(cgi_vars, "PATH_TRANSLATED", val);
    }

    if (val = pblock_findval("query", pRequest->reqpb)) {
	edir_kvtable_add(cgi_vars, "QUERY_STRING", val);
    }

    if (val = pblock_findval("auth-user", pRequest->vars)) {
	edir_kvtable_add(cgi_vars, "REMOTE_USER", val);
    }

    if (val = pblock_findval("method", pRequest->reqpb)) {
	edir_kvtable_add(cgi_vars, "REQUEST_METHOD", val);
    }


    /*
     * Get local addr, name, and port.
     */
    edir_sockaddr_to_str(local_addr, ipBuf);
    edir_kvtable_add(cgi_vars, "SERVER_ADDR", ipBuf);
    if (val = util_hostname()) {
	edir_kvtable_add(cgi_vars, "SERVER_NAME", val);
    } else if (val = uriData->virtHost) {
	edir_kvtable_add(cgi_vars, "SERVER_NAME", val);
    } else {
	edir_kvtable_add(cgi_vars, "SERVER_NAME", ipBuf);
    }
    sprintf(portBuf, "%d", conf_getglobals()->Vport);
    edir_kvtable_add(cgi_vars, "SERVER_PORT", portBuf);

    /*
     * Get remote addr, name, and port.
     */
    ipBuf[0] = '\0';
    portBuf[0] = '\0';
    edir_sockaddr_to_str(remote_addr, ipBuf);
    edir_kvtable_add(cgi_vars, "REMOTE_ADDR", ipBuf);
    if (val = session_dns(pSession)) {
	edir_kvtable_add(cgi_vars, "REMOTE_HOST", val);
    } else if (ipBuf[0]) {
	edir_kvtable_add(cgi_vars, "REMOTE_HOST", ipBuf);
    }

    if (val = pblock_findval("protocol", pRequest->reqpb)) {
	edir_kvtable_add(cgi_vars, "SERVER_PROTOCOL", val);
    }

    /*
     * Special processing for REQUEST_URI: Should contain path params.
     * Servlet spec seems to specify that it does NOT contain query
     * string, unlike CGI spec.
     */
    if (uriData->pathParams) {
	val = edir_asprintf("/%s;%s", request_uri, uriData->pathParams);
	edir_kvtable_add(cgi_vars, "REQUEST_URI", val);
	edir_free(val);
    } else {
	val = edir_asprintf("/%s", request_uri);
	edir_kvtable_add(cgi_vars, "REQUEST_URI", val);
	edir_free(val);
    }

    /*
     * Special processing for PATH_INFO: Should contain path params.
     */
    if (*path_info || uriData->pathParams) {
	if (uriData->pathParams) {
	    val = edir_asprintf("%s;%s", path_info, uriData->pathParams);
	    edir_kvtable_add(cgi_vars, "PATH_INFO", val);
	    edir_free(val);
	} else {
	    edir_kvtable_add(cgi_vars, "PATH_INFO", path_info);
	}
    }

    /*
     * Cheat a bit here.  We assume that at this point in history,
     * we're either using HTTP or HTTPS.  That being the case,
     * if we're secure, send SERVER_SCHEME as 'https'.  If we're
     * not secure, send 'http'.
     */
    if (security_active) {
	edir_kvtable_add(cgi_vars, "SERVER_SCHEME", "https");
    } else {
	edir_kvtable_add(cgi_vars, "SERVER_SCHEME", "http");
    }


    /*
     * Enumerate the request headers pblock into a more easily
     * used kvtable.
     */
    headers_to_kvtable(pRequest->headers, http_vars);

    /*
     * Send the CGI headers.  CGI values not sent by client
     */
    for (vals = edir_kvtable_first(cgi_vars, &key, &nvals); vals; ) {
	for (i=0; i<nvals; i++) {
	    val = vals[i];
	    edir_client_send_header(conn, key, strlen(key),
			    val ? val : "", val ? strlen(val) : 0,
			    EDIR_PKT_SUBTYPE_HEADER_HEADER,
			    outBuf, outBufSize, inBuf, inBufSize, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	vals = edir_kvtable_next(cgi_vars, &key, &nvals);
    }

    /*
     * Send the HTTP headers.  HTTP headers sent by client.
     * Do not modify the names as Servlets expect the header names
     * to be the same as if they are sent over the wire.
     */
    for (vals = edir_kvtable_first(http_vars, &key, &nvals); vals; ) {
	for (i=0; i<nvals; i++) {
	    val = vals[i];
	    edir_client_send_header(conn, key, strlen(key),
			    val ? val : "", val ? strlen(val) : 0,
			    EDIR_PKT_SUBTYPE_HEADER_HTTP,
			    outBuf, outBufSize, inBuf, inBufSize, exc);
	    if (*exc) EDIR_THROW(*exc);
	}
	vals = edir_kvtable_next(http_vars, &key, &nvals);
    }

    /*
     * Send the cookies.
     */
    if (uriData->cookies) {
	for (vals = edir_kvtable_first(uriData->cookies, &key, &nvals);vals;){
	    for (i=0; i<nvals; i++) {
		val = vals[i];
		edir_client_send_header(conn, key, strlen(key),
				val ? val : "", val ? strlen(val) : 0,
				EDIR_PKT_SUBTYPE_HEADER_COOKIE,
				outBuf, outBufSize, inBuf, inBufSize, exc);
		if (*exc) EDIR_THROW(*exc);
	    }
	    vals = edir_kvtable_next(uriData->cookies, &key, &nvals);
	}
    }

    /* Indicate end of headers to the server */
    edir_client_send_header(conn, "EDIR_CLIENT", sizeof("EDIR_CLIENT") - 1,
			    "NSAPI", sizeof("NSAPI") - 1,
			    EDIR_PKT_SUBTYPE_HEADER_LAST,
			    outBuf, outBufSize, inBuf, inBufSize, exc);
    if (*exc) EDIR_THROW(*exc);

    edir_kvtable_destroy(cgi_vars);
    cgi_vars = NULL;
    edir_kvtable_destroy(http_vars);
    http_vars = NULL;

    /*
     * If we have a request entity, send it to the server now.
     */
    for (remain = content_length; remain > 0; remain -= nbytes) {
	bytesToRead = EDIR_MIN(remain, outBufSize);
	nbytes = netbuf_getbytes(pSession->inbuf, outBuf, bytesToRead);
	if (nbytes == NETBUF_ERROR) {
	    *exc = edir_exception_create(EDIR_EXCEPT_IO_ERROR,
			    "Error reading from client.");
	    EDIR_THROW(*exc);
	}
	if (nbytes < 1) break; /* EOF */
	edir_client_send_data(conn, nbytes, 0, 
			      outBuf, outBufSize,
			      inBuf, inBufSize, exc);
	if (*exc) EDIR_THROW(*exc);
    }

    /*
     * Send "no more data" indication
     */
    edir_client_send_data(conn, 0, 1, outBuf,outBufSize,inBuf,inBufSize,exc);
    if (*exc) EDIR_THROW(*exc);

    /*
     * XXX - sending last request data automatically causes state machine
     * to receive first response header.  FIX ME  This should be
     * split up.  Especially if we want to split up the send
     * and receive portions of the state machine to allow true
     * full duplexing.
     * 
     * Also note that we don't check for write errors here.
     * Since we can't report errors to the client at this point,
     * it's not that important.  The ap_rwrite() function will
     * complain to the logs if an error occurs.
     */
    pblock_remove("content-type", pRequest->srvhdrs);
    pblock_remove("Content-type", pRequest->srvhdrs);
    pblock_remove("Content-Type", pRequest->srvhdrs);
    sbuf = edir_sbuf_create(); /* Cookie holder */
    sep = "";
    while (conn->state == EDIR_STATE_RECEIVE_HEADERS) {

        char *val = NULL;
	char *key = NULL;

        if (conn->packetSubType == EDIR_PKT_SUBTYPE_HEADER_COOKIE) {
	    /*
	     * We don't need to parse outgoing cookies.
	     * Just set them in one header
	     */
	    edir_sbuf_strcat(sbuf, sep);
	    sep = "; ";
	    edir_sbuf_bincat(sbuf, conn->headerName, conn->headerNameLen);
	    edir_sbuf_strcat(sbuf, "=");
	    edir_sbuf_bincat(sbuf, conn->headerValue, conn->headerValueLen);
	} else {
	    key = edir_strndup((char *)conn->headerName, conn->headerNameLen);
	    val = edir_strndup((char *)conn->headerValue, conn->headerValueLen);
	    if (!edir_strcasecmp(key, "Status")) {
	        cgi_status = atoi(val);
		protocol_status(pSession, pRequest, cgi_status, NULL);
	    } else if (!edir_strcasecmp(key, "Content-Type")) {
	        char *tmpv = NULL;
	        val = trim_inplace(val);
		tmpv = val;
		while (*tmpv) {
		  *tmpv = tolower(*tmpv);
		  tmpv++;
		}
		content_type_given_flag = 1;
		pblock_remove("content-type", pRequest->srvhdrs);
		pblock_remove("Content-type", pRequest->srvhdrs);
		pblock_remove("Content-Type", pRequest->srvhdrs);
		pblock_nvinsert("content-type", val, pRequest->srvhdrs);
	    } else if (!edir_strcasecmp(key, "Connection")) {
		/* Netscape server automatically handles keepalives */
		; /* discard */
	    } else if (!edir_strcasecmp(key, "Set-Cookie")) {
		edir_sbuf_appendf(sbuf, "%s%s", sep, val);
		sep = "; ";
	    } else {
		char *cp;
		for (cp=key; *cp; cp++) *cp = tolower(*cp);
		pblock_remove(key, pRequest->srvhdrs);
		pblock_nvinsert(key, val, pRequest->srvhdrs);
	    }
	}
	edir_client_next_header(conn, outBuf, 4096, inBuf, 4096, exc);
	if (*exc) EDIR_THROW(*exc);
    }

    /*
     * Set the cookies as one Set-Cookie header.
     */
    pblock_remove("Set-Cookie", pRequest->srvhdrs);
    if (edir_sbuf_length(sbuf) > 0) {
	val = edir_sbuf_tostring(sbuf, NULL);
	pblock_nvinsert("Set-Cookie", val, pRequest->srvhdrs);
    }

    /*
     * Set default content type if not already set.
     */
    if (content_type_given_flag == 0) {
        pblock_remove("Content-Type", pRequest->srvhdrs);
	pblock_nvinsert("Content-Type", "text/html", pRequest->srvhdrs);
    }

    edir_sbuf_free(sbuf);
    sbuf = NULL;

    if (protocol_start_response(pSession, pRequest) != REQ_NOACTION) {

	/* Send the response entity */

	while ((conn->state == EDIR_STATE_RECEIVE_RESPONSE) ||
	       (conn->state == EDIR_STATE_RECEIVE_RESPONSE_PACKET)) {

	    if (net_write(pSession->csd,
			  (char *) conn->receiveData,
			  conn->receiveDataLen) == IO_ERROR) {
		/* Don't log exception here -- this is a very
		 * common situation when users reload pages
		 * while they are still loading.
		 */
		EDIR_DEBUG_VERBOSE(("%s: @laststr@ (errnum=@lasterr@)", \
				    "Error sending response entity"));
		break;
	    }

	    edir_client_next_data(conn,outBuf,outBufSize,inBuf,inBufSize,exc);
	    if (*exc) EDIR_THROW(*exc);
	}

	/**
	 * Close the appserver connection.
	 */
	if (conn->state != EDIR_STATE_OPEN) {
	    EDIR_DEBUG_NORMAL(("Netscape server closed write pipe to %s", \
			       "enhydra director.  Aborting."));
	    edir_client_abort(conn, outBuf, outBufSize, exc);
	    if (*exc) edir_exception_free(exc);
	} else {
	    EDIR_DEBUG_VERBOSE(("Request finished.  Closing connection."));
	    edir_client_close(conn, outBuf, outBufSize, exc);
	    if (*exc) edir_exception_free(exc);
	}
    } else {
        /* NSAPI says we must return ASAP, is it an error? */
	/* No - This was probably a HEAD request. */
	EDIR_DEBUG_VERBOSE(("%s: %s() == REQ_NOACTION, %s", funcName, \
			    "protocol_start_response", \
			    "Can't send response.\n"));
    }
    if (cgi_vars) edir_kvtable_destroy(cgi_vars);
    if (http_vars) edir_kvtable_destroy(http_vars);
    if (conn) {
        edir_free(conn); /* XXX FIX probably need an edir_conn_free() func... */
    }
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;

except:
    if (cgi_vars) edir_kvtable_destroy(cgi_vars);
    if (http_vars) edir_kvtable_destroy(http_vars);
#ifdef NOTDEF
    if (conn) {
	edir_exception_t *_exc2 = NULL;
        edir_client_abort(conn, outBuf, outBufSize, &_exc2);
        if (_exc2) edir_exception_free(&_exc2);
        edir_free(conn); /* XXX FIX probably need an edir_conn_free() func... */
    }
#endif
    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
    return;
}

/**
 * Utility function used to extract the URI from an NSAPI Request structure.
 * <P>
 * This routine allocates a request termporary storage struct and fills it in
 * with all the URI information, including what virtual host the visitor
 * connected to.
 * <P>
 * Note that this routine is not complete with respect to filling out an URI.
 * It should probably be supplimented with a call to 
 *
 * @param pRequest [in] NSAPI 'Request' structure containing information about
 *	the visitor's request. It is from this structure that the URI is
 *    developed.
 */
static edir_uri_data_t *
create_uri_data(Request *pRequest)
{
    static char	    *funcName="create_uri_data";
    edir_uri_data_t *uriData = NULL;

    /* Initialize the uriData space */
    uriData = (edir_uri_data_t *) edir_malloc(sizeof(edir_uri_data_t));
    memset((char *)uriData, 0, sizeof(edir_uri_data_t));

    /* Parse the URI, looking for path parameters in particular. */
    uriData->uriBuf = pblock_findval("uri", pRequest->reqpb);
    uriData->uriBuf = edir_strdup( uriData->uriBuf ? uriData->uriBuf : "");

    EDIR_DEBUG_EVERYTHING(("%s: uriBuf = '%s'", funcName, uriData->uriBuf));

    edir_parse_uri(uriData->uriBuf, uriData, edir_nsapi_uri_parse_func);

    EDIR_DEBUG_EVERYTHING(("%s: PARSED pathinfo = '%s'", funcName, \
			uriData->pathInfo ? uriData->pathInfo : "(null)"));
    EDIR_DEBUG_EVERYTHING(("%s: PARSED queryString = '%s'", funcName, \
			uriData->queryString ? uriData->queryString:"(null)"));

    /* Fill in the vitual host information */
    uriData->virtHost = pblock_findval("host", pRequest->headers);
    if (uriData->virtHost) {
        char *tmpColon;
        uriData->virtHost = edir_strdup(uriData->virtHost);
        tmpColon = strchr(uriData->virtHost, ':');
        if (tmpColon)
	    *tmpColon ='\0';

	EDIR_DEBUG_EVERYTHING(("%s: PARSED virtHost = '%s'", funcName, \
			       uriData->virtHost ? uriData->virtHost:"(null)"));
    }

    /**
     * Gather the cookies, separate them into separate entries, and
     * put them in a table.  This is needed so we can look for the
     * application session cookie for load balancing purposes.
     */
    parse_cookie(&(uriData->cookies),
		 pblock_findval("cookie", pRequest->headers), 
		 &(uriData->sessionID));

    EDIR_DEBUG_EVERYTHING(("%s: PARSED sessionID = '%s'", funcName, \
			uriData->sessionID ? uriData->sessionID : "(null)"));

    return uriData;
}


/**
 * Handler function for the Enhydra Director NSAPI NameTrans directive.
 * <P>
 * Given an URI, determine if there corrisponds a valid application. If
 * there is, then say we do provide a handler for that under the <OBJECT>
 * name of "enhydra_director".
 * 
 * @param pb       [] NSAPI 'pblock' structure - ignored
 * @param pSession [] NSAPI 'Session' structure which is read to find out what
 *	the internet address of the client is.
 * @param pRequest [in] NSAPI 'Request' structure containing information about
 *	the visitor's request. It is from this structure that the URI is
 *    developed.
 * @return Integer NSAPI status code REQ_PROCEED if we do have an app that
 *	matches the given URI, otherwise send a REQ_NOACTION, in the hopes that
 *    the NS Server can find some other response for the visitor.
 */
#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int edir_nsapi_name_trans(pblock *pb, Session *pSession,
				       Request *pRequest)
{
    static char	       *funcName="edir_nsapi_name_trans";
    edir_load_sb_app_t *app = NULL;
    edir_uri_data_t    *uriData = NULL;
    struct sockaddr_in	local_addr, remote_addr;
    char	        buf[50];
    int			sinLen;
#if defined(XP_WIN32)
    SOCKET			sockfd;
#elif defined(XP_UNIX)
    int				sockfd;
#else
#error "Porting Needed"
#endif

    /* Set up per-thread logging context */
    edir_nsapi_log_set_session(pSession, pRequest);

    /* Parse the request to get the URI */
    uriData = create_uri_data(pRequest);

    if (uriData->pathInfo == NULL) {
        /* Not path? No application */
        return REQ_NOACTION;
    }

    /*
     * Get local addr, name, and port.
     */
#ifdef XP_WIN32
    sockfd = (SOCKET) net_native_handle(pSession->csd);
#else
    sockfd = net_native_handle(pSession->csd);
#endif
    sinLen = sizeof(struct sockaddr_in);
    memset(&local_addr, 0, sizeof(struct sockaddr_in));
    getsockname(sockfd, (struct sockaddr *) &local_addr, &sinLen);
    local_addr.sin_port = htons((uint16)(conf_getglobals()->Vport));
    local_addr.sin_family = AF_INET;

    /*
     * Get remote addr, name, and port.
     */
    sinLen = sizeof(struct sockaddr_in);
    memset(&remote_addr, 0, sinLen);
#ifndef XP_WIN32
    /* Win32 getpeername doesn't seem to work right here */
    if (!getpeername(sockfd,(struct sockaddr *)&remote_addr, &sinLen) == 0)
#endif
    {
	char *val = pblock_findval("ip", pSession->client);
	if (!edir_ip_addr(val ? val : "", &(remote_addr.sin_addr))) {
	    /* XXX remote addr unavailable */
	    remote_addr.sin_addr.s_addr = INADDR_ANY;
	}
    }
    /* Don't care about remote port */
    remote_addr.sin_port = 0;
    remote_addr.sin_family = AF_INET;

    /* Determine if there is an app that corrisponds to this URI */
    app = edir_load_check_path(scoreboard,
			       uriData->pathInfo,
			       &local_addr, &remote_addr,
			       uriData->virtHost);

    /* No app, let them find the path somewhere else */
    if ((app == NULL) && !edir_load_status_check_path(scoreboard,
			       uriData->pathInfo,
			       &local_addr, &remote_addr,
			       uriData->virtHost)) {
	EDIR_DEBUG_VERBOSE(("%s: URI http://%s/%s matches no application",
		    funcName, uriData->virtHost, uriData->pathInfo));
        return REQ_NOACTION;
    }

    EDIR_DEBUG_VERBOSE(("%s: URI /%s matches %s \"%s\"", \
			funcName,  uriData->pathInfo, \
			app ? "application" : "status prefix", \
			app ? app->appPrefix : ""));

    /* Pass these to the handler so it doesn't have to get these again */
    edir_sockaddr_to_str(&local_addr, buf);
    pblock_nvinsert("edir_server_addr", buf, pRequest->vars);
    sprintf(buf, "%d", ntohs(local_addr.sin_port));
    pblock_nvinsert("edir_server_port", buf, pRequest->vars);
    edir_sockaddr_to_str(&remote_addr, buf);
    pblock_nvinsert("edir_remote_addr", buf, pRequest->vars);

    /**
     * Let's set up a name for this type of app so we can use and issolated
     * <Object> block in the obj.conf config file
     */
    pblock_nvinsert("name", "enhydra_director", pRequest->vars);

    return REQ_PROCEED;
}

/**
 * Handler function for the Enhydra Director NSAPI module.
 * <P>
 * If the NameTrans function determines that a particular request is 
 * a servlet request, then this function is called to handle the request.
 * This function then does the initial work of processing the request.
 * 
 * @param pb       [] NSAPI 'pblock' structure - ignored
 * @param pSession [in] NSAPI 'Session' structure containing information about
 *	the visitor's session with the server.
 * @param pRequest [in] NSAPI 'Request' structure containing information about
 *	the visitor's request of the server.
 * @return Integer NSAPI status code which will be:
 *	REQ_PROCEED - if we are indeed answering the visitor's request.
 *	REQ_ABORTED - if there is some tragic problem with answering the request.
 *	REQ_NOACTION - if we were told to handle the request even though we don't
 *		actually have a application to match it.
 *	REQ_EXIT - if there is a fatal problem with the network.
 */
#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int edir_nsapi_handler(pblock *pb, Session *pSession, Request *pRequest)
{
    static char			*funcName="edir_nsapi_handler";
    edir_exception_t		*_exc=NULL, **exc=&_exc;
    char                        *stickyTag = NULL;
    char			*val, *cp, *dp;
    edir_nsapi_cfg_t		*cfg = globalCfg;
    edir_uri_data_t		*uriData;
    edir_load_sb_app_t		*app;
    edir_load_sb_server_t	*server = NULL;
    int				 releaseConn = 0;
    CRITICAL                     vLock = NULL;
    EDIR_IF_DEBUG(char		*trace = NULL; )
    int				cond_status = REQ_ABORTED;
    int				do_status = 0;
    int32			iv;
    struct sockaddr_in		local_addr, remote_addr;

    /* Set up per-thread logging context */
    edir_nsapi_log_set_session(pSession, pRequest);

    EDIR_DEBUG_FUNCS_IN(("func: edir_nsapi_handler("EDIR_PTR", "EDIR_PTR", "EDIR_PTR")",\
			 pb, pSession, pRequest));

    if (cfg == NULL) {
	edir_log_error("Global Enhydra Director Configuration unset");
	return REQ_ABORTED;
    }

    /*
     * Get the local and remote addresses that we discovered during
     * path translation.
     */
    val = pblock_findval("edir_server_addr", pRequest->vars);
    if (!edir_ip_addr(val ? val : "", &(local_addr.sin_addr))) {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
			    "edir_local_addr was lost!");
	EDIR_THROW(*exc);
    }
    val = pblock_findval("edir_remote_addr", pRequest->vars);
    if (!edir_ip_addr(val ? val : "", &(remote_addr.sin_addr))) {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
			    "edir_remote_addr was lost!");
	EDIR_THROW(*exc);
    }
    val = pblock_findval("edir_server_port", pRequest->vars);
    if (!edir_atoi(val ? val : "", &iv) || (iv < 0) || (iv > 65535)) {
	*exc = edir_exception_create(EDIR_EXCEPT_SERVLET, 
			    "edir_local_port was lost!");
	EDIR_THROW(*exc);
    }
    local_addr.sin_port = htons((unsigned short)iv);
    local_addr.sin_family = AF_INET;
    remote_addr.sin_port = 0;
    remote_addr.sin_family = AF_INET;

    /*
     * Parse the URI, looking for path parameters in particular.
     */
    uriData = create_uri_data(pRequest);

    if (uriData->pathInfo == NULL) {
	EDIR_DEBUG_FUNCS_OUT(("func: edir_nsapi_handler() returns REQ_ABORTED"));
	return REQ_ABORTED;
    }

    /** 
     * Remove the cookie tag from the headers, which was culled by the 
     * create_uri_data routine.
     */
    pblock_remove("cookie", pRequest->headers);

    /*
     * If we have a session ID, look for sticky session tag.
     */
    if ((uriData->sessionID) && (val = strchr(uriData->sessionID, '('))) {
	if (strlen(val) >= 9) stickyTag = val + 1; 
	EDIR_DEBUG_VERBOSE(("Sticky Session Tag: %s\n", stickyTag));
    }

    /*
     * see if we want to just display the status page.
     */
    if (edir_load_status_check_path(scoreboard, uriData->pathInfo, 
				    &local_addr, &remote_addr, 
				    uriData->virtHost)) {
	do_status = 1;
    }
    

    if (!do_status) {
	/**
	 * Lock the scoreboard briefly while we look for an app matching
	 * our prefix.
	 */
	/* ###### ENTER CRITICAL SECTION - scoreboard ###### */
	vLock = crit_init();
	crit_enter(vLock);
	edir_load_choose(scoreboard, uriData->pathInfo, stickyTag,
			 &app, &server, refCounts, numServers, 
			 &local_addr, &remote_addr, uriData->virtHost, exc);
	if (*exc) {
	    crit_exit(vLock);
	    EDIR_THROW(*exc);
	}
	EDIR_DEBUG_NORMAL(("Got server: %s:%d with sessionID = %s", \
			   server->hostName, server->port, \
			   uriData->sessionID ? uriData->sessionID : "(null)"));
	/* Up the use count on this server. */
	server->useCount++;
	/* Up the refcount */
	releaseConn = 1;
	refCounts[server->refIndex]++;
	/* Give up scoreboard lock, now that we have our server. */
	crit_exit(vLock);
	/* ###### LEAVE CRITICAL SECTION - scoreboard ###### */
    }

    if (!(edir_debug & EDIR_DEBUG_MASK_FAKE_IT) && !do_status) {
        edir_connect_to_enhydra(server,
				pRequest,
				pSession,
				cfg,
				uriData,
				app,
				stickyTag,
				&releaseConn,
				vLock,
				&local_addr,
				&remote_addr,
				exc);
	if (*exc) EDIR_THROW(*exc);
    } else { 
        /* Just display a status page */
	edir_kvtable_t *respHeaders;
	char		clength[30], *key, **vals;
	int32		len, nvals;
	int32		iv32;
	int		status;
	void		*sbuf = NULL;

	respHeaders = edir_kvtable_create(EDIR_KVTABLE_TO_LOWER_PRESERVE);
	sbuf = edir_load_status(scoreboard, uriData->pathInfo, 
				refCounts, numServers, respHeaders, exc);
	if (*exc) EDIR_THROW(*exc);

	val = edir_kvtable_find(respHeaders, "status");
	if (val) {
	    for (cp = val; *cp && isspace(*cp); cp++) ;
	    for (dp = cp; *dp && isdigit(*dp); dp++) ;
	    if (*dp) *dp++ = '\0';
	    if (edir_atoi(cp, &iv32)) {
		status = (int) iv32;
	    } else {
		status = PROTOCOL_OK;
	    }
	    edir_kvtable_remove(respHeaders, "status");
	} else {
	    status = PROTOCOL_OK;
	}

	val = edir_kvtable_find(respHeaders, "content-type");
	if (!val) val = "text/plain";
        pblock_remove("content-type", pRequest->srvhdrs);
	pblock_nvinsert("content-type", val, pRequest->srvhdrs);
	edir_kvtable_remove(respHeaders, "content-type");
	edir_kvtable_remove(respHeaders, "content-length");
	sprintf(clength, "%d", edir_sbuf_length(sbuf));
	pblock_nvinsert("content-length", clength, pRequest->srvhdrs);

	vals = edir_kvtable_first(respHeaders, &key, &nvals);
	while (vals) {
	    char kbuf[200];
	    strncpy(kbuf, key, 199);
	    kbuf[198] = kbuf[199] = '\0';
	    for (cp=kbuf; *cp; cp++) *cp = tolower(*cp);
	    pblock_remove(kbuf, pRequest->srvhdrs);
	    pblock_nvinsert(kbuf, vals[nvals-1], pRequest->srvhdrs);
	    vals = edir_kvtable_next(respHeaders, &key, &nvals);
	}
	edir_kvtable_destroy(respHeaders);
	protocol_status(pSession, pRequest, status, NULL);
	if (protocol_start_response(pSession, pRequest) != REQ_NOACTION) {
	    /* We weren't told to return immediately */
	    val = edir_sbuf_tostring(sbuf, &len);

	    if (net_write(pSession->csd, val, len) == IO_ERROR) {
		edir_free(val);
		edir_sbuf_free(sbuf);
		return REQ_EXIT;
	    }
	    edir_free(val);
	}
	if (sbuf) edir_sbuf_free(sbuf);
    } /* If not fake it */

    /*
     * Decrement the remote server's reference count.
     */
    if (releaseConn && server && !do_status) {
        crit_enter(vLock);
	if (refCounts[server->refIndex] > 0)
	    refCounts[server->refIndex]--;
        releaseConn = 0;
	crit_exit(vLock);
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns ok", funcName));
    return REQ_PROCEED;

except:
    if (*exc) {
	char *trace = edir_exception_stack_trace(*exc);
	edir_log_error("Request failed...");
	edir_log_error("%s", trace);
	edir_free(trace);
	edir_exception_free(exc);
    }
    if (releaseConn && server) {
        crit_enter(vLock);
        if (refCounts[server->refIndex] > 0)
	    refCounts[server->refIndex]--;
        releaseConn = 0;
	crit_exit(vLock);
    }
    if (vLock) {
      /**
       * I'm assumign that crit_termiante will also call crit_exit 
       * if we're still in a block of critical code, otherwise, 
       * we could be bummed
       */
        crit_terminate(vLock);
	vLock = NULL;
    }

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns error (%d)", funcName, cond_status));
    return cond_status;
}
