#ifdef XP_WIN32
#define NSAPI_PUBLIC __declspec(dllexport)
#include <winsock2.h>
#else /* !XP_WIN32 */
#define NSAPI_PUBLIC
#endif /* !XP_WIN32 */ 

#include "nsapi.h"
#include "edir.h"
#include "edir_kvtable.h"
#include "edir_nsapi.h"

/*
 * Default config file.
 */
edir_nsapi_cfg_t *globalCfg = NULL;
edir_conf_t	 *appCfg = NULL;

/**
 * Global reference count array to use instead of scoreboard itself.
 * <P>
 * The scoreboard itself has a reference count for each server entry,
 * but the actual scoreboard access code takes an array of references
 * in order to allow flexible handling of reference counts in multi-
 * process environments such as Apache.
 * <P>
 * It turns out that it is more efficient and less complex to just keep
 * a separate array of reference counts in parallel to the scoreboard.
 * XXX FIX A better solution will be to rearrange the scoreboard to
 * store the reference counts in a contiguous array, as that would also
 * allow non-systemV refcount methods to implement simpler code.
 */
int32 *refCounts = NULL;

/**
 * The number of servers (and thus reference counts).
 */
int32 numServers = 0;

/**
 * Pointer to the global scoreboard.  This scoreboard follows the
 * same format as the ISAPI and Apache modules.  It is position
 * independent, even though it does not need to be.  This allows for
 * maintenance of the scoreboard access code in only one place, which
 * should facilitate increased robustness and reliablility across all
 * platforms.
 */

edir_load_sb_offsets_t *scoreboard = NULL;

/**
 * Number of bytes allocated for the scoreboard.
 */

int32 sbSize = 0;

/**
 * Create a per-directory configuration record for mod_enhydra_director.
 * <P>
 * Note that this won't be called, since Enhydra Director is not configurable
 * on a per-directory basis.
 *
 * @param p       [in,out] Apache resource pool for temporarily allocated data.
 * @param dirspec [in      Directory under which this configuration resides.
 * @return The configuration record.
 */
static edir_nsapi_cfg_t *
create_server_config()
{
    edir_nsapi_cfg_t	*cfg;
    cfg = (edir_nsapi_cfg_t *) PERM_MALLOC(sizeof(edir_nsapi_cfg_t));
    cfg->cfg_Mode = EDIR_NSAPI_CFG_MODE_SERVER;
    cfg->inited = 0;
    cfg->cfg_Inheritable = 0;
    cfg->cfg_Dir = NULL;
    cfg->configFile = NULL;
    cfg->dataFile = NULL;
    cfg->lockFile = NULL;
    cfg->pathToDaemon = NULL;
    cfg->inBufSize = EDIR_CLIENT_DEFAULT_INBUF_SIZE;
    cfg->outBufSize = EDIR_CLIENT_DEFAULT_OUTBUF_SIZE;
    cfg->sessionIdPrefix = edir_strdup(EDIR_LOAD_DEFAULT_SID_PREFIX);
    return cfg;
}

/**
 * Module initializer callback.
 * <P>
 * This callback gets called to allow the Enhydra Director module to initialize
 * itself for the first time.  
 * We create a config info block.
 * 
 * @param pb [in] obj.conf command paramaters given to: Init fn="edir_init"
 * @param sNotUsed [-]
 * @param rNotUsed [-]
 */
#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int edir_init(pblock *pb, Session *sNotUsed, Request *rNotUsed)
{
    static char		*funcName = "edir_init";
    edir_exception_t	*_exc=NULL, **exc=&_exc;
    edir_nsapi_cfg_t	*cfg = NULL;
    char                *pathToDaemon = NULL;
    char		*confDir = NULL, *confFile = NULL;
    char		*dataDir = NULL, *dataFile = NULL, *lockFile = NULL;
    char                *trace = NULL;
    int                  i;

    edir_nsapi_log_init();
    edir_nsapi_log_set_session(sNotUsed, rNotUsed);

    EDIR_DEBUG_FUNCS_IN(("func: %s("EDIR_PTR")", funcName, pb));

    cfg = create_server_config();
    globalCfg = cfg;
    
    if (cfg && cfg->inited) {
	EDIR_DEBUG_VERBOSE(("%s: already initialized", funcName));
	EDIR_DEBUG_FUNCS_OUT(("func: %s() returns void", funcName));
	return REQ_ABORTED;
    }

    confFile = pblock_findval("conffile", pb);

    if (confFile == NULL) {
	edir_log_error("conffile= undefined in obj.conf's Init fn=\"edir_init\"");
	/* If we weren't given a location, use reasonable default */
#ifdef EDIR_AC_CONFIG_DIR
	confFile = edir_asprintf("%s/enhydra_director.conf",
	       		 	 EDIR_AC_CONFIG_DIR);
#else
        EDIR_DEBUG_FUNCS_OUT(("func: %s() returns REQ_ABORTED", funcName));
	return REQ_ABORTED;
#endif
    }

    /**
     * Try to open and process the EnhydraLink config file.
     */
    appCfg = edir_conf_process(confFile, exc);
    if (*exc) EDIR_THROW(*exc);

    /**
     * Set debugging mask if needed.
     */
    if (appCfg->debug) edir_debug |= appCfg->debugMask;

    /**
     * Allocate space for the scoreboard from the global memory
     * memory heap.  This memory is never freed until the DLL is
     * unloaded.  Note that we don't use edir_malloc because it's
     * set to the _temporary_ heap.
     */
    sbSize = edir_load_sb_compute_size(appCfg);
    scoreboard = (edir_load_sb_offsets_t *) edir_malloc(sbSize);

    /**
     * Convert the config data to a position-independent scoreboard.
     * We don't really need position-independent data here, but we
     * use it anyhow so that we can use the same scoreboard code
     * for all module implementations, including Apache.
     */
    edir_load_sb_from_config(appCfg, (void *)scoreboard);

    /*
     * Create a global reference count array.  We could use the
     * refcounts in the scoreboard, but since this is a single process,
     * it's more efficient to keep a separate array.
     */
    numServers = edir_load_sb_num_servers((void *)scoreboard);
    refCounts = (int32 *) edir_malloc(numServers * sizeof(int32));
    for (i=0; i<numServers; i++) refCounts[i] = 0;

    /* debug: print out the scoreboard contents */
    trace = edir_load_sb_debug_string((void *)scoreboard,
	   			      refCounts,
				      numServers);
    edir_log("Scoreboard:\n%s", trace);
    edir_free(trace);

    EDIR_DEBUG_FUNCS_OUT(("func: %s() returns REQ_PROCEED", funcName));
    return REQ_PROCEED;

except:

    edir_log_error("Error initializing scoreboard daemon.");
    if (*exc) {
	char *trace;
	trace = edir_exception_stack_trace(*exc);
	edir_log_error(trace);
	edir_free(trace);
	edir_exception_free(exc);
    }

    EDIR_DEBUG_FUNCS_IN(("func: %s() returns REQ_ABORTED", funcName));
    return REQ_ABORTED;
}
