/*
 * Enhydra Java Application Server 
 * The Initial Developer of the Original Code is Lutris Technologies Inc. 
 * Portions created by Lutris are Copyright (C) 1997-2000 Lutris Technologies
 * Inc. 
 * All Rights Reserved. 
 *
 * The contents of this file are subject to the Enhydra Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.enhydra.org/software/license/epl.html 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License. 
 *
 * $Id: edir_nsapi_log_line.c,v 1.1 2003/04/14 12:26:02 sinisa Exp $
 */

#ifdef XP_WIN32
#define NSAPI_PUBLIC __declspec(dllexport)
#include <winsock2.h>
#else /* !XP_WIN32 */
#define NSAPI_PUBLIC
#endif /* !XP_WIN32 */

#include "nsapi.h"
#include "edir.h"
#include "frame/log.h"

#ifndef EDIR_LOG_NOTICE
#define EDIR_LOG_NOTICE 0
#endif

#ifndef EDIR_LOG_ERR
#define EDIR_LOG_ERR 1
#endif

#ifndef EDIR_LOG_DEBUG
#define EDIR_LOG_DEBUG 2
#endif

/**
 * Per-thread session pointer index.
 */
static int tls_session_key = -1;

/**
 * Per-thread request pointer index.
 */
static int tls_request_key = -1;

/**
 * Log a single line of text to the correct log channel.
 * <P>
 * This function should only be called by the general logging functions
 * in 'edir_log.c'.  The common line logging function has been split out
 * into a separate module to allow specific client implementations to
 * integrate logging into the native web server's mechanism.
 * 
 * @param level [in]  The logging level, as per syslog.
 * @param line  [in]  Single line of pre-formatted log text.
 */
extern void
edir_log_line(int level, char *line)
{
    Session	*sn = NULL;
    Request	*rq = NULL;
    int		logFailure = 0;

    sn = (Session *) systhread_getdata(tls_session_key);
    rq = (Request *) systhread_getdata(tls_request_key);

    switch (level) {
	case EDIR_LOG_ERR:
#ifdef XP_WIN32
	    EDIR_IF_DEBUG( OutputDebugString(line); )
	    EDIR_IF_DEBUG( OutputDebugString("\n"); )
#endif
	    logFailure = log_error(LOG_FAILURE, "EnhydraDirector",
				   sn, rq, "%s", line);
	    break;
	case EDIR_LOG_DEBUG:
#ifdef XP_WIN32
	    EDIR_IF_DEBUG( OutputDebugString(line); )
	    EDIR_IF_DEBUG( OutputDebugString("\n"); )
#endif
	    logFailure = log_error(LOG_VERBOSE, "EnhydraDirector",
				   sn, rq, "%s", line);
	    break;
	case EDIR_LOG_NOTICE:
	default:
#ifdef XP_WIN32
	    EDIR_IF_DEBUG( OutputDebugString(line); )
	    EDIR_IF_DEBUG( OutputDebugString("\n"); )
#endif
	    logFailure = log_error(LOG_INFORM, "EnhydraDirector",
				   sn, rq, "%s", line);
	    break;
    }
    
    if (logFailure) {
	fprintf(stderr, "%s: %s\n", (level == EDIR_LOG_ERR) ? "ERROR" : "LOG",
		line);
	fflush(stderr);
    }
}

/**
 * Set whether the log system is in 'daemon' mode.
 * <P>
 * In the standard log module, this means "use syslog instead of printf".
 * For apache, it also means 'use syslog', but is only ever called
 * by the child process after forking the Enhydra Director daemon,
 * but before the exec().
 * 
 * @param progName [in]  The name of the program doing the logging.
 * @param isDaemon [in]  True for daemon mode, false for "normal".
 */
extern void
edir_log_setdaemon(char *progName, int isDaemon)
{
    return;
}

/**
 * Initialize the per-thread storage keys for the logger.
 */
extern void
edir_nsapi_log_init()
{
    tls_session_key = systhread_newkey();
    tls_request_key = systhread_newkey();
}

/**
 * Set the request and session for this thread.
 *
 * @param sn [in] Session in which to log for this thread.
 * @param rq [in] Request for which to log.
 */
extern void
edir_nsapi_log_set_session(Session *sn, Request *rq)
{
    systhread_setdata(tls_session_key, (void *) sn);
    systhread_setdata(tls_request_key, (void *) rq);
}

